package cia_components.ppc;
//import cia_components.reasoning.OWL_Query_Manager;

import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLIndividual;

import java.net.URI;
import java.util.*;
import java.util.regex.*;

/**
 * This class implements the algorithm for assigning CDG scores to the 
 * thematic relations between words or homonyms received from cdg. 
 * Whether scoring is between words or between homonyms depends on the
 * value of the variable PPC_INTEGRATION_MODE which is defined in the 
 * configuration script ~/cdg2/PPC/configure.
 * 
 * @author Patrick McCrae
 */

public class Score_Assigner
{   		
	OWL_Query_Manager query_manager;
	Sentence sentence;	
	
	HashMap <Word, Set <OWLIndividual>> crossmodal_anchoring;	
	// Maps a word object from the input sentence onto its corresponding 
	// OWLIndividual in the context model.
	
	HashSet <String> thematic_roles, thematic_relations, situation_roles;
	HashMap <String, String> score;
	
	HashMap <String, Set <String>> complement_roles; 
	// Maps each thematic role name onto the set of its complement roles, i.e. the roles
	// which can be excluded in case the first role has been detected.
	
	HashMap <OWLClass, Set<OWLClass>> compatible_context_concepts;
	// Maps each word concept onto the set of context concepts with which it is logically compatible.
	
	Vector <Word> contextually_compatible_words;
	/*	Contains all words that have a conceptualisation associated with them.
		We need a vector structure because the sequence in which words are added is of relevance in the
		anchoring check for whether a word is the last word in its slot which is performed in method 
		anchor_Words_in_Context_Individuals().
	*/
	
	Vector <String> descriptions;
	// Contains all description strings of the words in this sentence.
	
	HashMap <String, Float> context_compliance;
	// Maps each thematic role string onto the degree of compliance with cross-modal context as defined for that thematic role.
	
	HashMap <String, Float> veto_penalty;
	// Maps the name of a thematic role onto the veto penalty for that thematic role.
	


    
    public Score_Assigner( OWL_Query_Manager manager, Sentence sen )
	{
//    	long time1 = System.currentTimeMillis();
//    	long delta;
    	
		set_Query_Manager( manager );
		
//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Query_Manager': " + delta );
			
		set_Thematic_Roles();

//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Thematic_Roles': " + delta );

		
		set_Thematic_Relations();

//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Thematic_Relations': " + delta );

		set_Situation_Roles();
		
//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Situation_Roles': " + delta );
		
		set_Complement_Roles();
		
//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Complement_Roles': " + delta );
				
		set_Sentence( sen );
		
//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Sentence': " + delta );

		set_Description_Strings();

//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Description_Strings': " + delta );
		
		set_Context_Compliances_and_Veto_Penalties();
		
//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Perceptual_Certainties_and_Veto_Penalties': " + delta );

		ground_Words();

//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'ground_Words': " + delta );

		set_Contextually_Compatible_Word_Concepts();

//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'set_Contextually_Compatible_Word_Concepts': " + delta );
		
		anchor_Words_in_Context_Individuals();

//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'anchor_Words_in_Context_Individuals': " + delta );

		score = new HashMap <String, String> ();

		String ppc_integration_mode = System.getenv( "PPC_INTEGRATION_MODE" );
		
		if ( ppc_integration_mode.equals( "1" ) ) 
		{	
			assign_Predictions_Homonym_to_Homonym();
			// System.out.println( "Output Mode: homonym to homonym" );
		}
		else if ( ppc_integration_mode.equals( "2" ) ) 
		{	
			assign_Predictions_Homonym_to_Slot();
			// System.out.println( "Output Mode: homonym to slot" );
		}
		else if ( ppc_integration_mode.equals( "3" ) ) 
		{			
			assign_Predictions_Slot_to_Homonym();
			// System.out.println( "Output Mode: slot to homonym" );
		}
		else if ( ppc_integration_mode.equals ( "4" ) ) 
		{	
			assign_Predictions_Slot_to_Slot();
			// System.out.println( "Output Mode: slot to slot" );
		}
		
//		delta = System.currentTimeMillis() - time1;
//		System.out.println( "Time after 'Prediction assignment': " + delta );

	}
	

	/** 
     * This method sets the set of complement roles for each role on the INST level.
     * Complement roles are thematic roles which mutually exclude each other - 
     * at least according to the current design of the role-assigning grammar. 
     * As a result, if a dependant is found to take role X we can also place a veto 
     * on all other relations of this dependant with its complement roles.
     */
    
    private void set_Complement_Roles()
    {
    	complement_roles = new HashMap <String, Set <String>> ();

    	HashSet <String> set = new HashSet <String> ( thematic_roles );
    	set.remove( "AGENT" );	// Complement roles are only defined on the INST level.
    	set.remove( "THEME" );	// Complement roles are only defined on the INST level.
    	
    	for ( String role : set )
    	{
    		HashSet <String> set_copy = new HashSet <String> ( set );
    		set_copy.remove( role );
    		// System.out.println( "Setting complements for role '" + role + "'." );

    		complement_roles.put( role, set_copy );
    	}
    }


	/** This method iterates over all words and retrieves the set of concepts 
	 * 	that defines each word's meaning. For each of these word conceptualisations 
	 * 	in the input sentence this method determines which concepts instantiated 
	 *  in the context model are compatible with it. 
	 *  
     * 	The set of context concepts that are compatible with a given conceptualisation 
     * 	gets saved in the hashmap 'compatible_context_concepts' that takes a
     * 	word conceptualisation as key and the compatible context concepts as value.
     * 
     * 	Observe that the compatibility of word conceptualisations with context concepts 
     * 	is NOT a word property but a global property of the loaded ontology. However, so
     *  far this implementation does not use this fact to improve the efficiency of the
     *  matching algorithm.
     */
    
	private void set_Contextually_Compatible_Word_Concepts()
	{	
		contextually_compatible_words = new Vector <Word> ();

		for ( Word word : sentence.get_All_Word_Objects() )		// Iterate over all words in the sentence.
		{	
			for ( OWLClass word_concept : word.get_Conceptualisations() )	// Iterate over the current word's set of conceptualisations.
			{
//				System.err.println( "\nWord: '" + word.get_Description() );
//				System.err.println( "\tExamining conceptualisation '" + word_concept + "' for contextual compatibility." );
	
				for ( OWLClass context_concept : query_manager.get_Context_Concepts() )	// Iterate over all concepts instantiated in the context model.
				{
//					System.err.println( "\tChecking word concept '" + word_concept + "' and context concept '" + context_concept + "' for compatibility." );
	
					if ( query_manager.is_Compatible( word_concept, context_concept ) )
					{
						word.add_Contextually_Compatible_Conceptualisation( word_concept );

//						System.err.println( "\t\t Word concept '" + word_concept + "' is compatible with concept '" + context_concept + "' in the context model." );
						
						/* 
						 * 	Add the context concept to the set of context concepts that are compatible with the
						 *	current word conceptualisation. The set of context concepts compatible with the current word 
						 *	concept has been instantiated in the method ground_words() and hence may be empty but never 
						 *	returns 'null'.
						 */
						
						compatible_context_concepts.get( word_concept ).add( context_concept ); 
												
						
						/* 	Add the word to the set of Word objects that activate a concept that is 
						 * 	compatible with a concept instantiated in context.
						 * 	Note that we cannot use a HashSet as a data structure since we need the 
						 * 	sequence of elements in this data structure for a check in method 
						 * 	anchor_Words_in_Context_Individuals().
						*/

						if ( ! contextually_compatible_words.contains( word ) )
						{
							contextually_compatible_words.add( word );
//							System.err.println( "\tAdded the Word object '" + word.get_Description() + "' to the set of words whose concept is compatible with a contextually instantiated concept." );
						}
					}
//					else
//					{
//						System.err.println( "\t\t Word concept '" + word_concept + "' is not compatible with concept '" + context_concept + "' in the context model." );
//					}
				}
			}
		}
	}

	
	
	private void set_Situation_Roles()
	{
		situation_roles = new HashSet <String> ();
		situation_roles.add( "AGENT" );
		situation_roles.add( "THEME" );
		situation_roles.add( "RECIPIENT" );
	}


	private void set_Description_Strings()
	{
		descriptions = sentence.get_Descriptions();		
	}


	/**
	 * This method iterates over all thematic roles supported and assigns each one 
	 * the context compliance value defined for it in the configure script and the
	 * resulting CDG penalty score.
	 */
		
	private void set_Context_Compliances_and_Veto_Penalties()
	{
		context_compliance = new HashMap <String, Float>();
		veto_penalty = new HashMap <String, Float>();
		
		for ( String role : thematic_roles )	// Iterate over all thematic roles supported.
		{	
			// Observe that the cast from String to Float is syntactically somewhat involved.
			context_compliance.put( role, Float.valueOf( System.getenv( "CONTEXT_COMPLIANCE_" + role ) ) );
			veto_penalty.put( role, 1f - context_compliance.get( role ) );
						
			// System.err.println( "Set perceptual certainty for role " + role + " to " + perceptual_certainty.get( role ) );
			// System.err.println( "Set veto penalty for role " + role + " to " + veto_penalty.get( role ) );
		}
	}


	
	/**
	 * 	This method iterates over all word objects in the input sentence 
	 * 	that have a contextually compatible conceptualisation and tries 
	 * 	to map each of those Word objects onto OWLIndividuals asserted in 
	 * 	the context model which instantiate concepts that are compatible 
	 * 	with the set of concepts activated by the Word object.
	 * 
	 * 	An earlier version of this method included an algorithm for resolving
	 * 	ambiguous references for cases like
	 * 	Match(Hom.2.1) = {I1}
	 * 	Match(Hom.3.1) = {I1, I2}
	 * 	==> Match(Hom.3.1) = {I2}
	 * 
	 * 	This resolution of ambiguous reference has been commented out since it was 
	 * 	unclear how to proceed in cases like
	 * 	Match(Hom.1.1) = {I1}
	 * 	Match(Hom.2.1) = {I1}
	 * 	Match(Hom.3.1) = {I1, I2}
	 * 	==> ???
	 * 	 */
    
    private void anchor_Words_in_Context_Individuals()
    {
		crossmodal_anchoring = new HashMap <Word, Set <OWLIndividual>> ();
		
		HashMap <Word, Set<OWLIndividual>> multiple_candidate_set = new HashMap <Word, Set<OWLIndividual>> ();
		
		// Initialise the set of concept instantiations in the context model which instantiate a concept that 
		// is compatible with a given word.
		
		Set <OWLIndividual> matched_context_individuals = new HashSet <OWLIndividual> ();
		
		
		/* 
		 * Only words that activate a concept which is also compatible with a concept instantiated 
		 * in cross-modal context can benefit from cross-modal interaction. All other words activate 
		 * concepts that the context model cannot provide information about.
		 * 
		 * Note that in the current implementation this loop will keep overwriting the mapping
		 * from Word object to concept individual in visual context until there are no more concepts
		 * in the set of conceptualisations for that Word object. Since most words only have a
		 * conceptualisation of zero or one concepts, this works. 
		 * 
		 * Populate the set of all context individuals onto which a word is matched in cross-modal
		 * matching. Effectively, these concept instances are potential referents of the word.
		 * This set contains <all> contextually asserted concept instantiations in visual context 
		 * that instantiate a concept that is compatible with at least one of the concepts activated 
		 * by the current Word object.
		 */
		
		for ( Word word_to_anchor : contextually_compatible_words )	// Iterate over all Word objects (homonyms) in the sentence that have a contextually compatible conceptualisation.
		{
//			System.err.println( "\nTrying to anchor the Word object '" + word_to_anchor.get_Description() + "'." );
//			System.err.println( "Word '" + word_to_anchor.get_Description() + "' has conceptualisations: " + word_to_anchor.get_Conceptualisations() + "." );
//			System.err.println( "Word '" + word_to_anchor.get_Description() + "' is contextually compatible with: " + word_to_anchor.get_Contextually_Compatible_Conceptualisations() + "." );
			
//			Set <OWLIndividual> potential_individual_matches = new HashSet <OWLIndividual> ();
			
			for ( OWLClass compatible_word_concept : word_to_anchor.get_Contextually_Compatible_Conceptualisations() )	// Iterate over the current Word's contextually compatible concepts.
			{	
				for ( OWLClass compatible_context_concept : compatible_context_concepts.get( compatible_word_concept ) ) // Iterate over the context concepts that are compatible with the current conceptualisation from the current word's set of contextually compatible concepts.
				{
//					System.err.println( "The Word concept '" + compatible_word_concept + "' is compatible with the set of context concepts: " + compatible_context_concepts.get( compatible_word_concept ) + "." );

					// Add all individuals instantiating the concepts with which the current word is compatible to the set.
					// potential_individual_matches.addAll( query_manager.get_Context_Individuals( compatible_context_concept ) );
					add_Anchors( word_to_anchor, query_manager.get_Context_Individuals( compatible_context_concept ) );
				}
			}
				
//			System.err.println( "Word '" + word_to_anchor.get_Description() + "' maps to the following context individuals: " + get_Anchors( word_to_anchor ) + "." );

//			/* 
//			 * The mapping of Word object to context concept can be performed unambiguously
//			 * if the set of potential individual matches only contains one element.
//			 */
//
//			if ( potential_individual_matches.size() == 1 )
//			{
//				/* 
//				 * 	Accessing this OWLIndividual is a little tricky since the Set 
//				 * 	interface provides no method for accessing a specific individual 
//				 * 	element in a set. We therefore convert the Set to a Vector to 
//				 * 	access the only element in the set.
//				 */
//				
//				OWLIndividual individual = ( new Vector <OWLIndividual> ( potential_individual_matches ) ).firstElement();
//				
//				set_Anchors( word_to_anchor, individual );
//
//				 /* 	
//				  *	Only multiple words from the same slot may map onto the same
//				  * context individual. Hence only add the context individual to
//				  * the list of matched context individuals once the mapping against 
//				  * context individuals has been performed for all words from 
//				  * a given slot.
//				 */
//
//				if ( word_to_anchor == word_to_anchor.get_Parent_Slot().get_Last_Word() )
//				{
//					matched_context_individuals.add( individual );
//				}
//								
//				// System.out.println( "Anchored word '" + word_to_anchor.get_Description() + "' in context individual '" + individual.toString() + "'." );
//			}
//			else if( potential_individual_matches.size() > 1 )
//			{
//				multiple_candidate_set.put( word_to_anchor, potential_individual_matches );
//				// System.err.println( "Keeping the set of potentially matching context individuals '" + potential_individual_matches + "' for word '" + word_to_anchor.get_Description() + "' for later processing." );
//			}
//			else if ( potential_individual_matches.size() == 0 )
//			{
//				System.err.println( "The set of individual candidates in context for matching was empty for word '" + word_to_anchor.get_Description() + "'." );
//				// This case should never arise since we only iterate over the set of Word objects which 
//				// activate a concept that is compatible with a concept which has been instantiated in 
//				// visual context.
//			}
//		}
//		
//		
//		// Now remove all unambiguously mapped individuals from the multiple candidate sets.
//		// Note that this removal may result in a Word object which previously had cross-modal
//		// matches but now has none left.
//		
//		for ( Word word : multiple_candidate_set.keySet() )
//		{
//			Set <OWLIndividual> set = multiple_candidate_set.get( word );
//			
//			set.removeAll( matched_context_individuals );
//			System.err.println( "Word object '" + word.get_Description() + "': Removed the context individuals " + matched_context_individuals + " from set of cross-modal matches." );
//			if ( set.isEmpty() )
//			{
//				System.err.println( "The Word object '" + word.get_Description() + "' has no more matches to context individuals." );
//			}
//			else
//			{
//				set_Anchors( word, set );
//			}
		}
		
//		System.err.println( "Completed the anchoring of words in context individuals." );
	}
   
    
    /**	
     * 	This method prepares the cross-modal map from Word object to the concept instance
     * 	in the context model in case the Word object is compatible with only one concept
     * 	instance in the context model. The preparation essentially consists of creating a 
     * 	set of just one element and applying the method for set-based cross-modal anchoring.
     */
    
    private void add_Anchors( Word word, OWLIndividual individual )
    {
    	HashSet <OWLIndividual> set = new HashSet <OWLIndividual> ();
    	set.add( individual );
    	add_Anchors( word, set );
    	
    	// System.err.println( "Anchored word '" + word.get_Word_String() + "/" + word.get_Description() + "' in context individual '" + individual.toString() + "'." );
    }


    /**	
     * 	This method defines the cross-modal map from word object to the set of 
     * 	concept instances in the context model that the word object is compatible 
     * 	with.
     */

    private void add_Anchors( Word word, Set <OWLIndividual> set )
    {
    	if ( get_Anchors( word ) == null )
    	{
    		crossmodal_anchoring.put( word, new HashSet <OWLIndividual> () );
    	}
    	
    	get_Anchors( word ).addAll( set );	

//		System.err.println( "Anchored word '" + word.get_Word_String() + "/" + word.get_Description() + "' in the set of context individuals: " + get_Anchor( word ) + "." );
    }

    
//	/**
//     * This method returns the context concept which is most similar to a concept in the set of
//     * contextually compatible concepts in the linguistic modality. Note that methods for 
//     * determining conceptual similarity may differ. 
//     * 
//     * In this implementation, conceptual similarity is (rather crudely) approximated
//     * by the equality of the concepts' lexicalisation strings.
//     * 
//     * @param set
//     * The set of word concepts for which the most similar concept to a concept in the 
//     * context model is to be found.
//     * 
//     * @return
//     * The concept in the context model which is most similar to the input concept.
//     * 
//     */
//    
//	private OWLClass get_Closest_Context_Concept( Set<OWLClass> set )
//	{
//		OWLClass result = null; 
//		
//		for ( OWLClass word_concept : set )
//		{
//			for ( OWLClass context_concept : query_manager.get_Context_Concepts() )
//			{
//				// System.out.println( "Word concept '" + word_concept + "'" );
//				// System.out.println( "Context concept '" + context_concept + "'." );
//				
//				if ( query_manager.get_Lexicalisations( word_concept ).equals( query_manager.get_Lexicalisations( context_concept ) ) )
//				{
//					result = context_concept;
//					System.out.println( "Word concept '" + word_concept + "' is closest to visual context concept '" + context_concept + "'." );
//					break; 
//				}
//			}
//		}
//
//		return result;
//	}
	
	
	/**	This method assigns the prediction values for the PPC when invoked by CDG2 
	 * 	in homonym-to-homonym mode, i.e.: The entire list of homonyms for the input 
	 * 	sentence is received from CDG2 and the PPC attempts to generate predictions 
	 * 	for dependencies between homonyms rather than slots.
 	 */
	
	private void assign_Predictions_Homonym_to_Homonym()
	{		
		for ( Word dependant_word : sentence.get_All_Word_Objects() )	// Iterate over all word objects (homonyms) in the dependant slot.
		{		
			Iterator <Slot> regent_slot_iterator = sentence.get_All_Slot_Objects().iterator();	// We cannot iterate in a for-loop since we want to skip the first element.
			regent_slot_iterator.next();	// Skip slot no. 0 which contains the sentence's ID line that CDG2 hands over as well.

			Slot dependant_slot = dependant_word.get_Parent_Slot();

			Find_Regent:
				
			while ( regent_slot_iterator.hasNext() ) // Iterate over all regent slots other than slot no. 0. 
			{
				Slot regent_slot = regent_slot_iterator.next();	// We cannot iterate in a for-loop since we need the ability to skip elements.

				/*
				 * 	No slot can have a thematic relation with itself. Hence, prediction values 
				 * 	will only be calculated for those slots that mutually differ from each other. 
				 * 	CDG sets prediction values for all dependencies that have not received a score 
				 * 	from the predictor to 1.0 by default.
				*/
	
				if ( dependant_slot == regent_slot )	// Do not attempt to calculate scores for dependant_slot == regent_slot.
				{
					if ( regent_slot_iterator.hasNext() )
					{
//						System.err.println( "Dependant slot : " + dependant_slot.get_Slot_Number() + ": Skipping regent slot " + regent_slot.get_Slot_Number() );						
						regent_slot = regent_slot_iterator.next();
					}
					else	// Skip to the next dependant if there are no alternative regents to check for anymore.
					{						
//						System.err.println( "Leaving loop" );						
						break Find_Regent;	
					}
				}

				
				for ( Word regent_word : regent_slot.get_All_Word_Objects() )	// Iterate over all word objects (homonyms) in the regent slot.
				{
//					System.err.println( "Examining Word " + dependant_slot.get_Slot_Number() + "." + dependant_word.get_Position_in_Slot() + " and " + regent_slot.get_Slot_Number() + "." + regent_word.get_Position_in_Slot() );						
//					System.err.println( "Found " + dependant_slot.size() + " words in dependant slot " + dependant_slot.get_Slot_Number() );

				
					/*
					 * 	The PPC can only predict the score for dependencies between Word objects which 
					 * 	map to one or more OWLIndividuals in the context model. These Words must therefore
					 * 	fulfil the following coniditions:
					 * 
					 * 	(1)	The Word object must activate one or more concepts from the ontology.
					 * 
					 * 	(2)	The concepts activated by the Word object must be compatible with one or 
					 * 		more concecpts instantiated in the context model.
					 * 
					 * 	Hence relation scoring only needs to be performed for dependant-regent pairs for
					 * 	which both dependant and regent satisfy these two conditions. Scoring the 
					 * 	dependency between dependant and regent is performed based on the relation 
					 * 	detected between the individuals associated with the dependant and regent Word 
					 * 	objects. 
					 * 	Otherwise, the context model cannot provide cross-modal information on the 
					 * 	relation between entities and we hence cannot make any context-based predictions.
					 * 	In those cases the PPC will make no prediction, which results in CDG assigning 
					 * 	the default score of '1.0' to the relation upon reading in the prediction 
					 * 	results via its predictor interface, i.e.: no constraint is imposed on any 
					 * 	possible dependency.
					 */
							
					boolean a1 = ! dependant_word.get_Contextually_Compatible_Conceptualisations().isEmpty();
					boolean a2 = true; //! get_Anchor( dependant_word ).isEmpty();
					
					boolean b1 = ! regent_word.get_Contextually_Compatible_Conceptualisations().isEmpty();					
					boolean b2 = true; //! get_Anchor( regent_word ).isEmpty();
					
					if ( a1 && a2 && b1 && b2 )
					{
						score_Relation( dependant_word, regent_word );
					}
				}	
			}
		}
	}

	
	private void assign_Predictions_Homonym_to_Slot()
	{
//		Slot dependant_slot_object, slot_relation_target;
//		int dependant_slot_size;
//		
//		Word dependant_word_object, regent_word_object;
//		String regent_ID_string = "";
//		String key = "";
//		float value;
//		String prediction_string;
//		
//		
//		for ( int i = 1; i <= sentence_length; i++ )		// For all words to score iterate over all slots except for slot no. 0 which contains the ID line.
//		{
//			dependant_slot_object = sentence.get_Slot_Object( i );
//						
//			dependant_slot_size = dependant_slot_object.size();
//						
//			for ( int j = 1; j <= dependant_slot_size; j++ )		// For all words to score in a slot iterate over all homonyms in that slot.
//			{
//				dependant_word_object = dependant_slot_object.get_Word_Object( j );
//				prediction_string = "";				
//				
//				for ( int l = 1; l <= sentence_length; l++ )		// For the relation target iterate over all slots except for slot no. 0 containing the ID line.
//				{
//					if ( i != l )
//					{
//						slot_relation_target = sentence.get_Slot_Object( l );
//														
//						regent_word_object = slot_relation_target.get_Word_Object( 1 );
//									
//						regent_ID_string = ";" + Integer.toString( l );
//						
//						for ( String relation : thematic_relations )
//						{
//							key = relation + regent_ID_string;
//							value = score_Relation( dependant_word_object, regent_word_object, relation );
//							dependant_word_object.set_Prediction( key, value );	// This sets the prediction value in the word object's HashMap 'predictions'.
//							prediction_string = prediction_string + "\t" + key + "\t" + value + "          ".substring( ( (Float) value ).toString().length() );
//							// This builds the word to score's prediction string which holds all key-value pairs.
//						}	
//					}
//				}	
//				
//				dependant_word_object.set_Prediction_String( prediction_string );
//			}
//		}
	}
	
	
	private void assign_Predictions_Slot_to_Homonym()
	{
//		Slot dependant_slot_object, slot_relation_target;
//		int target_slot_size;
//		
//		Word dependant_word_object, regent_word_object;
//		String regent_ID_string = "";
//		String key = "";
//		float value;
//		String prediction_string;
//		
//		for ( int i = 1; i <= sentence_length; i++ )		// For all words to score iterate over all slots except for slot no. 0 which contains the ID line.
//		{
//			dependant_slot_object = sentence.get_Slot_Object( i );		
//			dependant_word_object = dependant_slot_object.get_Word_Object( 1 );
//			prediction_string = "";
//						
//			for ( int l = 1; l <= sentence_length; l ++ )		// For the relation target iterate over all slots except for slot no. 0 containing the ID line.
//			{
//				if ( i != l )
//				{
//					slot_relation_target = sentence.get_Slot_Object( l );			
//					target_slot_size = slot_relation_target.size();
//					
//					for ( int m = 1; m <= target_slot_size; m++ )
//					{
//						regent_word_object = slot_relation_target.get_Word_Object( m );
//						regent_ID_string = ";" + Integer.toString( l ) + ";" + regent_word_object.get_Description();
//												
//						for ( String relation : thematic_relations )
//						{
//							key = relation + regent_ID_string;
//							value = score_Relation( dependant_word_object, regent_word_object, relation );
//							dependant_word_object.set_Prediction( key, value );	// This sets the prediction value in the word object's HashMap 'predictions'.
//							prediction_string = prediction_string + "\t" + key + "\t" + value + "          ".substring( ( (Float) value ).toString().length() );
//							// This builds the word to score's prediction string which holds all key-value pairs.
//						}
//					}	
//				}
//			}
//			
//			dependant_word_object.set_Prediction_String( prediction_string );
//		}
	}

	

	private void assign_Predictions_Slot_to_Slot()
	{
//		throw new NoSuchMethodError("nict mehr nutzen");
//		Slot dependant_slot_object, slot_relation_target;
//		
//		Word dependant_word_object, regent_word_object;
//		String regent_ID_string = "";
//		String key = "";
//		float value;
//		String prediction_string;
//		
//	
//		for ( int i = 1; i <= sentence_length; i++ )		// For all words to score iterate over all slots except for slot no. 0 which contains the ID line.
//		{
//			dependant_slot_object = sentence.get_Slot_Object( i );		
//			dependant_word_object = dependant_slot_object.get_Word_Object( 1 );
//			prediction_string = "";
//						
//			for ( int l = 1; l <= sentence_length; l ++ )		// For the relation target iterate over all slots except for slot no. 0 containing the ID line.
//			{
//				if ( i != l )
//				{
//					slot_relation_target = sentence.get_Slot_Object( l );			
//					regent_word_object = slot_relation_target.get_Word_Object( 1 );
//					regent_ID_string = ";" + Integer.toString( l );
//					
//					for ( String role : thematic_roles )
//					{
//						String thematic_relation = "is_" + role + "_for";
//						key = role + regent_ID_string;
//						value = score_Relation( dependant_word_object, regent_word_object, thematic_relation );
//						dependant_word_object.set_Prediction( key, value );	// This sets the prediction value in the word object's HashMap 'predictions'.
//						prediction_string = prediction_string + "\t" + key + "\t" + value + "          ".substring( ( (Float) value ).toString().length() );
//						// This builds the word to score's prediction string which holds all key-value pairs.
//					
//						// System.out.println( "Prediction: " + prediction_string );
//						
//						/*
//						Create a unique hashkey under which to store the prediction in the 'predictions' HashMap.
//						For output mode 'homonym to homonym', the hashkey is in the format 'RELATION;description' 
//						where RELATION is the name of the thematic relation scored and 
//						description is the CDG-internal description string of the regent.
//						
//						For output mode 'slot to slot', this constructs the hashkey in the format 'RELATION;slot_number' where RELATION is the name of the thematic relation
//						scored and the slot_number is the sentence slot number of the word to which the thematic relation is rated.
//						*/
//					}	
//				}
//			}
//			
//			dependant_word_object.set_Prediction_String( prediction_string );
//		}
	}

	
	/**
	 * This method computes the CDG score for the thematic dependency 'relation' 
	 * between two word objects (homonyms) which both map to individuals in the
	 * context model. Note that the fact that they both map onto contextually 
	 * asserted individuals does not necessarily mean that in the context model 
	 * there is also a relation asserted between those individuals. It is for 
	 * those cases that we initially assign a default veto which subsequently 
	 * can be overwritten by concrete evidence from the model.
	 * 
	 * @param <code>dependant</code> is the dependant for which the relation to its 
	 * regent is to be scored.
	 * 
	 * @param <code>regent</code> is the regent to which the scorable relation holds.
	 * 
	 * @param <code>relation</code> is the thematic relation 'is_ROLE_for' that is 
	 * to be scored between word objects dependant and regent.
	 * 
	 * @return the CDG score for the thematic relation between dependant and regent.
	 * 
	 */
	
	private void score_Relation( Word dependant, Word regent )
	{
		String dependant_description = dependant.get_Description();
		int dependant_slot = dependant.get_Slot_Number();
		
		String regent_description = regent.get_Description();
		int regent_slot = regent.get_Slot_Number();

//		System.err.println( "\nAttempting to score dependencies between dependant '" + dependant.get_Word_String() + "/" + dependant.get_Description() + "' in slot " + dependant_slot + " and regent '" + regent.get_Word_String() + "/" + regent.get_Description() + "' in slot " + regent_slot + "." );

		/*
		 * Based on the closed-world assumption we can veto all thematic relations for 
		 * a given dependant-regent pair. Any relation that has been asserted between the 
		 * individuals in the context model will subsequently be detected and the veto on
		 * it will be overwriten by the actual score. If no relation is contextually asserted
		 * the veto score imposed now will remain since the closed-world assumption says:
		 * 
		 * "If a relation is to be considered in cross-modal interaction, it must be represented in the context model." 
		 * <==> "Any relation that is not represented in the context model must not contribute to the final analysis."
		 */
		
		for ( String role : thematic_roles )	
		{					
			// Place a veto for relations which have not been scored previously.
			if ( get_Score( dependant_description, dependant_slot, role, regent_description, regent_slot ).equals( "" ) )	
			{
				set_Score( dependant_description, dependant_slot, role, regent_description, regent_slot, veto_penalty.get( role ) );
				// System.err.println( "\t\tSetting the initial veto penalty of " + veto_penalty.get( role ) + " for " + dependant_description + " in Slot " + dependant_slot + " --" + role + "--> " + regent_description + " in Slot " + regent_slot + "." );
			}
//			else
//			{
//				System.err.println( "\t Found score for dependant '" + dependant_description + "' in Slot " + dependant_slot + "--" + role + "--> " + regent_description + " in Slot " + regent_slot + ": " + get_Score( dependant_description, dependant_slot, role, regent_description, regent_slot ) );
//			}
		}
		
		
		/*
		 *  Not only do the concepts associated with the dependant and regent word objects 
		 *  both need to be in the context model. The CDG score for the thematic relation 
		 *  between them is only affected if the corresponding concept individuals also have 
		 *  a thematic relation asserted between them in the context model. If there is no 
		 *  thematic relation between the individuals in the context model, then the thematic 
		 *  relation in CDG shall remain vetoed as set above.
		 */
		
		Set <OWLIndividual> dependant_context_individual_set = get_Anchors( dependant );		
		Set <OWLIndividual> regent_context_individual_set = get_Anchors( regent );
		
//		System.err.println( "Set of context individuals for regent '" + regent.get_Description() + "': " + regent_context_individual_set );
		
		for ( OWLIndividual dependant_individual : dependant_context_individual_set )
		{			
			for ( OWLIndividual regent_individual : regent_context_individual_set )
			{
				String relation_found = query_manager.get_Context_Relation( dependant_individual, regent_individual );
			
//				System.err.println( "Found context relation '" + relation_found + "' between dependant " + dependant_individual.toString() + " and regent " + regent_individual.toString() + "." );
				
				// Change the pre-set veto penalty only if a thematic relation has been asserted between individuals 
				// in the context model upon which these word objects have been grounded.
				
				if ( relation_found != "" )
				{
//					System.err.println( "Found a context relation for dependant: " + dependant_description + " in Slot " + dependant_slot + ", Regent: " + regent_description + " in Slot " + regent_slot + ": " + dependant_individual.toString() + " --" + relation_found + "--> " + regent_individual ) ;		
											
					Pattern role_pattern = Pattern.compile( "([A-Z]+)" );	// The role name in a relation only consists of capital letters.
					// Note that this regex will match regardless of whether the relation was asserted as 'is_ROLE_for' or as 'has_ROLE'.
					Matcher role_matcher = role_pattern.matcher( relation_found );
					
					/*
					 * Admit the thematic relation corresponding to the asserted context relation in CDG.
					 * 
					 * Note that this admittance of a thematic relation in CDG may even overwrite a previously 
					 * set veto penalty for this dependant-regent pair. If there was a pre-set veto it must have
					 * resulted from the vetoing process of another dependant-regent relation examined prior to 
					 * this relation as part of the current prediction process. Not skipping over previously set 
					 * scores here ensures we admit all thematic relations that we find active evidence for in 
					 * the context model. These admittances are protected and thus retained by skippig over 
					 * previously set scores in the veto process. As a result, vetoing will never overwrite 
					 * an admittance identified elsewhere in the prediction process.
					 */

					if ( role_matcher.find() ) // Check if a thematic relation has been asserted in the context model.					
					{
						String role_found = role_matcher.group( 1 );
						
//						System.err.println( "Admitting relation " + role_found + " between " + dependant.get_Description() + " in Slot " + dependant.get_Slot_Number() + " and " + regent.get_Description() + " in Slot " + regent.get_Slot_Number() + "." );
//						System.err.println( "Inverse score: " + score.get( regent_description + role_found + dependant_description ) );
												
						set_Score( dependant_description, dependant_slot, role_found, regent_description, regent_slot, new Float( 1 ) );
						
						
						/*==================
							***	VETOING ***							
						  ==================*/

						/*
						 *	SYMMETRY-BASED VETOING
						 *  =======================
						 *  
						 * The thematic relations detected are of the form is_ROLE_for and hence are not symmetric.
						 * Consequently, we can place a veto penalty on the relation B is_ROLE_for A if we have 
						 * positive evidence in the context model that A is_ROLE_for B. As with any veto scoring, 
						 * the veto score should only be imposed if the return relation has not been scored yet.
						 * 
						 * Example: Die Ärztin der Patientin singt.
						 * 			Patientin/NN --is_OWNER_for--> Ärztin/NN
						 * 			==> 	Ärztin/NN --//--is_OWNER_for--> Patientin/NN
						 * 
						 * Note that if this veto is placed prematurely it can always be overwritten by subsequent
						 * positive evidence for the return relation!
						 */						
	
						if ( get_Score( regent_description, regent_slot, role_found, dependant_description, dependant_slot ).equals( "" ) )
						{
							// Set the veto penalty for the inverse relation of the ROLE detected.
							set_Score( regent_description, regent_slot, role_found, dependant_description, dependant_slot, veto_penalty.get( role_found ) );
							
							// System.err.println( "\tReturn Relation Veto: Setting the veto score " + veto_penalty.get( role_found ) + " for " + regent_description + " -- " + role_found + "--> " + dependant_description + "." );
						}

						
						/*	ROW- AND COLUMN-VETOING BETWEEN DEPENDANT AND REGENT SLOTS
						 *  ==========================================================
						 *  
						 * CDG only assigns dependencies between slots. If we have scores a dependency between two
						 * homonyms we can hence veto all other thematic relations between the two slots.
						 * 
						 * Example: Die Ärztin der Patientin singt.
						 * 			Patientin/NN --is_OWNER_for--> Ärztin/NN
						 * 
						 * 			==> 	1. Row Veto:	Patientin --//--is_ROLE_for--> Ärztin (i.e.: any other homonym in the regent slot)
						 * 				and	2. Column Veto:	Any other word --//--is_OWNER_for--> Ärztin
						 * 
						 * As with all veoting, only set a veto penalty if no other score has previously been set for 
						 * the given dependant-regent pair. Like this, previous predictions for the given dependant-regent 
						 * pair, including the one scored above, will be preserved. 
						 */
	

						for ( Word word : sentence.get_All_Word_Objects() )	// Iterate over all words in the sentence.
						{										
							String word_description = word.get_Description();
							int word_slot_number = word.get_Slot_Number();

							for ( String role : thematic_roles )	// Iterate over all thematic roles.
							{	
								// Row vetoing: Iterate over all descriptions in the dependant slot.
								for ( Word alternative_dependant : dependant.get_Parent_Slot().get_All_Word_Objects() )
								{											
									if ( get_Score( alternative_dependant.get_Description(), alternative_dependant.get_Slot_Number(), role, word_description, word.get_Slot_Number()  ).equals( "" ) )
									{
										set_Score( alternative_dependant.get_Description(), alternative_dependant.get_Slot_Number(), role, word_description, word_slot_number, veto_penalty.get( role ) );
										// System.err.println( "\tSetting the row veto penalty of " + veto_penalty.get( role ) + " for: " + alternative_dependant.get_Description() + " in Slot " + alternative_dependant.get_Slot_Number() + " --" + role + "--> " + word_description + " in Slot " + word.get_Slot_Number() + "." );
									}									
								}
								
								// Column vetoing: Iterate over all descriptions in the regent slot.
								for ( Word alternative_regent : regent.get_Parent_Slot().get_All_Word_Objects() )
								{
									if ( get_Score( word_description, word_slot_number, role, alternative_regent.get_Description(), alternative_regent.get_Slot_Number() ).equals( "" ) )
									{
										set_Score( word_description, word_slot_number, role, alternative_regent.get_Description(), alternative_regent.get_Slot_Number(), veto_penalty.get( role ) );
										// System.err.println( "\tSetting the column veto penalty of " + veto_penalty.get( role ) + " for: " + word_description + " in Slot " + alternative_regent.get_Slot_Number() + " --" + role + "--> " + alternative_regent.get_Description() + " in Slot " + alternative_regent.get_Slot_Number() + "." );
									}
								}
							}
						}
					}
				}
//				else	// i.e.: in case there was no context relation asserted between the individuals that the dependant and regent words map to.
//				{
//					System.err.println( "\tFound no context relations between dependant '" + dependant_description + "' --" + relation_found + "--> " + regent_description + "." );
//				}
			}
		}
	}


	private void set_Score( String dependant, int dependant_slot, String relation, String regent, int regent_slot, Float number )
	{
		String value = number.toString();
		score.put( dependant + dependant_slot + relation + regent + regent_slot, value );
		
		// System.err.println( "Setting the score for " + dependant + " in Slot " + dependant_slot + "-- " + relation + "--> " + regent + " in Slot " + regent_slot + " " + " to " + value + "." );
	}


	public String get_Score( String dependant, int dependant_slot, String relation, String regent, int regent_slot )
	{			
		String result = score.get( dependant + dependant_slot + relation + regent + regent_slot );
		
		if (  result == null )
		{
			return "";
		}

		return result;
	}
	
	
	/** 
	 * 	This method returns the set of all OWLIndividuals asserted in the context
	 * 	model which instantiate a concept that is compatible with one of the concepts
	 * 	activated by the input Word object.
	 */

	public Set<OWLIndividual> get_Anchors( Word word )
	{			
		return (Set <OWLIndividual>) crossmodal_anchoring.get( word );
	}

	
	private void set_Sentence( Sentence sen )
	{
		sentence = sen;	
	}

	
	public Sentence get_Sentence()
	{
		return sentence;
	}		
	
	
	/** 
	 * 	This method defines the set of thematic roles based on the content
	 * 	of the environment variable SUPPORTED_THEMATIC_ROLES as set in the
	 * 	script ~/PPC/configure.
	 */
	
	private void set_Thematic_Roles()
	{
		thematic_roles = new HashSet <String> ();
		
		// System.out.println( "Number of roles supported: " + System.getenv( "SUPPORTED_THEMATIC_ROLES" ).split( " " ).length );
		
		for ( String element : System.getenv( "SUPPORTED_THEMATIC_ROLES" ).split( " " ) )
		{
			thematic_roles.add( element );	
			// System.out.println( "Added role: " + element + " to thematic roles.");
		}          
	}


	public Set <String> get_Thematic_Roles()
	{
		return thematic_roles;
	}	

	
	/** 
	 * 	This method defines the set of thematic relations based on the set
	 * 	of thematic roles defined. Thematic relations take the format 
	 * 	'is_ROLE_for'.
	 */
	
	private void set_Thematic_Relations()
	{
		thematic_relations = new HashSet <String> ();
		
		for ( String element : thematic_roles )
		{
			String relation = "is_" + element + "_for";
			thematic_relations.add( relation );	
			// System.out.println( "Relation: " + relation );
		}
	}


	public Set <String> get_Thematic_Relations()
	{
		return thematic_relations;
	}		
	
	
	private void set_Query_Manager( OWL_Query_Manager manager )
	{
		query_manager = manager;
	}
	
	
	/**
	 * This method iterates over all word objects (homonyms) in the input sentence
	 * and tries to map the words upon concepts from the conceptual hierarchy (i.e.:
	 * the T-box) to represent each word's meaning.
	 * 
	 * A word is mapped to the set of all concepts that have
	 * 		- the word's normalisation string as one of their lexicalisations,
	 * 		- the word's sem_val as their situation arity 
	 * and
	 * 		- a grammatical number that is compatible with the word's number. 
	 * 
	 * The set of concepts with matching lexicalisation, situation arity and compatible 
	 * grammatical number is added as the word's set of conceptualisations which is 
	 * stored under the key 'conceptualisations' in the word's hashmap 'word_features'.
	 * 
	 * This method also initialises the value in the hashmap 'compatible_context_concepts' 
	 * which maps every linguistically activated concept onto the set of context concepts 
	 * with which it is compatible. The actual set of compatible context concepts for each 
	 * linguistically activated concept will be populated at a later stage.
	 */
	 
	private void ground_Words()
	{
		
		compatible_context_concepts = new HashMap <OWLClass, Set<OWLClass>> ();
				
		for ( Word word : sentence.get_All_Word_Objects() )	// Iterate over all word objects (homonyms) in the sentence.
		{											
			/*
			 * Iterate over the set of concepts in the conceptual hierarchy (i.e.: the T-box) 
			 * that have the current word's normalisation as one of their lexicalisations. Recall 
			 * that the map from normalisation to concept may be 1:n, which allows to model polysemy.
			 */

//			System.err.println( "\tTrying to ground the word '" + word.get_Description()  + "' with sem_val '" + word.get_Sem_Val() + "', normalisation '" + word.get_Normalisation() + "' and number '" + word.get_Grammatical_Number() + "'." );

			for ( OWLClass concept : query_manager.get_Conceptualisations_from_Ontology( word.get_Normalisation() ) )
			{					
				/*
				 * Now check if the word's sem_val agrees with the concept's situation arity.
				 * 
				 * This is necessary to prevent the assignment of several concepts of different 
				 * situation arities to a homonym denoting a situation with a sepecific situation 
				 * arity (the latter appeared to be an attractive option for handling lexical 
				 * ambiguity initially). Each homonym shall only be related to the concepts which 
				 * match its semantic valence. 
				 * 
				 * Note that the handling of lexical ambiguity is still achieved since each slot 
				 * string may map onto several homonyms with a different semantic valence each.
				 * 
				 * The imprecise mapping of homonym to concept would subsequently lead to this 
				 * homonym passing the contextual consistency check with situation concepts of 
				 * situation arity different from its semantic valence. The homonym would thus be 
				 * considered further as a syntactic alternative even though its situation arity 
				 * should already have been dismissed at an earlier stage of the process. This is
				 * what can happen if the situation arity of the instance in the context model is 
				 * not specified precisely.
				 */
				
//				System.err.println( "Word '" + word.get_Word_String() + "' has sem_val '" + word.get_Sem_Val() + "'." );
//				System.err.println( "Checking for concept '" + concept + "'." );

				/*
				 *	Now check if the word has a sem_val that equals the situation arity of the current concept.
				 * 	This equality can be achieved in two ways: 
				 * 	1) The word has a sem_val defined and it is equal to concept's situation arity. 
				 * 	2) Neither of them has a value defined such that they both are NULL. 
				 * 	While situation concepts ground via a match in situation arity and sem_val, concepts that are 
				 * 	not verb-based match via a situation arity and sem_val value of NULL.
				 */		
				
				if ( word.get_Sem_Val().equals( query_manager.get_Situation_Arity( concept ) ) )
				{
					/*	
					 * If the word has a grammatical number defined (and most conceptually relevant words 
					 * do), the number of the concept it is to be grounded in must be compatible with that 
					 * number. This allows to distinguish between singular and plural concepts (e.g. 'X' vs.
					 * 'X & Plural').
					 * If the word has no grammatical number defined the number criterion is disregarded for 
					 * grounding. Recognised values for grammatical number are 'sg', 'pl' and 'bot'.
					 */

					String num ="";
					
					if ( word.has_Grammatical_Number() )
					{		
						num = word.get_Grammatical_Number();
						
						if ( num.equals( "sg" ) )
						{
							num = "Singular";
//							System.err.println( "Encountered grammatical number '" + num + "' for word '" + word.get_Word_String() + "." );
						}
						else if ( num.equals( "pl" ) )
						{
							num = "Plural";
//							System.err.println( "Encountered grammatical number '" + num + "' for word '" + word.get_Word_String() + "." );
						}
						else if ( num.equals( "bot" ) )
						{
							num = "Grammatical.Number";
//							System.err.println( "Encountered grammatical number '" + num + "' for word '" + word.get_Word_String() + "." );
						}
						else	
						{
							System.err.println( "Encountered unknown grammatical number '" + num + "' for word '" + word.get_Word_String() + "." );
						}
					
						OWLClass word_grammatical_number = query_manager.get_OWL_Ontology_Manager().getOWLDataFactory().getOWLClass( URI.create( query_manager.get_T_BOX_URI() + num ) );							
	
						
						if ( query_manager.is_Compatible( word_grammatical_number, query_manager.get_Grammatical_Number( concept ) ) )
						{
							word.add_Conceptualisation( concept );
											
//							System.err.println( "\tThe word '" + word.get_Description() + "' grounds the set of concepts '" + word.get_Conceptualisations() + "' with compatible grammatical number." );	
						
							/* 	Initialise the set of context concepts compatible with each linguistically activated concept
								as value in the hashmap 'compatible_context_features'. This hashmap is a record of which 
								linguistically activated concept is compatible with which context concept. The set of compatible 
								context concepts for each linguistically activated concept will be populated at a later stage.
							*/
							
							compatible_context_concepts.put( concept, new HashSet <OWLClass> () );	
							
							// System.out.println( "\tInitialised the compatible_context_concepts hash for concept '" + concept + "' with size " + compatible_context_concepts.get( concept ).size() + "." );
						}
						else
						{
//							System.err.println( "Word '" + "' and concept '" + concept + "' have incompatible grammatical number." );
						}
					}
					else	// If it does not have a grammatical number, ground without considering grammatical number.
					{
						word.add_Conceptualisation( concept );
						
//						System.err.println( "\tThe word '" + word.get_Description() + "' grounds set of concepts '" + word.get_Conceptualisations() + "' without number." );	
					
						/* 	Initialise the set of context concepts compatible with each linguistically activated concept
							as value in the hashmap 'compatible_context_features'. This hashmap is a record of which 
							linguistically activated concept is compatible with which context concept. The set of compatible 
							context concepts for each linguistically activated concept will be populated at a later stage.
						*/
						
						compatible_context_concepts.put( concept, new HashSet <OWLClass> () );	
						
						// System.out.println( "\tInitialised the compatible_context_concepts hash for concept '" + concept + "' with size " + compatible_context_concepts.get( concept ).size() + "." );
					}
				}
//				else	// Do not ground the concept in this word if sem_val and situation arity do not match.
//				{
//					System.err.println( "\t\tThe word '" + word.get_Description() + "' did not ground the concept '" + concept + "'." );
//				}					
			}
		}
		
//		System.err.println( "Completed the grounding of concepts in words." );
	}
	

	/**
	 * This method tells whether a given class has an instance
	 * in the visual context model or not.
	 * 
	 * @param OWLClass
	 * is the class for which to check whether it has an instance in the context model.
	 * 
	 * @return boolean
	 * the truth value of whether or not the given concept has an instance
	 * in the context model.
	 * 
	 */
	
	public boolean is_in_Context( OWLClass owl_class )
	{
		return query_manager.get_Context_Concepts().contains( owl_class );
	}
	
	
//	private void set_Score( String descriptor, float input )
//	{
//		String value = ( (Float) input ).toString(); 
//		value.format("%4d", value );
//		
//		score.put( descriptor, value );
//	}
}
