import java.io.PrintStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.inference.OWLClassReasoner;
import org.semanticweb.owl.inference.OWLReasonerException;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.util.ToldClassHierarchyReasoner;
import org.semanticweb.owl.vocab.OWLRDFVocabulary;

/**
 * Print a tab-indented plain-text hierarchy of OWL classes.
 * 
 * @author Roerd Hinrichsen <roerd.hinrichsen@public.uni-hamburg.de>
 * 
 */
public class ConceptHierarchyPrinter {

	private OWLClassReasoner _reasoner;

	private PrintStream _out;

	private Set<OWLClass> _seenClasses;

	private static boolean _shorten = true;

	public ConceptHierarchyPrinter(OWLClassReasoner reasoner, PrintStream out) {
		_reasoner = reasoner;
		_out = out;
		_seenClasses = new HashSet<OWLClass>();
	}

	private static OWLOntology loadOntologyFromFile(OWLOntologyManager manager,
			String filename) {
		URI uri = URI.create("file:" + filename);
		OWLOntology ontology = null;
		try {
			ontology = manager.loadOntologyFromPhysicalURI(uri);
		} catch (OWLOntologyCreationException e) {
			System.err.println("Could not load file: " + filename);
			e.printStackTrace();
			System.exit(1);
		}
		return ontology;
	}

	/**
	 * Print an indention of <code>level</code> tabs.
	 * 
	 * @param level
	 *            The indention level
	 */
	private void indent(int level) {
		for (int i = 0; i < level; ++i) {
			_out.print('\t');
		}
	}

	/**
	 * Print the name of <code>theClass</code> and its subclasses at indention
	 * level <code>level</code>.
	 * 
	 * @param theClass
	 *            The OWL class to be printed with its subclasses.
	 * @param level
	 *            The indention level representing the hierarchy level of the
	 *            class.
	 */
	private void printClass(OWLClass theClass, int level) {
		indent(level);
		_out.print(theClass);

		if (_shorten && _seenClasses.contains(theClass)) {
			_out.println(" [Subclasses: see above.]");
		} else {
			_out.println();

			Set<Set<OWLClass>> subClasses = null;
			try {
				subClasses = _reasoner.getSubClasses(theClass);
			} catch (OWLReasonerException e) {
				e.printStackTrace();
				System.exit(1);
			}

			List<OWLClass> subClassList = new ArrayList<OWLClass>();
			for (Set<OWLClass> subClassSubSet : subClasses) {
				subClassList.addAll(subClassSubSet);
			}
			Collections.sort(subClassList);

                        if (!subClassList.isEmpty()) {
                            _seenClasses.add(theClass);
                        }

			for (OWLClass subClass : subClassList) {
				printClass(subClass, level + 1);
			}
		}
	}

	/**
	 * Print a tab-intented plain-text representation of the asserted concept
	 * hierarchy in the OWL file.
	 * 
	 * @param args
	 *            the filenames of the OWL file to process, or these switches:
	 *            -l for full output, -s for shortened output (default).
	 */
	public static void main(String[] args) {
		OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
		OWLClassReasoner reasoner = new ToldClassHierarchyReasoner(manager);

		HashSet<OWLOntology> ontologies = new HashSet<OWLOntology>();
		for (String arg : args) {
			if (arg.equals("-l")) {
				_shorten = false;
			} else if (arg.equals("-s")) {
				_shorten = true;
			} else {
				ontologies.add(loadOntologyFromFile(manager, arg));
			}
		}
		try {
			reasoner.loadOntologies(ontologies);
			reasoner.classify();
		} catch (OWLReasonerException e) {
			e.printStackTrace();
			System.exit(1);
		}
		ConceptHierarchyPrinter printer = new ConceptHierarchyPrinter(reasoner,
				System.out);

		URI thingURI = OWLRDFVocabulary.OWL_THING.getURI();
		OWLClass thing = manager.getOWLDataFactory().getOWLClass(thingURI);

		printer.printClass(thing, 0);
	}

}
