import java.io.PrintStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.util.SimpleRenderer;
import org.semanticweb.owl.vocab.OWLRDFVocabulary;

/**
 * Print a tab-indented plain-text hierarchy of OWL classes.
 * 
 * @author Roerd Hinrichsen <roerd.hinrichsen@public.uni-hamburg.de>
 * 
 */
public class SimpleConceptHierarchyPrinter {

    private OWLOntology _ontology;

    private PrintStream _out;
    
    private SimpleRenderer _renderer = new SimpleRenderer(); 

    public SimpleConceptHierarchyPrinter(OWLOntology ontology, PrintStream out) {
	_ontology = ontology;
	_out = out;
    }

    private static OWLOntology loadOntologyFromFile(OWLOntologyManager manager,
	    String filename) {
	URI uri = URI.create("file:" + filename);
	OWLOntology ontology = null;
	try {
	    ontology = manager.loadOntologyFromPhysicalURI(uri);
	} catch (OWLOntologyCreationException e) {
	    System.err.println("Could not load file: " + filename);
	    e.printStackTrace();
	    System.exit(1);
	}
	return ontology;
    }

    /**
         * Print an indention of <code>level</code> tabs.
         * 
         * @param level
         *                The indention level
         */
    private void indent(int level) {
	for (int i = 0; i < level; ++i) {
	    _out.print('\t');
	}
    }

    /**
         * Print the name of <code>theClass</code> and its subclasses at
         * indention level <code>level</code>.
         * 
         * @param theClass
         *                The OWL class to be printed with its subclasses.
         * @param level
         *                The indention level representing the hierarchy level
         *                of the class.
         */
    private void printClass(OWLClass theClass, int level) {
	indent(level);
	_out.println(theClass);

	Set<OWLDescription> subClasses = theClass.getSubClasses(_ontology);
	Set<OWLClass> namedSubClasses = new HashSet<OWLClass>();
	for (OWLDescription subClass : subClasses) {
	    if (subClass.isAnonymous()) {
		indent(level + 1);
		_out.println(_renderer.render(subClass));
	    } else {
		namedSubClasses.add(subClass.asOWLClass());
	    }
	}

	List<OWLClass> subClassList = new ArrayList<OWLClass>(namedSubClasses);
	Collections.sort(subClassList);

	for (OWLClass subClass : subClassList) {
	    printClass(subClass, level + 1);
	}
    }

    /**
         * Print a tab-intented plain-text representation of the asserted
         * concept hierarchy in the OWL file.
         * 
         * @param args:
         *                first element should be the filename of the OWL file
         *                to process.
         */
    public static void main(String[] args) {
	OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
	String filename = args[0];

	SimpleConceptHierarchyPrinter printer = new SimpleConceptHierarchyPrinter(
		loadOntologyFromFile(manager, filename), System.out);

	URI thingURI = OWLRDFVocabulary.OWL_THING.getURI();
	OWLClass thing = manager.getOWLDataFactory().getOWLClass(thingURI);

	printer.printClass(thing, 0);
    }

}
