# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

package CDG::German;
use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK);
use BerkeleyDB;
use CDG::Utils qw(parse_string);

require Exporter;
@ISA = qw(Exporter);

# Export this by default
@EXPORT = qw( base_adj   is_adj
              base_noun  is_noun
              base_prep  is_prep
              base_verb  is_verb
              all_cats
              all_info
              normalize_word supernormalize
              deutsch_lc deutsch_uc);

# Export this upon request
@EXPORT_OK = qw();

# Figure out where we get the data from. Ideally CDGDIR is set, which
# answers all questions.
if(!exists $ENV{CDGDIR}) {
  $ENV{CDGDIR} = '/opt/pkg/cdg';
  if(! -d $ENV{CDGDIR}) {
    die "Please set CDGDIR."
  }
}

# The actual data are under lib/ or in share/cdg/grammar/deutsch/.
our $path = "$ENV{CDGDIR}/lib";
if(! -r "$path/Lexikon.db") {
  $path = "$ENV{CDGDIR}/grammar/deutsch";
}
if(! -r "$path/Lexikon.db") {
  $path = "$ENV{CDGDIR}/share/cdg/grammar/deutsch";
}
if(! -r "$path/Lexikon.db") {
  $path = "/opt/pkg/cdg/lib";
}

# verify all sources are present
if(! -r "$path/Lexikon.db") {
  die "Can't read $path/Lexikon.db!";
}
if(! -r "$path/deutsch-lexikon.db") {
  die "Can't read $path/deutsch-lexikon.db!";
}
if(! -r "$path/Lexikon.cdg") {
  die "Can't read $path/Lexikon.db!";
}
if(! -r "$path/deutsch-lexikon.cdg") {
  die "Can't read $path/deutsch-lexikon.db!";
}

# verify all indices are up to date
if( -M "$path/Lexikon.db" > -M "$path/Lexikon.cdg") {
  die "`$path/Lexikon.db' is outdated, please recreate!\n";
}
if( -M "$path/deutsch-lexikon.db" > -M "$path/deutsch-lexikon.cdg") {
  die "`$path/deutsch-lexikon.db' is outdated, please recreate!\n";
}


# The CDG lexicon of German is in two parts; `deutsch-lexikon.cdg'
# contains (mainly) the open word classes, and `Lexikon.cdg' contains
# the closed word classes.
our @sources =
(
 { db => "$path/Lexikon.db" },
 { db => "$path/deutsch-lexikon.db" }
);
for(@sources) {
  $_->{handle} = new BerkeleyDB::Btree {
    -Filename => $_->{db},
    -Flags => DB_RDONLY };
  ($_->{cdg} = $_->{db}) =~ s/db$/cdg/;
}

sub debug { print STDERR @_ if 1 };

sub deutsch_uc {
  my $x = shift;
  $x =~ s/^(.)//;
  my $c = $1;
  $c = uc $c;
  $c =~ tr///;
  return $c . $x;
}

sub deutsch_lc {
  my $x = shift;
  $x = lc $x;
  $x =~ tr///;
  return $x;
}

# Read the CDG lexicon and return the full definitions of all items
# with this key.
sub read_lexicon {
  my $key = shift;
  my @result;

  for my $s(@sources) {
    my $val;
    my $db = $$s{handle};

    my $cursor = $db->db_cursor() ;
    if(0 == $cursor->c_get($key, $val, DB_SET)) {
      do {
        my ($from,$to) = unpack('II', $val);
        my $cdgname = $$s{cdg};
        open(LEX, '<', $cdgname) or die "Can't read $cdgname: $!\n";
        my $x;
        seek(LEX, $from-1, 0);
        read(LEX, $x, $to-$from);
        close(LEX);
        push @result, $x;
      }  while 0 == $cursor->c_get($key, $val, DB_NEXT_DUP);
    }
  }
  return @result;
}

# Since actually seeking into the lexicon on disk is expensive, keep
# caches of everything that the utility functions have already computed.
our %cache;
our %ncache;
our %acache;
our %vcache;
our %pcache;

# Read all information about a word (like read_lexicon, but caches
# everything).
sub all_info {
  my $word = shift;
  if(exists $cache{$word}) {
    return $cache{$word};
  }
  my @result = read_lexicon $word;
  $cache{$word} = \@result;
  return \@result;
}



# Ask the database whether an entry for WORD with the category NN is
# present. If there is an entry with a baseform, the baseform is
# returned. If there is an entry with the category only, the word
# itself is returned. If there is no fitting entry, undef is returned.
sub query_noun {
  my $word = shift;

  if(exists $ncache{$word}) {
    return $ncache{$word};
  }

  my @items = grep {/\bcat: *NN\b/} read_lexicon($word);

  for(@items) {
    if(/base:([^,\]]+)/) {
      my $result = parse_string $1;
      $ncache{$word} = $result;
      return $result;
    }
  }
  $ncache{$word} = undef;
  return undef;

}


# Map `Fernverkehrsverbindungen' --> `Verbindung'
sub analyze_noun {
  my $word = shift;

  my $q = query_noun($word);
  if(defined $q) {
    return ($q,1);
  }

  # try to undo marked compounding
  $word =~ s/^.*-//;
  $q = query_noun($word);
  if(defined $q) {
    return ($q,1);
  }

  # try to undo unmarked compounding
  my $x = $word;
  $x =~ s/^...//;
  while (length($x) > 4) {
    $x =~ s/^.//;
    $x = deutsch_uc $x;
    $q = query_noun($x);
    if(defined $q) {
      return ($q,1);
    }
  }
  return ($word, 0);
}

sub base_noun {
  my $w = shift;
  my ($x,$y) = analyze_noun $w;
  return $x;
}

sub is_noun {
  my $w = shift;
  my ($x,$y) = analyze_noun $w;
  return $y;
}

# return tuple of P's base preposition form, and
# whether or not it is a sort of preposition. Examples:
#
#  analyze_prep('von') ==> ('von', 1)
#  analyze_prep('vom') ==> ('von', 1)
#  analyze_prep('Volk') ==> ('Volk', 0)
sub analyze_prep {
  my $p = deutsch_lc shift;
  if(exists $pcache{$p}) {
    if(!defined $pcache{$p}) {
      return ($p,0);
    } else {
      return ($pcache{$p},1);
    }
  }

  for(read_lexicon($p)) {
    if(/\bcat: *(APPO|APPR|APPRART|APZR|KOKOM|PWAV|PROAV)\b/ &&
       /\b(base|prep): *([a-zA-Z]+)/) {
      my $result = $2;
      $pcache{$p} = $result;
      return ($result,1);
    }
    if(/\bcat: *(APPO|APPR|APPRART|APZR|KOKOM|PROAV)\b/) {
      $pcache{$p} = $p;
      return ($p,1);
    }
  }
  $pcache{$p} = undef;
  return ($p,0);
}

sub base_prep {
  my $w = shift;
  my ($x,$y) = analyze_prep $w;
  return $x;
}

sub is_prep ($) {
  my $w = shift;
  my ($x,$y) = analyze_prep $w;
  return $y;
}


# Find the baseform of a verb.
#
# For ambiguous verb forms, this returns the full form itself. This
# has the following reason: when the form `zieht' occurs in free text,
# it cannot usually be said whether it is a form of `ziehen' or
# `zeihen' because no morphology is available. But since the first
# possibility is much more likely, it is better not to normalize and
# live with slightly sparser data, than to normalize wrongly, and
# perhaps consistently make a serious mistake.
sub analyze_verb {
  my $v = deutsch_lc shift;

  if(exists $vcache{$v}) {
    if(!defined $vcache{$v}) {
      return ($v,0);
    } else {
      return ($vcache{$v},1);
    }
  }

  my @items = read_lexicon($v);

  my (%bases, $b);
  for(@items) {
    if(/\bcat: *V/) {
      if(/\bbase: *([^,\] ]+)/) {
        $b = parse_string $1;
        $bases{$b}++;
      }
    }
  }
  if(!%bases) {
    $vcache{$v} = undef;
    return ($v,0);
  }

  # If there is only one base form, return this
  if(1 == scalar keys %bases) {
    $vcache{$v} = $b;
    return ($b,1);
  }

  # Otherwise, return the form itself
  $vcache{$v} = $v;
  return ($v,1);

}

sub base_verb {
  my $w = shift;
  my ($x,$y) = analyze_verb $w;
  return $x;
}

sub is_verb {
  my $w = shift;
  my ($x,$y) = analyze_verb $w;
  return $y;
}


sub analyze_adj {
  my $a = deutsch_lc shift;

  if(exists $acache{$a}) {
    if(!defined $acache{$a}) {
      return ($a,0);
    } else {
      return ($acache{$a},1);
    }
  }

  for(read_lexicon($a)) {
    if(/\bcat: *ADJ[AD]/) {
      if(/\bbase: *([^,\] ]+)/) {
        my $b = parse_string $1;
        $acache{$a} = $b;
        return ($b,1);
      } else {
        $acache{$a} = $a;
        return ($a,1);
      }
    }
  }
  $acache{$a} = undef;
  return ($a,0);
}

sub base_adj {
  my $w = shift;
  my ($x,$y) = analyze_adj $w;
  return $x;
}

sub is_adj {
  my $w = shift;
  my ($x,$y) = analyze_adj $w;
  return $y;
}

our %all_cache;
sub all_cats {
  my $w = shift;

  if(exists $all_cache{$w}) {
    return $all_cache{$w};
  }

  my %c;
  for (read_lexicon $w) {
    if(/\bcat: *([^,\] ]+)/) {
      my $cat = parse_string($1);
      $c{$cat}++;
    }
  }
  my $result = join('|', keys %c);
  $all_cache{$w} = $result;
  return $result;
}

# Compute the base form and the category of a German word.
sub normalize_word {

  my $w = shift;
  my $t = shift;

  if(defined $t) {
    if ('NN' eq $t) {
      $w = base_noun $w;
    } elsif ($t =~ /^V/) {
      $w = base_verb $w;
    } elsif ($t =~ /^(APPO|APPR|APPRART|PROAV|KOKOM)/) {
      $w = base_prep $w;
    } elsif($t =~ /^ADJ[AD]$/) {
      $w = base_adj $w;
    }
  } elsif($w =~ /^[A-Z]/) {
    return analyze_noun $w;
  } elsif(is_prep $w) {
    return analyze_prep $w;
  } elsif(is_verb $w) {
    return analyze_verb $w;
  } elsif(is_adj $w) {
    return analyze_adj $w;
  }
  return ($w,$t);
}

# Normalize even more aggressively; e.g., map 123 ==> `CARD'
# and `Montag' ==> `WOCHENTAG'.
sub supernormalize {

  my $w = shift;
  my $t = shift;

  if ($t =~ /^(TRUNC|NE|FM|CARD|PDS|PPER|PFR|PRELS|PPOSS|PWAV|PWS|PIS)$/) {
    $w = $t;
  }
  if ($w =~ /^(Montag|Dienstag|Mittwoch|Donnerstag|Freitag|Sonnabend|Samstag|Sonntag)$/) {
    $w = 'WOCHENTAG';
  }
  if ($w =~ /^(Januar|Februar|Mrz|April|Mai|Juni|Juli|August|September|Oktober|November|Dezember)$/) {
    $w = 'MONAT';
  }

  # normalize_word() doesn't do KOKOM, normalize the case by hand
  if ($t eq 'KOKOM') {
    $w = lc $w;
  }

  return $w;

}


1;
