#! /usr/bin/perl -w

#   This script reads the semantic valence definitions from the first invocation argument file 
#	(typically ~/cdg2/grammar/deutsch/Semantic.Valences) and writes the appropriate verbal 
#	information into the file ~/cdg2/grammar/deutsch/Verben.txt from which cdg generates 
#	deutsch.cdg and the lexicon files Lexikon.cdg and Lexikon.db.
#
#   Invocation Arguments:
#	=====================
#
#    1. File containing semantic valence definitions for insertion into the lexicon.
#
#
#	Assumptions:
#	============
#
#	This implementation rests on the following assumptions:
#	
#	(1) After running this script Verben.txt will only contain those sem_val definitions for inclusion into 
# 		the lexicon which have been listed in Semantic.Valences. Conversely, sem_val definitions not listed 
#		in Semantic.Valences will not be added to Verben.txt - or will be removed from it if they were included 
#		in Verben.txt before running this script.
#
#	(2)	Semantic.Valences can only define sem_vals for verb forms already listed in Verben.txt.
#	NOT SURE IF THIS STILL HOLDS

use strict;
use locale;
use File::Copy;

( my $verb_file = `echo \$HOME` ) =~ s/\n$/\/cdg2\/grammar\/deutsch\/Verben.txt/;

my $new_verb_file = $verb_file;
my $old_verb_file = $verb_file . "_old";

rename $verb_file, $old_verb_file;

my %valence_definition;

my $existing_semantic_valence;
my $lemma;
my $syntactic_valence;

my $reference;
my %print_definition;
my %avz_valence;

###
###		Read in the semantic valence definitions with their syntactic valence matches from the 
###	 	 file (typically ~/cdg2/grammar/deutsch/Semantic.Valences).
###

while( <> )
{ 
	next if /^\s*\/\//o;	# Skip comment lines starting with double slashes '//'.

	if ( /^(\S+)\s+([\w_]+)\s+([-?+\w]+)\n/o )	# Matches a regular sem_val definition line.
	{
		my $lemma = $1;		
		
		my $semantic_valence = $2;

		my $syntactic_valence = $3;
		
		my $key = $syntactic_valence . "#_#" . $semantic_valence;
		
		$valence_definition{ $lemma }{ $key } ++;
		# Note that this three-dimensional hash is required since in Semantic.Valences one 
		# lemma may have multiple entries corresponding to different semantic valences.
		
		$print_definition{ $lemma } = 1;	# This variable defines whether to print the valence definition for this lemma.
		
		# print "Semantic valence definition in Semantic.Valences $lemma \n\t Syntactic Valence: $syntactic_valence \n\t Semantic Valence: $semantic_valence \n";
	}
	elsif ( /^(\S+)\s+avz_valence:([-?+'\w]+)/o )	# Matches a avz_valence definition line.
	{
		# Also extract the lemma's avz_valence which is its old syntactic valence 
		# needed as a return value for the datamap AVZ which in CDG's standard 
		# grammar for German returns a syntactic valence.
		
		my $lemma = $1;		
		
		my $old_valence = $2;
		
		$avz_valence{ $lemma } = $old_valence;
		# print "AVZ Valence: $lemma --> $avz_valence{ $lemma } \n";
	}
}


### 
###	The new version of Veben.txt is generated by running through the standard version of Verben.txt and 
###
###	a. 		printing its line if its line is not a verbal line.
###	or b.	adding a sem_val definition for a lemma already found to have a sem_val definition.
###	or c.	overwriting an existing sem_val definition no longer endorsed by Semantic.Valences.
###	or d.	removing a sem_val definition no longer endorsed by Semantic.Valences.
###	

open( INPUT,  "<", $old_verb_file ) or die "Cannot open file $old_verb_file: $!\n";
open( OUTPUT,  ">", $new_verb_file ) or die "Cannot open file $new_verb_file: $!\n";

while ( my $line = <INPUT> )
{ 				
	if ( $line =~ /verb\s+(\S+)\s+([\w+?'-]+)( |$)/o )	# Check if line contains a verbal entry.
	{
		$lemma = $1;
		
		$syntactic_valence = $2;
		
		if ( $valence_definition{ $lemma } && $print_definition{ $lemma } ) 	# Check if there is a semantic valence definition in Semantic.Valences for that lemma.
		{			
			# Print all sem_val definitions for this lemma.
			
			for my $key ( sort keys %{ $valence_definition{ $lemma } } )	# Loop over all syn-sem valence pairs for this lemma.
			{
				my ( $new_syntactic_valence, $new_semantic_valence ) = split /#_#/, $key;	# Break up the fused hash keys made up of syn and sem valence.
				
				$line =~ /^(.*verb \S+)\s+[-+?'\w]+(( sem_val:[\w_]+|(.*)))?/;
				# The original line from Verben.txt may or may not already contain a sem_val definition. 
				# It may also contain further specifications such as a definition of objp etc.
				
				my $line_head = $1; 
				my $line_tail = $4;
				
				# Remove the objp definition from $line_tail if the new syntactic valence does not contain 'p'.
				$line_tail =~ s/ objp:[<>,\w\d]+// if ( $line_tail && $new_syntactic_valence !~ /p/ );
								
				printf OUTPUT "%-35s %s sem_val:%s%s avz_valence:%s\n", $line_head, $new_syntactic_valence, $new_semantic_valence, $line_tail, $avz_valence{ $lemma };
			}	
			
			$print_definition{ $lemma } = 0; 	# Setting this variable to '0' ensures that the valence definitions for this lemma does not get printed again.
		}
		elsif ( ! $valence_definition{ $lemma } )
		{
			# Remove any previous semantic valence definitions which may have been 
			# present in case there is no sem_val definition for this lemma.
			( my $cleaned_line = $line ) =~ s/ sem_val:[\w_]+//o;
			print OUTPUT $cleaned_line;
		}
		
		# Note that this way multiple lines of the same lemma in Verben.txt do not get printed as for them there is a
		# valence definition but the value for $print_definition has been set to FALSE after the first printed line.
		# This is good because the above loop prints all sem_val definitions for the lemma as declared in Semantic.Valences.
	}
	else	# i.e.: If the current line does not contain a verbal entry, then just print it.
	{
		print OUTPUT $line;
	}
}