#! /usr/bin/perl -w
	
###	This script tests the output of the CDG2-modfied version of make-verbs.pl 
###	for compliance with expectation.
###	The actual tests and their outcome are listed in the output of this script.

use strict;

my $directory = "/home/mccrae/cdg2/grammar/deutsch";

###	Read Verben.txt.

open INPUT, "<", "$directory/Verben.txt";

my %input_raw_lemma;
my %input_clean_lemma;
my %input_sem_vals;
my %input_syn_vals;
my %input_avz_vals;

while ( <INPUT> )
{ 
	next if /^$/;		# Ignore blank lines.
	next if /^\s*(\/\/|#)/;	# Ignore comment lines.
	
	next unless /verb/;	# Ignore lines that do not contain verbal declarations.
	
	# print "\n$_";
	
	if ( /verb\s+(\S+)\s+(\S+) sem_val:(\S+) .*avz_valence:(\S+)/ )	# Match verbal declaration lines with a sem_val definition.
	{	
		my $raw = &remove_special_characters( $1 );
		my $syn_val = &normalise_valence( $2 );
		my $sem_val = $3;
		my $avz_val = &normalise_valence( $4 );
		
		$input_raw_lemma{ $raw } ++;
		# print "Raw input lemma: >$raw< \n";
		
		my $clean = &clean_lemma ( $raw );
		
		$input_clean_lemma{ $clean } ++;
		# print "Clean input lemma for $raw: >$clean< \n";
		
		$input_syn_vals{ $clean }{ $syn_val } ++;
		# printf "All input syn_vals for $clean: >%s< \n", join ",", keys %{ $input_syn_vals{ $clean } };
			
		$input_sem_vals{ $clean }{ $sem_val } ++;
		# printf "All input sem_vals for $clean: >%s< \n", join ",", keys %{ $input_sem_vals{ $clean } };
	
		$input_avz_vals{ $clean }{ $avz_val } ++;
		# printf "All input avz_vals for $clean: >%s< \n", join ",", keys %{ $input_avz_vals{ $clean } };	
	}
		
	elsif ( /verb\s+(\S+)\s+(\S+)/ )	# Match verbal declaration lines without a sem_val and avz_valence definition.
	{
		my $raw =  &remove_special_characters( $1 );
		
		my $syn_val = &normalise_valence( $2 );
		
		$input_raw_lemma{ $raw } ++;
		# print "Raw input lemma: >$raw< \n";
		
		my $clean = &clean_lemma ( $raw );
		
		$input_clean_lemma{ $clean } ++;
		# print "Clean input lemma for $raw: >$clean< \n";
		
		$input_syn_vals{ $clean }{ $syn_val } ++;
		# printf "All input syn_vals for $clean: >%s< \n", join ",", keys %{ $input_syn_vals{ $clean } };
		
		$input_avz_vals{ $clean }{ $syn_val } ++;
		# printf "All input inferred avz_vals for $clean: >%s< \n", join ",", keys %{ $input_avz_vals{ $clean } };		
	}
	
	else
	{
		print "Line >$_< did not match the verbal declaration scheme. \n";
	}
}

close INPUT;


#	Read the output of `make-verbs.pl Verben.txt`.

open OUTPUT, "<", "$directory/Output";

my %output_clean_lemma;
my %output_sem_vals;
my %output_syn_vals;
my %output_avz_vals;

my $base_count = 0;
my $avz_count = 0;
my $no_avz_count = 0;

while ( <OUTPUT> )
{ 
	next if /^$/;		# Ignore blank lines.
	
	next unless /base:(.+?),/;	# All verbal declarations contain a base form. Ignore all other lines.
	
	$base_count ++;
	
	my $raw =  &remove_special_characters( $1 );
	my $syn_val;
	my $sem_val;
	my $avz_val;
	
	my $clean = &clean_lemma( $raw );
	# print "Clean output lemma: >$clean< \n";
	
	$output_clean_lemma{ $clean } ++;
		
	# print; 

	if ( /,valence:(.+?),/ )
	{
		$syn_val = $1;
		$syn_val =~ s/^'|'$//g;	# Strip surrounsing single quotes on syntactic valences.
		$syn_val = &normalise_valence( $syn_val );	# Ensure valence definitions are sorted as by make-verbs.pl.

		$output_syn_vals{ $clean }{ $syn_val } ++;
		# printf "All output syn_vals for $clean: >%s< \n", join ",", keys %{ $output_syn_vals{ $clean } };
	
		if ( /sem_val:(.+?),/ )
		{
			$sem_val = $1;
			$output_sem_vals{ $clean }{ $sem_val } ++;
			# printf "All output sem_vals for $clean: >%s< \n", join ",", keys %{ $output_sem_vals{ $clean } };
		}
		
		if ( /avz_valence:(.+?)(,|])/ )
		{
			$avz_val = $1;
			$avz_val =~ s/^'|'$//g;	# Strip surrounsing single quotes on syntactic valences.
			$avz_val = &normalise_valence( $avz_val );	# Ensure valence definitions are sorted as by make-verbs.pl.
			$avz_count ++;
			$output_avz_vals{ $clean }{ $avz_val } ++;
			# printf "All output avz_vals for $clean: >%s< \n", join ",", keys %{ $output_avz_vals{ $clean } };
		}
		else
		{
			$no_avz_count ++;
		}
	}

	# print "\n";
}
	
close OUTPUT;

###	Test_1
	
	print "\n===========================================================================================================\n\n";
	
	print "1) Coverage test: Show that every verb form in the lexicon has an avz_valence. \n\n";
	print "Output verb forms with base form: $base_count \n";
	print "Output Verb forms with avz_valence: $avz_count \n\n";
	print "Output verb forms with base form but no avz_avalence: $no_avz_count \n";
	
	print "\n===========================================================================================================\n\n";


###	Test_2: 
	
	print "2) Test that all output lemmata with different avz_valences correspond to \n";
	print "   an input lemma. \n\n";
		
	# As long as the input contains lemmata with different avz_valences which only differ by their 
	# articulatory stress the output will contain lemmata with different avz_valences as well.
	# As long as these output lemmata can be related to the stress-variants of the input lemmata
	# there is no problem.
	
	
	# Collect the cleaned input lemmata with more than one avz_valence.
	
	my %several_input_avz_valences;
	
	for my $input_lemma ( sort keys %input_clean_lemma )
	{
		if ( keys %{ $input_avz_vals{ $input_lemma } } > 1 )
		{	
			$several_input_avz_valences{ $input_lemma } = 1;
			# print $input_lemma, " ", join " ", keys %{ $input_avz_vals{ $input_lemma } }, "\n\n";
		}
	}	
	printf "Cleaned input lemmata with more than one avz_valence: %s. \n", scalar keys %several_input_avz_valences;
	
	
	# Collect the output lemmata with more than one avz_valence.
	my %several_output_avz_valences;
		
	for my $output_lemma ( sort keys %output_avz_vals )
	{
		if ( keys %{ $output_avz_vals{ $output_lemma } } > 1 )
		{	
			$several_output_avz_valences{ $output_lemma } = 1;
			# print $output_lemma, " ", join " ", keys %{ $output_avz_vals{ $output_lemma } }, "\n\n";
		}
	}	
	printf "Cleaned output lemmata with more than one avz_avalence: %s.\n", scalar keys %several_output_avz_valences;
		
		
	# Check which of the output lemmata with more than one avz_valence can be matched with 
	# an input lemma with more than one avz_valcence. Only those output lemmata with more 
	# than one avz_valence that cannot be matched to an input lemma with more than one 
	# avz_valence constitute a problem.
	
	my $several_avz_valences = 0;
	
	for my $lemma ( keys %several_output_avz_valences )
	{
		if ( defined $several_input_avz_valences{ $lemma } )
		{
			# print "Output lemma $lemma with more than one avz_valence could be matched with input lemma ";
			# print "with more than one avz_valence ", join " ", keys %{ $input_avz_vals{ $lemma } }, ".\n";
		}
		else
		{
			print "Output lemma $lemma with avz_valences ", join " ", keys %{ $output_avz_vals{ $lemma } }, " could not be matched with an input lemma with more than one avz_valence.";
			$several_avz_valences ++;
		}
	}
	
	print "\nOutput lemmata with different avz_valences that could not be matched to an input lemma: $several_avz_valences \n";
	
	print "\n===========================================================================================================\n\n";


##	Test_3: Test for correctness of avz_valence assignment

	print "3) Test for correctness: Show that the output avz_valences assigned in the lexicon \n";
	print "   have also been listed in Verben.txt. \n\n";
		
	my %generated_lemmata;
	my %mismatching_output_avz_valences; 
	
	for my $candidate ( sort keys %output_avz_vals )
	{	
		if ( defined $input_avz_vals{ $candidate } )	# Check if the current output lemma also was an input lemma.
		{
			for my $avz ( keys %{ $output_avz_vals{ $candidate } } ) # Check if the current output avz_valence also was an avz_valence of the corresponding input lemma.
			{	
				if ( defined $input_avz_vals{ $candidate }{ $avz } )
				{
					# print "Output lemma >$candidate< has avz_valence(s) >$avz< which maps to the input lemma's set of avz_valences >", join ",", keys %{ $input_avz_vals{ $candidate } }, "<.\n";
				}
				else
				{	
					$mismatching_output_avz_valences{ $candidate } = 1;
					# print "Output lemma >$candidate< has avz_valence >$avz< which is not contained in the input lemma's set of avz_valences >", join ",", keys %{ $input_avz_vals{ $candidate } }, "<.\n";
					# print $candidate, "\t", "Output: >", join "*", keys %{ $output_avz_vals{ $candidate } }, "<\tInput: ", join "*", keys %{ $input_avz_vals{ $candidate } }, "\n";
					next;
				}
			}
		}
		else
		{
			$generated_lemmata{ $candidate } ++;
			# print "Output lemma >$candidate< had an output avz_valence but was not listed as an input lemma with an avz_valence in Verben.txt. \n"
		}
	}
	
	printf "Output lemmata with an avz_valence: %s. \n\n", scalar keys %output_avz_vals;
	printf "Output lemmata with avz_valence that have no correspondence in Verben.txt: %s. \n\n", scalar keys %mismatching_output_avz_valences;
	printf "Output lemmata with an inferred avz_valence, i.e. one that is not in Verben.txt: %s. \n", scalar keys %generated_lemmata;

	print "\n===========================================================================================================\n\n";
	

##	Test_4

	print "4) Test for avz_val completeness: Show that every lemma-avz_valence pair \n";
	print "   defined in Verben.txt also occurs in the lexicon. \n\n";
		
	my %input_missing_in_lexicon; 
	
	for my $verb ( sort keys %input_avz_vals )
	{
		if ( defined $output_avz_vals{ $verb } )	# Check if the current output lemma was an input lemma.
		{
			for my $valence ( keys %{ $output_avz_vals{ $verb } } )
			{	
				if ( defined $output_avz_vals{ $verb }{ $valence } )
				{
					# print "Input lemma >$verb< has avz_valence >$valence< which is contained in the output lemma's set of avz_valences >", join ",", keys %{ $output_avz_vals{ $verb } }, "<.\n";
				}
				else
				{	
					$input_missing_in_lexicon{ $verb } = 1;
					# print "Output lemma >$verb< has avz_valence >$valence< which is not contained in the input lemma's set of avz_valences >", join ",", keys %{ $input_avz_vals{ $verb } }, "<.\n";
				}
			}
		}
		else
		{
			$generated_lemmata{ $verb } ++;
			# print "Output lemma >$candidate< was in generated lexicon but was not listed as an input lemma in Verben.txt. \n"
		}
	}	

	printf "Input lemmata with an avz_valence (defined or inferred): %s. \n\n", scalar keys %input_avz_vals;
	printf "Input lemmata with an avz_valence not listed in the lexicon: %s. \n", scalar keys %input_missing_in_lexicon;
	
	print "\n===========================================================================================================\n\n";


##	Test_5

	print "5) Test for sem_val completeness: Show that every lemma-sem_val pair \n";
	print "   defined in Verben.txt also occurs in the lexicon. \n\n";

	%input_missing_in_lexicon = ();
		
	for my $verb ( sort keys %input_sem_vals )
	{
		if ( defined $output_sem_vals{ $verb } )	# Check if the current output lemma was an input lemma.
		{
			for my $valence ( keys %{ $output_sem_vals{ $verb } } )
			{	
				if ( defined $output_sem_vals{ $verb }{ $valence } )
				{
					# print "Input lemma >$verb< has sem_val >$valence< which maps to the output lemma's set of sem_vals >", join ",", keys %{ $output_sem_vals{ $verb } }, "<.\n";
				}
				else
				{	
					$input_missing_in_lexicon{ $verb } = 1;
					# print "Input lemma >$verb< with sem_val >$valence< is not contained in the corresponding output lemma's set of sem_vals >", join ",", keys %{ $output_sem_vals{ $verb } }, "<.\n";
				}
			}
		}
		else
		{
			$input_missing_in_lexicon{ $verb } ++;
			# print "Input lemma >$verb< was in Verben.txt but not in the generated lexicon.\n"
		}
	}	

	printf "Input lemmata with a sem_val definition: %s. \n\n", scalar keys %input_sem_vals;
	printf "Input lemmata with a sem_val in Verben.txt that is not listed in the lexicon: %s. \n", scalar keys %input_missing_in_lexicon;
	#printf "Output lemmata that were not listed as input lemmata in Verben.txt: %s. \n", scalar keys %input_missing_in_lexicon;


	print "\n===========================================================================================================\n\n";


##	Test_6

	print "6) Test for correctness: Show that the sem_val assigned to each output \n";
	print "   lemma has also been listed for it in Verben.txt. \n\n";

	%generated_lemmata = ();
	my %mismatching_output_sem_vals; 
	
	for my $candidate ( sort keys %output_sem_vals )	# Loop over all output lemmata with a sem_val.
	{
		if ( defined $input_clean_lemma{ $candidate } )	# Check if the current output lemma was also a clean input lemma.
		{
			# print "output lemma $candidate has ", scalar keys %{ $output_sem_vals{ $candidate } }, " sem_vals, namely: ", keys %{ $output_sem_vals{ $candidate } }, "\n";
			# print "input lemma $candidate has ", scalar keys %{ $input_sem_vals{ $candidate } }, " sem_vals, namely: ", keys %{ $input_sem_vals{ $candidate } }, "\n";
			 
			for my $semantic ( keys %{ $output_sem_vals{ $candidate } } )
			{	
				if ( defined $input_sem_vals{ $candidate }{ $semantic } )	# Check if the current output sem_val was also an input sem_val.
				{
					# print "Output lemma >$candidate< has sem_valence(s) >$semantic< which maps to the input lemma's set of sem_vals >", join ",", keys %{ $input_sem_vals{ $candidate } }, "<.\n";		
				}						
				else
				{	
					$mismatching_output_sem_vals{ $candidate } = 1;
					# print "Output lemma >$candidate< has sem_val >$semantic< which does not map to the input lemma's set of sem_vals >", keys %{ $input_sem_vals{ $candidate } }, "<\n";
				}
			}
		}
		else
		{
			$generated_lemmata{ $candidate } ++;
			print "Output lemma >$candidate< with sem_val(s) >", keys %{ $output_sem_vals{ $candidate } }, "< was in the generated lexicon but was not listed in Verben.txt. \n";
		}
	}
	
	printf "Output lemmata with a sem_val: %s. \n", scalar keys %output_sem_vals;
	printf "Output lemmata also in Verben.txt whose sem_val was not listed in Verben.txt: %s. \n", scalar keys %mismatching_output_sem_vals;
	printf "Output lemmata with sem_val whose lemma was not listed in Verben.txt: %s. \n", scalar keys %generated_lemmata;

	print "\n===========================================================================================================\n\n";


sub normalise_valence
{
	#  	Ensure that the valence definitions in Verben.txt comply with the
	#	following requirements:
	#
	#	1. Valence blocks are sorted in alphabetical order (i.e.: d+a --> a+d).
	#	2. Constituents of a valence block are sorted in alphabetical order (i.e.: aics --> acis).
	#	3. Question marks indicating optionality are located at the end of a valence block (i.e.: a?d --> ad?).
	#	4. The reflexivity marker 'r' inseparably prefixes either valence 'a' oder 'd'. The resulting units appear in valence block initial position.

	my $raw_syntactic_valence = shift @_;
			
	my @sorted_blocks;
	
	for my $raw_block ( split ( /\+/, $raw_syntactic_valence ) )
	{
		my $sorted_block = $raw_block;

		# Encode reflexivity indicators for processing.
		$sorted_block =~ s/ra/r/go;
		$sorted_block =~ s/rd/R/go;
		
		# Create alphabetical order of block constituents.
		$sorted_block = join "", sort split ( "", $sorted_block );
		
		# Force optionality indicator question mark to the end of the block.
		$sorted_block .= "?" if ( $sorted_block =~ s/\?//o );
					
		# Force reflexivity indicators to the beginning of a block.
		$sorted_block = $1 . $sorted_block if ( $sorted_block =~ s/([rR])//o );			
						
		push @sorted_blocks, $sorted_block;
	}
	
	my $sorted_valence = join "+", sort @sorted_blocks;
	
	# Decode reflexivity indicators for processing.			
	$sorted_valence  =~ s/r/ra/go;
	$sorted_valence  =~ s/R/rd/go;
	
	# print STDERR "Sorted Valence: >$sorted_valence<\n";
	
	return $sorted_valence;
}


sub clean_lemma
{
	my $lemma = shift @_;
	$lemma =~ s/<|>//g;		# Remove markers of articulatory stress.
	
	$lemma = &remove_special_characters( $lemma );
	
	return $lemma;
}


sub remove_special_characters
{
	my $lemma = shift @_;
	$lemma =~ s/^'|'$//g;	# Remove surrounding single quotes as in '(fern)steuern'.
	$lemma =~ s/\\"//g;		# Remove backslash-escaped double quotes as in "POR"en.
	$lemma =~ s/\\'//g;		# Remove backslash-escaped apostrophes as in \'rberbringen.
	
	# print "Clean output lemma: >$lemma< \n";	
	
	return $lemma;
}