/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup Beamsearch Beamsearch - A Module for incremental beam search
 * @author Niels Beuck
 * @date 2.8.10
 *
 * This module implements an incremental beam search.
 *
 *
 * @{
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <string.h>
#include <blah.h>
#include "cdg.h"
#include "beamsearch.h"
#include "increment.h"
#include "constraintnet.h"
#include "hook.h"
#include "set.h"
#include "virtualnode.h"
#include "frobbing.h"

/* -- MACROS --------------------------------------------------------------- */

/* -- TYPE DEFINITIONS ----------------------------------------------------- */

/* -- VARIABLES ------------------------------------------------------------ */

int bsBeamWidth = 10;
int bsBottomUpPredictionDepth =2;
float bsBeamToleranceFactor = 0.3;
int bsVerbosity = 2; // 0= no intermediate output, 1= step end output, 2= entry end output, 3= as 2 plus warnings 4= all output
Boolean bsLockAllLevels = TRUE;
Boolean bsUnlockUnusedVNs = TRUE;

/* -- FUNCTIONS ------------------------------------------------------------ */


/* -- IMPLEMENTATION ------------------------------------------------------- */

/*
 * registers the set variables for this strategy
 */
void bsInitialize() {
	setRegister("bsBeamWidth",               SET_INT,   &bsBeamWidth,               NULL, NULL, NULL, NULL);
	setRegister("bsBottomUpPredictionDepth", SET_INT,   &bsBottomUpPredictionDepth, NULL, NULL, NULL, NULL);
	setRegister("bsBeamToleranceFactor",     SET_FLOAT, &bsBeamToleranceFactor,     NULL, NULL, NULL, NULL);
	setRegister("bsVerbosity", 	             SET_INT,   &bsVerbosity,               NULL, NULL, NULL, NULL);
	setRegister("bsLockAllLevels",           SET_BOOL,  &bsLockAllLevels,           NULL, NULL, NULL, NULL);
	setRegister("bsUnlockUnusedVNs",         SET_BOOL,  &bsUnlockUnusedVNs,         NULL, NULL, NULL, NULL);
}


/*
 * returns a beam entry for a given analysis
 *
 */
BeamEntry beFromAnalysis(Analysis a) {
	BeamEntry newBe = (BeamEntry) memMalloc(sizeof(BeamEntryStruct));
	newBe->best = a;
	newBe->p = analysis2parse(a);
	newBe->p->searchStrategy = strRegister("BeamSearch");
	newBe->virtualNodeUsed = NULL;      // we add it later
	newBe->virtualNodeAvailable = NULL; // same here
	return newBe;
}

/*
 * free the memory of the beam entry
 * including the analysis and the parse, if there is one
 */
void beDelete(BeamEntry be) {
	aDelete(be->best);

	if(be->p)
		parseDelete(be->p);

	if(be->virtualNodeAvailable) {
		bvDelete(be->virtualNodeAvailable);
	}

	if(be->virtualNodeUsed) {
		bvDelete(be->virtualNodeUsed);
	}

	memFree(be);
}

/*
 * get the score of the first entry in the beam
 * make sure the entries are sorted
 *
 * if the beam is empty, return 0.0
 */
Number beamGetHighestScore(List beam) {
	if(!beam) {
		return 0.0;
	}

	BeamEntry be = beam->item;
	Badness b = be->best->b;

	if(b->hard)
		return 0.0;

	return b->soft;
}

/*
 * get the score of the last entry in the beam
 * make sure the entries are sorted
 *
 * if the beam is empty, return 0.0
 */
Number beamGetLowestScore(List beam, int beamWidth) {
	if(!beam || listSize(beam) < beamWidth) {
		return 0.0;
	}

	BeamEntry be = listLastElement(beam);
	Badness b = be->best->b;

	if(b->hard)
		return 0.0;

	return b->soft;
}


/* ----------------------------------------------------------------------
   Print a parse that has no words or lexemgraph set to stdout
 */
void printAbstractParse(Parse parse, LexemGraph lg) {
	int i, j;
	String label;
	Arc arc;
	int nrLevels = inputCurrentGrammar->noOfLevels;
	Level level;

	for (i = 0; i < vectorSize(parse->verticesStructure); i++) {
		level = listNthElement(inputCurrentGrammar->levels, (i % nrLevels) + 1);
		if (!level || !level->showflag)
			continue;

		label = vectorElement(parse->verticesLabels, i);

		// skip uninteresting values of other levels and for virtual nodes
		if (!strcmp(label, "") &&
				(int) vectorElement(parse->verticesStructure, i) == -1 &&
				(level->no > 0 ||  (lg->ve && i/nrLevels >= lg->ve->vStartID) ) ) {
			continue;
		}

		if (lg->ve && i/nrLevels >= lg->ve->vStartID){
			cdgPrintf(CDG_DEFAULT, "   %03d %10s: %s --%s-->",
				i , level->id,
				"[virtual]", label);
		} else {
			arc = listNthElement(lg->lattice->arcs, i / nrLevels +1);

			cdgPrintf(CDG_DEFAULT, "   %03d %10s: %s --%s-->",
				arc->from * nrLevels + level->no, level->id,
				arc->word, label);
		}

		j = (int) vectorElement(parse->verticesStructure, i);
		if (j == -1) {
			cdgPrintf(CDG_DEFAULT, "NIL\n");
		} else if (j == -2) {
			cdgPrintf(CDG_DEFAULT, "NONSPEC\n");
		} else if (lg->ve && j >= lg->ve->vStartID){
			cdgPrintf(CDG_DEFAULT, "[virtual]\n");
		} else {
			arc = listNthElement(lg->lattice->arcs, j+1);
			cdgPrintf(CDG_DEFAULT, "%s\n", arc->word);
		}
	}

	cdgPrintf(CDG_DEFAULT, "\n");
}


/*
 * returns the best LevelValue from the given list,
 *
 */
LevelValue bestScoringLVfromList(List lvs) {
	List l= lvs;
	LevelValue best  = NULL;
	Number bestScore = 0.0;
	LevelValue lv;

	while (l) {
		lv= l->item;
		l=l->next;
		if(lv->score > bestScore) {
			best= lv;
			bestScore = lv->score;
		}
	}

	return best;
}

/**
 * returns a bytevector that says which lexemes are available in the given BeamEntry
 *
 * assumes:
 * - that only the first level is restrictive regarding lexemes
 * - that the net was already extended by one arc compared to the parse in the BeamEntry
 *
 */
ByteVector getAvailableLexemes(BeamEntry be, ConstraintNet net) {
	ByteVector ret, tmp;
	int i, j;
	int dep;
	int reg;
	String label;
	List lvs, l, l2;
	int noOfLevls= inputCurrentGrammar->noOfLevels;
	GraphemNode gn;
	LevelValue lv;
	Boolean b;

	ret= bvNew(bvSize(net->lexemgraph->isDeletedNode));
	bvSetAllElements(ret, TRUE);

	for(i=0; i < vectorSize(be->p->verticesStructure); i++ ) {
		if(i%noOfLevls)
			continue;

		dep= i/noOfLevls;
		reg= (int)vectorElement(be->p->verticesStructure, i);
		label= vectorElement(be->p->verticesLabels, i);

		// because we assume the net was already extended,
		// we have to do some stunts regarding virtual node indices
		if(dep >= net->lexemgraph->ve->vStartID -1) {
			dep+= 1;
		}
		if(reg >= net->lexemgraph->ve->vStartID -1) {
			reg+= 1;
		}

		// prepare a ByteVector for the local availability ...
		tmp = bvNew(bvSize(ret));
		bvSetAllElements(tmp,TRUE);

		// ... first mark all possible dependent lexemes as false,
		// so we can mark the actually occurring ones as true again
		gn = vectorElement(net->lexemgraph->graphemnodes, dep);
		l = gn->lexemes;
		while(l) {
			bvSetElement(tmp, FALSE, ((LexemNode)l->item)->no);
			l=l->next;
		}

		// same for the regents
		if(reg >= 0) {
			gn = vectorElement(net->lexemgraph->graphemnodes, reg);
			l = gn->lexemes;
			while(l) {
				bvSetElement(tmp, FALSE, ((LexemNode)l->item)->no);
				l=l->next;
			}
		}

		lvs= cnGetAccordingLVs(net,0,dep,label,reg);

		// now let's look at the actually occurring lexemes
		l = lvs;
		while(l) {
			lv = l->item;
			l2= lv->dependents;
			while(l2) {
				bvSetElement(tmp, TRUE, ((LexemNode)l2->item)->no);
				l2=l2->next;
			}

			l2= lv->regents;
			while(l2) {
				bvSetElement(tmp, TRUE, ((LexemNode)l2->item)->no);
				l2=l2->next;
			}

			l=l->next;
		}

		// ... and multiply it with the global vector
		for(j=0; j < bvSize(ret); j++) {
			b= bvElement(ret,j) && bvElement(tmp, j);
			bvSetElement(ret, b, j);
		}

		listDelete(lvs);
	}


	return ret;
}

/*
 * returns TRUE, if at least one of the given level values has a score above the threshold
 */
Boolean compatibleLvExists(ConstraintNet net, int lvlid, int dep, String label, int reg, Number threshold, ByteVector lexemAvailable) {

	List lvs;
	LevelValue lv;
	List l, l2;

	lvs = cnGetAccordingLVs(net, lvlid, dep, label, reg);

	// for all according LVs
	l   = lvs;
	while (l) {
		lv= l->item;
		if( significantlyGreater(lv->score, threshold) ) {
			// score is OK, next question: are the lexemes compatible?
			if(reg >= 0) {
				l2= lv->regents;
				while(l2) {
					if(bvElement(lexemAvailable, ((LexemNode)l2->item)->no)) {
						listDelete(lvs);
						return TRUE;
					}
					l2= l2->next;
				}
			} else {
				// regent is not not an actual graphemnode, so the is no issue of lexem compatibility
				listDelete(lvs);
				return TRUE;
			}
		}
		l=l->next;
	}

	listDelete(lvs);
	return FALSE;
}

/*
 * calculates a score estimation by multiplying the best score for every slot
 */
Number getScoreEstimation(Parse p, ConstraintNet net, Boolean virtOffset) {
	Number score = 1.0;
	int sid_i, wid_i;
	int wid_r;
	int noOfLevels = inputCurrentGrammar->noOfLevels;
	String label;
	int lvlno;
	List lvs;
	LevelValue lv;

	for(sid_i=0; sid_i< vectorSize(p->verticesStructure); sid_i++) {
		wid_i = sid_i / noOfLevels;
		wid_r = (int)vectorElement(p->verticesStructure, sid_i);
		label =      vectorElement(p->verticesLabels   , sid_i);
		lvlno = sid_i % noOfLevels;

		if(virtOffset && net->lexemgraph->ve && wid_i >= net->lexemgraph->ve->vStartID-1)
			wid_i +=1;

		if(virtOffset && net->lexemgraph->ve && wid_r >= net->lexemgraph->ve->vStartID-1)
			wid_r +=1;

		lvs   = cnGetAccordingLVs(net, lvlno, wid_i, label, wid_r);
		lv    = bestScoringLVfromList(lvs);

		listDelete(lvs);
		if(lv) {
			score = score * lv->score;
		} else {
			return 0.0;
		}

	}

	return score;
}

/*
 * compares two beam entries by the score of their respective analysis
 */
Boolean beCompare(BeamEntry a, BeamEntry b) {
	return bCompare(a->best->b, b->best->b);
}

/*
 * sort the beam entries by score and return a list containing the best n,
 * where n is the given beam size
 *
 * if a factor > 0 is given, there is an additional restriction:
 * no entries may be worse than the best score times the given factor
 *
 * deletes the given beam
 */
List pruneBeam(List oldBeam, int beamSize, Number f) {
	int i;
	List l= oldBeam;
	List ret = listNew();
	Number threshold;
	BeamEntry be;

	l= listSort(l, beCompare);

	// calculate the threshold from the best score
	be= l->item;
	threshold= be->best->b->soft * f;

	i= 0;
	while(l && i< beamSize) {
		be= l->item;
		if(be->best->b->soft < threshold) {
			break;
		}
		ret= listAppendElement(ret, l->item);
		i++;
		l= l->next;
	}

	// now delete all remaining entries
	while(l) {
		beDelete(l->item);
		l= l->next;
	}

	return ret;
}

/*
 * undeletes all LVs with virtual nodes as dependent, that are unused in the given parse
 */
void unlockUnusedVNs(ConstraintNet net, Parse p) {
	LexemGraph lg = net->lexemgraph;
	int i, j, k;
	int reg;
	int noOfLevels = inputCurrentGrammar->noOfLevels;
	String label;
	LevelValue lv;
	Vector lvs;

	if(!lg->ve)  {
		// no VNs, no problem
		return;
	}

	for(i=lg->ve->vStartID; i < lg->max; i++) {
		reg  = (int)vectorElement(p->verticesStructure, i* noOfLevels);
		label=      vectorElement(p->verticesLabels   , i* noOfLevels);

		if(reg==-1 && strcmp(label,"")==0) {
			// is unused, reactivate all lvs
			for(k=0; k<noOfLevels; k++) {
				lvs =((ConstraintNode)vectorElement(net->nodes, i*noOfLevels +k))->values;
				for(j=0; j< vectorSize(lvs); j++) {
					lv = vectorElement(lvs, j);
					if(lv->score > 0.0) {
						lv->isDeleted = FALSE;
						lv->limit = lv->score;
					}
				}
			}
		}
	}

	for(i=0; i< vectorSize(net->nodes); i++) {
		incrUpdateNumberCaches(vectorElement(net->nodes, i));
	}
}

/*
 * takes a list of parses and generates a list of beam entries of the given beam size
 */
List addParseToBeam(Parse p, int beamSize, Number f, List oldBeam, ConstraintNet net) {
	List beam = oldBeam;
	BeamEntry newBE;
	Problem P;
	Number threshold;

	threshold=  max(
			beamGetHighestScore(beam) * bsBeamToleranceFactor,
			beamGetLowestScore(beam, bsBeamWidth));

	// lexeme selection is done by frobbing
	// so for all possible parses ...

	if(bsVerbosity>=4) {
		cdgPrintf(CDG_INFO, "------------------------------------\n");
		printAbstractParse(p, net->lexemgraph);
		cdgPrintf(CDG_INFO, "\n");
	}

	if(bsVerbosity < 4) {
		if(bsVerbosity < 3) {
			hkDeactivate(CDG_WARNING);
		}
		hkDeactivate(CDG_INFO);
	}

	// we have to limit the search space, only lexeme variations are allowed
	cnResetToThreshold(net,threshold);
	if(bsLockAllLevels) {
		cnLockToParse(net, p, NULL);
	} else {
		cnLockToParse(net, p, inputCurrentGrammar->levels->item);
	}
	if(bsUnlockUnusedVNs) {
		unlockUnusedVNs(net, p);
	}

	// To make an initial analysis from a parse,
	// we would need a decorated path,
	// and for decorating a path we would need words with descriptions
	// parseDecorate(p, net->lexemgraph, NULL);
	// now we generate a problem so we can call frobbing on it
	//P = net2problem(net, 0.0, p); // parse2analysis is called in here
	// instead we will not give a parse for now and hope for the locking to do the job
	P = net2problem(net, 0.0, NULL);

	// let frobbing find the best lexeme selection
	if(P) {
		combined(P);

		hkRestoreVerbosity();
		if(bsVerbosity>=4)
			cdgPrintf(CDG_INFO, "Score: %.5f\n", P->best->b->hard?0.0:P->best->b->soft);
		if(P->best->b->hard || significantlyGreater(threshold, P->best->b->soft)) {
			// this variant's score is too bad, delete it
			aDelete(P->best);

		} else {
			newBE = beFromAnalysis(P->best);
			beam = listAppendElement(beam, newBE);

			beam = pruneBeam(beam, beamSize, f);
		}

		// either way, we don't need the problem anymore, so:
		aDelete(P->current);
		problemDelete(P);
	} else {
		// this parse did not work,
		// probably as the lvs compatible with one of the new edges
		// are not compatible lexeme wise with the rest of the parse
		hkRestoreVerbosity();
		if(bsVerbosity>=3)
			cdgPrintf(CDG_WARNING, "Invalid parse, skipping \n");
	}

	return beam;
}


/* ---------------------------------------------------------------------------
 * Make variants of the given parse by attaching the new arc
 * - to every existing node
 * - with every existing label
 * in the first level, in all other levels it will be attachet to nil
 *
 * only choose combinations with a according LV in the net,
 * that has a score above the threshold
 *
 * @param arc	the arc to be attached
 * @param p 	the parse to be varied, may differ from the one in the BE,
 *                                      this is needed for recursive calls for virtual nodes
 * @param be	beam entry, contains the net and virtual node information
 * @param insert, whether the arc has to be inserted or is already part of the parse
 * @param childTracking, list containing transitive children for circle detection
 *
 * @returns a list of variants of the given parse
 */
List appendArcToParse(Arc arc, Parse p, ConstraintNet net, BeamEntry be, List oldBeam, Boolean insert, List childTracking) {
	List results = oldBeam;
	Number threshold = 0.0;
	int i, j, pos, slot, r, wid_r;
	int noOfLvls = inputCurrentGrammar->noOfLevels;
	List lvs;
	String label;
	Parse p2;
	Vector tracker; // vector of lists, indexed by slot ids
	LevelValue lv;
	Boolean vReg;
	GraphemNode gn;

	if(listSize(childTracking) > bsBottomUpPredictionDepth) {
		return results;
		// more than n steps in the unknown are cut off
	}

	Number estimatedScore = getScoreEstimation(p, net, insert);
	if(estimatedScore == 0.0) {
		// the given slots already prevent a soft solution, so we can stop right here
		// then again, why is this pare in the beam? probably a score fluctuation effect
		return results;
	}

	ByteVector availableLexemes = getAvailableLexemes(be, net);

	tracker = vectorNew(vectorSize(net->nodes) +2); // +2 for nil and nonspec
	for(i=0; i < vectorSize(net->nodes); i++) {
		vectorSetElement(tracker, listNew(), i);
	}

	pos= arc->from;
	slot= pos* noOfLvls;

	lvs= vectorToList(((ConstraintNode)vectorElement(net->nodes, slot))->values);

	threshold=  max(
			beamGetHighestScore(results) * bsBeamToleranceFactor,
			beamGetLowestScore(results, bsBeamWidth));

	// try all levelValues
	for( ; lvs; lvs= lvs->next) {

		lv= lvs->item;

		if(lv->score * estimatedScore < threshold) {
			// no need to check this levelValue, as it is too bad,
			// as the lvs are sorted by score, there is also no need to go any further
			break;
		}

		label= lv->label;
		if(root(lv->regents)) {
			wid_r= -1;
			vReg= FALSE;
		} else if(nonspec(lv->regents)) {
			wid_r= -2;
			vReg= FALSE;
		} else {
			gn= lvGetRegentGN(lv);
			wid_r = gn->no;
			vReg= gn->isVirtual;
		}

		if(!insert && childTracking && listContains(childTracking, (void *)wid_r) ) {
			// circle detection!
			// the word we are about to append is an virtual one
			// and i currently points to one of the reasons we need the new virtual word
			// so skip the word i to prevent a circle
			continue;
		}

		if(listContains(vectorElement(tracker,wid_r +2), label)) {
			// we already had that combination
			continue;
		}

		// add combination to tracker
		vectorSetElement(tracker,
				listAppendElement(
						vectorElement(tracker, wid_r +2),
						label),
				wid_r +2);

		if(!insert && wid_r==-1 && strcmp(label,"")==0) {
			// We are currently introducing a virtual word because it is needed.
			// A dependency to NIL with the empty label would result in a
			// partially used virtual word, that is forbidden.
			// To save time it is better to stop it here,
			// instead of letting the grammar handle it.
			continue;
		}

		// be->virtualNodeAvailable uses old indices
		if(vReg && !bvElement(be->virtualNodeAvailable, wid_r-1)) {
			// that virtual word is not for use in this entry, skip it
			// there should be another free virtual word of the same kind
			continue;
		}

		if(compatibleLvExists(net, 0, pos, label, wid_r, threshold/estimatedScore, availableLexemes) ) {
			// found a LevelValue above the threshold,

			// so add the current combination to the copy of the parse
			p2= parseClone(p, FALSE); // No need to decorate the parse
			if(insert) {
				vectorInsertElement(p2->verticesLabels, label, slot);
				vectorInsertElement(p2->verticesStructure, (void *)wid_r, slot);
				for(j= 1; j< noOfLvls; j++) {
					// add default values for other parses
					vectorInsertElement(p2->verticesLabels, "", slot+j);
					vectorInsertElement(p2->verticesStructure, (void *)-1, slot+j);
				}
				if(useVirtualNodes) {
					// push references to virtual nodes
					for(j= 0; j< vectorSize(p2->verticesStructure); j++) {
						if(j== slot) // skip the word itself
							continue;
						r = (int) vectorElement(p2->verticesStructure, j);
						if(r >= pos) {
							vectorSetElement(p2->verticesStructure, (void *)r+1, j);
						}
					}
				}
			} else {
				vectorSetElement(p2->verticesLabels, label, slot);
				vectorSetElement(p2->verticesStructure, (void *)wid_r, slot);
			}

			if(vReg && !bvElement(be->virtualNodeUsed, wid_r-1)) {
				results = appendArcToParse(((GraphemNode)vectorElement(net->lexemgraph->graphemnodes,wid_r))->arc,
						p2, net, be, results, FALSE, listAppendElement(listClone(childTracking), (void *)pos) );
			} else {
				// now go from parse to an actual analysis, that we can evaluate, and add that to the beam
				results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net);
			}
			parseDelete(p2);

			// update threshold
			threshold=  max(
					beamGetHighestScore(results) * bsBeamToleranceFactor,
					beamGetLowestScore(results, bsBeamWidth));
		}
	} // end for all lvs

	for(i=0; i< vectorSize(tracker); i++) {
		listDelete(vectorElement(tracker, i));
	}
	vectorDelete(tracker);

	listDelete(childTracking);
	return results;
}


/* ---------------------------------------------------------------------------
 * Old version of the function
 *
 * Make variants of the given parse by attaching the new arc
 * - to every existing node
 * - with every existing label
 * in the first level, in all other levels it will be attachet to nil
 *
 * only choose combinations with a according LV in the net,
 * that has a score above the threshold
 *
 * @param arc	the arc to be attached
 * @param p 	the parse to be varied, may differ from the one in the BE,
 *                                      this is needed for recursive calls for virtual nodes
 * @param be	beam entry, contains the net and virtual node information
 * @param insert, whether the arc has to be inserted or is already part of the parse
 * @param childTracking, list containing transitive children for circle detection
 *
 * @returns a list of variants of the given parse
 */
List appendArcToParse2(Arc arc, Parse p, ConstraintNet net, BeamEntry be, List oldBeam, Boolean insert, List childTracking) {
	List results = oldBeam;
	Number threshold = 0.0;
	int i, j, s, pos, slot, r, old_i;
	int noOfLvls = inputCurrentGrammar->noOfLevels;
	List labels= ((Level)inputCurrentGrammar->levels->item)->labels;
	List l;
	String label;
	Parse p2;

	ByteVector availableLexemes = getAvailableLexemes(be, net);

	Number estimatedScore = getScoreEstimation(p, net, insert);
	if(estimatedScore == 0.0) {
		// the given slots already prevent a soft solution, so we can stop right here
		// then again, why is this pare in the beam? probably a score fluctuation effect
		return results;
	}

	pos= arc->from;
	slot= pos* noOfLvls;

	if(listSize(childTracking) > bsBottomUpPredictionDepth) {
		return results;
		// more than n steps in the unknown are cut off
	}

	// starting value, nil is -1, nonspec -2
	if(cnUseNonSpec ) {
		s= -1; //s= -2;
		// TODO NONSPEC not supported yet, we would have to be able to replace it like a virtual node
		cdgPrintf(CDG_WARNING, "WARNING: BeamSearch does not support NONSPEC yet \n");
	} else {
		s= -1;
	}

	// attach to old words
	for(i= s; i< pos; i++) {

		if(!insert && childTracking && listContains(childTracking, (void *)i) ) {
			// circle detection!
			// the word we are about to append is an virtual one
			// and i currently points to one of the reasons we need the new virtual word
			// so skip the word i to prevent a circle
			continue;
		}

		for(l= labels ;l; l= l->next) {
			label= l->item;

			if(!insert && i==-1 && strcmp(label,"")==0) {
				// We are currently introducing a virtual word because it is needed.
				// A dependency to NIL with the empty label would result in a
				// partially used virtual word, that is forbidden.
				// To save time it is better to stop it here,
				// instead of letting the grammar handle it.
				continue;
			}

			threshold=  max(
					beamGetHighestScore(results) * bsBeamToleranceFactor,
					beamGetLowestScore(results, bsBeamWidth));

			if(threshold> estimatedScore) {
				// we can stop right here, threshold is too high
				return results;
			}

			if(compatibleLvExists(net, 0, pos, label, i, threshold/estimatedScore, availableLexemes) ) {
				// found a LevelValue above the threshold,

				// so add the current combination to the copy of the parse
				p2= parseClone(p, FALSE); // No need to decorate the parse
				if(insert) {
					vectorInsertElement(p2->verticesLabels, label, slot);
					vectorInsertElement(p2->verticesStructure, (void *)i, slot);
					for(j= 1; j< noOfLvls; j++) {
						// add default values for other parses
						vectorInsertElement(p2->verticesLabels, "", slot+j);
						vectorInsertElement(p2->verticesStructure, (void *)-1, slot+j);
					}
					if(useVirtualNodes) {
						// push references to virtual nodes
						for(j= 0; j<slot ; j++) {
							r = (int) vectorElement(p2->verticesStructure, j);
							if(r >= pos) {
								vectorSetElement(p2->verticesStructure, (void *)r+1, j);
							}
						}
					}
				} else {
					vectorSetElement(p2->verticesLabels, label, slot);
					vectorSetElement(p2->verticesStructure, (void *)i, slot);
				}

//				if(insert) {
//					cdgPrintf(CDG_INFO, "This parse was generated by appendArcToParse, case 1\n");
//				} else {
//					cdgPrintf(CDG_INFO, "This parse was generated by appendArcToParse, case 1+2\n");
//				}
//				printAbstractParse(p2, net->lexemgraph);
//				cdgPrintf(CDG_INFO, "------------------------------------\n");

				// now go from parse to an actual analysis, that we can evaluate, and add that to the beam
				results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net);
				parseDelete(p2);
			}
		}
	}

	if(useVirtualNodes) {
		// attach to virtual nodes, either to
		// - already used ones or
		// - new ones and then attach that one as well
		for(i= net->lexemgraph->ve->vStartID; i< net->lexemgraph->max; i++) {

			// ATTENTION: i is a new LG index, but the old parse points to the old LG
			old_i = i-1;

			if(!insert && childTracking && listContains(childTracking, (void *)i) ) {
				// circle detection!
				continue;
			}

			// be->virtualNodeAvailable uses old indices
			if(!bvElement(be->virtualNodeAvailable, old_i)) {
				// that virtual word is not for use in this entry, skip it
				// there should be another free virtual word of the same kind
				continue;
			}

			if(!insert && arc->from == i) {
				// the arc is an virtual word and we are about to connect it to itself, so skip it
				continue;
			}

			for(l=labels ; l; l= l->next) {
				label= l->item;

				threshold=  max(
						beamGetHighestScore(results) * bsBeamToleranceFactor,
						beamGetLowestScore(results, bsBeamWidth));

				if(threshold> estimatedScore) {
					// we can stop right here, threshold is too high
					return results;
				}

				if(compatibleLvExists(net, 0, pos, label, i, threshold/estimatedScore, availableLexemes) ) {
					// found a LevelValue above the threshold,
					// so add the current combination to the copy of the parse
					p2= parseClone(p, FALSE); // no need to decorate it
					if(insert) {
						vectorInsertElement(p2->verticesLabels, label, slot);
						vectorInsertElement(p2->verticesStructure, (void *)i, slot);

						for(j= 1; j< noOfLvls; j++) {
							vectorInsertElement(p2->verticesLabels, "", slot+j);
							vectorInsertElement(p2->verticesStructure, (void *)-1, slot+j);
						}
						// push references to virtual nodes
						for(j= 0; j<slot ; j++) {
							r = (int) vectorElement(p2->verticesStructure, j);
							if(r >= pos) {
								vectorSetElement(p2->verticesStructure, (void *)r+1, j);
							}
						}
					} else {
						vectorSetElement(p2->verticesLabels, label, slot);
						vectorSetElement(p2->verticesStructure, (void *)i, slot);
					}

					if(!bvElement(be->virtualNodeUsed, old_i)) {
						results = appendArcToParse(((GraphemNode)vectorElement(net->lexemgraph->graphemnodes,i))->arc,
								p2, net, be, results, FALSE, listAppendElement(listClone(childTracking), (void *)pos) );
					} else {

						// now go from parse to an actual analysis, that we can evaluate, and add that to the beam
						results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net);

//						if(insert) {
//							cdgPrintf(CDG_INFO, "This parse was generated by appendArcToParse, case 2\n");
//						} else {
//							cdgPrintf(CDG_INFO, "This parse was generated by appendArcToParse, case 2+2\n");
//						}
//						printAbstractParse(p2, net->lexemgraph);
//						cdgPrintf(CDG_INFO, "------------------------------------\n");

						parseDelete(p2);
					}
				}

			}
		}
	}

	listDelete(childTracking);
	return results;
}

/* -------------------------------------------------------------------------
 * Make variants of the given parse by replacing the virtual nodes with the given arc.
 * If there are no virtual nodes, the list is empty.
 *
 * @param arc	the arc to be inserted
 * @param p 	the parse to be varied
 * @returns a list of variants of the given parse
 */
List insertArcIntoParse(Arc arc, BeamEntry be, ConstraintNet net, List oldBeam) {
	List results = oldBeam;
	int wid_v, wid_n, old_wid_v; 	// word IDs
	int sid_i, sid_n;       		// slot IDs
	int i, r;
	Parse p = be->p;
	Parse p2;
	List lvs;
	Boolean toobad = FALSE;
	LevelValue lv;
	Number score;
	Number threshold;

	int noOfLvls= inputCurrentGrammar->noOfLevels;

	wid_n= arc->from;
	sid_n= wid_n * noOfLvls;

	// for all virtual nodes
	for(wid_v= net->lexemgraph->ve->vStartID; wid_v< net->lexemgraph->max ; wid_v++) {

		// the virtual nodes in the LexemGraph are already pushed back, but we still need the old index at some points
		old_wid_v = wid_v -1;

		if(! bvElement(be->virtualNodeAvailable, old_wid_v) ||
				! bvElement(be->virtualNodeUsed, old_wid_v)) {
			// the node is not available or not used, so there is not point in replacing it with the new word
			continue;
		}

		threshold=  max(
				beamGetHighestScore(results) * bsBeamToleranceFactor,
				beamGetLowestScore(results, bsBeamWidth));

		p2= parseClone(p, FALSE); // no need to decorate it

		// push references to virtual nodes
		for(i= 0; i< vectorSize(p2->verticesStructure) ; i++) {
			r = (int) vectorElement(p2->verticesStructure, i);
			if(r >= wid_n) {
				vectorSetElement(p2->verticesStructure, (void *)r+1, i);
			}
		}

		toobad=FALSE;
		score= 1.0;

		// we have to
		// 1. replace the regent of all edges going to the node to be replaced
		// 2. replace the edge going out from the virtual node with a parking edge
		// 3. add an edge going out from the new node

		// we do step 1 and 2 in one for loop
		// for all old slots
		for(sid_i=0; sid_i< vectorSize(p->verticesStructure); sid_i++) {

			if( ((int)vectorElement(p->verticesStructure, sid_i)) == old_wid_v ) {
				// step 1:
				// This edge points towards the virtual node to-be-replaced.
				// Change its head so it points to the new word instead.
				vectorSetElement(p2->verticesStructure, (void *)wid_n,
						sid_i); // virtual nodes are not pushed back yet in this vector, so we can still use the old id

				// keep the label the same, nothing to do there

				// check max score
				// parse->verticesStructure starts at 0, as do the GraphemNode numbers, no need to add or subtract something
				lvs = cnGetAccordingLVs( net, // net has already the new IDs,
						sid_i % noOfLvls, 		// level
						(sid_i/noOfLvls >= wid_n)?(sid_i/noOfLvls+1):(sid_i/noOfLvls) , 	// dependent (wID), we get the new id for virtual nodes by adding 1
						vectorElement(p->verticesLabels, sid_i),	// label
						wid_n);					// regent (wID)

				lv= bestScoringLVfromList(lvs);
				listDelete(lvs);
				// we don't know the best possible combination of lv's,
				// but we can compare the ideally best combination to the threshold
				score= score * (lv?lv->score:0.0);
				if(score < threshold || score == 0.0) {
					// local score already too bad, abort this variant
					parseDelete(p2);
					toobad=TRUE;
					break; // for all old slots, we break for all virtual nodes later via 'toobad'
				}
			} else if (sid_i / noOfLvls == old_wid_v) { // virtual nodes are not pushed back yet in this vector, so have to use the old id
				// step 2:
				// This edge's dependent is the virtual node to-be-replaced.
				// We replace this edge, so that the virtual node is now unused,
				// i.e., the regent becomes NIL and the label is empty.
				vectorSetElement(p2->verticesStructure, (void *) -1 , sid_i); // the NIL regent, its wID is -1 here
				vectorSetElement(p2->verticesLabels   , "", sid_i);			  // empty label
			}
		}

		if(toobad) {
			// the multiplied best scores are below the threshold, skip to next virtual word
			continue;
		}

		// step 3:
		// finally we insert the new edges for the new word

		// first check max score, we only check level 0 here
		lvs = cnGetAccordingLVs( net, // net has already the new IDs,
				0, 		// level
				wid_n, 	// dependent (wID),
					  vectorElement(p->verticesLabels,    old_wid_v*noOfLvls +0),	// label
				(int) vectorElement(p->verticesStructure, old_wid_v*noOfLvls +0) );	// regent (wID)
		lv= bestScoringLVfromList(lvs);
		listDelete(lvs);
		if(!lv || lv->score * score < threshold) {
			continue;
		}

		// for all levels
		for(i=0; i < noOfLvls; i++) {
			vectorInsertElement(
				p2->verticesStructure,
				vectorElement(p->verticesStructure, old_wid_v*noOfLvls +i),
				wid_n* noOfLvls + i);
			vectorInsertElement(
				p2->verticesLabels,
				vectorElement(p->verticesLabels, old_wid_v*noOfLvls +i),
				wid_n* noOfLvls + i);
		}

		// OK, we got a valid parse, add it to the beam
		results = addParseToBeam(p2, bsBeamWidth, bsBeamToleranceFactor, results, net);

//		cdgPrintf(CDG_INFO, "This parse was generated by insertArcIntoParse\n");
//		printAbstractParse(p2, net->lexemgraph);
//		cdgPrintf(CDG_INFO, "------------------------------------\n");

		parseDelete(p2);

	} // end for all virtual nodes

	return results;
}

/* ---------------------------------
 * Make variants of the given parse.
 *
 * @param arc	the arc to be appended
 * @param p 	the parse to be varied
 * @returns a list of variants of the given parse
 */
List beAddArc(BeamEntry be, Arc arc, ConstraintNet net, List oldBeam) {
	List results = oldBeam;

	if(useVirtualNodes) {
		results = insertArcIntoParse(arc, be, net, results);
		results = appendArcToParse(arc, be->p, net, be, results, TRUE, NULL);

	} else {
		results = appendArcToParse(arc, be->p, net, be, results, TRUE, NULL);
	}

	return results;
}


/*
 * returns a new beam search state
 * the constraintNet has to be set separately
 * virtual node specifications are also not set here
 */
BeamSearchState bssNew() {
	BeamSearchState bss = memMalloc( sizeof(BeamSearchStateStruct));
	bss->entries = vectorNew(bsBeamWidth);
	bss->net = NULL;
	bss->virtualNodeTypes = vectorNew(10);
	return bss;
}

/*
 * This method marks as available
 * - all used entries
 * - and the first unused occurrence
 * for each virtual node specification
 *
 */
void markAvailableVirtualNodes(ConstraintNet net, BeamEntry be) {
	int i;
	Boolean foundUnusedOne;
	List l;
	Vector vns = net->lexemgraph->ve->virtualNodes;
	GraphemNode gn;

	be->virtualNodeAvailable = bvNew(net->lexemgraph->max);
	// bvSetAllElements(be->virtualNodeAvailable, TRUE);

	for(i= 0; i < vectorSize(vns); i++ ) {
		VirtualNode vn = vectorElement(vns, i);
		l= vn->graphemNodes;
		foundUnusedOne= FALSE;
		while(l) {
			gn = l->item;
			l  = l->next;
			if(!bvElement(be->virtualNodeUsed, gn->no)) {
				if(!foundUnusedOne) {
					foundUnusedOne= TRUE;
					// this last one is still available
					bvSetElement(be->virtualNodeAvailable, TRUE, gn->no);
				} else {
					// we don't need any more of those!
					bvSetElement(be->virtualNodeAvailable, FALSE, gn->no);
				}
			} else {
				bvSetElement(be->virtualNodeAvailable, TRUE, gn->no);
			}
		}
	}

}


/* -------------------------------------------------------------------------
 * Execute one step of beam search in the following way:
 * For every parse in the beam, all possible continuations are generated and evaluated via the grammar
 * the best n are used for the next beam front
 *
 * configure n via bsBeamWidth
 *
 */
BeamSearchState beamSearchStep(BeamSearchState prevBss, Arc newArc, Boolean complete)
{
	if(bsVerbosity>=1)
		cdgPrintf(CDG_INFO, "---------------------------------------------------------\n");

	List newEntries = listNew();
	int i,j;
	BeamEntry be;
	List l;

	BeamSearchState bss = (BeamSearchState) memMalloc(sizeof(BeamSearchStateStruct));

	bss->entries = NULL; // we will fill it in the end
	bss->virtualNodeTypes = vectorClone(prevBss->virtualNodeTypes);
	bss->net = prevBss->net;

	// handle virtual nodes in the net and in the beam
	if(useVirtualNodes) {

		// first detect, whether we need more virtual node, depending on the beam entries
		for(i=0; i< vectorSize(prevBss->entries); i++) {
			be = vectorElement(prevBss->entries, i);
			cnManageVirtualNodes(bss->net, be->best, NULL, NULL);
		}

		// then determine which ones are used in which beam entry
		for(i=0; i< vectorSize(prevBss->entries); i++) {
			GraphemNode gn;
			be = vectorElement(prevBss->entries, i);
			be->virtualNodeUsed = bvNew(bss->net->lexemgraph->max);
			// we start at zero, but it doesn't really matter what value we enter for non-virtual words
			for(j=0; j < vectorSize(bss->net->lexemgraph->graphemnodes); j++) {
				gn = vectorElement(bss->net->lexemgraph->graphemnodes, j);
				if(gnIsUnusedVirtualNode(gn,be->best)) {
					bvSetElement(be->virtualNodeUsed, FALSE, j);
				} else {
					bvSetElement(be->virtualNodeUsed, TRUE, j);
				}
			}
		}

		// and we mark which VNs should be available for each beam entry, as they share the net but might not need
		// the additional entries
		for(i=0; i< vectorSize(prevBss->entries); i++) {
			be = vectorElement(prevBss->entries, i);
			markAvailableVirtualNodes(bss->net, be);
		}
	}

	if(complete) {
		bss->net->lexemgraph->complete = TRUE;
	}
	// now we add the new arc to the net
	bss->net = extendNet(prevBss->net, newArc, NULL);
	bss->net = postExtendNetCleanup(bss->net, NULL);
	// ATTENTION: expanding the net made some indices in the old parses invalid,
	// namely the indices of virtual nodes, as the new arc is inserted in front of them

	// for all entries in the beam ...
	for(i=0; i< vectorSize(prevBss->entries); i++) {
		if(bsVerbosity>=2) {
			cdgPrintf(CDG_INFO, "----------------------\n");
			cdgPrintf(CDG_INFO, "Beam entry #%d\n", i+1);
		}

		be = vectorElement(prevBss->entries, i);

		// ... find all possible ways to add the new arc to the current parse.
		newEntries = beAddArc(be, newArc, bss->net, newEntries);
	}

	bss->entries = listToVector(newEntries);

	if(!bss->entries) {
		cdgPrintf(CDG_WARNING,"WARNING: empty beam \n");
		bss->entries= vectorNew(10);
	}

	l= newEntries;

	if(bsVerbosity>=1) {
		cdgPrintf(CDG_INFO, "------------------------------------\n");
		while(l) {
			be = l->item;
			parsePrint2(be->p, TRUE, frobIgnoreThreshold);
			l=l->next;
		}
		int n = vectorSize(bss->entries);
		cdgPrintf(CDG_INFO,"INFO: There %s %d entr%s in the beam \n", n==1?"is":"are", n, n==1?"y":"ies");
		if(newEntries)
			cdgPrintf(CDG_INFO," Best  score: %.5f \n", bScore(((BeamEntry)newEntries->item)->best->b));
		cdgPrintf(CDG_INFO," Worst score: %.5f \n", bScore(((BeamEntry)listLastElement(newEntries))->best->b));
	}
	listDelete(newEntries);
	return bss;
}

/**
 * same as beamSearchStep but without a previous state
 */
BeamSearchState beamSearchInitialStep(Arc arc, Boolean complete) {
	if(bsVerbosity>=1)
		cdgPrintf(CDG_INFO, "---------------------------------------------------------\n");

	List l, entries;
	int i, j;
	BeamSearchState bss;
	ConstraintNet cn;
	Lattice lat;
	Parse p;
	VirtualNodeSpec spec;

	// make initial state
	bss = bssNew();

	// select the virtual nodes to be used
	if(useVirtualNodes) {
		l= inputCurrentGrammar->vnspecs;
		while(l) {
			spec = l->item;
			for(i= 0; i < spec->copies; i++) {
				vectorAddElement(bss->virtualNodeTypes, spec);
			}
			l= l->next;
		}
	}

	// make lattice
	lat = (Lattice) memMalloc(sizeof (LatticeStruct));
	lat->id = strPrintf("wordgraph0");
	lat->filename = strRegister("*BS*");
	lat->lineNo = 0;
	lat->lookahead = listNew();
	lat->arcs = listAppendElement(listNew(), arc);
	inputScanLattice(lat);

	if(bsVerbosity>=3)
		cdgPrintf(CDG_INFO,"Building initial CNet\n");
	cn = cnBuild(lat, TRUE, complete);
	if(useVirtualNodes) {
		cnInitVirtualNodes(cn, inputCurrentGrammar->vnspecs, NULL, NULL);
	}
	cn = postExtendNetCleanup(cn, NULL);
	bss->net= cn;

	BeamEntry initialEntry = memMalloc(sizeof(BeamEntryStruct));
	initialEntry->best=NULL;
	initialEntry->virtualNodeAvailable = bvNew(listSize(inputCurrentGrammar->vnspecs));
	bvSetAllElements(initialEntry->virtualNodeAvailable, TRUE);
	initialEntry->virtualNodeUsed = bvNew(listSize(inputCurrentGrammar->vnspecs));
	bvSetAllElements(initialEntry->virtualNodeUsed, FALSE);

	p= parseNew(NULL);
	p->verticesLabels= vectorNew(10);
	p->verticesStructure= vectorNew(10);

	p->words= vectorNew(0);
//	w = newWord();
//	w->from = ln->arc->from;
//	w->to = ln->arc->to;
//	w->word = ln->lexem->word;
//	w->description = strRegister(ln->lexem->description);
//	vectorAddElement(p->words, w);

	initialEntry->p=p;

	if(useVirtualNodes) {
		// add edges for the virtual nodes to the parse
		for(i=0; i< vectorSize(bss->virtualNodeTypes); i++) {
			for(j=0; j< inputCurrentGrammar->noOfLevels; j++) {
				vectorAddElement(p->verticesLabels, "");
				vectorAddElement(p->verticesStructure, (void *)-1);
			}
		}
	}

	entries = appendArcToParse(arc, p, cn, initialEntry, listNew(), TRUE, NULL);

	bss->entries = listToVector(entries);
	if(!bss->entries) {
		cdgPrintf(CDG_WARNING,"WARNING: empty beam \n");
		bss->entries= vectorNew(10);
	}

	l= entries;
	while(l) {
		parsePrint2(((BeamEntry)l->item)->p, TRUE, frobIgnoreThreshold);
		l=l->next;
	}

	if(bsVerbosity>=1) {
		int n = vectorSize(bss->entries);
		cdgPrintf(CDG_INFO,"INFO: There %s %d entr%s in the beam \n", n==1?"is":"are", n, n==1?"y":"ies");
		if(entries)
			cdgPrintf(CDG_INFO," Best  score: %.5f \n", ((BeamEntry)entries->item)->best->b->hard?0.0:((BeamEntry)entries->item)->best->b->soft);
		cdgPrintf(CDG_INFO," Worst score: %.5f \n", beamGetLowestScore(entries, 0));
	}
	listDelete(entries);

	return bss;
}


/*
 * this function implements the incremental beam search strategy
 */
BeamSearchState beamSearch(Lattice lat) {
	List arcs = lat->arcs;
	BeamSearchState bss;

	bss  = beamSearchInitialStep(arcs->item, (arcs->next || !incrAutoFinalize)?FALSE:TRUE);

	arcs = arcs->next;
	while( arcs ) {
		Arc arc = arcs->item;

		bss = beamSearchStep(bss ,arc, (arcs->next || !incrAutoFinalize)?FALSE:TRUE);

		arcs= arcs->next;
	}
	return bss;
}


/*
 * returns a string containing a sequence of 1s and 0s
 */
void bvPrint(ByteVector bv) {
	int i;
	for(i=0;i < bvSize(bv); i++) {
		cdgPrintf(CDG_INFO, "%s",bvElement(bv,i)?"1":"0");
	}
	cdgPrintf(CDG_INFO, "\n");
}
/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */
