/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup Cdgdb Cdgdb - Get lexical entries out of a berkeley database
 * @author Othello Maurer
 * @date unknown
 *
 * This file provides an interface for retrieving lexical entries out of a
 * berkeley database. The database file contains only references to an
 * underlying cdg file which actually holds the entries. For loading the
 * requested lexical entries, a temp file will be created out of the cdg
 * file, containing only these entries. This file will then be parsed with
 * \b inpuLoad() into the memory.
 * @{
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <config.h>
#include <unistd.h>
#include <time.h>
#include <libgen.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/resource.h>
#include <blah.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <db.h>
#include <sys/times.h>
#include "cdgdb.h"
#include "cdg.h"
#include "hook.h"
#include "input.h"
#include "parsing.h"

/* -- VARIABLES ------------------------------------------------------------ */

/** database handle. The handle to the database struct */
static DB *database;

/** the underlying cdg file. This is actually the file, where the lexical
  * entries are stored in. */
static FILE *cdgstream;

/** file name of the file with CDG input in it */
static String dbFileName;

/** file name of the index */
static String dbIndexName;

/** age of that file at the time of the last opening */
static time_t dbAge;

/** the filename of the temporary file. This temporary file will contain
  * a set of lexical entries that can then be parsed into the memory. */
static String tmpFilename = NULL;

/** table of words already loaded. Holds the word which are yet retrieved
  * from the database, so that the database won't be questioned for them */
static Hashtable done = NULL;

/** a database entry.
  * This is an entry for the database which holds a key and the byte position
  * of its lexical entry in the underlying cdg file. */
typedef struct {
  String key;  /**< the key of the lexical entry. */
  int oBegin; /**< offset of its Beginning in the file */
  int oEnd; /**< offset of its End in the file */
} entryStruct;

/** a pointer to the entryStruct */
typedef entryStruct *dbEntry;


/* -- FUNCTIONS ------------------------------------------------------------ */

/* -- IMPLEMENTATION ------------------------------------------------------- */

/* -------------------------------------------------------------------------
 * opens a database.
 * this function opens a database, specified by the parameter 'filename'
 * and stores the handle in the static variable 'database'. If there is no
 * database file with this filename, a new database is created.
 * @param filename the filename of the database file
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbOpenIndexFile(String filename)
{
  int ret = 0;		     // return code for database operations

  // create database handle
  if ((ret = db_create(&database, NULL, 0)) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR (db_create): %s\n", db_strerror(ret));
    return FALSE;
  }
  // setting database to accept more than one data item for one key
  // sort order is lexical (this is the default)
  database->set_flags(database, DB_DUP);

  /* open database 'filename' */
#if (DB_VERSION_MAJOR < 4 || \
     DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 0)
  if ((ret =
       database->open(database, filename, NULL, DB_BTREE, DB_RDONLY, 0664)) != 0) {
#else
  if ((ret =
       database->open(database, NULL, filename, NULL, DB_BTREE, DB_RDONLY, 0664)) != 0) {
#endif
    cdgPrintf(CDG_ERROR, "ERROR (database->open): %s\n", db_strerror(ret));
    database = NULL;
    return FALSE;
  }
  dbIndexName = strRegister(filename);
  return TRUE;
}


/* -------------------------------------------------------------------------
 * open the cdg file.
 * opens the cdg file, which the database is indexing
 * you have to ensure that it is the right one, otherwise you will get crap
 * @param filename the name of the cdg file
 * @returns TRUE if success, else FALSE
 */

Boolean dbOpenCdgFile(String filename) {
  struct stat buf;

  cdgstream = fopen(filename, "r");
  if(!cdgstream) {
    database = NULL;
  }

  stat(filename, &buf);
  dbAge = buf.st_mtime;
  dbFileName = strRegister(filename);

  return (cdgstream != NULL);
}

/* -------------------------------------------------------------------------
 * close the database.
 * writes database to disk and closes it.
 * @returns TRUE if success, FALSE otherwise
 */

Boolean dbClose(void)
{
  int ret = 0;		     // return code for database operations

  if (database == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: No open database found for closing\n");
    return FALSE;
  }

  if ((ret = database->close(database, 0)) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR (database->close): %s\n", db_strerror(ret));
    return FALSE;
  }
  database = NULL;
  return TRUE;
}

/* ----------------------------------------------------------------------
   Open the data base.

   NAME is the base name of the cdg and the index file.
*/
Boolean dbOpen(String name) {
  struct stat buffer;

  dbFileName = strAppend(name, ".cdg", NULL);
  dbIndexName = strAppend(name, ".db", NULL);

  /* Try very hard to find the files. If they aren't in the current
     directory, maybe they are in another directory we loaded files
     from. */
  if(stat(dbFileName, &buffer)) {
    /* remove directory component */
    String filebase = strCopy(basename(dbFileName));
    String indexbase = strCopy(basename(dbIndexName));
    String separator = strRegister("/");
    List l;
    for(l = inputCurrentGrammar->files; l != NULL; l = l->next) {
      String file = strCopy(l->item);
      String dir = dirname(file);
      String filealt = strAppend(dir,separator,filebase, NULL);
      String indexalt = strAppend(dir,separator,indexbase, NULL);
      if(!stat(filealt, &buffer) && !stat(indexalt, &buffer)) {
        dbFileName = filealt;
        dbIndexName = indexalt;
        break;
      }
    }
  }

  /* open the cdg file */
  if (!dbOpenCdgFile(dbFileName)) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot open %s\n", dbFileName);
    return FALSE;
  }
  /* open the database file */
  if (!dbOpenIndexFile(dbIndexName)) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot open %s\n", dbIndexName);
    return FALSE;
  }
  return TRUE;
}


/* -------------------------------------------------------------------------
 * checks if there is a database handle available.
 * @returns TRUE if there is a database opened, else FALSE
 */
Boolean dbAvailable(void) {
  struct stat buf;
  time_t age;

  if(!database) {
    return FALSE;
  }

  /* If the file that we read data from is newer than it was
     when we last opened it, this means it has been recreated
     while we ran. To be safe we re-open the data base. */
  stat(dbFileName, &buf);
  age = buf.st_mtime;
  if(age > dbAge) {
    cdgPrintf(CDG_WARNING, "WARNING: data base has changed!\n");
    cdgPrintf(CDG_WARNING, "WARNING: re-opening data base.\n");
    dbClose();
    dbOpenIndexFile(dbIndexName);
    dbOpenCdgFile(dbFileName);
  }

  return TRUE;

}

/* -------------------------------------------------------------------------
 * constructor for dbEntry objects.
 * @returns a new dbEntry
 */
dbEntry newDbEntry() {
  dbEntry entry = memMalloc(sizeof(entryStruct));
  entry->key = NULL;
  entry->oBegin = 0;
  entry->oEnd = 0;
  return entry;
}


/* -------------------------------------------------------------------------
 * gets lexical entries from the database.
 * gets a result list of LexiconItem matching the given word (the key) from
 * the database. This function succeeds only once for a given word form.
 * @param word the word which is the key
 * @returns a list of dbEntry struct, all of them having the same key and
 *          NULL if there was an Error
 */
List dbGetEntries(String word)
{
  int ret = 0;		     // return code for database operations
  DBT key, value;	     // DBT structs for key and value
  DBC *cursorp;		     // cursor for getting duplicates
  dbEntry ent;
  List entries = NULL;	     // the return list

  if (!database) {
    cdgPrintf(CDG_ERROR, "ERROR: No database opened\n");
    return entries;
  }

  // create cursor for getting duplicates
  database->cursor(database, NULL, &cursorp, 0);

  memset(&key, 0, sizeof (key));
  memset(&value, 0, sizeof (value));
  key.data = word;
  key.size = strlen(word);

  // position the cursor at the right key and get the first item
  if ((ret = cursorp->c_get(cursorp, &key, &value, DB_SET)) != 0) {
    if (ret == DB_NOTFOUND) {
      return NULL;
    } else {
      cdgPrintf(CDG_ERROR, "ERROR (cursorp->c_get): %s\n", db_strerror(ret));
      return NULL;
    }
  }
  ent = newDbEntry();
  ent->key = strCopy(key.data);
  ent->oBegin = *((int *)value.data);
  ent->oEnd = *((int *)value.data + 1);
  entries = listAppendElement(entries, ent);
  // now get all duplicates
  while ((ret =
	  cursorp->c_get(cursorp, &key, &value, DB_NEXT_DUP)) != DB_NOTFOUND) {
    if (ret != 0) {
      cdgPrintf(CDG_ERROR, "ERROR (cursorp->c_get): %s\n", db_strerror(ret));
      break;
    }

    ent = newDbEntry();
    ent->key = strCopy(word);
    ent->oBegin = *((int *)value.data);
    ent->oEnd = *((int *)value.data + 1);
    entries = listAppendElement(entries, ent);
  }
  return entries;
}

/* -------------------------------------------------------------------------
 * loads lexical entries into the memory.
 * generates a temp file which holds lexical entries and parses them into
 * the memory
 * @param entries a list of dbEntry structs you want to parse
 * @returns True if success, else FALSE
 */
Boolean dbLoadEntries(List entries)
{
  dbEntry ent;
  List l;
  FILE *tmpstream;
  unsigned long mode;
  tmpFilename = strPrintf("/tmp/dbload.%d.%d.cdg", getpid(), getuid());

  /* if there are no entries for loading*/
  if(!entries) {
    return FALSE;
  }

  /* open a temporary file */
  tmpstream = fopen(tmpFilename, "w");



  if(tmpstream == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR (dbLoadEntries): cannot generate tempfile, errno: %d\n", errno);
    return FALSE;
  }
  for (l = entries; l != NULL; l = l->next) {
    ent = l->item;
    fseek(cdgstream, (ent->oBegin), SEEK_SET); // seek to begin of entry
    while(ftell(cdgstream) != ent->oEnd) {
      putc(getc(cdgstream), tmpstream);
    }
    putc((int)'\n', tmpstream);
  }
  listForEachDelete(entries, memFreeFunction);
  fclose(tmpstream);

  /* suppress uninteresting `loading 58749857.cdg' message */
  mode = hkVerbosity;
  hkVerbosity = hkVerbosity & (~CDG_INFO);
  inputLoad(tmpFilename);
  hkVerbosity = mode;

  /* promote new data to current data */
  mergeInput(inputCurrentGrammar, parseResult);
  inputCacheInput(inputCurrentGrammar);
  memFree(parseResult);
  parseResult = NULL;

  return TRUE;
}

/* -------------------------------------------------------------------------
 * wrapper for dbLoadEntries().
 * generates a temp file which holds lexical entries and parses them into
 * the memory
 * @param key the key of the lex entries which have to be parsed
 * @returns TRUE if success, else FALSE
 */

Boolean dbLoad(String key) {
  List records = NULL;
  if(!dbAvailable()) {
    return FALSE;
  }

  /* detect duplicate query */
  if(hashGet(done, strRegister(key))) {
    return TRUE;
  }
  records = dbGetEntries(key);

  /* provide interesting `loading form "Schnabeltier"' message */
  if(records) {
    cdgPrintf(CDG_INFO, "INFO: loading form `%s'\n", key);

    /* remember never to read this key again */
    hashSet(done, strRegister(key), (Pointer)TRUE);
    return dbLoadEntries(records);
  }
  return TRUE;
}

/* -------------------------------------------------------------------------
   Another wrapper for dbLoadEntries().

   It loads all known lexicon items for words in the list FORMS. This has
   exactly the same effect as calling dbLoad() repeatedly, but is faster
   because there will be only one file system operation instead of n.
*/
Boolean dbLoadAll(List forms) {
  List records = NULL;
  List l;
  int i = 0;

  if(!dbAvailable()) {
    return FALSE;
  }
  for(l = forms; l != NULL; l = l->next) {
    String form = l->item;

    /* detect duplicate query */
    if(hashGet(done, form)) {
      continue;
    }

    records = listAppendList(records, dbGetEntries(form));
    hashSet(done, strRegister(form), (Pointer)TRUE);
    if(0 == (++i % 1000)) {
      cdgPrintf(CDG_PROGRESS, "%d forms...\n", i);
    }
  }
  cdgPrintf(CDG_INFO, "INFO: loading %d items\n", listSize(records));

  return dbLoadEntries(records);
}


/* -------------------------------------------------------------------------
 * Initialize
 */

void dbInitialize(void) {
  done = hashNew( 500, 0.8,
                  hashStringHashFunction,
                  hashStringEqualFunction );
}

/* -------------------------------------------------------------------------
 * finalize
 * close the database, delete \b done hash, remove temporary file
 */
void dbFinalize(void)
{
  if (dbAvailable())
    dbClose();
  if (tmpFilename) {
    remove(tmpFilename);
  }
  hashDelete(done);
}


/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */
