/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*
   File: frobbing.c
   Author: Kilian Foth
   Purpose: heuristical construction of solutions
   Birth: 25.9.98
 */

/*

   This module provides a heuristic solution method for constraint nets
   based on the stepwise transformation of an invalid analysis to a valid
   one. It uses the constraints violated by an analysis to select LVs for
   better analysis. It also performs a restricted kind of pruning on the
   net that never makes mistakes.

   ``frobnicate /frob'ni-kayt/ /vt./ [usually abbreviated to frob]
   To manipulate or adjust, to tweak. [...]

   Usage: frob, twiddle, and tweak sometimes connote points along a
   continuum. `Frob' connotes aimless manipulation; `twiddle' connotes
   gross manipulation, often a coarse search for a proper setting;
   `tweak' connotes fine-tuning. If someone is turning a knob on an
   oscilloscope, then if he's carefully adjusting it, he is probably
   tweaking it; if he is just turning it but looking at the screen, he
   is probably twiddling it; but if he's just doing it because turning
   a knob is fun, he's frobbing it.''

   [from the Hackers Jargon File]

   This is the first module (others will follow, it is to be hoped) in
   the CDG system to construct analyses of lattices by successive
   repair. The principal reference on it is my own thesis, but since
   it does not go into technical detail, this file contains what may
   be more illuminating information about the actual implementation.

   Different strategies of transformation are available. Each of them
   resides as an entry in the array frobMethods[]. All strategies use
   the same way of building an initial Analysis for a Problem, but
   the actual transformation is done by the respective function *repair.

*/

 /*

    TO DO:

    optimizeDomains() could probably save a lot of time if it didn't
    explicitly pJudge() every Patch it considers. That way, if there
    are 321 patches that make the same lexeme swap, the swap has to be
    computed 321 times. Probably a two-level strategy would be better:
    if a Patch introduces new lexemes, compute and store the Analysis
    with the new lexemes, then use that to judge the Patch. The
    problem is how to quickly find Patches with the same lexeme swaps.

    enhance combined() to perform ! steps on the solution until
    optimality has been proven (or... some other condition applies
    that I can't quite define offhand.) That should remove any
    annoying suboptimal solutions.

    perhaps change dSearch() so that it will prune the space on its
    own by breaking up the list of domains?

    We should seriously study what effect keeping the history even
    after global improvements has on the batch search. Apparently
    a good one. But is it possible that the advantage disappears if
    taken too far?

  */

/* ---------------------------------------------------------------------- */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cdg.h"
#include "chunker.h"
#include "constraintnet.h"
#include "eval.h"
#include "frobbing.h"
#include "hook.h"
#include "lexemgraph.h"
#include "netsearch.h"
#include "parse.h"
#include "set.h"
#include "tagger.h"
#include "timer.h"
#include "transform.h"
#include "write.h"

/* -- FORWARD DECLARATIONS ---------------------------------------------- */
void minConflict(Problem P);
void errorDriven(Problem P);
void combined(Problem P);
void manual(Problem P);
void genet(Problem P);
void dynamic(Problem P);
void threefold(Problem P);

/* -- VARIABLES --------------------------------------------------------- */

/* Various strategies are defined for transforming analyses heuristically.
   Each of these strategies is modeled as a structure of type FrobMethod.

   Pretty much the only ones that are any good are `combined', which
   combines the repair process with an error-driven version of pruning, and
   `manual', which lets you do everything to an Analysis that is not
   totally objectionable. */
FrobMethod frobMethods[] = {
  {"min-conflict", minConflict, "Minton92"},
  {"error-driven", errorDriven, "error-based repair"},
  {"combined", combined, "repair + pruning + search"},
  {"manual", manual, "user-driven"},
  {"genetic", genet, "evolutionary"},
  {"dynamic", dynamic, "combined applied to subproblems"},
  {"threefold", threefold, "three-phase transformation"},

  /* end of array marker */
  {NULL, NULL, NULL}
};

int stepcount = 0;
void frobResetStepcount() {
	stepcount=0;
}

long int evaluations = 0;      /* # of evaluations done in the entire net */

/* FALSE during batch frobbing, TRUE during interactive search. */
Boolean interactive = TRUE;

/* governs level of debug output */
int debug = 0;

/* Sometimes we want to measure the distance of an Analysis to the
   optimal Analysis. Since this optimal Analysis is generally unknown,
   we first have to construct one Analysis to measure another against
   it. As the module does not export any complete Analyses, we use
   this global variable to hold the Analysis between calls to
   frobbing(). It is merely a list of LV indices. */
List benchmark = NULL;


/* pressure to apply to all problems */
Number globalPressure = 0.0;

/* limit beyond which violations are not to be attacked */
float frobIgnoreThreshold = 1.0;

/* frobbing stops if this goal is not reachable anymore */
Number frobbingGoal = 0.0;

/* commands to pass to manual frobbing mode */
String frobCommands = NULL;

/* The following variables only control the behaviour of dynamic(). */

/* how should we define subproblems? */
String dynamicBoundaries = NULL;

/* should we freeze trees computed by dynamic frobbing? */
Boolean freezeSubTrees = TRUE;

/* must subtrees stay within their subrange? */
Boolean strictSubproblems = TRUE;

/* minimal size of subproblems during dynamic() */
int minSize = 6;

/* maximal size of subproblems during dynamic() */
int maxSize = 20;

/* Should we be greedy, i.e. should we use OD_IMMEDIATE? */
Boolean greedy = TRUE;

/* How many frobs are allowed per conflict? */
int max_steps = 0;


/* -- FUNCTIONS --------------------------------------------------------- */

/* Functions setting speedup options
   ================================= */

void frobSetAgendaSize(int x) { localSize = x; }
void frobSetBeamWidth(int x) { beamWidth = x; }
void frobSetPressure(Number x) { globalPressure = x; }




/* Functions dealing with Analyses
   =============================== */

/* --------------------------------------------------------------------
   Store an Analysis so that distance measurements can be made
   against a second run.

   This function creates a List holding the indexWRTNet of every
   LV in a and assigns it to the global variable benchmark.
   If benchmark is already set, it is reset instead.
 */
void aStore(Analysis a) {

  int i, j, k;
  LevelValue lv;

  /* Actually, what this does is to toggle the status of benchmarking.
     If there already is a benchmark analysis, it is erased. */
  if (benchmark) {
    listDelete(benchmark);
    benchmark = NULL;
    cdgPrintf(CDG_INFO, "Deleting benchmark analysis\n");
    return;
  }

  /* But if there is none, the current analysis becomes the benchmark. */
  cdgPrintf(CDG_INFO, "Storing: ");

  for(k = 0; k < a->P->nrLevels; k++) {
    for (i = 0; i < vectorSize(a->slots); i++) {
      lv = (LevelValue) vectorElement(a->slots, i);
      if (!lv)
        continue;
      if(lv->level->no != k) continue;

      j = lv->indexWRTNet;
      benchmark = listAppendElement(benchmark, (Pointer) j);
      cdgPrintf(CDG_INFO, "#%d ", j);
    }

  }

  cdgPrintf(CDG_INFO, "\n");

}

/* ----------------------------------------------------------------------
   Show distance to the benchmark.

   This is simply the number of LVs in a that are not also in the benchmark
   Analysis.
*/
void printDistance(unsigned long mode, Analysis a) {

  int i, j, result = 0;
  LevelValue lv;

  if (!benchmark)
    return;

  for (i = 0; i < vectorSize(a->slots); i++) {
    lv = (LevelValue) vectorElement(a->slots, i);
    if (!lv)
      continue;

    j = lv->indexWRTNet;
    if (listContains(benchmark, (Pointer) j)) {
      result++;
    }
  }

  cdgPrintf(mode, " distance %d after %lu ms",
            vectorSize(a->slots) - result, timerElapsed(a->P->profileTime));
}

/* ----------------------------------------------------------------------
   Print # of links differing between two Analyses
 */
void aReport(Analysis a, Analysis b) {

  int i;
  LevelValue lva,lvb;
  int links = 0, correct = 0;

  /* can't compare Analyses of different Problems */
  if(a->P != b->P) {
    return;
  }

  for(i = 0; i < b->P->nrSlots; i++) {
    lvb = vectorElement(b->slots,i);
    if(!lvb) {
      continue;
    }
    links++;
    lva = vectorElement(a->slots,i);
    if(lva != NULL &&
       lva->indexWRTNet == lvb->indexWRTNet) {
      correct++;
    }
  }
  cdgPrintf(CDG_INFO, "  %d of %d links correct after %lu ms",
            correct, links, timerElapsed(a->P->profileTime));
}

/* ----------------------------------------------------------------------
   Show time used so far.

   Displays the time spent before the algorithm terminated, how long it
   took to find the optimal Analysis, and how many steps were tried
   out and executed in the process.

 */
void showTime(Problem P)
{
  double totalDuration, bestFoundDuration, softFoundDuration;

  totalDuration = timerElapsed(P->profileTime) / 1000.0;
  bestFoundDuration = P->lastSolFound / 1000.0;
  softFoundDuration = P->firstSoftFound / 1000.0;

  cdgPrintf(CDG_PROFILE, "PROFILE: %.3fs spent frobbing \n",
            totalDuration);
  cdgPrintf(CDG_PROFILE, "PROFILE: %.3fs spent before finding this analysis\n",
            bestFoundDuration);
  cdgPrintf(CDG_PROFILE, "PROFILE: %.3fs spent before finding a soft analysis\n",
            softFoundDuration);
  cdgPrintf(CDG_PROFILE, "PROFILE: %d repair steps and %d repair attempts\n",
            stepcount, P->judgeCounter);

  if(P->net) {
    cdgPrintf(CDG_PROFILE, "PROFILE: frobbing performed %li constraint evaluations.\n",
              P->net->evalUnary + P->net->evalBinary - evaluations);
  }
}




/* ----------------------------------------------------------------------
   Find a Patch that removes the conflict at position WHERE in the list of
   current conflicts from A.

   The result can be NULL.

   FLAGS determines the behaviour of this function and the helper functions
   it calls (see frobbing.h).
*/
Patch attackConflict(Analysis a, ConstraintViolation cv, int flags) {

  Vector which = NULL;
  Patch result = NULL;
  int w, x, y, z;

  /* allow all kinds of changes */
  flags |= OD_ALLOWALL;

  if(debug) {
    cdgPrintf(CDG_DEBUG, "Attacking ");
    cvPrint(CDG_DEBUG, cv);
    cdgPrintf(CDG_DEBUG, "... ");
  }

  x = lvIndex(cv->lv1);
  if (cv->lv2) {
    y = lvIndex(cv->lv2);
    if (listSize(a->P->domains[y]) < listSize(a->P->domains[x])) {
      w = x;
      x = y;
      y = w;

    }
  } else {
    y = -1;
  }


if (cv->lv3) {
    z = lvIndex(cv->lv3);

	
	if (listSize(a->P->domains[z]) < listSize(a->P->domains[y])) 
	{
		w = y;
		y = z;
		z = w;
	}

		if (listSize(a->P->domains[y]) < listSize(a->P->domains[x])) 
	{
		w = x;
		x = y;
		y = w;
	
	}
  } else {
    z = -1;
  }  

  /* With a context-sensitive constraint, we can not guarantee
     that changing *any* LV will fulfill it. It might actually need
     several different LVs in the structure.

     But we assume that most of them just check for some kind of existence,
     so that it is worthwhile to search all unary patches. */

  a->P->removedConflicts = listAppendElement(a->P->removedConflicts, cvClone(cv));

  if(cv->constraint->is_context_sensitive) {
    int levelno = cv->lv1->level->no;
    int i;
    which = vectorNew(20);
    for(i = a->subStart; i < a->subStop; i++) {
      int index = a->P->nrLevels * i + levelno;
      vectorAddElement(which, (Pointer)index);
    }
    if(cv->lv2) {
      vectorAddElement(which, (Pointer)lvIndex(cv->lv2));
    }
      if(cv->lv3) 
	{
		vectorAddElement(which, (Pointer)lvIndex(cv->lv3));
	}
    result = optimizeDomains(a, which, (flags | OD_SINGLE));
  }

  /* otherwise it's easy: frob those places where the error manifests */
  else {

if (cv->lv3) {

		which = vectorNew(3);
		vectorAddElement(which, (Pointer) x);
		vectorAddElement(which, (Pointer) y);
		vectorAddElement(which, (Pointer) z);

		result = optimizeDomains(a, which, flags);
	}

    else if (cv->lv2) {
      which = vectorNew(2);
      vectorAddElement(which, (Pointer) x);
      vectorAddElement(which, (Pointer) y);
      result = optimizeDomains(a, which, flags);
    } else {
      result = optimizeDomains(a, (Pointer)x, flags | OD_INT);
    }
  }

  if(debug) {
    if(result) {
      cdgPrintf(CDG_DEBUG, "DEBUG: returning: ");
      pPrint(CDG_DEBUG, result, TRUE);
      cdgPrintf(CDG_DEBUG, "\n");
    } else {
      cdgPrintf(CDG_DEBUG, "DEBUG: no solution..\n");
    }

  }

  if(which) {
    vectorDelete(which);
  }
  return (result);

}



/* ----------------------------------------------------------------------
   Find next conflict to be attacked.

   This is simply the first conflict in the list that isn't unremovable.
   Because of the way the list ->conflicts is handled, this means that it
   is the hardest recent conflict, or simply the hardest conflict if there
   are no recent conflicts.
*/
int selectConflict(Problem P) {

  int result = 0;
  ConstraintViolation cv;
  List l;

  sortConflicts(P->current);

  for (result = 1, l = P->current->conflicts; l; l = l->next, result++) {
    cv = l->item;
    if (isUnremovable(P, cv))
      continue;
    return result;
  }

  return 0;

}

/* ----------------------------------------------------------------------
   Produce XML performance statistics on frobbing.
 */
void frobWriteXmlStatistics(Problem P) {
  double totalDuration, bestFoundDuration, softFoundDuration;

  totalDuration = timerElapsed(P->profileTime);
  bestFoundDuration = P->lastSolFound;
  softFoundDuration = P->firstSoftFound;

  writeXmlStatistics(1, "frobbing", "general statistics on frobbing",
      "score", (P->best->b->hard) ? (double) 0.0: P->best->b->soft,
      "totalTime", totalDuration,
      "softTime", softFoundDuration,
      "solutionTime", bestFoundDuration,
      "steps", (Number) stepcount,
      NULL);
}

/* Functions implementing FrobMethods
   ================================== */


/* ----------------------------------------------------------------------
   Repair � la Minton92.

   This function implements the transformation strategies with the same
   name. At every step, one variable is selected by selectVariable(). The
   new value that produces the best Analysis (under bCompare()) is
   substituted.

 */
void minConflict(Problem P) {

  Patch p;
  int where;

  Boolean oldinteractive = interactive;
  interactive = FALSE;

  /* clear history list */
  resetProblem(P);

  /* give info */
  bPrint(CDG_INFO, P->current->b);
  cdgPrintf(CDG_INFO, "\n");

  /* iterate */
  while (P->best->b->hard ||
         significantlyGreater(P->globalMaximum, P->best->b->soft)) {

    /* step */
    where = selectVariable(P);

    if (where == -1) {
      break;
    }

    p = optimizeDomains(P->current, (Pointer)where,
                        OD_INT | OD_ALLOWALL | OD_PREDICT | OD_SINGLE);
    stepcount++;

    cdgPrintf(CDG_INFO, "changing variable @%d: ", where);

    /* user break? */
    if (cdgCtrlCTrapped) {
      cdgCtrlCTrapped = FALSE;
      cdgPrintf(CDG_INFO, "\nOK, you take over.\n");
      break;
    }

    /* analysis impossible? */
    if (P->globalMaximum == 0.0) {
      return;
    }

    /* correction possible? */
    if (p) {

      {
        LevelValue lvold = vectorElement(P->current->slots, where);
        LevelValue lvnew = vectorElement(p->selection, 0);

        cdgPrintf(CDG_INFO, "replacing #%d by #%d\n",
                  lvold->indexWRTNet, lvnew->indexWRTNet);
      }

      apply(P->current, p);

      /* new maximum? */
      if (bCompare(P->current->b, P->best->b)) {
        newMaximum(P->current);
        cdgPrintf(CDG_INFO, "*");
      }
    } else {
      break;
    }

    /* show progress */
    bPrint(CDG_INFO, P->current->b);
    printAmbiguity(CDG_INFO, P);
    cdgPrintf(CDG_INFO, "\n");
    cdgFlush();

  }

  interactive = oldinteractive;

}


/* ----------------------------------------------------------------------
   Repair with memory: always insert the value that minimizes
   conflicts without repeating any errors.
 */
void errorDriven(Problem P) {

  Patch p;
  ConstraintViolation cv;
  ConstraintViolation primaryConflict = NULL;
  int where;

  Boolean oldinteractive = interactive;
  interactive = FALSE;

  bPrint(CDG_INFO, P->current->b);
  cdgPrintf(CDG_INFO, "\n");

  /* iterate */
  while (P->best->b->hard ||
         significantlyGreater(P->globalMaximum, P->best->b->soft)) {

    aCheck(P->current, 3);

    /* step */
    where = selectConflict(P);

    /* no progress possible? */
    if (where == 0) {
      cdgCtrlCTrapped = FALSE;
      backTrace(P);
      cdgPrintf(CDG_INFO, "No removable conflicts remain.\n");
      break;
    }

    cv = listNthElement(P->current->conflicts, where);
    if (!primaryConflict) {
      primaryConflict = cvClone(cv);
    }

    p = attackConflict(P->current, cv, OD_PREDICT | OD_NARROW);
    stepcount++;

    /* user break? */
    if (cdgCtrlCTrapped) {
      cdgCtrlCTrapped = FALSE;
      backTrace(P);
      cdgPrintf(CDG_INFO, "OK, you take over.\n");
      break;
    }

    /* analysis impossible? */
    if (P->globalMaximum == 0.0) {
      return;
    }

    /* correction possible? */
    if (p) {

      apply(P->current, p);

      /* new maximum? */
      if (bCompare(P->current->b, P->best->b)) {

        newMaximum(P->current);
        cdgPrintf(CDG_INFO, "*");

        cvDelete(primaryConflict);
        primaryConflict = NULL;

      }
    } else {

      /* skip the conflict whose removal led us here... */
      cdgPrintf(CDG_DEBUG, "DEBUG: couldn't remove ");
      cdgPrintf(CDG_DEBUG, "%s", primaryConflict->constraint->id);
      cdgPrintf(CDG_DEBUG, ", ignoring...\n");

      P->unremovable = listAppendElement(P->unremovable, primaryConflict);
      primaryConflict = NULL;

      /* ...and return to the best Analysis */
      backTrace(P);
    }

    /* show progress */
    bPrint(CDG_PROGRESS, P->current->b);
    printAmbiguity(CDG_PROGRESS, P);
    cdgPrintf(CDG_PROGRESS, "\n");
    cdgFlush();

  }

  backTrace(P);
  interactive = oldinteractive;

}

/*
 * Checks if the Problem can still reach the given score goal
 * criteria checked:
 *
 * global Maximum
 *
 * product of unremovable conflic penalty
 *
 */
Boolean goalReachable(Number goal, Problem P) {
	if(!P->best->b->hard && P->best->b->soft > goal)
		return TRUE; // obviously

	if(P->globalMaximum < goal)
		return FALSE; // limits say it's not possible

	Number score = 1.0;
	List l;
	for(l=P->unremovable; l; l= l->next) {
		ConstraintViolation cv = l->item;
		score *= cv->penalty;
	}

	if(score < goal)
		return FALSE; // unremovable conflicts alone prevent the goal to be reachable

	return TRUE; // otherwise, assume it's possible
}

/* ----------------------------------------------------------------------
   Repair enhanced with pruning: like errorDriven, but restricted to
   Patches with an optimal limit.
 */
void combined(Problem P) {

  Patch p;
  ConstraintViolation cv;
  ConstraintViolation primaryConflict = NULL;
  RepairAttempt currentAttempt = NULL;
  int where;

  Boolean oldinteractive = interactive;
  interactive = FALSE;

  cdgCtrlCTrapped = FALSE;
  bPrint(CDG_INFO, P->current->b);
  printAmbiguity(CDG_INFO, P);
  cdgPrintf(CDG_INFO,"\n");

  /* report on distance to the annotation, if any */
  if(nsAutoCompare && P->annotation) {
    aReport(P->current, P->annotation);
  }

  P->repairAttempts= listNew();

  /* iterate */
  while (P->best->b->hard ||
         significantlyGreater(P->globalMaximum, P->best->b->soft)) {

    /* check for internal errors */
    aCheck(P->current, 3);
#ifdef DEBUGCOMBINED
    cdgPrintf(CDG_DEBUG, "DEBUG: Have \n");
    aPrint(CDG_DEBUG, P->current);
    printConflicts(CDG_DEBUG, P->current);
#endif

    // check if a score goal is still reachable, otherwise stop frobbing
    if(!goalReachable(frobbingGoal, P)) {
    	cdgPrintf(CDG_INFO, "Frobbing goal not reachable, aborting frobbing! \n");
    	break;
    }


    /* step */
    where = selectConflict(P);

    /* no progress possible? */
    if (where == 0) {
      cdgCtrlCTrapped = FALSE;
      backTrace(P);
      cdgPrintf(CDG_INFO, "No removable conflicts remain.\n");
      break;
    }

    cv = listNthElement(P->current->conflicts, where);

    /* no serious penalties remaining? */
    if(cv->penalty >= frobIgnoreThreshold) {
      cdgCtrlCTrapped = FALSE;
      backTrace(P);
      cdgPrintf(CDG_INFO, "No serious conflicts remain.\n");
      break;
    }

    if (!primaryConflict) {
    	P->patchAttempts++;
    	primaryConflict = cvClone(cv);
		currentAttempt = repairAttemptNew(primaryConflict);
    } else {
    	currentAttempt->otherConflicts= listAppendElement(
    			currentAttempt->otherConflicts,
    			cv->constraint);
    }

    stepcount++;
    cdgPrintf(CDG_DEBUG, "DEBUG: attacking %s\n", cv->constraint->id);

    p = attackConflict(P->current, cv,
                       OD_LIMIT | (greedy ? OD_IMMEDIATE : 0) | OD_PRUNE | OD_NARROW);

    /* user break? */
    if (cdgCtrlCTrapped) {
      backTrace(P);
      cdgPrintf(CDG_INFO, "\nOK, you take over.\n");
      break;
    }

    /* time limit? */
    if(timerExpired) {
      cdgPrintf(CDG_INFO, "timer expired!\n");

      currentAttempt->successful = FALSE;
      currentAttempt->time = timerElapsed(currentAttempt->timer);
      timerFree(currentAttempt->timer);
      currentAttempt->timer=NULL;
      P->repairAttempts= listAppendElement(P->repairAttempts,currentAttempt);
      currentAttempt=NULL;

      pDelete(p);
      if (primaryConflict) {
        cvDelete(primaryConflict);
      }
      backTrace(P);
      P->killedByTimelimit=TRUE;
      break;
    }

    /* analysis impossible? */
    if (P->globalMaximum == 0.0) {
      currentAttempt->successful = FALSE;
      currentAttempt->time = timerElapsed(currentAttempt->timer);
      timerFree(currentAttempt->timer);
      currentAttempt->timer=NULL;
      P->repairAttempts= listAppendElement(P->repairAttempts,currentAttempt);
      currentAttempt=NULL;
      if(!reducePressure(P)) {
        break;
      }
      primaryConflict = NULL;
      continue;
    }

    /* step limit reached? */
    if(max_steps && listSize(P->removedConflicts) > max_steps) {

      cdgPrintf(CDG_DEBUG, "DEBUG: step limit %d exceeded.\n", max_steps);

      currentAttempt->successful = FALSE;
      currentAttempt->time = timerElapsed(currentAttempt->timer);
      timerFree(currentAttempt->timer);
      currentAttempt->timer=NULL;
      P->repairAttempts= listAppendElement(P->repairAttempts,currentAttempt);
      currentAttempt=NULL;

      P->unremovable = listAppendElement(P->unremovable, primaryConflict);
      primaryConflict = NULL;
      backTrace(P);
      continue;
    }

    /* correction possible? */
    if (p) {
#ifdef DEBUGCOMBINED
      cdgPrintf(CDG_DEBUG, "DEBUG: got ");
      pPrint(CDG_DEBUG, p, FALSE);
#endif
      apply(P->current, p);
      aCheck(P->current,3);

      /* new maximum? */
      if (bCompare(P->current->b, P->best->b)) {

        cdgPrintf(CDG_INFO, "*");
        bPrint(CDG_INFO, P->current->b);
        printAmbiguity(CDG_INFO, P);
        if(nsAutoCompare && P->annotation) {
          aReport(P->current, P->annotation);
        }
        cdgPrintf(CDG_INFO, " (%d step%s)\n",
                  listSize(P->removedConflicts),
                  listSize(P->removedConflicts)>1?"s":"");
        cdgFlush();

        newMaximum(P->current);
        cvDelete(primaryConflict);
        primaryConflict = NULL;

        // finalize repair attempt
        currentAttempt->successful = TRUE;
        currentAttempt->time = timerElapsed(currentAttempt->timer);
        timerFree(currentAttempt->timer);
        currentAttempt->timer=NULL;
        P->repairAttempts= listAppendElement(P->repairAttempts,currentAttempt);
        currentAttempt=NULL;

        //cdgPrintf(CDG_INFO, "continuing\n"); //NB
        P->successfulPatches++;
        continue;
      }
    } else {

      /* skip the conflict whose removal led us here... */
      cdgPrintf(CDG_DEBUG, "DEBUG: couldn't remove ");
      cdgPrintf(CDG_DEBUG, "%s", primaryConflict->constraint->id);
      cdgPrintf(CDG_INFO, " (%d step%s)\n",
                listSize(P->removedConflicts),
                listSize(P->removedConflicts)>1?"s":"");

      // finalize repair attempt
      currentAttempt->successful = FALSE;
      currentAttempt->time = timerElapsed(currentAttempt->timer);
      timerFree(currentAttempt->timer);
      currentAttempt->timer=NULL;
      P->repairAttempts= listAppendElement(P->repairAttempts,currentAttempt);
      currentAttempt=NULL;


      /* Do not accept that a hard conflict cannot be removed.
         Reduce pressure instead. */
      if(0.0 == primaryConflict->constraint->penalty) {
        if(!reducePressure(P)) {
          break;
        }
        primaryConflict = NULL;
        continue;
      }

      P->unremovable = listAppendElement(P->unremovable, primaryConflict);
      primaryConflict = NULL;


      /* ...and return to the best Analysis */
      backTrace(P);
    }

    /* show progress */
    bPrint(CDG_PROGRESS, P->current->b);
    printAmbiguity(CDG_PROGRESS, P);
    cdgPrintf(CDG_PROGRESS, "\n");
    cdgFlush();

  } // end while (P->best->b->hard || significantlyGreater(P->globalMaximum, P->best->b->soft))

  P->steps = stepcount;
  backTrace(P);
  interactive = oldinteractive;

} // end combined


/* ----------------------------------------------------------------------
   Return list of time points that constitute subproblems to be searched in
   isolation.

   For instance, if the sentence `Steuermann, la� die Wacht!' carries the
   expected time information (time points 0 through 6) and only commas are
   considered indicators of subproblems, the list [0,2,6] is returned. Each
   element is either the first time point in the sentence, or the end of a
   subproblem.

   The behaviour of this function is controlled by the variable
   dynamicBoundaries.

*/
List findSubproblems(Problem P) {

  List result = listPrependElement(NULL, (Pointer)P->minTimepoint);
  Lattice lat = P->lg->lattice;
  int this, last;
  int i;

  last = P->minTimepoint;

  /* method `punc' (default): break at punctuation */
  if(!dynamicBoundaries || !strcmp(dynamicBoundaries,"punc")) {
    List l;
    for(l = lat->arcs; l != NULL; l = l->next) {
      Arc arc = l->item;
      this = arc->to;
      if(this - last <= minSize) continue;
      if(this - last > maxSize ||
         (strlen(arc->word) == 1 && strspn(arc->word,",:;-()"))) {
        result = listAppendElement(result, (Pointer) this);
        last = this;
        continue;
      }
    }
  }

  /* method `chunks': break at chunk boundaries */
  else if(!strcmp(dynamicBoundaries,"chunks")) {
    List l;
    for(l = P->lg->chunks; l != NULL; l = l->next) {
      Chunk c = l->item;
      this = c->to->arc->to;
      if(this < P->maxTimepoint) {
        result = listAppendElement(result, (Pointer)this);
      }
    }
  }

  /* method `everyN': break at regular intervals */
  else if(sscanf(dynamicBoundaries,"every%d",&i)) {
    this = last;
    while(this+i < P->maxTimepoint) {
      this += i;
      result = listAppendElement(result, (Pointer)this);
    }
  }

  /* method `exactN,M,O': break at specified places */
  else if(!strncmp(dynamicBoundaries,"exact",5)) {
    String positions=dynamicBoundaries+5;
    while(strlen(positions) &&
          isdigit((int)positions[0])) {
      sscanf(positions,"%d",&this);
      if(this > last &&
         this > P->minTimepoint &&
         this < P->maxTimepoint) {
        result = listAppendElement(result, (Pointer)this);
      }
      /* skip numbers and commas */
      positions += strspn(positions,"0123456789");
      positions += strspn(positions,",");
    }
  }

  /* unknown method: complain and fail */
  else {
    cdgPrintf(CDG_ERROR,
              "ERROR: can't use fragmentation method `%s'\n",
              dynamicBoundaries);
    listDelete(result);
    return NULL;
  }

  result = listAppendElement(result, (Pointer)P->maxTimepoint);
#ifdef DEBUGFRAGMENT
  {
    List l;
    cdgPrintf(CDG_ERROR, "Clauses == ");
    for(l = result; l != NULL; l = l->next) {
      int i = (int)l->item;
      cdgPrintf(CDG_ERROR, "%d ", i);
    }
    cdgPrintf(CDG_ERROR, "\n");

  }
#endif

  return result;

}

/* ----------------------------------------------------------------------
   Apply combined frobbing to time points FROM through TO.

   Invokes combined() on the specified subrange of timepoints of P. It sets
   P->current->subStart and P->current->subStop and temporarily deletes all
   LVs that cross the borders of this subproblem.

*/
void subFrob(Problem P, int from, int to, int ms) {
  int i;
  List l;

  cdgPrintf(CDG_INFO, "\nsubFrob: `");
  for(i = from; i < to; i++) {
    cdgPrintf(CDG_INFO, "%s ",
              ((LexemNode)vectorElement(P->lexemes, i * P->lgWidth))->lexem->word);
  }
  if(ms) {
    cdgPrintf(CDG_INFO, "' for %d ms\n\n", ms);
  } else {
    cdgPrintf(CDG_INFO, "'\n");
  }

  if(strictSubproblems) {

    /* temporarily delete all LVs that leave the specified zone. */
    for(i = inputCurrentGrammar->noOfLevels * from;
        i < inputCurrentGrammar->noOfLevels * to;
        i++) {
      for(l = P->domains[i]; l != NULL; l = l->next) {
        LevelValue lv = l->item;
        if(lv && spec(lv->regents)) {
          LexemNode ln = lv->regents->item;
          if(ln->arc->from >= to ||
             ln->arc->from < from) {
            lv->isDeleted = TRUE;
          }
        }
      }
    }
  }

  /* make sure that resets due to excessive pressure
     return to this state, and not to the global one. */
  rememberProblem(P);

  /* remove deprecated LVs from the Analysis */
  for(i = inputCurrentGrammar->noOfLevels * from;
      i < inputCurrentGrammar->noOfLevels * to;
      i++) {
    LevelValue lv = slotFiller(P->current, i);
    if(lv && spec(lv->regents)) {
      LexemNode ln = lv->regents->item;
      if(ln->arc->from >= to ||
         ln->arc->from < from) {
        aRedirect(P->current, i, -1);
      }
    }
  }

  if(P->best) {
    aDelete(P->best);
  }
  P->best = aClone(P->current);

  /* Set a shorter timer */
  timerStopAlarm(); timerSetAlarm(ms);

  /* (pretty) normal frobbing */
  aRestrict(P->current, from, to);
  aRestrict(P->best, from, to);
  P->globalMinimum = globalPressure;
  aJudge(P->current);
  aJudge(P->best);
  combined(P);

  aPrint(CDG_DEBUG,P->current);
  printConflicts(CDG_DEBUG, P->current);

  /* now judge the best Analysis so it won't spook
     the check in combined() with its good score. */
  aUnrestrict(P->current);
  aUnrestrict(P->best);
  aJudge(P->current);
  aJudge(P->best);

  restoreProblem(P,P->savedState);
  listForEachDelete(P->unremovable, cvDelete);
  P->unremovable = NULL;

}


/* ----------------------------------------------------------------------
   Maybe delete unselected LVs permanently.

   This function calls freezeDomain() on all domains of P that do not cause
   serious conflicts.

*/
void freezeSubtrees(Analysis a) {
  int i;
  Level mainlevel = inputGetMainlevel(inputCurrentGrammar);
  Vector byDomain = aAnalyze(a);

  cdgPrintf(CDG_INFO, "\nCandidates for reassembly:\n\n");

  /* Good candidates for attachments are LVS on the mainlevel
     that point to NIL and have bad scores. */
  for (i = 0; i < vectorSize(a->slots); i++) {
    Badness b = vectorElement(byDomain,i);
    LevelValue lv;
    if(!slotActive(a,i) ||
       NULL == (lv = (LevelValue) vectorElement(a->slots, i)) ||
       lv->level != mainlevel) {
      continue;
    }
    if(b->soft < 0.2) {
      cdgPrintf(CDG_INFO, "@%d: ", i);
      lvPrint(CDG_INFO,lv,NULL,0);
      cdgPrintf(CDG_INFO, "\n");
    } else {
      if(freezeSubTrees) {
        freezeDomain(a, i);
      }
    }
  }

  determineAmbiguity(a->P);
  rememberProblem(a->P);
  vectorDelete(byDomain);
}


/* ----------------------------------------------------------------------
   Simulate sentence fragments by deleting all LVs that cross one of the
   boundaries given in CLAUSES.
*/
void separateRanges(Problem P, List clauses) {
  int i;
  List l;
  for(i = inputCurrentGrammar->noOfLevels * P->minTimepoint;
      i < inputCurrentGrammar->noOfLevels * P->maxTimepoint;
      i++) {
    LevelValue lv = slotFiller(P->current, i);
    if(lv && lv->regents) {
      int down = ((LexemNode)(lv->dependents->item))->arc->from;
      int up = ((LexemNode)(lv->dependents->item))->arc->from;
      for(l = clauses; l != NULL; l = l->next) {
        int boundary = (int)l->item;
        if((down < boundary && up >= boundary) ||
           (down >= boundary && up < boundary)) {
          aRedirect(P->current, i, -1);
          break;
        }
      }
    }
  }
}

/* ----------------------------------------------------------------------
   Apply combined() to subproblems and recombine.

   This function implements the transformation strategy with the same name.
   It will first call subFrob() on partial problems delimited by
   punctuation, then combined() on the entire problem. If a time limit is
   in force, half of the allowed time will be alloted to the search within
   subproblems and half to the final combined search.
*/
void dynamic(Problem P) {

  int i,j;
  List l, clauses;
  int time;
  ProblemState orig;
  Number oldPressure = P->globalMinimum;

  /* If the lattice branches, call combined() instead */
  if(latticeBranches(P->lg->lattice)) {
    cdgPrintf(CDG_WARNING, "WARNING: lattice has branches, using `combined' instead.\n");
    combined(P);
    return;
  }

  /* break sentence into subclauses */
  clauses = findSubproblems(P);
  if(!clauses) return;


  /* if there aren't any, just call combined() */
  if(2 == listSize(clauses)) {
    cdgPrintf(CDG_WARNING, "WARNING: no subclauses present, using `combined' instead.\n");
    combined(P);
    return;
  }

  /* turn all LVs that cross boundaries to NIL, so they won't
     satisfy nonlocal constraints across subproblems. */
  if(strictSubproblems) {

    separateRanges(P, clauses);

    /* the same for P->best, because backTrace() returns to P->best */
    if(P->best) {
      aDelete(P->best);
      P->best = aClone(P->current);
    }

  }

  /* remember the state the Problem is now in,
     because analysing subproblems totally messes up
     the limit/deletion information. */
  orig = psNew(P);

  /* call frobbing on each subproblem in turn */
  i = (int)clauses->item;
  for(l = clauses->next; l != NULL; l = l->next) {
    j = (int)l->item;
    time = (int)(cdgTimeLimit * (float)(j - i) /
                 ((P->maxTimepoint - P->minTimepoint)));

    /* 0 is a sentinel value that means `no time limit'. Therefore we
       absolutely must not allow rounding to convert 0.3 to 0. */
    if(0 == time && cdgTimeLimit > 0) {
      time = 1;
    }
    subFrob(P,i,j,time);
    restoreProblem(P, orig);
    i = j;
  }

  P->globalMinimum = oldPressure;

  /* now try to join the fragments */
  aJudge(P->current);
  aJudge(P->best);
  restoreProblem(P,orig);
  freezeSubtrees(P->current);

  /* recalculate score for the entire structure */
  aJudge(P->current);
  aJudge(P->best);

  /* If the entire structure is now valued at >0, we count this as
     the first `soft' solution. There might actually have been a previous
     solution that had a soft score globally, but since we do not do global
     evaluation in phase 1, we could not have detected that. */
  newMaximum(P->current);
  cdgPrintf(CDG_INFO, "*");

  /* give frobbing another stab at the entire problem */
  time = (int)(cdgTimeLimit / 2.0);
  if(0 == time && cdgTimeLimit > 0) {
    time = 1;
  }
  timerStopAlarm();
  timerSetAlarm(time);
  combined(P);


}

/* ----------------------------------------------------------------------
   Apply the most precise transformation method. The effect is exactly the
   same as calling `frobbing dynamic', `renewnet', and `frobbing combined',
   only you need not type three commands.
*/
void threefold(Problem P) {

  ProblemState ps = psNew(P);
  dynamic(P);
  restoreProblem(P, ps);
  psDelete(ps);

  addParse(P->best ? P->best : P->current);
  if(P->net) {
    cnSortLVs(P->net);
  }
  listForEachDelete(P->unremovable, cvDelete);
  listForEachDelete(P->removedConflicts, cvDelete);
  listDelete(P->history);
  P->unremovable = P->removedConflicts = P->history = NULL;
  P->globalMaximum = 1.0;
  P->globalMinimum = globalPressure;
  if (!P->best->b->hard) {
    P->globalMinimum = max(globalPressure, P->best->b->soft);
  }
  determineAmbiguity(P);
  aJudge(P->current);
  aJudge(P->best);

  combined(P);

}

/* ----------------------------------------------------------------------
   This function implements the transformation strategy `genetic'.

   This is a reimplementation of genetic::genetic() within the frobbing
   framework, i.e. with automatic maintenance of the path criterion.
*/
void genet(Problem P) {
  Vector herd = vectorNew(30);
  int i, j;
  int POPULATION = 30;
  long GENERATIONS = 30;

  /* Build population of 30 Analyses */
  vectorAddElement(herd, P->current);
  for(i = 1; i < POPULATION; i++) {
    Analysis a = aClone(P->current);

    for(j = 0; j < P->nrSlots; j++) {
      LevelValue lv;
      Patch p0;
      Patch p1;
      if(!slotActive(P->current,j)) {
        continue;
      }
      lv = listNthElement(P->domains[j], 1 + (rand() % listSize(P->domains[j])));
      p0 = pNew(a, 1, OD_DELETED);
      p1 = pNext(p0, lv, OD_ALLOWALL | OD_DELETED);
      if(p1) {
        (void)pLexicalize(p1,0,FALSE);
        apply(a, p1);
      }
      pDelete(p0);
    }
    vectorAddElement(herd, a);
  }

  /* evolve 30 times */
  for(i = 0; i < GENERATIONS; i++) {
    Vector v = vectorClone(herd);
    vectorSort(herd, aCompare);

    cdgPrintf(CDG_INFO, "Generation %d: ", i);
    bPrint(CDG_INFO, P->current->b);
    cdgPrintf(CDG_INFO, "\n");

    /* retain best Analysis */
    P->current = vectorElement(herd, 0);

    /* cross and mutate other Analyses */
    for(j = 1; j < vectorSize(herd); j++) {
      Analysis a = vectorElement(herd, j);
      Analysis b = vectorElement(herd, rand() % POPULATION);
      vectorSetElement(v, aCross(a, b), j);
    }
    for(j = 1; j < vectorSize(herd); j++) {
       aDelete(vectorElement(herd, j));
    }
    vectorCopy(herd, v);
    vectorSetElement(herd, P->current, 0);
    vectorDelete(v);

    if (cdgCtrlCTrapped) {
      cdgCtrlCTrapped = FALSE;
      break;
    }
  }
}

/* ----------------------------------------------------------------------
   Changes the LV at position WHERE in A so that its value for the feature
   ATT is VALUE.

   It changes the actual lexical item, which permanently alters the
   constraint net, makes the cached unary score of the LV invalid and has
   assorted other undesirable side effects. In fact, calling this function
   will probably cause a crash later on. Only use it during manual frobbing
   for what-if experiments during parsing runs that you don't care about.

   In case that wasn't clear enough: Do Not Call This.

*/
Boolean aHack(Analysis a, int where, String att, String val) {

  List path = scanPath(att);
  int index = inputLookupPath(path, inputCurrentGrammar);
  String s = strRegister(val);
  LevelValue lv = vectorElement(a->slots, where);
  LexiconItem le;

  if(!lv) {
    cdgPrintf(CDG_ERROR, "No such LV.\n");
    return FALSE;
  }

  le = ((LexemNode)(lv->dependents->item))->lexem;

  /* check if the coast is clear */
  if(-1 == index) {
    cdgPrintf(CDG_WARNING, "WARNING: improper hack attempted for attribute `%s'.\n", att);
    return FALSE;
  }

  /* hack the victim */
  le->values[index]->data.string = s;
  peekValue(le->value, path)->data.string = s;

  /* clean away the blood */
  listDelete(path);

  /* act innocent */
  return TRUE;
}


/* ----------------------------------------------------------------------
   Repair by hand: always choose the value that the user wants.

   This function implements the transformation strategy with the same name.
   In contrast to the other strategies it runs fully interactively after
   building the initial Analysis. (However, there is a command to switch
   manual to automatic processing, and an automatic method can be switched
   to manual with C-c.)

   This mode provides a command-line sub-prompt to accept individual
   commands from the user. Various transformations can be performed on the
   Analysis, and the result is printed after each change. In particular,
   the ! command provides a method to search a subset of the domains in the
   problem jointly. This command can be used to prove optimality of an
   Analysis if the ambiguity in the Problem is small enough.

   Type `h' in manual frobbing mode to see all available commands.

*/
void manual(Problem P) {

  Boolean display = TRUE;
  Boolean record = FALSE;
  ConstraintViolation cv;
  Level mainlevel = inputGetMainlevel(inputCurrentGrammar);
  LevelValue lv;
  LexemGraph lg;
  LexemNode lna, lnb;
  List l, m, n;
  Number num;
  Patch p, q;
  Vector v;
  int debugflag = debug ? OD_DEBUG : 0;

  char *bufptr;
  char *token;
  char buffer[BUFFERSIZE];
  char s[BUFFERSIZE];

  while (1) {

    cdgCtrlCTrapped = FALSE;

    {
      Vector prev = vectorClone(P->current->slots);
      /* new record? */
      if(bCompare(P->current->b, P->best->b)) {
        int i;

        /* make list of changed slots */
        for(i = 0; i < vectorSize(P->best->slots); i++) {
          lv = vectorElement(P->best->slots, i);
          if(lv) {
            vectorSetElement(prev, (Pointer)lv->indexWRTNet, i);
          }
        }

        newMaximum(P->current);
        cdgPrintf(CDG_INFO, "*");
        record = TRUE;
        display = TRUE;
      }

      if (display) {
        display = FALSE;
        aCalibrate(P->current);
        printProblem(CDG_INFO, P);
      }

      if(record) {
        int i;
        record = FALSE;
        cdgPrintf(CDG_INFO, "INFO: new global maximum found (changed");

        for(i = 0; i < vectorSize(P->current->slots); i++) {
          lv = vectorElement(P->current->slots, i);
          if(lv && lv->indexWRTNet != (int)vectorElement(prev,i)) {
            cdgPrintf(CDG_INFO, " @%d", i);
          }
        }
        cdgPrintf(CDG_INFO, ")\n");
        vectorDelete(prev);
      }
    }

    /* check for internal errors */
    aCheck(P->current, 3);

    /* get input */
    strcpy(buffer, "");

    /* from execute=XXX specification */
    if(frobCommands && strcmp(frobCommands,"")) {
      int i = strcspn(frobCommands, ",");
      strncpy(buffer, frobCommands, i);
      buffer[i] = '\0';
      if(i == strlen(frobCommands)) {
        frobCommands = NULL;
      } else {
        frobCommands = &frobCommands[i+1];
      }

    } else {

      /* from terminal */
      cdgPrintf(CDG_INFO, "frobbing> ");
      cdgFlush();
      cdgGetString(buffer, BUFFERSIZE);
    }

    /* remove trailing whitespace */
    while (isspace((int)buffer[strlen(buffer) - 1])) {
      buffer[strlen(buffer) - 1] = '\0';
    }

    /* remove leading whitespace */
    bufptr = buffer;
    while (isspace((int)*bufptr)) {
      bufptr++;
    }


    /* The two next options require some explanation. They are intended
       to be used for transforming the output of depsy.pl into
       Annotations that have as few errors as possible.

       There are two problems that can occur. (1) Output from older
       versions contains only the main level, so the other levels had to be
       filled in arbitrarily and are usually wrong. (2) It often contains
       no morphological information, so the same goes for lexical
       selection.

       Here we try to fill in the missing parts by asking the grammar
       which choice would be best in each case.
    */


    /* `*' means optimise lexical selection */
    if (!strcmp("*", bufptr) && taggerCategoryIndex >=0) {

      /* This takes somewhat more care than immediately apparent. It is not
         enough to perform unary optimization on any word. Consider the
         example sentence

         er lie� sich in_dat seinen_dat_pl Lieblingssessel_nom_pl fallen

         A unary optimization would first check `in' and find that the
         dative reading is, in fact, unarily optimal. To find the global
         optimum would require changes to `in', `seinen' and `Sessel'.
         A global search is actually required. But since this could be
         really expensive, we compromise: any place in the word graph where
         serious conflicts occur is optimized, and if that doesn't help, it
         is again optimized together with its neighbour words, and so on
         until a maximum width of neighbours has been tried.

      */

      int i;
      Vector byDomain = NULL;
      Boolean progress;

      /* carefully undelete only those lexemes that have the same category
         as the selected one */
      bvSetAllElements(P->lg->isDeletedNode, TRUE);
      for(i = 0; i < vectorSize(P->lg->nodes); i++) {
        LexemNode ln = vectorElement(P->lg->nodes, i);
        LexemNode selected = selectedAt(P->current, ln->arc->from);
        Value cat1 = ln->lexem->values[taggerCategoryIndex];
        Value cat2 = selected->lexem->values[taggerCategoryIndex];
        if(cat1->type == VTString &&
           cat2->type == VTString &&
           cat1->data.string == cat2->data.string) {
          bvSetElement(P->lg->isDeletedNode, FALSE, ln->no);
        }
      }

      do {

        progress = FALSE;
        if(byDomain) {
          vectorDelete(byDomain);
        }
        byDomain = aAnalyze(P->current);

        for(i = 0; i < vectorSize(byDomain); i++) {
          Badness b = vectorElement(byDomain, i);
          if(b->hard || b->soft < frobIgnoreThreshold) {
            int j;
            Patch p = NULL;
            int width;
            int timepoint = i / P->nrLevels;

            for(width = 0; width < 3; width++) {
              Vector which = vectorNew(10);
              cdgPrintf(CDG_DEBUG, "DEBUG: checking");
              for(j = -width; j <= width; j++) {
                vectorAddElement(which, (Pointer)timepoint + j);
                cdgPrintf(CDG_DEBUG, " %d", timepoint+j);
              }
              cdgPrintf(CDG_DEBUG, "\n");
              p = optimizeLexemes(P->current, which);
              vectorDelete(which);
              if(p && bCompare(p->predicted, P->current->b)) {
                apply(P->current, p);
                progress = TRUE;
                break;
              } else {
                pDelete(p);
              }
            }
            if(progress) {
              break;
            }
          }
        }
      } while(progress);
    }


    /* `_' means optimize auxiliary levels */
    else if (!strcmp("_", bufptr)) {

      /* The tactic here is to take all conflicts on non-main
         levels and attack them, using only orientation changes. */

      int i, j;
      LevelValue new;
      Boolean progress;

      /* However, there is a catch: to use dSearch(), the LVs we want to
         try out must already be present in the Problem! So we compute all
         alternative LVs and insert them into the Problem first. */
      for(i = 0; i < vectorSize(P->current->slots); i++) {
        LevelValue lv = vectorElement(P->current->slots, i);
        if(lv->level == mainlevel) continue;

        for(j = -1; j < P->maxTimepoint; j++) {
          if(!lv->regents && j == -1) continue;
          if(spec(lv->regents) &&
             ((LexemNode)(lv->regents->item))->arc->from == j)
            continue;

          /* add new LV */
          new = lvClone(lv);
          listDelete(new->regents);
          new->regents = NULL;
          if(j != -1) {
            LexemNode ln = selectedAt(P->current, j);
            new->regents = listPrependElement(NULL, ln);
            evalUnary(new, NULL, NULL, FALSE, NULL, NULL, TRUE);
            if(new->score == 0.0) {
              lvDelete(new);
              continue;
            }
            new->limit = new->score;

            /* tell addLVToProblem() to add a better index number */
            new->indexWRTNet = -1;

            addLVToProblem(P, new);
          }
        }
      }

      /* Now try to remove those conflicts that have to do with
         non-main-level domains. */

      do {
        progress = FALSE;
        for(l = P->current->conflicts; l != NULL; l = l->next) {
          ConstraintViolation cv = l->item;
          Constraint c = cv->constraint;
          if(c->sig->l1 != mainlevel ||
             (c->sig->l2 && c->sig->l2 != mainlevel)) {

            if((p = attackConflict(P->current, cv, OD_ORIENT | OD_PREDICT))) {
              if(bCompare(p->predicted, P->current->b)) {
                pPrint(CDG_DEBUG, p, TRUE);
                apply(P->current, p);
                progress = TRUE;
                display = TRUE;
                break;
              } else {
                listForEachDelete(P->removedConflicts, cvDelete);
                P->removedConflicts = NULL;
              }
            } else {
              pDelete(p);
              listForEachDelete(P->removedConflicts, cvDelete);
              P->removedConflicts = NULL;
            }
          }
        }
      } while(progress);
    }

    /* `+' means more debug info */
    else if (!strcmp("+", bufptr)) {
      cdgPrintf(CDG_DEBUG, "debug level %d\n", ++debug);
      debugflag = debug ? OD_DEBUG : 0;
    }

    /* `^' means hand the parse up to xcdg */
    else if (!strcmp("^", bufptr)) {
      Parse p = analysis2parse(P->current);
      p->searchStrategy = strRegister("manual frobbing");
      cdgExecHook(HOOK_PARTIALRESULT, p);
    }

    /* `-' means less debug info */
    else if (!strcmp("-", bufptr)) {
      cdgPrintf(CDG_DEBUG, "debug level %d\n", --debug);
      debugflag = debug ? OD_DEBUG : 0;
    }

    /* %<number> means swap lexeme node */
    else if (bufptr[0] == '%') {

      /* get %-separated list of numbers */
      l = NULL;
      token = strtok(bufptr, " %");
      while (token) {

        /* parse number */
        int no = -1;
        sscanf(token, "%d", &no);
        if (no > -1 && no < P->nrLexemes) {
          lna = vectorElement(P->lg->nodes, no);
          if (!listContains(l, lna)) {
            l = listPrependElement(l, lna);
          }
        }
        token = strtok(NULL, " %");
      }

      /* Ugly hack: lgMakePath() will balk if passed a deleted
         lexeme node as an argument. But during manual frobbing, we
         may actually want to switch a deleted lexeme into the
         Analysis, presumably to find out why it was deleted (``Hey,
         wouldn't the score be much better if we just used the
         accusative reading?''). Therefore we must make sure that none
         of the lexemes passed into it is actually deleted at the
         time. Solution: we temporarily undelete all nodes and restore
         the original state immediately after the call. */
      lg = P->lg;

      {
        /* zip! */   ByteVector oldDeletions = lg->isDeletedNode;
        /* clone! */ lg->isDeletedNode = bvClone(lg->isDeletedNode);
        /* flush! */ bvSetAllElements(lg->isDeletedNode, FALSE);
        /* call! */  m = lgMakePath(lg, l);
        /* wham! */  bvDelete(lg->isDeletedNode);
        /* zap! */   lg->isDeletedNode = oldDeletions;
      }

      /* if lgMakePath() still balked, we know that there is truly no
         path at all that includes the requested lexemes. This means
         that either some of them overlap or at least one of them is a
         dangling lexeme. Such requests deserve no pity. */
      if (!m) {
        cdgPrintf(CDG_INFO, "Inconsistent lexemes requested.\n");
        continue;
      }
      /* Now we have a complete path that includes all of the
         requested lexemes. But we would also like to swap as little
         as possible. So we replace all of the additional (not
         explicitly requested) lexemes in the path with their
         currently selected lexical alternatives, if there are any. */
      for (n = m; n; n = n->next) {
        lna = n->item;
        if (listContains(l, lna))
          continue;
        lnb = selectedAt(P->current, lna->arc->from);
        if (lna->arc->from == lnb->arc->from && lna->arc->to == lnb->arc->to) {
          n->item =lnb;
        }
      }

      /* now swap all of these lexemes in */
      selectPath(P->current, m);
      aJudge(P->current);
      sortConflicts(P->current);
      listDelete(l);
      listDelete(m);
      display = TRUE;
    }

    /* ?<number>{,<number>} means search list of domains for optima */
    else if (bufptr[0] == '?') {

      /* get space-separated list of numbers */
      v = vectorNew(10);
      token = strtok(bufptr, " ?@");
      while (token) {

        /* parse number */
        int no = -1;
        sscanf(token, "%d", &no);
        if (no > -1 && no < P->nrSlots) {
          if(!slotActive(P->current,no)) {
            cdgPrintf(CDG_INFO, "Domain @%d not used. \n", no);
          }
          vectorAddElement(v, (Pointer) no);
        }
        token = strtok(NULL, " ?@");
      }
      if (vectorSize(v) > 0) {

        /* give feedback */
        int i;
        cdgPrintf(CDG_INFO, "searching: ");
        for (i = 0; i < vectorSize(v); i++) {
          cdgPrintf(CDG_INFO, "@%d ", (int)vectorElement(v, i));
        }
        cdgPrintf(CDG_INFO, "\n");

        /* search */
        l = dSearch(P->current, v, OD_ALLOWALL | OD_PRUNE | debugflag);
        vectorDelete(v);

        /* display best results */
        if (l) {
          num = ((Patch) l->item)->limit;
          for (m = l; m; m = m->next) {
            p = m->item;
            if (p->limit < num)
              break;
            pPrint(CDG_INFO, p, TRUE);
          }
          listForEachDelete(l, pDelete);
        }
      }
    }

    /* `!' means optimize all violated/changed domains */
    else if (strcmp("!", bufptr) == 0) {

      /* find all pertinent domains */
      v = vectorNew(10);
      for (l = P->history; l; l = l->next) {
        vectorAddElement(v, l->item);
      }
      for (l = P->current->conflicts; l; l = l->next) {
        cv = l->item;
        if (!vectorContains(v, (Pointer) cv->nodeBindingIndex1)) {
          vectorAddElement(v, (Pointer) cv->nodeBindingIndex1);
        }
        if (cv->nodeBindingIndex2 != -1 &&
            !vectorContains(v, (Pointer) cv->nodeBindingIndex2)) {
          vectorAddElement(v, (Pointer) cv->nodeBindingIndex2);
        }
      }

      if (vectorSize(v) > 0) {
        int i;

        /* give feedback */
        cdgPrintf(CDG_INFO, "searching: ");
        for (i = 0; i < vectorSize(v); i++) {
          cdgPrintf(CDG_INFO, "@%d ", (int)vectorElement(v, i));
        }
        cdgPrintf(CDG_INFO, "\n");

        /* search */
        p = optimizeDomains(P->current, v,
                            OD_ALLOWALL | OD_LIMIT |
                            OD_INTERACTIVE | OD_PRUNE | debugflag);
        if (p) {
          apply(P->current, p);
          display = TRUE;
        }
        vectorDelete(v);
      }
    }

    /* `!!' means optimize all domains */
    else if (strcmp("!!", bufptr) == 0) {
      int i;

      /* find all domains */
      v = vectorNew(P->nrSlots);
      for (i = 0; i < vectorSize(P->current->slots); i++) {
        if(!slotActive(P->current,i)) {
          continue;
        }
        vectorAddElement(v, (Pointer) i);
      }

      if (vectorSize(v) > 0) {

        /* search */
        p = optimizeDomains(P->current, v,
                            OD_ALLOWALL | OD_PREDICT |
                            OD_PRUNE | OD_INTERACTIVE |
                            OD_CONTEXT | debugflag);
        if (p) {
          apply(P->current, p);
          display = TRUE;
        }
      }
      vectorDelete(v);
    }

    /* { #<number> } means replace LVs */
    else if (bufptr[0] == '#') {

      /* get #-separated list of numbers */
      l = NULL;

      token = strtok(bufptr, " #");
      while (token) {

        /* parse number */
        int no = -1;
        sscanf(token, "%d", &no);
        if (no > -1 && no < P->nrLVs) {
          l = listAppendElement(l, (Pointer) no);
        }

        /* if no number was parsed, stop scanning. This allows the
           output of pPrint() to be pasted into the command line. */
        if (no == -1) {
          break;
        }

        token = strtok(NULL, " #");
      }

      if (!l) {
        continue;
      }

      p = pNew(P->current, listSize(l), OD_DELETED);

      /* build a patch from these LVs */
      for (m = l; m; m = m->next) {
        int i = (int)m->item;
        if ((lv = lvByNumber(P, i)) == NULL)
          continue;
        if(!lv->level->useflag) {
          cdgPrintf(CDG_INFO, "LV #%d not used -- level inactive.\n", i);
          continue;
        }
        q = pNext(p, lv, OD_ALLOWALL | OD_DELETED | debugflag);
        pDelete(p);
        p = q;

        if (!p) {
          cdgPrintf(CDG_INFO, "Inconsistent lexemes requested.\n");
          break;
        }
      }

      if (p) {
        if(pLexicalize(p,OD_DELETED,FALSE)) {
          apply(P->current, p);
          display = TRUE;

          /* note down conflicts removed by this step */
          for (l = P->current->conflicts; l; l = l->next) {
            if (!cvContains(P->current->conflicts, l->item)) {
              P->removedConflicts =
                listAppendElement(P->removedConflicts, cvClone(l->item));
            }
          }
        }
      }
    }

    /* `$' means revert to annotation */
    else if (bufptr[0] == '$') {
      if(P->annotation != NULL) {
        aDelete(P->current);
        P->current = aClone(P->annotation);
        display = TRUE;
      } else {
        cdgPrintf(CDG_INFO, "No annotation available.\n");
      }
    }


    /* `@' means give ambiguity by domain */
    else if (bufptr[0] == '@') {
      int i;
      for(i = 0; i < P->nrSlots; i++) {
        if(slotActive(P->current,i)) {
          cdgPrintf(CDG_INFO, "@%d: %d values\n", i, dSize(P, i));
        }
      }
    }

    /* `>' means re-direct an LV */
    else if (bufptr[0] == '>') {
      int no = -1;
      int i = -1;
      sscanf(bufptr, ">%d %d", &no, &i);

      if (no > -1 &&
          no <= vectorSize(P->current->slots) &&
          i > P->lg->min && i <= P->lg->max) {

        if(!slotActive(P->current,no)) {
          cdgPrintf(CDG_INFO, "Domain @%d not used -- level inactive.\n", no);
          continue;
        }

        aRedirect(P->current, no, i);
        aJudge(P->current);
        display = TRUE;
      }
    }

    /* `~' means manipulate a lexical item */
    else if (bufptr[0] == '~') {
      char a[1024], b[1024];
      int no = -1;
      sscanf(bufptr, "~%d %s %s", &no, a, b);

      /* if no number is given, i remains -1.
         aRedirect() can handle this. */
      if (no > -1 &&
          no <= vectorSize(P->current->slots)) {

        if(!slotActive(P->current,no)) {
          cdgPrintf(CDG_INFO, "Domain @%d not used -- level inactive.\n", no);
          continue;
        }

        aHack(P->current, no, a, b);
        aJudge(P->current);
        sortConflicts(P->current);
        aCalibrate(P->current);
        display = TRUE;
      }
    }

    /* `a' means frob on auto pilot */
    else if (!strcmp("a", bufptr)) {
      combined(P);
      if (!P->current)
        break;
      display = TRUE;
    }

    /* `b' means abandon this branch */
    else if (strcmp("b", bufptr) == 0) {
      backTrace(P);
      display = TRUE;
    }

    /* c<number> means call attackConflict */
    else if (bufptr[0] == 'c') {
      int no = -1;
      ConstraintViolation cv;
      sscanf(bufptr, "c%d", &no);
      if (no < 0 || no >= listSize(P->current->conflicts))
        continue;
      cv = listNthElement(P->current->conflicts, no + 1);

      /* success ? */
      if ((p = attackConflict
           (P->current, cv,
             OD_LIMIT | OD_INTERACTIVE | OD_PRUNE | debugflag))) {
        apply(P->current, p);
        display = TRUE;
      }
    }

    /* `d' means print a domain */
    else if (bufptr[0] == 'd') {
      int no = -1;
      sscanf(bufptr, "d%d", &no);
      if (no > -1 && no < P->nrSlots) {
        if(!slotActive(P->current,no)) {
          cdgPrintf(CDG_INFO, "Domain @%d not used -- level inactive.\n", no);
          continue;
        }
        dPrint(CDG_INFO, P, no);
      }
    }

    /* `e' means change localSize */
    else if (bufptr[0] == 'e') {
      int no = -1;
      sscanf(bufptr, "e%d", &no);
      if (no > 0) {
        localSize = no;
        cdgPrintf(CDG_INFO,
                  "local search space limited to %d alternatives\n", no);
      }
    }

    /* `f' means flush `obstinate' tags */
    else if (bufptr[0] == 'f') {
      listForEachDelete(P->unremovable, cvDelete);
      P->unremovable = NULL;
      display = TRUE;
    }

    /* `g' means print lexeme graph */
    else if (bufptr[0] == 'g') {
      lgPrint(CDG_INFO, P->lg);
    }

    /* `h' means help user */
    else if (bufptr[0] == 'h') {
      cdgPrintf(CDG_INFO, "\n"
                "commands during manual frobbing:\n"
                "\n"
                "%%{<n>}              select arc n\n"
                "*                   optimise lexeme selection\n"
                "_                   optimise auxiliary levels\n"
                "^                   hand the parse up to Tcl\n"
                "@                   list ambiguity by domain\n"
                "$                   cheat (by looking at the annotation)\n"
                "?{<n>}              search domains jointly (expensive)\n"
                "!                   search all violated domains jointly (EXTREMELY expensive!)\n"
                "!!                  search ALL domains (DON'T EVEN THINK OF USING THIS!)\n"
                "#{<n>}              substitute LV #m into the analysis\n"
                "><n> <m>            let LV #n point to word #m\n"
                "~<n> <s> <s>        change value of attribute in modifier of LV #n\n"
                "a                   switch to Automatic mode (C-c interrupts)\n"
                "b                   return to Best analysis\n"
                "c<n>                suggest patch for a Conflict\n"
                "d<n>                print a Domain\n"
                "e<n>                Extend local agenda\n"
                "f                   remove `obstinate' tags\n"
                "g                   show lexeme Graph\n"
                "h                   Help about commands\n"
                "i                   Ignore conflict #n\n"
                "j                   re-compute conflicts (`Judge')\n"
                "l <n> <s>           set Label for domain n to the string s\n"
                "m <f>               set Minimal acceptable score\n"
                "n                   aNalyse analysis for trouble spots\n"
                "o{<n>}              Optimise domains in the current context\n"
                "p                   Print current analysis\n"
                "q                   Quit to cdgp shell\n"
                "r                   Renew constraint net\n"
                "s                   Store this analysis\n"
                "t <n>               set ignore-Threshold\n"
                "u                   show the Unary best LVs in the analysis\n"
                "v                   verify current analysis\n"
                "w<n> <n>            Work on subrange\n"
                "x<n>                eXplain how a conflict arises\n"
                "y                   this slot can be yours for just 45$ a day!\n"
                "z                   save as annotation\n");
    }

    /* i<number> means disregard a conflict */
    else if (bufptr[0] == 'i') {
      int no = -1;
      sscanf(bufptr, "i%d", &no);
      if (no < 0 || no >= listSize(P->current->conflicts))
        continue;
      P->unremovable =
        listAppendElement(P->unremovable,
                       cvClone(listNthElement(P->current->conflicts, no + 1)));
      display = TRUE;
    }

    /* `j' means rejudge entire Analysis */
    else if (!strcmp("j", bufptr)) {
      cdgPrintf(CDG_INFO, "recomputing... ");
      cdgFlush();
      aJudge(P->current);
      sortConflicts(P->current);
      bPrint(CDG_INFO, P->current->b);
      cdgPrintf(CDG_INFO, "\n");
      continue;
    }

    /* `l' means re-label an LV */
    else if (bufptr[0] == 'l') {
      int no = -1;
      sscanf(bufptr, "l%d %s", &no, s);
      if (no > -1 && no < vectorSize(P->current->slots)) {
        if(!slotActive(P->current,no)) {
          cdgPrintf(CDG_INFO, "Domain @%d not used -- level inactive.\n", no);
          continue;
        }

        if(aRelabel(P->current, no, strRegister(s))) {
          aJudge(P->current);
          display = TRUE;
        }
      }
    }

    /* m<number> means enforce a new global limit */
    else if (bufptr[0] == 'm') {
      float f;
      sscanf(bufptr, "m%f", &f);
      if (f <= 1.0 && f >= 0.0) {
        if (f > P->globalMinimum) {
          P->globalMinimum = f;
          cdgPrintf(CDG_INFO, "global limit set to %4.3e\n", P->globalMinimum);
        }
      }
    }

    /* `n' means analyse hot spots */
    else if (!strcmp("n", bufptr)) {
      Vector byDomain = aAnalyze(P->current);
      int i;

      /* print individual badnesses by domain */
      for (i = 0; i < vectorSize(byDomain); i++) {
        Badness bb = vectorElement(byDomain, i);
        if (bb->no == 0)
          continue;

        cdgPrintf(CDG_INFO, "@%03d: ", i);
        bPrint(CDG_INFO, bb);
        cdgPrintf(CDG_INFO, "\n");
      }

      for (i = 0; i < vectorSize(byDomain); i++) {
        bDelete(vectorElement(byDomain,i));
      }
      vectorDelete(byDomain);
    }

    /* o<number>{,<number>} means optimize domains WRT current Analysis */
    else if (bufptr[0] == 'o') {

      /* get space-separated list of numbers */
      v = vectorNew(10);
      token = strtok(bufptr, " o@");
      while (token) {

        /* parse number */
        int no = -1;
        sscanf(token, "%d", &no);
        if (no > -1 && no < P->nrSlots) {
          if(slotActive(P->current,no)) {
            vectorAddElement(v, (Pointer) no);
          } else {
            cdgPrintf(CDG_INFO, "Domain @%d not used -- level inactive.\n", no);
          }
        }
        token = strtok(NULL, " o@");
      }
      if (vectorSize(v) > 0) {
        int i;

        /* give feedback */
        cdgPrintf(CDG_INFO, "optimizing: ");
        for (i = 0; i < vectorSize(v); i++) {
          cdgPrintf(CDG_INFO, "@%d ", (int)vectorElement(v, i));
        }
        cdgPrintf(CDG_INFO, "\n");

        if ((p =
             optimizeDomains(P->current, v,
                             OD_ALLOWALL | OD_PREDICT| OD_PRUNE |
                             OD_INTERACTIVE | debugflag))
            != NULL) {
          apply(P->current, p);
          display = TRUE;
        }
      }
      vectorDelete(v);
    }

    /* `p' means print */
    else if (!strcmp("p", bufptr)) {

      display = TRUE;
    }

    /* `q' means end */
    else if (!strcmp("q", bufptr)) {
      return;
    }

    /* `r' means renew the constraint net */
    else if (!strcmp("r", bufptr)) {
      cnRenew(P->net);
      P->globalMinimum = 0.0;
      P->globalMaximum = 1.0;
      determineAmbiguity(P);
      aJudge(P->current);
      aJudge(P->best);
      display = TRUE;
    }

    /* `s' means store this Analysis */
    else if (!strcmp("s", bufptr)) {
      aStore(P->current);
    }

    /* `t' means set frobIgnoreThreshold */
    else if (buffer[0] == 't') {
      float f;
      sscanf(bufptr, "t%f", &f);
      if (f <= 1.0 && f >= 0.0) {
        frobIgnoreThreshold = f;
        cdgPrintf(CDG_INFO, "ignore threshold set to %4.3e\n", f);
      } else {
        cdgPrintf(CDG_INFO, "%f invalid!\n", f);

      }
    }

    /* `u' means show unary best LVs */
    else if (!strcmp("u", bufptr)) {
      showUnary(P->current);
    }

    /* `v' means show wrong values */
    else if (bufptr[0] == 'v') {
      aVerify(P->current);
    }

    /* `w' means manual subfrobbing */
    else if (bufptr[0] == 'w') {
      ProblemState orig;
      int from = -1, to = -1;
      sscanf(bufptr, "w%d %d", &from, &to);
      if(from >= P->minTimepoint && from < P->maxTimepoint &&
         to >= P->minTimepoint && to <= P->maxTimepoint &&
         from < to) {
        List boundaries = listAppendElement(NULL, (Pointer)to);
        boundaries = listAppendElement(boundaries, (Pointer)from);
        orig = psNew(P);
        separateRanges(P, boundaries);
        subFrob(P, from, to, INT_MAX);
        restoreProblem(P, orig);
        listDelete(boundaries);
      }
    }

    /* `x' means explain a conflict */
    else if (bufptr[0] == 'x') {
      int no = -1;
      sscanf(bufptr, "x%d", &no);
      if (no > -1 && no < listSize(P->current->conflicts)) {
        cvAnalyse(listNthElement(P->current->conflicts, no + 1), P->current->slots);
      }
    }

    /* `z' means write to file as Annotation */
    else if (bufptr[0] == 'z') {
      char id[128];
      String name = NULL;
      sscanf(bufptr, "z%100s", id);
      if(1 == strlen(buffer)) {
        name = P->lg->lattice->id;
      } else {
        name = strRegister(id);
      }

      writeAnnotation(parse2annotation(analysis2parse(P->current)), name);
      cdgPrintf(CDG_INFO,"OK.\n");
    }
  }
}



/* ----------------------------------------------------------------------
   Finalize the module.
*/
void frobbingFinalize(void) {

  /* This list is used to hold results between two calls
     to frobbing(), so it can't be freed in frobbing() itself. */
  if (benchmark) {
    listDelete(benchmark);
    benchmark = NULL;
  }

}


/* ----------------------------------------------------------------------
   Set module-global variables to their default values.
*/
void set_default_values() {
  beamWidth = INT_MAX;
  dynamicBoundaries = NULL;
  freezeSubTrees = TRUE;
  frobCommands = NULL;
  globalPressure = 0.0;
  localSize = 3000;
  minSize = 6;
  maxSize = 20;
  strictSubproblems = TRUE;
  greedy = TRUE;
  max_steps = 0;
}


/* ----------------------------------------------------------------------
   Build a solution heuristically.

   This is the module entry point. It is called by cmdFrobbing().

   Each command word must be either the name of a constraint net, which
   will then be searched, or a string of the form key=value. The
   following keys and values are recognized:

   agenda=N: sets localSize.

   batch=B: resets interactive.

   beam=N: sets beamWidth.

   execute=S: specifies commands to be executed
   (interpreted as in manual()).

   freeze=B: sets freezeSubtrees.

   method=S: selects frobbing method.

   minsize=N: sets minimum size of subproblems for `dynamic'

   net=S: selects constraint net.

   parse=S: start out from specific parse.

   pressure=N: sets globalPressure.

   strict=B: sets strictSubproblems.

   subproblem=S: sets dynamicBoundaries.

   Either a constraint net or a parse must be specified; however, frobbing
   from a parse almost never makes sense.

*/
Boolean frobbing(int no, char **args) {
  Problem P = NULL;
  int i, j;
  int method = -1;
  float pressure = -1.0;
  ConstraintNet net = NULL;
  Parse parse = 0;
  Lattice lat = NULL;

  set_default_values();

  for(i = 0; i < no; i++) {
    String s = strRegister(args[i]);
    char *key, *value;

    /* find named assignments */
    if(NULL != (key = strtok(s, "=")) &&
       NULL != (value = strtok(NULL, "="))) {

      if(!strcmp(key, "agenda")) {
        sscanf(value, "%d", &localSize);
      }
      else if(!strcmp(key, "batch")) {
        if(!strcmp(value, "no")) {
          /* do nothing */
        }
        else if(!strcmp(value, "yes")) {
          interactive = FALSE;
        }
        else {
          cdgPrintf(CDG_ERROR, "ERROR: `batch' must be `yes' or `no'.\n");
          return FALSE;
        }
      }
      else if(!strcmp(key, "beam")) {
        sscanf(value, "%d", &beamWidth);
        cdgPrintf(CDG_INFO, "Beam width set to %d\n", beamWidth);
      }
      else if(!strcmp(key, "execute")) {
        frobCommands = strCopy(value);
      }
      else if(!strcmp(key, "freeze")) {
        if(!strcmp(value, "yes")) {
          freezeSubTrees = TRUE;
        }
        else if(!strcmp(value, "no")) {
          freezeSubTrees = FALSE;
        }
        else {
          cdgPrintf(CDG_ERROR, "ERROR: `freeze' must be `yes' or `no'.\n");
          return FALSE;
        }
      }
      else if(!strcmp(key, "greedy")) {
        if(!strcmp(value, "yes")) {
          greedy = TRUE;
        }
        else if(!strcmp(value, "no")) {
          greedy = FALSE;
        }
        else {
          cdgPrintf(CDG_ERROR, "ERROR: `greedy' must be `yes' or `no'.\n");
          return FALSE;
        }
      }
      else if(!strcmp(key, "steps")) {
        sscanf(value, "%d", &max_steps);
        cdgPrintf(CDG_INFO,
                  "INFO: frobbing limited to %d steps per conflict\n",
                  max_steps);
      }
      else if(!strcmp(key, "strict")) {
        if(!strcmp(value, "yes")) {
          strictSubproblems = TRUE;
        }
        else if(!strcmp(value, "no")) {
          strictSubproblems = FALSE;
        }
        else {
          cdgPrintf(CDG_ERROR, "ERROR: `strict' must be `yes' or `no'.\n");
          return FALSE;
        }
      }
      else if(!strcmp(key, "method")) {
        for (j = 0; frobMethods[j].id != NULL; j++) {
          if (strcmp(value, frobMethods[j].id) == 0) {
            method = j;
            break;
          }
        }
        if(method == -1) {
          cdgPrintf(CDG_ERROR,
                    "ERROR: frobbing method `%s' not known.\n",
                    value);
          return FALSE;
        }
      }
      else if(!strcmp(key, "net")) {
        if(!(net = cnFindNet(value))) {
          cdgPrintf(CDG_ERROR, "ERROR: net `%s' not known.\n", value);
          return FALSE;
        }
      }
      else if(!strcmp(key, "lattice")) {
        if(!(lat = findLattice(value))) {
          cdgPrintf(CDG_ERROR, "ERROR: lattice `%s' not known.\n", value);
          return FALSE;
        }
      }
      else if(!strcmp(key, "parse")) {
          if(!strcmp(value, "latest")) {
              parse = parseMostRecentlyCreatedParse;
          }
          else if(!(parse = parseFind(value))) {
          cdgPrintf(CDG_ERROR, "ERROR: parse `%s' not known.\n", value);
          return FALSE;
        }
      }
      else if(!strcmp(key, "pressure")) {
        sscanf(value, "%f", &pressure);
      }
      else if(!strcmp(key, "subproblem")) {
        dynamicBoundaries = strRegister(value);
      }
      else if(!strcmp(key, "minsize")) {
        sscanf(value, "%d", &minSize);
      }
      else if(!strcmp(key, "maxsize")) {
        sscanf(value, "%d", &maxSize);
      }
      else {
        cdgPrintf(CDG_ERROR, "ERROR: can't use named parameter `%s'\n", key);
        return FALSE;
      }
      continue;
    }

    /* find old-style batch flag */
    if(!strcmp(s, "batch")) {
      interactive = FALSE;
      continue;
    }

    /* find old-style frob method */
    if(method == -1) {
      for (j = 0; frobMethods[j].id != NULL; j++) {
        if (strcmp(s, frobMethods[j].id) == 0) {
          method = j;
          break;
        }
      }
      if(method > 0) {
        continue;
      }
    }

    /* find old-style net */
    if(NULL != (net = cnFindNet(s))) {
      continue;
    }

    /* find old-style lattice */
    if(!(lat = findLattice(s))) {
      cdgPrintf(CDG_ERROR, "ERROR: can't use parameter `%s'\n", s);
      return FALSE;
    }
  }

  if(method == -1) {
    method = 6;
    cdgPrintf(CDG_INFO,
              "INFO: using frob method `%s' as a default\n",
              frobMethods[method].id);
  }

  if(pressure == -1.0) {
    pressure = 0.1;
  }
  if (pressure < 0.0 || pressure > 1.0) {
    cdgPrintf(CDG_ERROR,
              "ERROR: make sure that 0 <= pressure <= 1\n");
    return FALSE;
  }

  if(net) {
    evaluations = net->evalUnary + net->evalBinary;
  }
  stepcount = 0;
  debug = 0;

  /* yet another way of setting the pressure is to set it globally */
  pressure = max(pressure, globalPressure);
  globalPressure = pressure;

  /* Figure out which Problem to use. Since there are many ways of
     specifying this, we need many conditions. */

  /* If nothing was specified, but a lattice was already created,
     use that lattice. */
  if(!lat && !P && !net && !parse && inputMostRecentlyCreatedLattice) {
    lat = inputMostRecentlyCreatedLattice;
  }

  /* If we now have a lattice, find or build a net for it. */
  if(lat) {
    net = cnFindNetForLattice(lat->id);
    if(!(net = cnBuild(lat, TRUE, TRUE))) {
      cdgPrintf(CDG_ERROR,
                "ERROR: Can't build constraint net from lattice `%s'!\n",
                lat->id);
    }
  }

  /* If we have no net, but a net was already built, take that net. */
  if(!net && cnMostRecentlyCreatedNet) {
    net = cnMostRecentlyCreatedNet;
    cdgPrintf(CDG_INFO,
              "INFO: using most recently created net `%s'\n", net->id);
  }

  /* If we now have a net, find or build a Problem for it. */
  if(net) {
    P = findProblem(net);

    /* Be cautious about the state of information that transform has added
       to the actual constraint net. We store Problems in a global list so
       they can be reused between invocations of frobbing; but if the
       constraint net changes between those calls, we have no way of
       knowing that, since the ConstraintNet does not know it has an
       associated Problem. Therefore we now re-compute the ambiguity of the
       Problem .*/
    if(P) {
      P->globalMinimum = 0.0;
      P->globalMaximum = 1.0;
      determineAmbiguity(P);
    }
  }

  /* If we still have nothing, but a Parse was already created, use that. */
  if(!P && !net && !parse && parseMostRecentlyCreatedParse) {
    parse = parseMostRecentlyCreatedParse;
    cdgPrintf(CDG_INFO,
              "INFO: using most recently created parse `%s'\n", parse->id);
  }

  /* If we still have no Problem... */
  if(!P) {

    /* ...build one from a ConstraintNet */
    if (net) {
      P = net2problem(net, pressure, parse);
    }

    /* ...or from a parse */
    else if(parse) {
      P = parse2problem(parse, pressure);
    }
  }

  /* give up exhausted */
  if (!P) {
    cdgPrintf(CDG_ERROR, "ERROR: Don't know what problem to solve.\n");
    return FALSE;
  }

  aJudge(P->best);
  aJudge(P->current);
  aJudge(P->annotation);
  P->globalMinimum = pressure;
  if (!P->best->b->hard) {
    P->globalMinimum = max(pressure, P->best->b->soft);
  }

  cdgPrintf(CDG_INFO,"Start frobbing! \n");

  /* batch search */
  frobMethods[method].repair(P);

  /* If an interruption occurred, we switch to interactive search even if
     interactive is off. Obviously someone is using the program
     interactively, otherwise there wouldn't have been a C-c. */
  if(cdgCtrlCTrapped) {
    cdgCtrlCTrapped = FALSE;
    interactive = TRUE;
  }

  /* interactive search */
  if (interactive && strcmp("manual", frobMethods[method].id)) {
    manual(P);
  }
  /* register result */
  addParse(P->best ? P->best : P->current);

  if (aValid(P->best)) {
    cdgPrintf(CDG_INFO, "INFO: found solution with score %4.3e\n", P->best->b->soft);
  } else {
    cdgPrintf(CDG_INFO, "INFO: found an invalid solution\n");
  }

  /* re-sort the LVs in the net */
  if(P->net) {
    cnSortLVs(P->net);
  }

  /* show statistics */
  showTime(P);
  frobWriteXmlStatistics(P);


  if(net) {
    cnPrintInfo(net);
    writeXmlNetInfo(net);
  }

  /* clean up information that only applies to this parsing run */
  listForEachDelete(P->unremovable, cvDelete);
  listForEachDelete(P->removedConflicts, cvDelete);
  listDelete(P->history);
  P->unremovable = P->removedConflicts = P->history = NULL;

  return TRUE;

}

/* ----------------------------------------------------------------------
 * Register CDG variables.
 */
void frobInitialize()
{
  setRegister("ignorethreshold",
              SET_FLOAT, &frobIgnoreThreshold, NULL, NULL, NULL, NULL);
  setRegister("globalPressure",
			  SET_FLOAT, &globalPressure, NULL,NULL,NULL,NULL);
  setRegister("maxsteps",
			  SET_INT, &max_steps, NULL,NULL,NULL,NULL);
  set_default_values();
}


/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */

