/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* ---------------------------------------------------------------------------
 * @addtogroup Lexemgraph
 *
 */

#ifndef LEXEMGRAPH_H
#define LEXEMGRAPH_H

/* ------------------------------------------------------------------------ */
#include <cdg.h>
#include <input.h>

/* ------------------------------------------------------------------------ */

#ifndef SWIG
/* ----------------------------------------------------------------------------
   A LexemGraph is a word graph enriched with lexical information. In
   particular, it contains several lexemes for each arc of the Lattice
   that is lexically ambiguous. Each of these pairs of lexical entry and time
   span is called a lexeme node.  The field \a lattice points to the
   underlying word graph.

   The Vectors \a graphemnodes and \a nodes contains all grapheme nodes
   and lexeme nodes. Grapheme nodes are an intermediary data structure between
   word arcs and lexeme nodes that is not strictly necessary.

   The fields \a min and \a max correspond to the fields with the
   same names in the Lattice.

   The field \a distance holds an Array of the distance between any two
   lexeme nodes. The distance is measured in words, hence any two adjacent
   lexeme nodes have distance~1. The distance array is also used to check
   whether two lexeme nodes are compatible with each other, i.e.  whether there
   is a path through the lexeme graph that includes them both.

   The fields \a noOfPathsFromStart and \a noOfPathsToEnd hold arrays
   that map each grapheme node to the number of complete paths from the start
   or to the end of the entire graph that go through it. This is used to
   determine whether a lexeme node can be deleted or not.

   The field \a noOfPaths holds the total number of paths from start to end
   possible in the word graph. This should reflect the state of the Vector
   \a isDeletedNode.

   \b WARNING: The last three fields use GNU's \c long \c long \c int type
   as a cheap way to get 64-bit integers because the number of paths in
   realistic word graphs really does need that. However, the tool SWIG cannot
   deal with this type, and as a result XCDG cannot access the
   LexemGraph structure at all. Several functions in this library exist
   only to work around this restriction. Ultimately these fields should be
   converted to a proper Bigint type.

   The Vector \a isDeletedNode marks those lexeme nodes that have been
   deleted and should be ignored. For instance, when an LV is added to a
   partial solution, all lexeme nodes that are not lgCompatibleNodes()
   to its lexeme nodes should be marked as deleted.
 */
typedef struct {
  Lattice lattice;                      /**< underlying word graph */
  Vector graphemnodes;                  /**< vector of grapheme nodes */
  Vector nodes;                         /**< vector of lexeme nodes */
  int min;                              /**< minimum start position */
  int max;                              /**< maximum end position */
  VirtualLGExtension ve;				/**< virtual nodes for anticipatory incremental parsing */

  /** the following parts are considered dynamic */

  Array distance;                       /**< matrix of distances */
  long long *noOfPathsFromStart;        /**< vector of numbers of paths */
  long long *noOfPathsToEnd;            /**< vector of numbers of paths */
  long long noOfPaths;                  /**< total number of paths through graph */
  ByteVector isDeletedNode;             /**< vector of boolean flags */
  ByteVector isFrozenNode;				// used for incremental parsing, is persistent among steps, in contrast to isDeletedNode
  List chunks;							// set of all chunks of the lattice
  Boolean complete;						// whether this lexemgraph represents a complete sentence, in contrast to a partial sentence like a prefix
} LexemGraphStruct;
typedef LexemGraphStruct *LexemGraph;

/* ----------------------------------------------------------------------------
   A grapheme node represents the hypothesis of a specific phonetic form for a
   specific time interval. Thus, it corresponds biuniquely to an Arc in
   the underlying Lattice. However, it also holds information about the
   state of processing by a particular grammar.

   The field \a no is an index into the field \a graphemnodes of the
   enclosing lexeme graph.

   The field \a lexemgraph points to this graph.

   The field \a arcpoints to the corresponding Arc.

   The field \a lexemes contains all LexemeNode objects built from lexical
   entries with the same phonetic form as the word on the Arc.

   The field predictions contains key-value pairs that were assigned by
   external programs by observing the sequence of ord forms.

*/
struct GraphemNodeStruct {
  int no;                               /**< index in lg->graphemnodes */
  LexemGraph lexemgraph;                /**< pointer back to the lexem graph */
  Arc arc;                              /**< arc in word lattice */
  List lexemes;                         /**< list of disambiguated lexeme nodes */
  Boolean live;                         /**< are there any lexemes left undeleted? */
  Vector predictions;
  Hashtable specialPredictions;
  Chunk chunk;							/**< to which chunk do we belong */
  Boolean isVirtual;					/**< if true, the gn does not correspond to a word from the lattice */
};
#else
typedef struct LexemGraphStruct *LexemGraph;
#endif

/* ----------------------------------------------------------------------------
   A LexemNode represents the hypothesis of a specific lexical variant
   for a specific time interval.

   The field \a no is the index of the lexeme node in the field
   \a nodes of the enclosing lexeme graph.

   The field \a lexemgraph point to the enclosing lexeme graph.

   The field \a arc points to the underlying Arc.

   The field \a lexem points to the lexical entry postulated.

   The field \a grapheme points to the grapheme node used to build the
   lexeme node.

   The field \a limit corresponds to the field \a limit in an LV: a
   lexeme node with limit \a x can only appear in solutions not better than
   \a x (see Frobbing)
 */
struct LexemNodeStruct {
  int no;		     /**< index in LexemGraph::nodes */
  LexemGraph lexemgraph;     /**< pointer back to the lexem graph */
  Arc arc;		     /**< arc in word lattice */
  LexiconItem lexem;	     /**< lexical entry */
  GraphemNode grapheme;	     /**< pointer to the original grapheme */
  Number limit;		     /**< limit calculated during frobbing */
};

/* ---------------------------------------------------------------------- */
extern LexemGraph lgNew(Lattice lat);
extern String gnFormatPredictions(GraphemNode gn);

#ifndef SWIG
extern Boolean lgCompactLVs;

extern Boolean lgIsDeletableNode(GraphemNode gn);
extern Boolean lgAreDeletedNodes(LexemGraph lg, List lexemes);
extern Boolean lgCompatibleNodes(LexemGraph lg, LexemNode a, LexemNode b);
extern Boolean lgCompatibleSets(LexemGraph lg, List a, List b);
extern Boolean lgContains(LexemGraph lg, String form);
extern Boolean lgCopySelection(LexemGraph destination, LexemGraph source);
extern Boolean lgForbiddenBy(LexemGraph lg, LexemNode a, List b);
extern Boolean lgIntersectingSets(List a, List b);
extern Boolean lgIsDeletedNode(LexemGraph lg, LexemNode n);
extern Boolean lgIsEndNode(GraphemNode n);
extern Boolean lgIsStartNode(GraphemNode n);
extern Boolean lgLexemeInLexemNodeList(LexiconItem le, List list);
extern Boolean lgMayModify(LexemGraph lg, GraphemNode down, GraphemNode up);
extern Boolean lgMember(LexemNode ln, List lexemes);
extern Boolean lgNewFinal(LexemGraph lg);
extern Boolean lgTag(LexemGraph lg);
extern void    lgPredict(LexemGraph lg);
extern LexemNode lgAddLexemNode(LexemGraph lg, GraphemNode gn, LexiconItem le);
extern Boolean lgNewIter(LexemGraph lg, Arc arc);
extern Boolean lgOverlap(LexemNode a, LexemNode b);
extern Boolean lgSpuriousUppercase(LexemGraph lg, Arc arc);
extern Boolean lgSubset(List a, List b);
extern Boolean lgUpdateArcs(LexemGraph lg, Lattice lat, List listArcs);
extern LexemGraph lgClone(LexemGraph lg);
extern LexemGraph lgNewInit();
extern List lgMakePath(LexemGraph lg, List nodes);
extern List lgMostProbablePath(LexemGraph lg);
extern List lgPartitions(GraphemNode gn, BitString features);
extern List lgQueryCat(LexemGraph lg, GraphemNode gn);
extern Value gnReadPrediction(GraphemNode gn, int slot, String key);
extern Value gnReadPredictionSpecial(GraphemNode gn, String predictor, String description, String key);
extern String gnHighestPrediction(GraphemNode gn);
extern Number lgTagscore(LexemNode ln);
extern int lgDistanceOfNodes(LexemGraph lg, LexemNode a, LexemNode b, Boolean virtualAsNonspec);
extern int lgWidth(LexemGraph lg);
extern void gnWriteNumber(GraphemNode gn, int slot, String key, Number n);
extern void gnWriteNumberSpecial(GraphemNode gn, String predictor, String description, String key, Number x);
extern void gnWriteString(GraphemNode gn, int slot, String key, String s);
extern void gnWriteStringSpecial(GraphemNode gn, String predictor, String description, String key, String s);
extern void lgComputeDistances(LexemGraph lg);
extern void lgComputeNoOfPaths(LexemGraph lg, Boolean realloc);
extern void lgDelete(LexemGraph lg);
extern void lgDeleteNode(LexemGraph lg, LexemNode n);
extern void lgDeleteNodes(LexemGraph lg, List nodes);
extern void lgInitialize();
extern void lgPrint(unsigned long, LexemGraph lg);
extern void lgPrintNode(unsigned long mode, LexemNode ln);
extern void lgRequireLexeme(LexemGraph lg, ByteVector v, LexemNode ln);
extern void lgRequireLexemes(LexemGraph lg, ByteVector v, List which);
extern GraphemNode gnNew(void);
#endif

/* ---------------------------------------------------------------------- */

#endif /* don't insert anything after this #endif */
