/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 6/5/97
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup Netsearch Netsearch - A best first search
 * \author Ingo Schroeder (see also AUTHORS and THANKS for more)
 *
 * 
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <blah.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#include "cdg.h"
#include "input.h"
#include "eval.h"
#include "lexemgraph.h"
#include "levelvalue.h"
#include "netsearch.h"
#include "increment.h"
#include "scache.h"
#include "hook.h"
#include "timer.h"
#include "parse.h"
#include "write.h"
#include "set.h"

/* -- MACRO DEFINITIONS ------------------------------------------------- */
#define AGENDASIZEDEFAULT 3000

/* -- FORWARD DECLARATIONS ---------------------------------------------- */
List nsBranchBound(ConstraintNet net, int no, char **args);
List nsFullSearch(ConstraintNet net, int no, char **args);

/* -- VARIABLES --------------------------------------------------------- */
SearchMethod search_methods[] =
{
   {"branchbound", nsBranchBound, "branch and bound search"},
   {"bb", nsBranchBound, "branch and bound search"},
   {"fullsearch", nsFullSearch, "bestfirst full search"},
   {"fs", nsFullSearch, "bestfirst full search"},

  /* end of array marker */
   {(String) NULL, (ListFunction *) NULL, (String) NULL}
};

Boolean nsAutoCompare = FALSE;
Boolean nsSearchModifiesNetFlag = FALSE;
NsNormalizationType nsNormalization = NSNSquare;

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * performs a search for a constraint net
 */
int netsearch(ConstraintNet net, int no, char **args)
{
   int i, index;
   List parses;

   /* find the search method */
   if (no > 0 && args[0] != NULL) {
      index = -1;
      for (i = 0; search_methods[i].id != NULL; i++) {
	 if (strcmp(args[0], search_methods[i].id) == 0) {
	    index = i;
	    args++;
	    no--;
	    break;
	 }
      }
      if (index == -1) {
	 cdgPrintf(CDG_ERROR, "ERROR: search method `%s' unknown\n", args[0]);
	 return (-1);
      }
   } else {
     /* default, this should be settable */
     index = 0;
   }

   /* perform search */
   parses = search_methods[index].doIt(net, no, args);
   if (!parses) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: search failed for method `%s'\n",
	       search_methods[index].id);
     return (-1);
   }

   /* register parses to constraint net */
   nsRegisterParses(net, parses);
   listDelete(parses);

   return (0);
}

/* ----------------------------------------------------------------------
 * agenda normalization
 */
Number nsComputeAgendaScore(NetSearchState nss)
{
  Number as;

  switch (nsNormalization) {
    case NSNOff:
      as = nss->badness->soft - nss->badness->hard;
      break;
    case NSNLinear:
      as = (nss->badness->soft - nss->badness->hard) * (nss->noOfBindings+1);
      break;
    case NSNSquare:
      as = pow(nss->badness->soft - nss->badness->hard, 1 / (Number) (nss->noOfBindings+1));
      break;
    case NSNDepth:
      as = (Number) nss->noOfBindings;
      break;
    case NSNBreadth:
      as = (Number) -nss->noOfBindings; 
      break;
    default:
      exit(0);
  }
#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: %d/%4.3e = %4.3e\n",
      nss->badness->hard, nss->badness->soft, as);
#endif

  return as;
}

/* ----------------------------------------------------------------------
 * deletes net search state
 */
void nsDeleteNetSearchState(NetSearchState nss)
{
  /* Shouldn't happen, but anyway... */
  if (!nss) {
#ifdef DEBUG_DELETENETSEARCHSTATE
    cdgPrintf(CDG_DEBUG, "DEBUG: nsDeleteNetSearchState: nss==NULL\n");
#endif
    return;
  }
  if (nss->noOfUndeletedChildren != 0) {
#ifdef DEBUG_DELETENETSEARCHSTATE
     cdgPrintf(CDG_DEBUG, "DEBUG: nsDeleteNetSearchState: noOfUndeletedChildren==%d\n",
      nss->noOfUndeletedChildren);
#endif
    return;
  }

  if (nss->isDeletedLexemNode)
    bvDelete(nss->isDeletedLexemNode);
  if (nss->nodeBinding)
    memFree(nss->nodeBinding);

  /* Am I the first one? */
  if (nss->parent && 0 == --nss->parent->noOfUndeletedChildren)
    /* Nope. Tell my mom. */
    nsDeleteNetSearchState(nss->parent);

  bDelete(nss->badness);

  memFree(nss);
}

/* ----------------------------------------------------------------------
 * is there an undeleted lexeme in the list?
 */
Boolean nsLexemesValid(List lexemes, NetSearchState nss)
{
   List p;
   LexemNode ln;

   for (p = lexemes; p != NULL; p = p->next) {
      ln = p->item;
      if (!bvElement(nss->isDeletedLexemNode, ln->no))
	 return (TRUE);
   }

   return (FALSE);
}

/* ----------------------------------------------------------------------
 * computes score for list of node bindings and new node binding
 */
Badness nsCompatibleBindings(NetSearchState nss, LevelValue lv, ByteVector isDeletedLexemNode)
{
   NodeBinding nb;
   static BadnessStruct badnessStruct;
   Badness badness = &badnessStruct;
   NetSearchState lnss;
   int found;
   List l;

   badness->no = 0; badness->hard = 0; badness->soft = 1.0;

   /* new dependents must not all be deleted */
   found = 0;
   for(l = lv->dependents; l != NULL; l = l->next) {
     if(!bvElement(isDeletedLexemNode, ((LexemNode)(l->item))->no)) {
       found = 1;
       break;
     }
   }
   if(!found) {
#ifdef DEBUG_NSCOMPATIBLEBINDINGS
    cdgPrintf(CDG_DEBUG,
       "DEBUG:  ...the modifier is deleted already.\n");
#endif
    badness->no = -1;
    return badness;
   }

   /* new regents must not all be deleted */
   if (spec(lv->regents)) {
     found = 0;
     for(l = lv->regents; l != NULL; l = l->next) {
       if(!bvElement(isDeletedLexemNode, ((LexemNode)(l->item))->no)) {
	 found = 1;
	 break;
       }
     }

     if(!found) {
#ifdef DEBUG_NSCOMPATIBLEBINDINGS
      cdgPrintf(CDG_DEBUG,
	 "DEBUG:  ...the modifiee is deleted already.\n");
#endif
      badness->no = -1;
      return badness;
     }
   }

   /* alright, evaluate the grammar */
   for (lnss = nss; lnss; lnss = lnss->parent) {
      if (lnss->isDeletedState)
	 continue;
      nb = lnss->nodeBinding;
      evalBinary(lv, nb->value, nb->node->net, NULL, FALSE, badness, NULL);
   }

   return badness;
}


/* ----------------------------------------------------------------------
   Eval a complete structure in context
*/
void nsEvalInContext(NetSearchState nss) {
  Vector context = vectorNew(20);
  NetSearchState lnss;
  LevelValue lv;
  NodeBinding nb;
  
  /* create context */
  for (lnss = nss; lnss; lnss = lnss->parent) {
    if (lnss->isDeletedState)
      continue;
    nb = lnss->nodeBinding;
    lv = nb->value;
    vectorSetElement(context, lv, lvIndex(lv));
  }
   

  /* eval context-sensitive constraints */
  evalInContext(context, context, nss->badness, NULL);

  vectorDelete(context);
}

/* ----------------------------------------------------------------------
 * checks whether a node binding is valid
 */
Boolean nsIsValidSolution(NetSearchState nss)
{
   NetSearchState lnss, kmss;
   List levels = NULL;
   ConstraintNode cn;
   int i, count;
   LevelValue lv;

#if 0
   /* if the solution has score 0, it is not valid.
      This can happen if a hard context-sensitive
      constraint failed. */
   if(nss->badness->hard > 1) {
     return FALSE;
   }
#endif

   /* collect levels */
   for (lnss = nss; lnss; lnss = lnss->parent) {
      if (lnss->isDeletedState)
	 continue;
      lv = lnss->nodeBinding->value;
      if (listContains(levels, lv->level))
	 continue;
      levels = listPrependElement(levels, lv->level);
   }

   /* added this kind of test because of problems with uselevel */
   for (i = 0; i < vectorSize(nss->net->nodes); i++) {
      cn = (ConstraintNode) vectorElement(nss->net->nodes, i);
      if (!listContains(levels, cn->level)) {
	 /* level not available */
#ifdef DEBUG_NSISVALIDSOLUTION
	 cdgPrintf(CDG_DEBUG,
	    "DEBUG: level not available\n");
#endif
	 listDelete(levels);
	 return (FALSE);
      }
   }

   /* check whether it's a legal path */
   i = nss->net->lexemgraph->min;
   count = 0;
   while (i < nss->net->lexemgraph->max) {
      count++;
      for (lnss = nss; lnss; lnss = lnss->parent) {
	 if (lnss->isDeletedState)
	    continue;
	 lv = lnss->nodeBinding->value;
	 if (i == ((LexemNode)(lv->dependents->item))->arc->from) {
	    i = ((LexemNode)(lv->dependents->item))->arc->to;
	    break;
	 }
      }
      if (lnss == NULL) {
	 /* invalid path */
#ifdef DEBUG_NSISVALIDSOLUTION
	 cdgPrintf(CDG_DEBUG,
	    "DEBUG: nsIsValidSolution: invalid path\n");
#endif
	 listDelete(levels);
	 return (FALSE);
      }
   }

   if (listSize(levels) * count != nss->noOfBindings) {
      /* no of words times no of levels doesn't equal no of bindings */
#ifdef DEBUG_NSISVALIDSOLUTION
      cdgPrintf(CDG_DEBUG,
	 "DEBUG: nsIsValidSolution: #words*#levels=%d, but #bindings=%d\n",
	 listSize(levels)*count, nss->noOfBindings);
#endif
      listDelete(levels);
      return (FALSE);
   }
   for (lnss = nss; lnss; lnss = lnss->parent) {
      if (lnss->isDeletedState)
	 continue;
      lv = lnss->nodeBinding->value;
      i = 0;
      while (i <= nss->noOfBindings && spec(lv->regents)) {
	 for (kmss = nss; kmss; kmss = kmss->parent) {
	    if (kmss->isDeletedState)
	       continue;
	    if (lgIntersectingSets(lv->regents, kmss->nodeBinding->value->dependents)
	       &&
	       lv->level == kmss->nodeBinding->value->level) {
#ifdef DEBUG_NSISVALIDSOLUTION
	       cdgPrintf(CDG_DEBUG,
		  "DEBUG: %d %s %s\n",
		  i, ((LexemNode)(lv->dependents->item))->lexem->description,
		  ((LexemNode)(kmss->nodeBinding->value->dependents->item))->lexem->description);
#endif
	       lv = kmss->nodeBinding->value;
	       i++;
	       break;
	    }
	 }
	 if (kmss == NULL) {
	    /* not connected */
#ifdef DEBUG_NSISVALIDSOLUTION
	    cdgPrintf(CDG_DEBUG,
	       "DEBUG: nsIsValidSolution: not connected\n");
#endif
	    listDelete(levels);
	    return (FALSE);
	 }
      }
      if (i > nss->noOfBindings) {
	 /* some kind of cycle */
#ifdef DEBUG_NSISVALIDSOLUTION
	 cdgPrintf(CDG_DEBUG,
	    "DEBUG: cycle\n");
#endif
	 listDelete(levels);
	 return (FALSE);
      }
   }

   listDelete(levels);
   return (TRUE);
}

/* ----------------------------------------------------------------------
 * nsSearchStateContainsNode: check whether a constraint node is already 
 *                        bound within a searchstate
 */
Boolean nsSearchStateContainsNode(NetSearchState nss, ConstraintNode node)
{
   NetSearchState lnss;

   for (lnss = nss; lnss; lnss = lnss->parent) {
      if (!lnss->isDeletedState &&
	 lnss->nodeBinding->node == node)
	 return (TRUE);
   }

   return (FALSE);
}

/* ----------------------------------------------------------------------
 * skips nextNodeIndex to next undeleted and unused node
 */
void nsSkipOverDeletedAndUsedNodes(NetSearchState nss)
{
   ConstraintNode cn;
   ConstraintNet net = nss->net;

   for (; nss->nextNodeIndex < vectorSize(net->nodes); nss->nextNodeIndex++) {
      cn = (ConstraintNode) vectorElement(net->nodes, nss->nextNodeIndex);
      if (nsLexemesValid(cn->gn->lexemes, nss)
	 && !nsSearchStateContainsNode(nss, cn))
	 break;
   }
}

/* ----------------------------------------------------------------------
 * skips nextNodeIndex to next undeleted node
 */
void nsSkipOverDeletedNodes(NetSearchState nss)
{
   ConstraintNode cn;
   ConstraintNet net = nss->net;

   for (; nss->nextNodeIndex < vectorSize(net->nodes); nss->nextNodeIndex++) {
      cn = (ConstraintNode) vectorElement(net->nodes, nss->nextNodeIndex);
      if (nsLexemesValid(cn->gn->lexemes, nss))
	 break;
   }
}

/* ----------------------------------------------------------------------
 * nsPrint:
 *      print bindings of netsearchstate
 */
int nsPrint(long unsigned int mode, NetSearchState nss, int no)
{
   if (nss == NULL || !nss->parent)
      return 0;

   no = nsPrint(mode, nss->parent, no);

   if (!nss->isDeletedState) {
      cdgPrintf(mode, "%02d  ", no);
      lvPrint(mode, nss->nodeBinding->value, nss->isDeletedLexemNode, 0);
      cdgPrintf(mode, "\n");
   }
   return no + 1;
}

/* ----------------------------------------------------------------------
 * nsNSState2Parse
 */
Parse nsNSState2Parse(ConstraintNet net, NetSearchState nss)
{
  Parse p;
  LexemGraph lg = net->lexemgraph;
  NetSearchState lnss;
  Vector lvs;
  LevelValue lv;

  p = parseNew(NULL);
  p->latticeId = lg->lattice->id;
  p->searchStrategy = strRegister("netsearch");
  p->comment = NULL;
  p->solutionNo = 1;
  p->noOfSolutions = 1;
  p->score = (nss->badness->hard>0)?0.0:nss->badness->soft;
  p->optimal = TRUE;
      
  /* find all lvs used */
  lvs = vectorNew(20);
  for (lnss = nss; lnss; lnss = lnss->parent) {
    if (lnss->isDeletedState)
      continue;
    lv = lnss->nodeBinding->value;
    vectorAddElement(lvs, lv);
  }
  vectorSort(lvs,lvCompareByNaturalOrder);
  parseComputeStructure(p, lvs, nss->isDeletedLexemNode);
  vectorDelete(lvs);
    
  p->evalUnary = net->evalUnary;
  p->evalBinary = net->evalBinary;
  p->numberOfHits = 0; /* FIXME: where shall we find this number? */
  p->evaluationTime = 0; /* FIXME: profileTime is local to nsSearch()! */
  p->statisticComment = NULL;
  parseDecorate(p, net->lexemgraph, NULL);
  parseEval(p);

  return p;
}

/* ----------------------------------------------------------------------
 * nsRegisterParses
 */
void nsRegisterParses(ConstraintNet net, List parses)
{
  LevelValue lv;
  List l, m;
  Parse p;
  int i, j = 0;
  
  /* delete all lvs not contained in the search result:
     (a) delete everything... */
  if (nsSearchModifiesNetFlag) {
    for (i = 0; i < vectorSize(net->nodes); i++) {
      ConstraintNode cn = vectorElement(net->nodes, i);
      cn->noValidValues = 0;
      for (j = 0; j < vectorSize(cn->values); j++) {
        lv = (LevelValue) vectorElement(cn->values, j);
        lv->isDeleted = TRUE;
      }
    }
  }

  /* add parses */
  i = 0;
  for (l = parses; l != NULL; l = l->next) {
    p = (Parse)l->item;
    p->solutionNo = ++i;
    p->noOfSolutions = listSize(parses);
      
    /* can't fail here, the grammar is definitely present */
    parseEval(p);
      
    /* add to net if it is a new solution */
    for(m = net->parses; m != NULL; m = m->next) {
      if(parseEquals(m->item, p)) {
	net->parses = listDeleteElement(net->parses, p);
        parseDelete(p);
        p = NULL;
        break;
      }
    }

    if(p) {
      net->parses = listPrependElement(net->parses, p);

      /* add to global List */
      parseRegister(p);
      cdgPrintf(CDG_INFO, "INFO: parse %s added to the global parses list.\n", p->id);
    }
  }      
}

/* ----------------------------------------------------------------------
   Returns a new NetSearchState
 */
NetSearchState nssNew(ConstraintNet net, NetSearchState nss)
{

  NetSearchState result = memMalloc(sizeof (NetSearchStateStruct));

  /* id should be set by the caller */
  /* score should be set by the caller */

  result->net = net;
  result->nextNodeIndex = nss ? nss->nextNodeIndex + 1 : 0;
  result->isDeletedLexemNode =
    bvClone(nss ? nss->isDeletedLexemNode : net->lexemgraph->isDeletedNode);
  result->nodeBinding = NULL;
  result->parent = nss;
  result->noOfUndeletedChildren = 0;
  if (nss) {
    nss->noOfUndeletedChildren++;
  }

  /* isDeletedState should be set by the caller */
  result->noOfBindings = nss ? nss->noOfBindings : 0;
  result->noOfDeletedBindings = 0;

  return result;
}

/* ----------------------------------------------------------------------
 * performs a search
 */
List nsSearch(ConstraintNet net, SearchType mode, int agendasize, Number threshold)
{
  int i, j;
  Badness bestBadnessSoFar, badness;
  List bestStatesSoFar = NULL;
  ConstraintNode cn;
  LevelValue lv;
  NodeBinding nb;
  LexemNode ln;
  LexemGraph lg = net->lexemgraph;
  List l;
  int stepcount = 0;
  int statecount = 0;
  NetSearchState nss, newstate;
  Timer profileTime;
  unsigned long solutionProfile = 0, totalProfile = 0;
  unsigned long softProfile = 0;
  Boolean softFlag = FALSE;
  int solutionSteps = 0;
  ByteVector originalDeletions = lg->isDeletedNode;
  List result = NULL;
  AnnoEntry ae=findAnnoForLattice(net->lexemgraph->lattice->id, TRUE);
  
  if (threshold != -1.0) {
    bestBadnessSoFar = bNew(0,0,threshold);
  } else {
    bestBadnessSoFar = worstBadness();
  }

  cdgPrintf(CDG_INFO, "INFO: annotation %s found\n", ae ? ae->id : " NOT ");

  /*
   * Search is cancelled after cdgTimeLimit milliseconds.
   * TODO: A problem might be that a system call may be interrupted
   * when the SIGALARM signal is delivered. Check this.
   */
  if ( cdgTimeLimit != 0 ) {
    timerSetAlarm(cdgTimeLimit);
  } else {
    timerExpired = FALSE;
  }
  profileTime = timerNew();

  net->searchagenda = agNew(agendasize, nsDeleteNetSearchState);
  agSetVerbosity(net->searchagenda, hkVerbosity & CDG_INFO);

#if 0
  net->evalUnary = 0;
  net->evalBinary = 0;
  if (net->cache)
    net->cache->hits = 0;
#endif

  /* create initial state */
  nss = nssNew(net, NULL);
  nss->id = statecount = 1;
  nss->badness = bestBadness();
  nss->isDeletedState = TRUE;

  nsSkipOverDeletedNodes(nss);

  if (hkVerbosity & CDG_HOOK)
    cdgExecHook(HOOK_NSSEARCH, nss, 1);

  /* It really doesn't matter which score to take,
     it's removed in the next line anyway. */
  cdgAgInsert(net->searchagenda, nsComputeAgendaScore(nss), nss);
  while (!agIsEmpty(net->searchagenda)) {
    nss = (NetSearchState) agRemoveBest(net->searchagenda);

    if (cdgCtrlCTrapped || timerExpired) {
      /* ok, clean up */
      nsDeleteNetSearchState(nss);
      continue;
    }

    stepcount++;

    if (stepcount % 1000 == 0) {
      cdgPrintf(CDG_PROGRESS, "step %d, %d%%, %d/%4.3e after %lums\r",
		stepcount,
		100 * nss->nextNodeIndex / vectorSize(nss->net->nodes),
		nss->badness->hard, nss->badness->soft, timerElapsed(profileTime));
    }

#if 0
    cdgPrintf(CDG_DEBUG, "\nDEBUG: search step (%d)\n", stepcount);
    cdgPrintf(CDG_DEBUG, "DEBUG: retrieved search state [%d]\n", nss->id);
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: deletions: ");
    for (j = 0; j < vectorSize(lg->nodes); j++)
      cdgPrintf(CDG_DEBUG, "%s\n",
		bvElement(nss->isDeletedLexemNode, j) ? "0" : "1");
#endif
    cdgPrintf(CDG_DEBUG, "DEBUG: next node: %d, state score: %d/%f \n",
	      nss->nextNodeIndex, nss->badness->hard, nss->badness->soft);
    nsPrint(CDG_DEBUG, nss, 0);
#endif

    cdgFlush();

    /* we already have a better solution */
    if (mode == STBranchBound && !bCompare(nss->badness, bestBadnessSoFar)) {
      /* free state */
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: we already have a better solution %d/%4.3e than %d/%4.3e\n",
	  bestBadnessSoFar->hard, bestBadnessSoFar->soft,
	  nss->badness->hard, nss->badness->soft);
#endif
      nsDeleteNetSearchState(nss);
      continue;
    }
    if (nss->nextNodeIndex >= vectorSize(nss->net->nodes)) {
      /* complete solution found */

      /* now eval context-sensitive constraints */
      nsEvalInContext(nss);

      if (!nsIsValidSolution(nss)) {
#if 0
	cdgPrintf(CDG_DEBUG,
		  "DEBUG: invalid solution, deleting net search state\n");
#endif
#if 0
	nsPrint(CDG_DEBUG, nss, 0);
	abort();
#endif
	nsDeleteNetSearchState(nss);
	continue;
      }
      /* it's valid, i.e. it's a tree etc. */
      if (bCompare(nss->badness, bestBadnessSoFar)) {
        Parse p;

        solutionProfile = timerElapsed(profileTime);
	solutionSteps = stepcount;
	cdgPrintf(CDG_INFO,
		  "INFO: solution with better score %d/%4.3e found after %d steps",
		  nss->badness->hard, nss->badness->soft, solutionSteps);
	if (hkVerbosity & CDG_PROFILE) {
	  cdgPrintf(CDG_PROFILE, ", %lums\n", solutionProfile);
	} else {
	  cdgPrintf(CDG_INFO, "\n");
	}
	listForEachDelete(bestStatesSoFar, nsDeleteNetSearchState);
	bestStatesSoFar = listPrependElement(NULL, nss);
	bCopy(bestBadnessSoFar, nss->badness);

	/* capture first solution profile */
	if (!softFlag) {
	  softFlag = TRUE;
	  softProfile = solutionProfile;
	}

	if (hkVerbosity & CDG_HOOK)
	  cdgExecHook(HOOK_NSSEARCH, nss, 2);

        net->lexemgraph->isDeletedNode = originalDeletions;
        p = nsNSState2Parse(net, nss);
        p->searchStrategy = strRegister("netsearch");
        cdgExecHook(HOOK_PARTIALRESULT, p);
        
        if (nsAutoCompare && ae) {
	  parseVerificationDelete(parseVerify(p, ae));
	}

      } else if (bEqual(nss->badness, bestBadnessSoFar)) {
	bestStatesSoFar = listPrependElement(bestStatesSoFar, nss);

	if (hkVerbosity & CDG_HOOK)
	  cdgExecHook(HOOK_NSSEARCH, nss, 2);

	cdgPrintf(CDG_INFO,
		  "INFO: solution %d with score %d/%4.3e found after %d steps",
		  listSize(bestStatesSoFar), nss->badness->hard, nss->badness->soft, stepcount);
	if (hkVerbosity & CDG_PROFILE) {
	  cdgPrintf(CDG_PROFILE, ", %lums\n", timerElapsed(profileTime));
	} else {
	  cdgPrintf(CDG_INFO, "\n");
	}
      } else {
	nsDeleteNetSearchState(nss);
      }
      continue;
    }
    cn = (ConstraintNode) vectorElement(net->nodes, nss->nextNodeIndex);

#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: trying to bind node #%d ", nss->nextNodeIndex);
    cnPrintNode(CDG_DEBUG, cn);
    cdgPrintf(CDG_DEBUG, " with %d values\n", vectorSize(cn->values));
#endif

    /* temporarily install a different state of lexeme deletions */      
    lg->isDeletedNode = nss->isDeletedLexemNode;
    lgComputeNoOfPaths(lg, FALSE);

    /* try to delete node, maybe it's a legal state, too */
    /* check whether the lexemes are needed for the lexeme graph
       or for a previous binding */
    if (lgIsDeletableNode(cn->gn)) {

#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG:  trying to delete node #%d\n",
	      nss->nextNodeIndex);
#endif

      /* allocate new state with this node deleted */
      newstate = nssNew(net, nss);
      newstate->noOfDeletedBindings++;
      newstate->id = ++statecount;
      newstate->badness = bClone(nss->badness);
      newstate->isDeletedState = TRUE;

      /* delete all of the node's lexemes */
      for (l = cn->gn->lexemes; l != NULL; l = l->next) {
        ln = l->item;
        bvSetElement(newstate->isDeletedLexemNode, TRUE, ln->no);
      }

      /* skip over deleted nodes */
      nsSkipOverDeletedNodes(newstate);

#if 0
      cdgPrintf(CDG_DEBUG,
		"DEBUG:  inserting new search state [%d] with score %d/%4.3e, nss score=%4.3e\n",
		newstate->id, newstate->badness->hard, newstate->badness->soft,
		nsComputeAgendaScore(newstate));
#endif

      if (hkVerbosity & CDG_HOOK)
	cdgExecHook(HOOK_NSSEARCH, newstate, 4);

      cdgAgInsert(net->searchagenda, nsComputeAgendaScore(newstate), newstate);
    } else {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG:  node is not deletable\n");
#endif
    }
    for (i = vectorSize(cn->values)-1; i >= 0; i--) {
      lv = (LevelValue) vectorElement(cn->values, i);

      if (lv->isDeleted) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG:  ... value is deleted already.\n");
#endif
	continue;
      }
      /* check compatibility with existing node bindings */
      badness = nsCompatibleBindings(nss, lv, nss->isDeletedLexemNode);
      if (badness->no == -1) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG:  ... value is incompatible.\n");
#endif
	continue;
      }
      badness = bAdd(bAddBadness(badness, nss->badness),lv->score);
      if (mode == STBranchBound && !bCompare(badness, bestBadnessSoFar)) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG: we already have a better solution %d/%4.3e than %d/%4.3e\n",
	    bestBadnessSoFar->hard, bestBadnessSoFar->soft,
	    badness->hard, badness->soft);
#endif
	continue;
      }
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG:  trying value #%d: ", i + 1);
      lvPrint(CDG_DEBUG, lv, nss->isDeletedLexemNode);
      cdgPrintf(CDG_DEBUG, "\n");
#endif

      newstate = nssNew(net, nss);
      newstate->id = ++statecount;
      newstate->badness = bClone(badness);
#ifdef DEBUG_NSSEARCH
      cdgPrintf(CDG_DEBUG, "DEBUG: %d/%4.3e + %d/%4.3e + %4.3e = %d/%4.3e\n",
	  nss->badness->hard, nss->badness->soft,
	  badness->hard, badness->soft, lv->score,
	  newstate->badness->hard, newstate->badness->soft);
#endif
      newstate->isDeletedState = FALSE;
      newstate->noOfBindings++;
      /* delete all incompatible lexemes */
      for (j = 0; j < vectorSize(lg->nodes); j++) {
	ln = (LexemNode) vectorElement(lg->nodes, j);
        if(lgForbiddenBy(lg, ln, lv->regents)
           || lgForbiddenBy(lg, ln, lv->dependents)) {
          bvSetElement(newstate->isDeletedLexemNode, TRUE, j);
#if 0
	  cdgPrintf(CDG_DEBUG, "DEBUG: forbidding lexem %s\n",
	      ln->lexem->description);
#endif
        }
      }

      nb = (NodeBinding) memMalloc(sizeof (NodeBindingStruct));
      nb->node = cn;
      nb->value = lv;
      newstate->nodeBinding = nb;

      /* skip over deleted nodes */
      nsSkipOverDeletedNodes(newstate);

      /* Add search state to agenda */
#if 0
      cdgPrintf(CDG_DEBUG,
		"DEBUG:  inserting new search state [%d] with score %d/%4.3e, nss score=%4.3e\n",
		newstate->id, newstate->badness->hard, newstate->badness->soft,
		nsComputeAgendaScore(newstate));
#endif
      if (hkVerbosity & CDG_HOOK) {
	cdgExecHook(HOOK_NSSEARCH, newstate, 3);
      }
      
      cdgAgInsert(net->searchagenda, nsComputeAgendaScore(newstate), newstate);
    }

    /* free memory for state */
    nsDeleteNetSearchState(nss);
  }

  /* measure total profile */
  totalProfile = timerElapsed(profileTime);
  cdgPrintf(CDG_INFO, "\n");
  cdgPrintf(CDG_PROFILE, "PROFILE: the search took %lums\n", totalProfile);
  cdgPrintf(CDG_INFO, "INFO: the search took %d steps\n", stepcount);
  if ( timerExpired == TRUE ) {
    cdgPrintf(CDG_INFO,
	      "INFO: search aborted because CPU time limit of %d ms was exceeded\n",
	      cdgTimeLimit );    
  }
  
  /* restore original state of the lexeme graph */
  lg->isDeletedNode = originalDeletions;
  lgComputeNoOfPaths(lg, FALSE);

  /* compute parses from netsearch states */
  for (l = bestStatesSoFar; l; l = l->next) {
    nss = (NetSearchState)l->item;
    result = listPrependElement(result, nsNSState2Parse(net, nss));
  }

  /* free bestStatesSoFar */
  listForEachDelete(bestStatesSoFar, nsDeleteNetSearchState);

#if 0
  cdgPrintf(CDG_INFO,
	    "INFO: agenda size: %d/%d\n",
	    agMaxSizeSoFar(net->searchagenda), agMaxSize(net->searchagenda));
#endif
  /* delete agenda */
  agDelete(net->searchagenda);
  net->searchagenda = NULL;

  timerStopAlarm();

  /* write statistics in xml format */
  writeXmlStatistics(1, "netsearch", "general statistics on netsearch",
      "score", bestBadnessSoFar->hard?0.0:bestBadnessSoFar->soft,
      "totalTime", (Number) totalProfile,
      "solutionTime", (Number) solutionProfile,
      "softTime", (Number) softProfile,
      "steps", (Number) stepcount, 
      NULL);

  if (profileTime != NULL) {
    timerFree(profileTime);
    profileTime = NULL;
  }

  bDelete(bestBadnessSoFar);
    
  return result;
}

/* ----------------------------------------------------------------------
 * performs a branch and bound search
 */
List nsBranchBound(ConstraintNet net, int no, char **args)
{
   int i;
   float f = -1.0;

   if (no > 0 && args[0] != NULL) {
      if (sscanf(args[0], "%d", &i)) {
	 cdgPrintf(CDG_INFO, "INFO: agenda size set to %d\n", i);
      } else {
	 cdgPrintf(CDG_WARNING,
	    "WARNING: nsBranchBound: invalid agenda size `%s',\n", args[0]);
	 cdgPrintf(CDG_WARNING,
	    "         using default of %d\n", AGENDASIZEDEFAULT);
	 i = AGENDASIZEDEFAULT;
      }
      if (no > 1 && args[1] != NULL) {
	 if (sscanf(args[1], "%f", &f)) {
	    cdgPrintf(CDG_INFO, "INFO: threshold set to %4.3e\n", f);
	 } else {
	    cdgPrintf(CDG_WARNING,
	       "WARNING: nsBranchBound: invalid threshold `%s'\n", args[1]);
	    f = -1.0;
	 }
      }
   } else {
      i = AGENDASIZEDEFAULT;
   }
   return (nsSearch(net, STBranchBound, i, (Number) f));
}

/* ----------------------------------------------------------------------
 * performs a full search
 */
List nsFullSearch(ConstraintNet net, int no, char **args)
{
   int i;

   if (no > 0 && args[0] != NULL) {
      if (sscanf(args[0], "%d", &i)) {
	 cdgPrintf(CDG_INFO, "INFO: agenda size set to %d\n", i);
      } else {
	 cdgPrintf(CDG_WARNING,
	    "WARNING: nsFullSearch: invalid agenda size `%s',\n", args[0]);
	 cdgPrintf(CDG_WARNING,
	    "         using default of %d\n", AGENDASIZEDEFAULT);
	 i = AGENDASIZEDEFAULT;
      }
   } else {
      i = AGENDASIZEDEFAULT;
   }
   return (nsSearch(net, STFullSearch, i, -1));
}

/* ----------------------------------------------------------------------
 * nsInitialize
 */
void nsInitialize()
{
  setRegister("normalization", SET_ENUM, &nsNormalization,
	      NULL, NULL, NULL,
	      "off", NSNOff,
	      "linear", NSNLinear,
	      "square", NSNSquare,
	      "depth", NSNDepth,
	      "breadth", NSNBreadth,
	      NULL);
  setRegister("autocompare", SET_BOOL, &nsAutoCompare,
	      NULL, NULL, NULL,
	      NULL);
  setRegister("searchmodifiesnet", SET_BOOL, &nsSearchModifiesNetFlag,
	      NULL, NULL, NULL,
	      NULL);
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
