/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 26/11/99
 *
 */

/* ---------------------------------------------------------------------------
 @addtogroup Parse Parse - Everything about parses.
 \author Kilian Foth (see also AUTHORS and THANKS for more)


 This module exports a generic data structure to represent a dependency
 analysis, independent of a particular solution method. Indeed, the module
 parse is expressly intended to be independent of the internal data
 structures, instead representing all information in terms of Strings and
 numbers for easier communication with the outside world. All modules that
 create or process dependency analyses are required to be able to convert
 their proprietary data structures from and to Parse structures.

 A Parse can exist independent of any other grammar element, even when no
 lexicon items, levels etc. are loaded, and it can be displayed and changed
 even so. Such Parses are called `abstract'.

 However, if you want to evaluate a CDG on a parse, the corresponding
 lexicon items, levels, and constraints must be loaded, and the Parse must
 be explicitly `decorated' with parseDecorate(). This more or less means
 that it contains an array of LevelValue structures, which can be passed
 to evalConstraint() and its variants.

 CAVEAT

 If a Parse contains LVs, they are always owned by the Parse, and not by
 any ConstraintNet. For instance, calling parseShiftEdge() on a decorated
 Parse will always delete the old LV that the Parse held and generate a
 new one. This is different from the behaviour of other modules, which
 generally assume that normal LVs are owned by the ConstraintNet that
 created them, and only delete those with an indexWRTNet of -1. Therefore,
 functions intended for use with ConstraintNets such as cvClone() may not
 do the expected thing when used on decorated Parses.

 BUGS

 The parallel existence of Parses, AnnoEntries and Analyses is redundant;
 the AnnoEntry data structure should be abolished.

 @{
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <blah.h>
#include <limits.h>
#include "parse.h"
#include "cdg.h"
#include "eval.h"
#include "hook.h"
#include "input.h"
#include "set.h"
#include "tagger.h"
#include "timer.h"
#include "lexemgraph.h"
#include "virtualnode.h"

#define MARK_SUFFIX "_disabled"

/* -- VARIABLES --------------------------------------------------------- */
Parse parseMostRecentlyCreatedParse = NULL;
Boolean parseVerifyPartially = FALSE;
Boolean parseTryOutLexemeVariants = TRUE;

/* -- MACROS ------------------------------------------------------------ */

/* -- TYPE DEFINITIONS -------------------------------------------------- */

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 Return an unused, newly allocated parse id.
 */
String parseMakeId() {
	static int parseCounter = 0;
	return strPrintf("parse%d", parseCounter++);
}

/* ----------------------------------------------------------------------
 Return a new Parse structure with nothing set. Take the argument as the
 id of the new parse, or create an old-style parse0, parse1, ... id.
 */
Parse parseNew(String id) {
	Parse result = memMalloc(sizeof (ParseStruct));
	if (id == NULL) {
		result->id = parseMakeId(); /* no strRegister needed */
	} else {
		result->id = strRegister(id);
	}

	result->userName = cdgUser;
	result->latticeId = NULL;
	result->dateOfCreation = time(NULL);
	result->grammarFiles = listClone(inputCurrentGrammar->files);
	result->searchStrategy = NULL;
	result->comment = NULL;
	result->solutionNo = -1;
	result->noOfSolutions = -1;
	result->score = 0.0;
	result->optimal = FALSE;
	result->violatedConstraints = NULL;
	result->words = NULL;
	result->levels = NULL;
	result->labels = NULL;
	result->verticesStructure = NULL;
	result->verticesLabels = NULL;
	result->evalUnary = -1;
	result->evalBinary = -1;
	result->numberOfHits = -1;
	result->evaluationTime = -1;
	result->statisticComment = NULL;
	result->LVs = NULL;
	result->lg = NULL;
	result->violations = NULL;
	result->badness = NULL;
	result->map = NULL;
	result->rmap = NULL;

	return result;
}

/* ----------------------------------------------------------------------
 Delete a Parse.
 */
void parseDelete(Parse p) {

	int i;

	cdgFreeString(p->userName);
	listDelete(p->grammarFiles);
	cdgFreeString(p->searchStrategy);
	cdgFreeString(p->comment);
	listForEachDelete(p->violatedConstraints, cdgFreeString);

	for (i = 0; i < vectorSize(p->words); i++) {
		deleteWord(vectorElement(p->words, i));
	}
	vectorDelete(p->words);

	listForEachDelete(p->levels, cdgFreeString);
	listForEachDelete(p->labels, cdgFreeString);

	if (p->verticesStructure) {
		vectorDelete(p->verticesStructure);
	}
	if (p->verticesLabels) {
		vectorDelete(p->verticesLabels);
	}
	cdgFreeString(p->statisticComment);

	if (!parseIsAbstract(p)) {
		parseStrip(p);
	}
	if (p == parseMostRecentlyCreatedParse) {
		parseMostRecentlyCreatedParse = NULL;
	}
	memFree(p);

}

/* ----------------------------------------------------------------------
 Compare two words by time index
 */
Boolean compareWord(Word a, Word b) {
	return (a->from < b->from);
}

/* ----------------------------------------------------------------------
 Given a parse, the name of a level, and the index of a word,
 return the corresponding index into the field verticesStructure.
 */
int parseIndex(Parse p, String levelname, int wordno) {
	List l;
	int nrLevels = listSize(p->levels);
	int levelindex = 0;
	String s = NULL;

	levelname = strRegister(levelname);
	for (l = p->levels; l != NULL; l = l->next) {
		s = l->item;
		if (s == levelname) {
			break;
		}
		levelindex++;
	}
	if (s != levelname) {
		cdgPrintf(CDG_ERROR, "ERROR: no level `%s' in parse %s\n", levelname,
				p->id);
		return -1;
	}
	if (wordno > vectorSize(p->words)) {
		cdgPrintf(CDG_ERROR, "ERROR: no %dth word in parse %s with %d words\n",
				wordno, p->id, vectorSize(p->words));
		return -1;
	}
	return nrLevels * wordno + levelindex;
}

/* ----------------------------------------------------------------------
 Compute p->map and p->rmap.
 */
Boolean parseComputeIndices(Parse p, LexemGraph lg) {

	int max = listSize(inputCurrentGrammar->levels) * lg->max;
	int i, vertexno, lvindex;
	Word w;
	List l;

	p->map = vectorNew(max);
	p->rmap = vectorNew(max);
	vectorSetAllElements(p->map, (Pointer) -1);
	vectorSetAllElements(p->rmap, (Pointer) -1);

	vertexno = 0;
	for (i = 0; i < vectorSize(p->words); i++) {
		w = vectorElement(p->words, i);
		for (l = p->levels; l != NULL; l = l->next) {
			String levelname = l->item;
			Level level = findLevel(levelname);
			if (!level) {
				cdgPrintf(CDG_INFO, "ERROR: no level `%s' defined.\n",
						levelname);
				return FALSE;
			}
			lvindex = inputCurrentGrammar->noOfLevels * w->from + level->no;
			vectorSetElement(p->map, (Pointer) lvindex, vertexno);
			vectorSetElement(p->rmap, (Pointer) vertexno, lvindex);

			vertexno++;
		}
	}
	return TRUE;
}

/* ----------------------------------------------------------------------
 Find the lv index corresponding to a vertex number
 */
int parseMap(Parse p, int vertexno) {
	if (vertexno == -1) {
		return -1;
	}
	if (!p->map) {
		return -1;
	}
	return (int) vectorElement(p->map, vertexno);
}

/* ----------------------------------------------------------------------
 Find the vertex number corresponding to an lv index
 */
int parseRmap(Parse p, int lvindex) {
	if (lvindex == -1) {
		return -1;
	}
	if (!p->rmap) {

		return -1;
	}

	return (int) vectorElement(p->rmap, lvindex);
}

/* ----------------------------------------------------------------------
 Calculate the structure of a Parse, i.e. the fields words, levels
 etc., from the Vector of LVs.
 */
void parseComputeStructure(Parse p, Vector LVs, ByteVector isDeleted) {
	Vector map;
	LexemNode ln, ln2;
	LevelValue lv = NULL;
	List l, m;
	Word w;
	int i, j;
	int firstLevel = -1;
	List found = NULL;
	LexemGraph lg = NULL;

	/* grant this please */
	if (!LVs || vectorSize(LVs) == 0) {
		/* never reach */
		cdgPrintf(CDG_ERROR,
				"ERROR: no levelvalues given to parseComputeStructure\n");
		abort();
	}

	/* build words and levels */

	/* we're going to find all used words by selecting the correct modifier
	 * from all LVs of the first level found. If the lexical path is
	 * ambiguous, the user has already been warned. */

	/* search first value and level */
	for (i = 0; i < vectorSize(LVs); i++) {
		lv = vectorElement(LVs, 0);
		if (lv) {
			if (!lg) {
				lg = ((LexemNode) (lv->dependents->item))->lexemgraph;
			}
			firstLevel = lv->level->no;
			break;
		}
	}
	if (!lv) {
		cdgPrintf(CDG_ERROR,
				"ERROR: no levelvalues given to parseComputeStructure\n");
		abort();
	}

	map = vectorNew(((LexemNode) (lv->dependents->item))->lexemgraph->max);
	p->levels = NULL;
	p->words = vectorNew(0);
	j = 0;
	for (i = 0; i < vectorSize(LVs); i++) {
		lv = vectorElement(LVs, i);
		if (!lv)
			continue;

		/* register level */
		p->levels = listAddUniqueElement(p->levels, lv->level->id);

		/* register label */
		p->labels = listAddUniqueElement(p->labels, lv->label);

		/* check for ambiguous LVs */
		for (l = lv->dependents; l != NULL; l = l->next) {
			ln = l->item;
			if (listContains(found, ln) || bvElement(isDeleted, ln->no)) {
				continue;
			}
			for (m = l->next; m != NULL; m = m->next) {
				ln2 = m->item;
				if (listContains(found, ln2) || bvElement(isDeleted, ln2->no)) {
					continue;
				}

				found = listAddUniqueElement(found, ln);
				found = listAddUniqueElement(found, ln2);
				cdgPrintf(CDG_WARNING, "WARNING: ");
				lgPrintNode(CDG_WARNING, ln);
				cdgPrintf(CDG_WARNING, " chosen, but\n         ");
				lgPrintNode(CDG_WARNING, ln2);
				cdgPrintf(CDG_WARNING, " is undeleted as well!\n");
			}
		}

		/* register word */
		if (lv->level->no != firstLevel) {
			continue;
		}

		ln = NULL;
		for (l = lv->dependents; l != NULL; l = l->next) {
			ln = l->item;
			if (!bvElement(isDeleted, ln->no)) {
				break;
			}
		}

		w = newWord();
		w->from = ln->arc->from;
		w->to = ln->arc->to;
		w->word = ln->lexem->word;
		w->description = strRegister(ln->lexem->description);
		vectorAddElement(p->words, w);
	}
	listDelete(found);
	vectorSort(p->words, compareWord);

	/* compute mapping from time index to word index */
	j = 0;
	for (i = 0; i < vectorSize(p->words); i++) {
		w = vectorElement(p->words, i);
		vectorSetElement(map, (Pointer) j++, w->from);
	}

	/* Note that parseComputeIndices() cannot fail here.
	 It only returns FALSE if inputCurrentGrammar does
	 not contain a level named in p, and that can't happen
	 because we only just got the level's name from there. */
	(void) parseComputeIndices(p, lg);

	/* compute structure */
	p->verticesStructure = vectorNew(vectorSize(LVs));
	p->verticesLabels = vectorNew(vectorSize(LVs));

	for (i = 0; i < vectorSize(LVs); i++) {
		lv = vectorElement(LVs, i);
		if (!lv)
			continue;

		if (lv->regents == NULL) {
			j = -1;
		} else if (lv->regents == NONSPEC) {
			j = -2;
		} else {
			j = (int) vectorElement(map,
					((LexemNode) (lv->regents->item))->arc->from);
		}
		vectorAddElement(p->verticesStructure, (void *) j);
		vectorAddElement(p->verticesLabels, lv->label);
	}

	vectorDelete(map);

}

/* ----------------------------------------------------------------------
 * Generate and return an LV that corresponds to the binding at position
 * where in the Parse.
 *
 * There could be more than one fitting LV, if there is more than one fitting lexeme for the word at where
 */
List/*<LevelValue>*/ parseFindLV(Parse p, int where, AnnoEntry ae) {
	LevelValue lv = NULL;
	List lvs = listNew();
	int nrLevels = listSize(p->levels);
	List dependents = NULL;
	List regents = NULL;

	String levelId;
	Level level;
	String label;
	Word modifierWord;
	LexemNode modifier;
	List modifierCandidates;
	Word modifieeWord;
	LexemNode modifiee;
	List modifieeCandidates;
	int j;
	String s = NULL;
	List l,m;

	if (0 == nrLevels) {
		return NULL;
	}

	levelId = strRegister(listNthElement(p->levels, (where % nrLevels) + 1));
	level = findLevel(levelId);
	label = strRegister(vectorElement(p->verticesLabels, where));
	modifierWord = vectorElement(p->words, where / nrLevels);
	modifierCandidates= parseFindLn(p, modifierWord, ae);
	//modifier = modifierCandidates? modifierCandidates->item : NULL;

	j = (int) vectorElement(p->verticesStructure, where);

	if (j == -1) {
		modifieeWord = NULL;
		// modifiee = NULL;
		modifieeCandidates = listAppendElement(listNew(), NULL);
	} else if (j == -2) {
		modifieeWord = NULL;
		// modifiee = NONSPEC;
		modifieeCandidates = listAppendElement(listNew(), NONSPEC);
	} else {
		modifieeWord = vectorElement(p->words, j);
		modifieeCandidates = parseFindLn(p, modifieeWord, ae);
		// modifiee = modifieeCandidates? modifieeCandidates->item : NULL;
	}

	if (!level) {
		cdgPrintf(CDG_ERROR, "ERROR: can't find level `%s'.\n", levelId);
		cdgPrintf(CDG_ERROR, "ERROR: can't decorate this parse.\n");
		return NULL;
	}

	/* check whether the level has this label */
	for (l = level->labels; l != NULL; l = l->next) {
		s = l->item;
		if (s == label) {
			break;
		}
	}

	if (s != label) {
		cdgPrintf(CDG_WARNING,
				"WARNING: level `%s' has no label `%s', using `%s'.\n",
				levelId, label, s);
		label = s;
	}

	if (!modifierCandidates) {
		cdgPrintf(CDG_ERROR, "ERROR: can't find word `%s'.\n",
				modifierWord->word);
		cdgPrintf(CDG_ERROR, "ERROR: Can't decorate this parse.\n");
		return NULL;
	}

	if (!modifieeCandidates) {
		cdgPrintf(CDG_ERROR, "ERROR: can't find word `%s'.\n",
				modifieeWord->word);
		cdgPrintf(CDG_ERROR, "ERROR: Can't decorate this parse.\n");
		return NULL;
	}

	/* Fabricate LVs */
	for(l= modifierCandidates; l ; l=l->next) {
		for(m= modifieeCandidates; m ; m=m->next) {
			modifier = l->item;
			modifiee = m->item;
			dependents = listPrependElement(NULL, modifier);
			if (spec(modifiee)) {
				regents = listPrependElement(NULL, modifiee);
			} else {
				regents = (List) modifiee;
			}
			lv = lvNew(dependents, level, label, regents);
			lvs= listAppendElement(lvs, lv);
		}
	}

	return lvs;
}

/* ----------------------------------------------------------------------
 Check whether a Parse is abstract.
 */
Boolean parseIsAbstract(Parse p) {
	return (p->lg == NULL || p->LVs == NULL || p->badness == NULL);
}

/* ----------------------------------------------------------------------
 * How well does an match le?
 */
int anCompare(Annotation an, LexiconItem le, Boolean verbose) {

	List l, path;
	Value v;
	String c = NULL;
	int result = 0;
	Hierarchy h = strlen(inputFeatureHierarchy) ? findHierarchy(
			inputFeatureHierarchy) : NULL;

	for (l = an->specs; l; l = l->next) {
		Specification s = (Specification) l->item;
		if (s->type != STTag) {
			continue;
		}
		v = le->values[inputRegisterPath(path = scanPath(s->kind),
				inputCurrentGrammar)];
		listDelete(path);
		if (v == ErrorValue) {
			if (verbose) {
				cdgPrintf(CDG_WARNING,
						"WARNING: feature `%s' does not exist\n", s->kind);
			}
			result++;
		} else {
			int i = inputCompareFeatures(s->name, (c = sprintValue(v)), h);
			if (i > 1 && verbose) {
				cdgPrintf(
						CDG_WARNING,
						"WARNING: feature `%s' has value `%s', should be `%s'\n",
						s->kind, c, s->name);
			}
			result += i;
		}
	}

	/* penalize variants that differ in case */
	if (an->word != le->word) {
#ifdef DEBUGANNO2LE
		cdgPrintf(CDG_DEBUG,
				"WARNING: annotation says `%s', lexicon says `%s'\n",
				an->word, le->word);
#endif
		result++;
	}

	return result;
}

/* ----------------------------------------------------------------------
 Translate Annotation to LexiconItem.

 Among the currently loaded LexiconItem structures, the one that matches
 A best is selected and returned. If no entry even has the same word
 form, NULL may be returned.
 */
LexiconItem anno2le(Annotation an) {
	List l, entries;
	LexiconItem result = NULL;
	int i, record = INT_MAX;
	String lower;

#ifdef DEBUGANNO2LE
	cdgPrintf(CDG_INFO, "DEBUG: translating `%s'\n", an->word);
#endif

	/* Pick the lexical entry that best matches the specified tags */
	entries = inputLexiconGet(an->word);

	/* allow lower case form of a word */
	lower = strCopy(an->word);

	{
		/* skip initial ( or " */
		int index = 0;
		while (((unsigned char *) lower)[index] == '"'
				|| ((unsigned char *) lower)[index] == '(') {
			index++;
		}
		lower[index] = tolower(lower[index]);
	}
	entries = listAppendList(entries, inputLexiconGet(lower));

	for (l = entries; l != NULL; l = l->next) {
		LexiconItem le = l->item;
#ifdef DEBUGANNO2LE
		i = anCompare(an, le, TRUE);
		cdgPrintf(CDG_DEBUG, "DEBUG:  `%s' has distance %d\n", le->description, i);
#else
		i = anCompare(an, le, FALSE);
#endif
		if (i < record) {
			record = i;
			result = le;
		}
	}

	/* if there is no word in the lexicon that satisfies the annotation,
	 use the one that matches best, but warn the user about the differences */
	if (record && result) {
		(void) anCompare(an, result, TRUE);
		cdgPrintf(CDG_WARNING, "WARNING: `%s' differs in %d feature%s.\n",
				result->description, record, record == 1 ? "" : "s");
	}

	if (result) {
#ifdef DEBUGANNO2LE
		cdgPrintf(CDG_INFO, "DEBUG: selected `%s'\n", result->description);
#endif
		result = inputCloneLI(result);
	}

	listForEachDelete(entries, freeLexiconItem);
	return result;

}

/* ----------------------------------------------------------------------
 Translate Annotation to LexemNode.

 Among the nodes of LG, the one that matches A best is selected and
 returned.

 An arbitrary number of feature mismatches are allowed if necessary, but
 if no node even has the same word form, NULL may be returned.
 */
List/*<LexemNode>*/ anno2ln(Annotation an, LexemGraph lg) {
	List result = listNew();
	int i, record = INT_MAX;

#ifdef DEBUGANNO2LE
	cdgPrintf(CDG_INFO, "DEBUG: translating `%s(%d,%d)'\n", an->word, an->from, an->to);
#endif

	for (i = 0; i < vectorSize(lg->nodes); i++) {
		LexemNode ln = vectorElement(lg->nodes, i);
		int j;
		if (ln->arc->from != an->from || ln->arc->to != an->to) {
			continue;
		}

#ifdef DEBUGANNO2LE
		j = anCompare(an, ln->lexem, TRUE);
		cdgPrintf(CDG_INFO, "DEBUG:  `%s' has distance %d\n", ln->lexem->description, j);
#else
		j = anCompare(an, ln->lexem, FALSE);
#endif
		if (j < record) {
			record = j;
			result = listAppendElement(listNew(), ln);
		} else if (j == record) {
			result = listAppendElement(result, ln);
		}
	}

	/* If a result is found but that satisfies the annotation,
	 use the one that matches best,
	 but warn the user about the differences */
	if (record && result) {
		// (void) anCompare(an, result->lexem, TRUE);
		cdgPrintf(CDG_WARNING, "WARNING: %s: `%s' differs in %d feature%s.\n",
				lg->lattice->id, ((LexemNode)result->item)->lexem->description,
				record, record == 1 ? "" : "s");
	}

	if (result) {
#ifdef DEBUGANNO2LE
		cdgPrintf(CDG_INFO, "DEBUG: selected `%s'\n", ((LexemNode)result->item)->lexem->description);
#endif
	}

	return result;

}

/* ----------------------------------------------------------------------
 Find the lexeme(s) node that corresponds to w best.

 If AE is not NULL, it is used to decide which of several homonyms was
 meant.
 */
List/*<LexemNode>*/ parseFindLn(Parse p, Word w, AnnoEntry ae) {
	Annotation an = NULL;
	List l;

	/* We cannot use parseIsAbstract() here,
	 because this function is used by parseDecorate() itself,
	 when the parse is still semi-decorated. */
	if (!p->lg) {
		return listNew();
	}

	/* If we have an AnnoEntry, find the relevant Annotation in it */
	if (ae) {
		for (l = ae->annos; l != NULL; l = l->next) {
			Annotation a = l->item;
			if (a->from == w->from && a->to == w->to) {
				an = a;
				break;
			}
		}
	}

	/* If no annotation is available, just return a vaguely similar
	 lexeme node */
	if (!an) {
		List result = listNew();
		int i;
		for (i = 0; i < vectorSize(p->lg->nodes); i++) {
			LexemNode ln = vectorElement(p->lg->nodes, i);
			if (ln->arc->from != w->from || ln->arc->to != w->to) {
				continue;
			}
			result = listAppendElement(result, ln);
			if (ln->lexem->description == w->description) {
				return listAppendElement(listNew(), ln);
			}
		}
		if (!result) {
			cdgPrintf(CDG_DEBUG, "DEBUG: lexeme `%s' not found!\n",
					w->description);
		}
		return result;
	}

	/* If we have an annotation, use it to select the desired variant.  */
	return anno2ln(an, p->lg);

}


/**
 * compares two lists of LexemNodes for if they have a common entry
 */
Boolean haveCommonLN(List l1, List l2) {
	List l;
	LexemNode ln;

	for (l= l1; l ; l=l->next) {
		ln = l->item;
		if(listContains(l2, ln)) {
			return TRUE;
		}
	}

	return FALSE;
}

/*
 * given a list of lists of alternative LVs,
 * this function spell the combinations out recursively
 */
List/*<Vector<LevelValue>>*/ spellOutVariants(Vector soFar, List alternatives, ByteVector validLNs) {
	List variants = listNew();

	if(! alternatives) {
		// base case, just return a new list containing only the vector soFar
		return listAppendElement(variants, soFar);
	}

	List lvs       = alternatives->item;
	List remaining = alternatives->next;
	List l, m;
	LevelValue lv; //lvlast;
	LexemNode ln;
	Vector continuation;
	Boolean compatibleD, compatibleR, contains;
	ByteVector validLNsCopy;


	for (l= lvs; l; l=l->next) {
		lv = l->item;

//		// first cut out variants where on different levels different lexemes are selected for the same word,
//		// otherwise we get too many variants
//		if(vectorSize(soFar) > 0 ) {
//			lvlast= vectorElement(soFar, vectorSize(soFar)-1);
//		} else {
//			lvlast = NULL;
//		}
//		if(lvlast && lvGetDependentGN(lv) == lvGetDependentGN(lvlast)
//				&& !haveCommonLN(lv->dependents, lvlast->dependents)) {
//			continue;
//		}

		validLNsCopy = bvClone(validLNs);

		compatibleD = FALSE;
		for(m = lvGetDependentGN(lv)->lexemes; m ; m=m->next) {
			ln= m->item;
			contains = listContains(lv->dependents, ln);
			if(contains && bvElement(validLNsCopy, ln->no)) {
				compatibleD = TRUE;
			} else {
				bvSetElement(validLNsCopy, FALSE, ln->no);
			}
		}

		GraphemNode reg =  lvGetRegentGN(lv);
		if(reg) {
			compatibleR = FALSE;
			for(m = reg->lexemes; m ; m=m->next) {
				ln= m->item;
				contains = listContains(lv->regents, ln);
				if(contains && bvElement(validLNsCopy, ln->no)) {
					compatibleR = TRUE;
				} else {
					bvSetElement(validLNsCopy, FALSE, ln->no);
				}
			}
		} else {
			compatibleR = TRUE;
		}

		if(compatibleD && compatibleR) {
			continuation = vectorClone(soFar);
			vectorAddElement(continuation, lv);

			variants = listAppendList(variants, spellOutVariants(continuation, remaining, validLNsCopy));
		}
	}
	return variants;
}

/* ----------------------------------------------------------------------
 Decorate a Parse with the optional elements required for eval'ing.
 The parse is not actually evaluated by this function.

 A clone of the lexeme graph passed is inserted into P, and fitting LVs
 are allocated and installed.

 Which lexical entry is used for each of the words depends on the third
 argument. If AE is not NULL, the lexeme from LG that corresponds most
 closely to the specifications in AE is used. Otherwise that lexeme that
 has the same name as the field w->description is used. If no such lexeme
 exists, one is chosen at random.

 */
Boolean parseDecorate(Parse p, LexemGraph lg, AnnoEntry ae) {
	LevelValue lv = NULL;
	List lvs;
	List l;
	int i;
	LexemNode ln;
	List lns;
	Word w;

	/* clone a private lexeme graph all of our own */
	p->lg = lgClone(lg);
	bvSetAllElements(p->lg->isDeletedNode, TRUE);
	for (i = 0; i < vectorSize(p->words); i++) {
		w = vectorElement(p->words, i);
		if (!(lns = parseFindLn(p, w, ae))) {
			parseStrip(p);
			return FALSE;
		}
		ln = lns->item;
		bvSetElement(p->lg->isDeletedNode, FALSE, ln->no);
		w->description = ln->lexem->description;
	}

	if (!parseComputeIndices(p, lg)) {
		parseStrip(p);
		return FALSE;
	}

	/* allocate vector of LVs */
	if (p->LVs) {
		vectorDelete(p->LVs);
	}

	/* Note well: although we allocate p->LVs to the same size
	 as p->verticesStructure, it may actually become larger later!
	 p->verticesStructure is as large as the number of constraint edges.
	 p->LVs may be larger if there are `long' words in the parse.
	 p->LVs is always indexed by lvIndex(). */
	p->LVs = vectorNew(vectorSize(p->verticesStructure));


	if(parseTryOutLexemeVariants) {
		p->violations = NULL;
		p->badness = bestBadness();

		/* find all combinations and select the best one */
		// first we identify the alternatives
		List alternatives = listNew(); // a list of lists of LVs
		for (i = 0; i < vectorSize(p->verticesStructure); i++) {
			lvs = parseFindLV(p, i, ae);
			if (!lvs) {
				parseStrip(p);
				return FALSE;
			}
			alternatives = listAppendElement(alternatives, lvs);
		}

		// then we try them all out
		ByteVector validLNs = bvNew(vectorSize(lg->nodes));
		bvSetAllElements(validLNs, TRUE);
		List variants = spellOutVariants(vectorNew(vectorSize(p->verticesStructure)), alternatives, validLNs);
		Vector best = NULL;
		Vector current;
		Badness bestB = worstBadness();

		cdgPrintf(CDG_INFO, "Trying out %d lexeme variant%s \n", listSize(variants), listSize(variants) == 1?"":"s");

		for (l= variants; l; l=l->next) {
			current = l->item;
			p->LVs = current;
			parseEval(p);

			if(bCompare(p->badness, bestB) ) {
				best = current;
				bestB = bClone(p->badness);
			}
		}

		p->LVs= best;
		parseEval(p);

	} else {
		/* fill it with LVs owned by nobody else but us */
		for (i = 0; i < vectorSize(p->verticesStructure); i++) {
			lv = parseFindLV(p, i, NULL)->item;
			if (!lv) {
				parseStrip(p);
				return FALSE;
			}
			vectorSetElement(p->LVs, lv, lvIndex(lv));
		}
	}

	return TRUE;
}

/* ----------------------------------------------------------------------
 * Make a decorated Parse abstract again.
 */
void parseStrip(Parse p) {
	int i;
	LevelValue lv;

	if (p->LVs) {
		for (i = 0; i < vectorSize(p->LVs); i++) {
			lv = vectorElement(p->LVs, i);
			if (lv) {
				lvDelete(lv);
			}
		}
		vectorDelete(p->LVs);
		p->LVs = NULL;
	}

	if (p->violations) {
		listForEachDelete(p->violations, cvDelete);
		p->violations = NULL;
	}

	if (p->lg) {
		freeLattice(p->lg->lattice);
		lgDelete(p->lg);
		p->lg = NULL;
	}
	if (p->badness) {
		bDelete(p->badness);
	}
	if (p->map) {
		vectorDelete(p->map);
	}
	if (p->rmap) {
		vectorDelete(p->rmap);
	}
}

/* ----------------------------------------------------------------------
 *   Compute the score of a Parse.
 */
void parseEval(Parse p) {
	ConstraintViolation cv;
	List l;
	int i, j;
	LevelValue lva, lvb;

	/* must have the LVs to work! */
	if (parseIsAbstract(p)) {
		cdgPrintf(CDG_WARNING, "WARNING: no levelvalues in parse `%s'\n", p->id);
		return;
	}

	/* retract previous info */
	if (p->violations != NULL) {
		listForEachDelete(p->violations, cvDelete);
		p->violations = NULL;
	}
	if (p->violatedConstraints != NULL) {
		listForEachDelete(p->violatedConstraints, cdgFreeString);
		p->violatedConstraints = NULL;
	}

	p->score = 1.0;
	p->badness->no = 0;
	p->badness->hard = 0;
	p->badness->soft = 1.0;

	/* seek unary violations */
	for (i = 0; i < vectorSize(p->LVs); i++) {
		lva = vectorElement(p->LVs, i);
		if (lva) {
			Number penalty = evalUnary(lva, NULL, p->LVs, FALSE, p->badness,
					&p->violations, FALSE);
			if (cdgMultiply == cdgAggregation) {
				p->score *= penalty;
			} else {
				p->score = min(p->score, penalty);
			}
		}
	}

	/* seek binary violations */
	for (i = 0; i < vectorSize(p->LVs); i++) {
		lva = vectorElement(p->LVs, i);
		if (!lva)
			continue;
		for (j = i + 1; j < vectorSize(p->LVs); j++) {
			lvb = vectorElement(p->LVs, j);
			if (lvb) {
				Number penalty = evalBinary(lva, lvb, NULL, p->LVs, FALSE,
						p->badness, &p->violations);
				if (cdgMultiply == cdgAggregation) {
					p->score *= penalty;
				} else {
					p->score = min(p->score, penalty);
				}
			}
		}
	}

	/* sort the list of violations */
	l = listSort(p->violations, cvCompare);
	listDelete(p->violations);
	p->violations = l;

	/* compute constraint list */
	for (l = p->violations; l != NULL; l = l->next) {
		cv = l->item;
		p->violatedConstraints = listAddUniqueElement(p->violatedConstraints,
				cv->constraint->id);
	}

}

/* ----------------------------------------------------------------------
 Recompute conflicts and score after a change at the points indicated by
 CHANGES. The list is then deallocated, on the assumption that you
 created it specifically for calling this function.
 */
void parseReEval(Parse p, List changes) {
	ConstraintViolation cv;
  List l, m, newViolations = NULL;
  int i, j;
  LevelValue lva, lvb;
	List indices;

	/* must have the LVs to work! */
	if (!p->LVs) {
		return;
	}

	/* build list of lv indices */
	indices = NULL;
	for (l = changes; l != NULL; l = l->next) {
		int i = (int) l->item;
		indices = listPrependElement(indices, vectorElement(p->map, i));
	}
	listDelete(changes);

	/* retract conflicts that may have disappeared */
	for (l = p->violations; l != NULL; l = l->next) {
		cv = l->item;

		if (!cv->constraint->is_context_sensitive && !listContains(indices,
				(Pointer) cv->nodeBindingIndex1) && (cv->nodeBindingIndex2
				== -1
				|| !listContains(indices, (Pointer) cv->nodeBindingIndex2))) {
			newViolations = listAppendElement(newViolations, cv);

		} else {
			if (cdgMultiply == cdgAggregation) {
				bSubtract(p->badness, cv->penalty);
			}
			cvDelete(cv);
		}
	}
	listDelete(p->violations);
	p->violations = newViolations;
	newViolations = NULL;

	/* Under the minimum strategy, Badnesses cannot be computed in reverse.
	 Therefore, recompute the partial Badness from scratch. */
	if (cdgMinimum == cdgAggregation) {
		p->badness->no = 0;
		p->badness->hard = 0;
		p->badness->soft = 1.0;
		for (l = p->violations; l != NULL; l = l->next) {
			cv = l->item;
			bAdd(p->badness, cv->penalty);
		}
	}

	for (l = indices; l != NULL; l = l->next) {
		i = (int) l->item;

		lva = vectorElement(p->LVs, i);

		if (!lva) {
			continue;
		}

		/* seek new unary violations */
		(void) evalUnary(lva, NULL, NULL, FALSE, p->badness, &newViolations,
				FALSE);

		/* seek binary violations */

		/* (A) binary conflicts between changed and unchanged LVs */
		for (j = 0; j < vectorSize(p->LVs); j++) {
			if (listContains(indices, (Pointer) j)) {
				continue;
			}

			lvb = vectorElement(p->LVs, j);
			if (lvb)
				(void) evalBinary(lva, lvb, NULL, NULL, FALSE, p->badness,
						&newViolations);
		}

		/* (B) binary conflicts between two changed LVs */
		for (m = l->next; m != NULL; m = m->next) {

			j = (int) m->item;
			lvb = vectorElement(p->LVs, j);
			if (lvb)
				(void) evalBinary(lva, lvb, NULL, NULL, FALSE, p->badness,
						&newViolations);
		}
#ifdef TERNARY
        if(inputCurrentGrammar->no_ternary) {
		/* seek ternary violations */
		/* (A) ternary conflicts between one changed and two unchanged LVs */
		for (j = 0; j < vectorSize(p->LVs); j++) {
			if (listContains(indices, (Pointer) j)) {
				continue;
			}
			lvb = vectorElement(p->LVs, j);
			if (lvb) {
            int k;
				for (k = j + 1; k < vectorSize(p->LVs); k++) {
					if (listContains(indices, (Pointer) k)) {
						continue;
					}

                LevelValue lvc = vectorElement(p->LVs, k);
					if (lvc) {
						(void) evalTernary(lva, lvb, lvc, NULL, NULL, FALSE,
								p->badness, &newViolations);
					}
				}
			}
		}

		/* (B) ternary conflicts between two changed LVs and one unchanged LV */
		for (m = l->next; m != NULL; m = m->next) {

			j = (int) m->item;
			lvb = vectorElement(p->LVs, j);
			if (lvb) {
            int k;
				for (k = 0; k < vectorSize(p->LVs); k++) {
					if (listContains(indices, (Pointer) k)) {
						continue;
					}

                LevelValue lvc = vectorElement(p->LVs, k);
					if (lvc) {
						(void) evalTernary(lva, lvb, lvc, NULL, NULL, FALSE,
								p->badness, &newViolations);
					}
				}
			}
		}

		/* (c) ternary conflicts between three changed LVs*/
		for (m = l->next; m != NULL; m = m->next) {

			j = (int) m->item;
			lvb = vectorElement(p->LVs, j);
			if (lvb) {
            List n;
				for (n = m->next; n != NULL; n = n->next) {
                int k = (int) n->item;
                LevelValue lvc = vectorElement(p->LVs, k);
					if (lvc) {
						(void) evalTernary(lva, lvb, lvc, NULL, NULL, FALSE,
								p->badness, &newViolations);
					}
				}
			}
		}
        }
#endif
	}

	/* find context-sensitive conflicts */
	(void) evalInContext(p->LVs, p->LVs, p->badness, &newViolations);

	/* insert sorted the new violations into the remaining violations */
	for (l = newViolations; l; l = l->next) {
		p->violations = listInsertSorted(p->violations, l->item, cvCompare);
	}
	listDelete(newViolations);

	/* compute score */
	p->score = 1.0;
	for (l = p->violations; l != NULL; l = l->next) {

		if (cdgMultiply == cdgAggregation) {
			p->score *= ((ConstraintViolation) l->item)->penalty;
		} else {
			Number penalty = ((ConstraintViolation) l->item)->penalty;
			p->score = min(p->score, penalty);
		}
	}

	/* compute constraint list */
	listDelete(p->violatedConstraints);
	p->violatedConstraints = NULL;
	for (l = p->violations; l != NULL; l = l->next) {
		cv = l->item;
		p->violatedConstraints = listAddUniqueElement(p->violatedConstraints,
				cv->constraint->id);
	}
	listDelete(indices);

}

/* ----------------------------------------------------------------------
 Change one label in a Parse.

 The label of the edge at index WHERE is changed to LABEL.
 If EVAL is specified, the grammar is immediately reevaluated.
 */
void parseSwapLabel(Parse p, int where, String label, Boolean eval) {
	List l;
	LevelValue new, old;
	String oldLabel = vectorElement(p->verticesLabels, where);

	label = strRegister(label);

	/* swap label */
	vectorSetElement(p->verticesLabels, label, where);
	p->labels = listAddUniqueElement(p->labels, label);
	if (!listContains(p->labels, oldLabel)) {
		p->labels = listDeleteElement(p->labels, oldLabel);
	}

	if (!parseIsAbstract(p)) {

		/* change LV to suit */
		old = vectorElement(p->LVs, (int) vectorElement(p->map, where));
		new = parseFindLV(p, where, NULL)->item;
		if (old == new) {
			return;
		}
		lvDelete(old);
		vectorSetElement(p->LVs, new, lvIndex(new));

		/* sanity check */
		if (!listContains(new->level->labels, label)) {
			cdgPrintf(CDG_WARNING,
					"WARNING: label `%s' not declared for level `%s'!\n",
					label, new->level->id);
		}

		if (eval) {
			/* re-eval grammar */
			l = listAppendElement(NULL, (Pointer) where);
			parseReEval(p, l);
		}
	}
}

/* ----------------------------------------------------------------------
 Change the governor of one edge in a Parse.

 The edge at index WHERE is changed so that its new governor is word TO.
 If EVAL is specified, the grammar is immediately reevaluated.
 */
void parseShiftEdge(Parse p, int where, int to, Boolean eval) {
	List l;
	LevelValue old, new;

	/* shift edge */
	vectorSetElement(p->verticesStructure, (void *) to, where);

	if (!parseIsAbstract(p)) {

		/* change LV to suit */

		old = vectorElement(p->LVs, (int) vectorElement(p->map, where));
		new = parseFindLV(p, where, NULL)->item;
		if (old == new) {
			return;
		}
		lvDelete(old);
		vectorSetElement(p->LVs, new, lvIndex(new));

		/* re-eval grammar */
		if (eval) {
			l = listAppendElement(NULL, (Pointer) where);
			parseReEval(p, l);
		}
	}
}

/* ----------------------------------------------------------------------
 Exchange a lexical reading in a Parse.

 The word number WHERE is changed to the form with the description ID.
 If EVAL is specified, the grammar is immediately reevaluated.
 */
void parseSwapWord(Parse p, int where, String id, Boolean eval) {
	LexemNode oldln = NULL, newln;
	LevelValue new, old;
	List changed;
	int nrLevels = listSize(p->levels);
	int i;
	Word w = vectorElement(p->words, where);
	String oldDescription = w->description;

	if (!parseIsAbstract(p)) {
		oldln = parseFindLn(p, w, NULL)->item;
	}

	/* swap id */
	w->description = strRegister(id);

	if (!parseIsAbstract(p)) {

		newln = parseFindLn(p, w, NULL)->item;

		if (!newln) {
			cdgPrintf(CDG_ERROR, "ERROR: lexeme node `%s' doesn't exist.\n", id);
			w->description = strRegister(oldDescription);
			return;
		}

		/* maybe change reading */
		w->word = newln->lexem->word;

		/* prevent insidious bug that would happen in the next line. */
		if (newln == oldln) {
			return;
		}

		/* select node */
		bvSetElement(p->lg->isDeletedNode, FALSE, newln->no);
		lgDeleteNode(p->lg, oldln);

		/* update LVs */
		changed = NULL;
		for (i = 0; i < vectorSize(p->verticesStructure); i++) {
			if (i / nrLevels == where || (int) vectorElement(
					p->verticesStructure, i) == where) {
				new = parseFindLV(p, i, NULL)->item;
				if (!new) {
					cdgPrintf(CDG_ERROR, "Error: parseFindLV() failed.\n");
					return;
				}
				old = vectorElement(p->LVs, lvIndex(new));
				if (old != new) {
					vectorSetElement(p->LVs, new, lvIndex(new));
					changed = listPrependElement(changed, (Pointer) i);
					lvDelete(old);
				}
			}
		}

		/* re-eval */
		if (eval) {
			parseReEval(p, changed);
		} else {
			listDelete(changed);
		}
	}
}

/* ----------------------------------------------------------------------
 Manipulate a Parse so that all mirror edges are unarily optimal.

 If EVAL is specified, the grammar is immediately reevaluated.
 */
void parseMirror(Parse p, Boolean eval) {
	Badness best;
	List l, m;
	String label;
	int i, j, k;
	int nrLevels = listSize(p->levels);
	int nrWords = vectorSize(p->words);
	List troublespots = NULL;

	/* refuse to work without LVs */
	if (parseIsAbstract(p)) {
		return;
	}

	/* find instances of hard violations */
	for (l = p->violations; l != NULL; l = l->next) {
		ConstraintViolation cv = l->item;
		if (!cv->penalty) {
			troublespots = listAddUniqueElement(troublespots,
					(Pointer) cv->nodeBindingIndex1);
			if (cv->nodeBindingIndex2 != -1) {
				troublespots = listAddUniqueElement(troublespots,
						(Pointer) cv->nodeBindingIndex2);
			}
		}
	}

	/* iterate over levels */
	for (i = 0, l = p->levels; l != NULL; l = l->next, i++) {
		String s = l->item;

		label = NULL;
		for (m = inputCurrentGrammar->levels; m != NULL; m = m->next) {
			Level level = m->item;
			if (s == level->id && (label = getLevelProperty(level, "mirror")))
				break;
		}
		/* find mirror levels */
		if (!label) {
			continue;
		}

		best = bClone(p->badness);

		/* iterate over words */
		for (j = 0; j < nrWords; j++) {
			int index = nrLevels * j + i;
			LevelValue lv = vectorElement(p->LVs, (int) vectorElement(p->map,
					index));
			int record = (int) vectorElement(p->verticesStructure, index);

			if (!lv) /* never reach */
				continue;

			/* only search if this LV causes hard conflicts */
			if (!listContains(troublespots, (Pointer) index)) {
				continue;
			}

			bCopy(best, p->badness);

			/* iterate over attachment points */
			for (k = -1; k < nrWords; k++) {
				if (k == j) {
					continue;
				}
				parseShiftEdge(p, index, k, eval);

				if (bCompare(p->badness, best)) {
					bCopy(best, p->badness);
					record = k;
				}
			}
			parseShiftEdge(p, index, record, eval);
		}
	}
	listDelete(troublespots);
}

/* ----------------------------------------------------------------------
 Computes the height of level id in Parse p.
 */
int parseHeight(Parse p, String levelname) {
	int wordNo, height, treeHeight = 0;

	/* get the maximum height over all words */
	for (wordNo = 0; wordNo < vectorSize(p->words); wordNo++) {
		height = parseWordHeight(p, levelname, wordNo);

		/* cycle? */
		if (height > vectorSize(p->verticesStructure)) {
			return -1;
		}
		treeHeight = max(treeHeight, height);
	}
	return treeHeight;
}

/* -------------------------------------------------------------------------
 * get all child nodes of a word on a level tree
 *
 * @param parse the current parse that we examine
 * @param wordNo the word index in the word vector of the current parse
 * @param level a Level
 * @returns a list of word indices or NULL if there are no dependents.
 *
 * @note you become the owner of the returned @ref List,
 * so deallocate (@ref ListFree) it after you've consumed the result.
 */
List parseGetModifiers(Parse parse, Level level, int wordIndex) {
	int i;
	List dependents = NULL;
	int nrLevels = listSize(parse->levels);

	for (i = 0; i < vectorSize(parse->words); i++) {
		int modifieeIndex = (int) vectorElement(parse->verticesStructure,
				nrLevels * i + level->no);
		if (wordIndex == modifieeIndex) {
			dependents = listAppendElement(dependents, (Pointer) i);
		}
	}

	return dependents;
}

/* ----------------------------------------------------------------------
 Computes the depth of a word on a level in Parse p, that is
 the maximum number of hops to the leaf nodes
 */
int parseWordDepth(Parse parse, Level level, int wordIndex) {
	int maxDepth = 0;
	List dependents = parseGetModifiers(parse, level, wordIndex);

	if (listSize(dependents) != 0) {
		List l;
		int depth;

		for (l = dependents; l; l = l->next) {
			int modifierIndex = (int) l->item;
			depth = parseWordDepth(parse, level, modifierIndex) + 1;
			if (depth > maxDepth)
				maxDepth = depth;
		}
	}

	return maxDepth;
}

/* ----------------------------------------------------------------------
 Computes the height of a word on a level in Parse p, that is
 the number of hops to the root
 */
int parseWordHeight(Parse p, String levelname, int wordNo) {
	int k, height = 0;
	int max = vectorSize(p->words);

	/* find the longest path from a leaf to the root */
	while (wordNo >= 0 && height <= max) {
		k = parseIndex(p, levelname, wordNo);
		if (k < 0) {
			return 0;
		}

		if (parseIsMarked(p, levelname, wordNo)) {
			// marked edges are to be treated as root edges
			wordNo = -1;
		} else {
			wordNo = (int) vectorElement(p->verticesStructure, k);
		}

		height++;
	}

	/* cycle? */
	if (height > max) {
		height = INT_MAX;
	}

	return height;
}

/* ----------------------------------------------------------------------
 Print a parse on stdout.
 */
void parsePrint2(Parse parse, Boolean skipUninteresting, float threshold) {
	int i, j;
	String label, levelId;
	Word word;
	int nrLevels = listSize(parse->levels);
	Level level;
	ConstraintViolation cv;
	Vector violations;

	/*
	 cdgPrintf(CDG_DEFAULT, "\n"
	 "   parse id    : %s\n"
	 "   lattice id  : %s\n"
	 "   username    : %s\n"
	 "   date        : %s"
	 "   strategy    : %s\n"
	 "   comment     : %s\n"
	 "   score       : %4.3e\n"
	 "   optimal     : %s\n\n   bindings:\n",
	 parse->id,
	 parse->latticeId,
	 parse->userName,
	 ctime(&parse->dateOfCreation),
	 parse->searchStrategy,
	 parse->comment ? parse->comment : "<undefined>",
	 parse->score,
	 parse->optimal ? "yes" : "no");
	 */

	for (i = 0; i < vectorSize(parse->verticesStructure); i++) {
		levelId = listNthElement(parse->levels, (i % nrLevels) + 1);
		level = findLevel(levelId);
		if (!level || !level->showflag)
			continue;

		label = vectorElement(parse->verticesLabels, i);
		if (skipUninteresting &&
				!strcmp(label, "")
				&& (int) vectorElement(parse->verticesStructure, i) == -1 &&
				(level->no > 0 || (parse->lg->ve && i/nrLevels >= parse->lg->ve->vStartID)) )
			continue;

		word = vectorElement(parse->words, i / nrLevels);

		cdgPrintf(CDG_DEFAULT, "   %03d %10s: %s(%d-%d)--%s-->", word->from
				* inputCurrentGrammar->noOfLevels + level->no, levelId,
				word->description, word->from, word->to, label);

		j = (int) vectorElement(parse->verticesStructure, i);
		if (j == -1) {
			cdgPrintf(CDG_DEFAULT, "NIL\n");
		} else if (j == -2) {
			cdgPrintf(CDG_DEFAULT, "NONSPEC\n");
		} else {
			word = vectorElement(parse->words, j);
			cdgPrintf(CDG_DEFAULT, "%s(%d-%d)\n", word->description,
					word->from, word->to);
		}
	}
	if (parse->violations) {
		cdgPrintf(CDG_DEFAULT, "\n   Violations:\n");
		violations = listToVector(parse->violations);
		vectorSort(violations, cvCompare);
		for (i = 0; i < vectorSize(violations); i++) {
			cv = vectorElement(violations, i);
			if (cv->penalty > threshold)
				continue;
			levelId = listNthElement(parse->levels, (cv->nodeBindingIndex1
					% nrLevels) + 1);
			level = findLevel(levelId);
			if (!level || !level->showflag)
				continue;
			cvPrint(CDG_DEFAULT, cv);
			cdgPrintf(CDG_DEFAULT, "\n");
		}
		vectorDelete(violations);
	}

	cdgPrintf(CDG_DEFAULT, "\n");
}

/* ----------------------------------------------------------------------
 Print a parse on stdout.
 */
void parsePrint(Parse parse) {
	parsePrint2(parse, FALSE, 1.0);
}

/* ----------------------------------------------------------------------
 Return a new Word structure with nothing set.
 */
Word newWord(void) {

	Word result = memMalloc(sizeof (WordStruct));

	result->from = -1;
	result->to = -1;
	result->word = NULL;
	result->description = NULL;

	return result;
}

/* ----------------------------------------------------------------------
 Delete a WordStruct.
 */
void deleteWord(Word w) {

	cdgFreeString(w->word);
	cdgFreeString(w->description);

	memFree(w);
}

/* ----------------------------------------------------------------------
 Compute the mapping of words from vector a onto words from vector b;
 the two are vectors of Word
 */
void parseComputeMap(Vector map, Vector a, Vector b) {
	int i, j;
	int la = vectorSize(a);
	int lb = vectorSize(b);

	Vector d = vectorNew((la + 1) * (lb + 1));

	/* Levenshtein distance matrix */
	Vector dir = vectorNew((la + 1) * (lb + 1));

	/* directions matrix, used to determine the mapping */

#define pos(x,y) ((x)*(lb+1)+(y))

	vectorSetElement(d, (void *) 0, pos(0, 0));
	for (i = 1; i <= la; i++)
		vectorSetElement(d, (void *) i, pos(i, 0));
	for (j = 1; j <= lb; j++)
		vectorSetElement(d, (void *) j, pos(0, j));
	for (i = 1; i <= la; i++)
		for (j = 1; j <= lb; j++) {
			Word wa = (Word) vectorElement(a, i - 1);
			Word wb = (Word) vectorElement(b, j - 1);
			int m1 = (int) vectorElement(d, pos(i - 1, j - 1)) + (strcmp(
					wa->word, wb->word) == 0 ? 0 : 1);
			int m2 = (int) vectorElement(d, pos(i, j - 1)) + 1;
			int m3 = (int) vectorElement(d, pos(i - 1, j)) + 1;

			if (m1 <= m2 && m1 <= m3) {
				vectorSetElement(d, (void *) m1, pos(i, j));
				vectorSetElement(dir, (void *) 1, pos(i, j));
			}
			if (m2 <= m1 && m2 <= m3) {
				vectorSetElement(d, (void *) m2, pos(i, j));
				vectorSetElement(dir, (void *) 2, pos(i, j));
			}
			if (m3 <= m1 && m3 <= m2) {
				vectorSetElement(d, (void *) m3, pos(i, j));
				vectorSetElement(dir, (void *) 3, pos(i, j));
			}
		}

#if 0
	for (i = 0; i <= la; i++) {
		for (j = 0; j <= lb; j++)
		cdgPrintf(CDG_INFO, "%d ", (int)vectorElement(d, pos(i, j)));
		cdgPrintf(CDG_INFO, "\n");
	}
#endif

	for (i = 0; i < la; i++)
		vectorSetElement(map, (void *) -1, i);
	i = la;
	j = lb;
	while (i >= 1 && j >= 1) {
		Word wa = (Word) vectorElement(a, i - 1);
		Word wb = (Word) vectorElement(b, j - 1);

		if (strcasecmp(wa->word, wb->word) == 0)
			vectorSetElement(map, (void *) (j - 1), i - 1);
		switch ((int) vectorElement(dir, pos(i, j))) {
		case 1:
			i--;
			j--;
			break;
		case 2:
			j--;
			break;
		case 3:
			i--;
			break;
		}
	}
	for (i = la - 1; i >= 0 && (int) vectorElement(map, i) == -1; vectorSetElement(
			map, (void *) -2, i--))
		;

	vectorDelete(d);
	vectorDelete(dir);
}

/* ----------------------------------------------------------------------
 Return the lexical category of a Word as a String, or "-".

 This function returns the syntactic category of the word at position
 WORDNO in P as a String. It can fail for many reasons
 (taggerCategoryIndex not set, parse not decorated, etc.). If it fails,
 the String `-' is returned.
 */
String wordCat(Parse p, int wordno) {

	/*  if we don't know which attribute encodes the syntactic category,
	 we can't say anything */
	if (-1 != taggerCategoryIndex) {
		if (!parseIsAbstract(p)) {
			int index = wordno * inputCurrentGrammar->noOfLevels;
			if (index < vectorSize(p->LVs)) {
				LevelValue lv = vectorElement(p->LVs, index);
				if (lv) {
					Value
							v =
									((LexemNode) (lv->dependents->item))->lexem->values[taggerCategoryIndex];
					if (v->type == VTString) {
						return v->data.string;
					}
				}
			}
		}
	}

	/* give up */
	return strRegister("-");

}

/* ----------------------------------------------------------------------
   A faade for parseCompare(): simply answers yes or no quickly.
 */
Boolean parseEquals(Parse a, Parse b) {
	ParseCmpStatStruct stat;

	return parseCompare(a, b, &stat, PC_SILENT, PC_STRICT) == 0;
}

/* ----------------------------------------------------------------------
 Compare two parses.

 VERBOSITY ranges from PC_SILENT to PC_ALL and determines how much screen
 output is produced.

 STRICT determines whether or not words found only in the reference parse
 count as errors, i.e. whether incomplete parses are punished or not.

 The total number of errors is returned.
 */
int parseCompare(Parse ref, Parse test, ParseCmpStat stat, int verbosity,
		int strict) {
	if (ref == NULL) {
		cdgPrintf(CDG_ERROR, "ERROR: NULL reference parse\n");
		return -1;
	} else if (test == NULL) {
		cdgPrintf(CDG_ERROR, "ERROR: NULL test parse\n");
		return -1;
	} else {
		Vector map; /* mapping of words in first to words in second parse */
		int nrwordsprefixref; /* number of words in prefix of reference */

		/*
		 * values of the mapping:
		 * -1  = word not in the second wordgraph
		 * -2  = out of the common time span
		 * >=0 = position in the second wordgraph
		 */
		List l;
		int i, j, k1, k2, errors, index;
		int node1, node2; /* absolute number of current word in the two wordgraphs */
		int level1, level2; /* number of the current level in the two parses */
		int nodegov1, nodegov2; /* number of the current governor in the two parses */
		String strword; /* the current word */
		String strcat = NULL; /* its category */
		String strnode1, strnode2; /* the current word, including lexical info, in both parses */
		String strlevel; /* the current level */
		String strwordgov1, strwordgov2; /* the current governors in both parses */
		String strnodegov1, strnodegov2; /* same, including lexical info */
		String strcatgov1, strcatgov2; /* their lexical category */
		String strlabel1, strlabel2; /* the current labels in both parses */
		Word word1, word2; /* the current words in both parses */
		Word gov1, gov2; /* the current governors in both parses */
		Vector levels_in_grammar; /* the level ids in the current grammar */
		Vector totaledgecorrect_level; /* total number of correct edges per level      */
		Vector totaledgetest_level; /* total number of edges per level in test      */
		Vector totaledgeref_level; /* total number of edges per level in reference */

		int nrwords1 = vectorSize(ref->words);
		int nrwords2 = vectorSize(test->words);
		int nrlevels1 = listSize(ref->levels);
		int nrlevels2 = listSize(test->levels);
		int nrlevels = listSize(inputCurrentGrammar->levels);

		if (verbosity == PC_SUMMARY || verbosity == PC_ALL) {
			cdgPrintf(CDG_INFO, "\n");
			cdgPrintf(CDG_INFO, "INFO: Compare parses %s and %s\n\n", ref->id,
					test->id);
		}

		map = vectorNew(nrwords1);
		parseComputeMap(map, ref->words, test->words);

#if 0
		cdgPrintf(CDG_INFO, "%d %d %d %d\n", nrwords1, nrwords2, nrlevels1,
				nrlevels2);

		for (i = 0; i < vectorSize(map); i++)
		cdgPrintf(CDG_INFO, "%d ", (int)vectorElement(map, i));
		cdgPrintf(CDG_INFO, "\n");

		for (i = 0; i < nrwords1; i++)
		cdgPrintf(CDG_INFO, "%s ", ((Word) vectorElement(ref->words, i))->word);
		cdgPrintf(CDG_INFO, "\n");

		for (i = 0; i < nrwords2; i++)
		cdgPrintf(CDG_INFO, "%s ", ((Word) vectorElement(test->words, i))->word);
		cdgPrintf(CDG_INFO, "\n");
#endif

		stat->errorextrawordref = 0;
		stat->errorextrawordtest = 0;
		stat->errorwordlex = 0;
		stat->errorgovword = 0;
		stat->errorgovlex = 0;
		stat->erroredgelabel = 0;
		stat->matchedgebase = 0;
		stat->matchedgegov = 0;
		stat->matchedgetot = 0;
		stat->totalextrawordref = 0;
		stat->totalextrawordtest = 0;
		stat->totalwordlex = 0;
		stat->totalgovword = 0;
		stat->totalgovlex = 0;
		stat->totaledgelabel = 0;
		stat->totaledgebase = 0;
		stat->totaledgegov = 0;
		stat->totaledgetot = 0;

		totaledgecorrect_level = vectorNew(nrlevels);
		totaledgetest_level = vectorNew(nrlevels);
		totaledgeref_level = vectorNew(nrlevels);
		vectorSetAllElements(totaledgecorrect_level, (Pointer) 0);
		vectorSetAllElements(totaledgetest_level, (Pointer) 0);
		vectorSetAllElements(totaledgeref_level, (Pointer) 0);

		stat->totaledgetest = 0;
		stat->totaledgeref = 0;

		levels_in_grammar = vectorNew(nrlevels);
		for (l = inputCurrentGrammar->levels, i = 0; l; l = l->next, i++) {
			Level level = (Level) l->item;
			vectorSetElement(levels_in_grammar, (Pointer) level->id, i);
		}

		/* errors: word in test but not in reference */
		for (i = 0; i < nrwords2; i++) {
			stat->totalextrawordtest++;
			if (vectorIndexOf(map, (void *) i, 0) == -1) {
				Word w = (Word) vectorElement(test->words, i);

				if (verbosity == PC_ALL)
					cdgPrintf(
							CDG_INFO,
							"ERROR: word %s found in test, but not in reference\n",
							w->word);
				stat->errorextrawordtest++;
			}
		}

		/* errors: word in reference but not in test */
		for (i = 0; i < nrwords1; i++) {
			node1 = i;
			node2 = (int) vectorElement(map, node1);
			word1 = (Word) vectorElement(ref->words, node1);
			strword = word1->word;
			strnode1 = word1->description;

			if (strict == PC_STRICT) {
				stat->totalextrawordref++;
				if (node2 == -2) {
					if (verbosity == PC_ALL)
						cdgPrintf(
								CDG_INFO,
								"ERROR: word %s found in reference, but not in test\n",
								strword);
					stat->errorextrawordref++;
					continue;
				}
			} else {
				if (node2 == -2)
					continue;
			}

			stat->totalextrawordref++;
			if (node2 == -1) {
				if (verbosity == PC_ALL)
					cdgPrintf(
							CDG_INFO,
							"ERROR: word %s found in reference, but not in test\n",
							strword);
				stat->errorextrawordref++;
				continue;
			}
		}

		/* compute nrwordsprefixref */
		if (strict == PC_STRICT) {
			nrwordsprefixref = nrwords1;
		} else {
			int mapind;
			for (mapind = vectorSize(map) - 1; mapind >= 0
					&& (int) vectorElement(map, mapind) < 0; mapind--)
				;
			nrwordsprefixref = mapind + 1;
		}

		/* do the comparisons */
		for (i = 0; i < nrwords1; i++) {
			node1 = i;
			node2 = (int) vectorElement(map, node1);

			if (node2 == -1 || node2 == -2)
				continue;

			word1 = (Word) vectorElement(ref->words, node1);
			word2 = (Word) vectorElement(test->words, node2);
			strnode1 = word1->description;
			strnode2 = word2->description;
			strword = word1->word;

			stat->totalwordlex++;
			if (strcmp(strnode1, strnode2) != 0) {
				if (verbosity == PC_ALL)
					cdgPrintf(
							CDG_INFO,
							"ERROR: Word with different lexical reading: %s should be %s\n",
							strnode1, strnode2);
				stat->errorwordlex++;
			}

			for (j = 0; j < nrlevels1; j++) {
				level1 = j;
				strlevel = strRegister(listNthElement(ref->levels, level1 + 1));
				level2 = listIndex(test->levels, strlevel) - 1;
				if (level2 == -1)
					continue; /* level present in the first, but not in
					 * the second parse */

				/* index of level in the vector of levels in the grammar */
				index = vectorIndexOf(levels_in_grammar, strlevel, 0);
				if (-1 == index) {
					continue;
				}

				k1 = node1 * nrlevels1 + level1;
				k2 = node2 * nrlevels2 + level2;
				strcat = wordCat(ref, node1);

				nodegov1 = (int) vectorElement(ref->verticesStructure, k1);
				nodegov2 = (int) vectorElement(test->verticesStructure, k2);
				strlabel1 = vectorElement(ref->verticesLabels, k1);
				strlabel2 = vectorElement(test->verticesLabels, k2);
				if (nodegov1 == -1) {
					gov1 = NULL;
					strwordgov1 = strRegister("NIL");
					strcatgov1 = strRegister("-");
					strnodegov1 = strRegister("NIL");
				} else if (nodegov1 == -2) {
					gov1 = NONSPEC;
					strwordgov1 = strRegister("NONSPEC");
					strcatgov1 = strRegister("-");
					strnodegov1 = strRegister("NONSPEC");
				} else {
					gov1 = (Word) vectorElement(ref->words, nodegov1);
					strwordgov1 = gov1->word;
					strcatgov1 = wordCat(test, nodegov1);
					strnodegov1 = gov1->description;
				}
				if (nodegov2 == -1) {
					gov2 = NULL;
					strwordgov2 = strRegister("NIL");
					strcatgov2 = strRegister("-");
					strnodegov2 = strRegister("NIL");
				} else if (nodegov2 == -2) {
					gov2 = NONSPEC;
					strwordgov2 = strRegister("NONSPEC");
					strcatgov2 = strRegister("-");
					strnodegov2 = strRegister("NONSPEC");
				} else {
					gov2 = (Word) vectorElement(test->words, nodegov2);
					strwordgov2 = gov2->word;
					strcatgov2 = wordCat(ref, nodegov2);
					strnodegov2 = gov2->description;
				}

				stat->totalgovword++;
				if (nodegov1 != nodegov2) {
					if (verbosity == PC_ALL)
						cdgPrintf(
								CDG_INFO,
								"ERROR: Governor of `%s/%s' on level `%s' is `%s/%s', should be `%s/%s'\n",
								strword, strcat, strlevel, strwordgov1,
								strcatgov1, strwordgov2, strcatgov2);
					stat->errorgovword++;
				} else {
					stat->totalgovlex++;
					if (strcmp(strnodegov1, strnodegov2) != 0) {
						if (verbosity == PC_ALL)
							cdgPrintf(
									CDG_INFO,
									"ERROR: Governor's lexical reading of node `%s' on level `%s' is `%s', should be `%s'\n",
									strword, strlevel, strnodegov1, strnodegov2);
						stat->errorgovlex++;
					}
				}
				stat->totaledgelabel++;
				if (strcmp(strlabel1, strlabel2) != 0) {
					if (verbosity == PC_ALL)
						cdgPrintf(
								CDG_INFO,
								"ERROR: Label of edge above node `%s/%s' on level `%s' is `%s', should be `%s'\n",
								strword, strcat, strlevel, strlabel1, strlabel2);
					stat->erroredgelabel++;
				}
				stat->totaledgebase++;
				if (strcmp(strnode1, strnode2) == 0 && strcmp(strlabel1,
						strlabel2) == 0) {
					stat->matchedgebase++;
				}
				stat->totaledgegov++;
				if (strcmp(strnode1, strnode2) == 0 && strcmp(strlabel1,
						strlabel2) == 0 && strcmp(strwordgov1, strwordgov2)
						== 0) {
					stat->matchedgegov++;
				}
				stat->totaledgetot++;
				if (strcmp(strnode1, strnode2) == 0 && strcmp(strlabel1,
						strlabel2) == 0 && strcmp(strnodegov1, strnodegov2)
						== 0) {
					int currentvalue = (int) vectorElement(
							totaledgecorrect_level, index);
					vectorSetElement(totaledgecorrect_level, (Pointer) (1
							+ currentvalue), index);
					stat->matchedgetot++;
				}
			}
		}

		for (j = 0; j < nrlevels2; j++) {
			level2 = j;
			strlevel = listNthElement(test->levels, level2 + 1);
			index = vectorIndexOf(levels_in_grammar, strlevel, 0);
			if (index != -1) {
				vectorSetElement(totaledgetest_level, (Pointer) nrwords2, index);
			}
			stat->totaledgetest += nrwords2;
		}

		for (j = 0; j < nrlevels1; j++) {
			level1 = j;
			strlevel = listNthElement(ref->levels, level1 + 1);
			index = vectorIndexOf(levels_in_grammar, strlevel, 0);
			if (index != -1) {
				vectorSetElement(totaledgeref_level,
						(Pointer) nrwordsprefixref, index);
			}
			stat->totaledgeref += nrwordsprefixref;
		}

		errors = stat->errorextrawordref + stat->errorextrawordtest
				+ stat->errorwordlex + stat->errorgovword + stat->errorgovlex
				+ stat->erroredgelabel;

		if (verbosity == PC_SUMMARY || verbosity == PC_ALL) {
			if (errors != 0) {
				cdgPrintf(CDG_INFO, "\nINFO: Total of %d errors found\n\n",
						errors);
			} else {
				cdgPrintf(CDG_INFO, "INFO: No errors found\n\n");
			}
			cdgPrintf(CDG_INFO,
					"INFO: Summary of errors/checks in parseCompare\n");
			cdgPrintf(CDG_INFO, "\n");
			cdgPrintf(
					CDG_INFO,
					"INFO: Extra words in reference:                 errors  %5d, checks %5d\n",
					stat->errorextrawordref, stat->totalextrawordref);
			cdgPrintf(
					CDG_INFO,
					"INFO: Extra words in test:                      errors  %5d, checks %5d\n",
					stat->errorextrawordtest, stat->totalextrawordtest);
			cdgPrintf(
					CDG_INFO,
					"INFO: Words with different lexical reading      errors  %5d, checks %5d\n",
					stat->errorwordlex, stat->totalwordlex);
			cdgPrintf(
					CDG_INFO,
					"INFO: Governors with different words:           errors  %5d, checks %5d\n",
					stat->errorgovword, stat->totalgovword);
			cdgPrintf(
					CDG_INFO,
					"INFO: Governors with different lexical reading: errors  %5d, checks %5d\n",
					stat->errorgovlex, stat->totalgovlex);
			cdgPrintf(
					CDG_INFO,
					"INFO: Edges with different label:               errors  %5d, checks %5d\n",
					stat->erroredgelabel, stat->totaledgelabel);
			cdgPrintf(
					CDG_INFO,
					"INFO: Identical edges (governors not counting): matches %5d, checks %5d\n",
					stat->matchedgebase, stat->totaledgebase);
			cdgPrintf(
					CDG_INFO,
					"INFO: Identical edges (same governors words):   matches %5d, checks %5d\n",
					stat->matchedgegov, stat->totaledgegov);
			cdgPrintf(
					CDG_INFO,
					"INFO: Identical edges (same gov. lex. read.):   matches %5d, checks %5d\n",
					stat->matchedgetot, stat->totaledgetot);
			cdgPrintf(CDG_INFO, "\n");
		}

		if (verbosity == PC_SUMMARY || verbosity == PC_MEASURES || verbosity
				== PC_ALL) {
			cdgPrintf(CDG_INFO, "INFO: Total\tAccuracy = ");
			if (stat->totaledgetest == 0) {
				stat->accuracy = -1;
				cdgPrintf(CDG_INFO, "%s\tRecall = ", "n/a");
			} else {
				stat->accuracy = (float) stat->matchedgetot
						/ (float) stat->totaledgetest;
				cdgPrintf(CDG_INFO, "%7.5f\tRecall = ", stat->accuracy);
			}
			if (stat->totaledgeref == 0) {
				stat->recall = -1;
				cdgPrintf(CDG_INFO, "%s\tF-measure = ", "n/a");
			} else {
				stat->recall = (float) stat->matchedgetot
						/ (float) stat->totaledgeref;
				cdgPrintf(CDG_INFO, "%7.5f\tF-measure = ", stat->recall);
			}
			if (stat->totaledgeref + stat->totaledgetest == 0) {
				stat->fmeasure = -1;
				cdgPrintf(CDG_INFO, "%s\n", "n/a");
			} else {
				stat->fmeasure = 2 * (float) stat->matchedgetot
						/ ((float) stat->totaledgetest
								+ (float) stat->totaledgeref);
				cdgPrintf(CDG_INFO, "%7.5f\n", stat->fmeasure);
			}

			for (i = 0; i < nrlevels; i++) {
				String strLevel = (String) vectorElement(levels_in_grammar, i);
				int totaledgecorrect = (int) vectorElement(
						totaledgecorrect_level, i);
				int totaledgetest = (int) vectorElement(totaledgetest_level, i);
				int totaledgeref = (int) vectorElement(totaledgeref_level, i);
				cdgPrintf(CDG_INFO, "INFO: %s\tAccuracy = ", strLevel);
				if (totaledgetest == 0) {
					cdgPrintf(CDG_INFO, "%s\tRecall = ", "n/a");
				} else {
					cdgPrintf(CDG_INFO, "%7.5f\tRecall = ",
							(float) totaledgecorrect / (float) totaledgetest);
				}
				if (totaledgeref == 0) {
					cdgPrintf(CDG_INFO, "%s\tF-measure = ", "n/a");
				} else {
					cdgPrintf(CDG_INFO, "%7.5f\tF-measure = ",
							(float) totaledgecorrect / (float) totaledgeref);
				}
				if (totaledgeref + totaledgetest == 0) {
					cdgPrintf(CDG_INFO, "%s\n", "n/a");
				} else {
					cdgPrintf(CDG_INFO, "%7.5f\n", 2 * (float) totaledgecorrect
							/ ((float) totaledgetest + (float) totaledgeref));
				}
			}
		}
		/*
		 * F-measure = 2 a r / (a + r)
		 * a = total_correct / total_test
		 * r = total_correct / total_ref
		 *
		 * ==> F-measure = 2 total_correct / (total_test + total_ref)
		 */

		vectorDelete(map);
		vectorDelete(levels_in_grammar);
		vectorDelete(totaledgecorrect_level);
		vectorDelete(totaledgetest_level);
		vectorDelete(totaledgeref_level);
		return errors;
	}
}

/* ----------------------------------------------------------------------
 Clone a Parse.

 This function creates a deep copy of all elements of a Parse, including
 LVs and lexeme graph if they are present. Only the Strings in the Parse
 are, as always, shallow copies.

 if decorate is false the copy is not decorated,
 otherwise it depends on the decorated state of the

 */
Parse parseClone(Parse p, Boolean decorate) {
	Parse p_copy;
	List l;

	p_copy = memMalloc(sizeof (ParseStruct));

	if (p->userName == NULL)
		p_copy->userName = NULL;
	else
		p_copy->userName = strRegister(p->userName);

	if (p->latticeId == NULL)
		p_copy->latticeId = NULL;
	else
		p_copy->latticeId = strRegister(p->latticeId);

	p_copy->dateOfCreation = p->dateOfCreation;
	p_copy->grammarFiles = NULL;
	for (l = p->grammarFiles; l; l = l->next) {
		String s = strRegister((String) l->item);

		p_copy->grammarFiles = listAppendElement(p_copy->grammarFiles, s);
	}
	if (p->searchStrategy == NULL)
		p_copy->searchStrategy = NULL;
	else
		p_copy->searchStrategy = strRegister(p->searchStrategy);

	if (p->comment == NULL)
		p_copy->comment = NULL;
	else
		p_copy->comment = strRegister(p->comment);

	p_copy->solutionNo = p->solutionNo;
	p_copy->noOfSolutions = p->noOfSolutions;
	p_copy->score = p->score;
	p_copy->optimal = p->optimal;
	p_copy->violatedConstraints = NULL;
	for (l = p->violatedConstraints; l; l = l->next) {
		String s = strRegister((String) l->item);

		p_copy->violatedConstraints = listAppendElement(
				p_copy->violatedConstraints, s);
	}
	if (p->words == NULL)
		p_copy->words = NULL;
	else {
		int s = vectorSize(p->words);
		int i;

		p_copy->words = vectorNew(s);
		for (i = 0; i < s; i++) {
			Word w0 = (Word) vectorElement(p->words, i);
			Word w = newWord();

			w->from = w0->from;
			w->to = w0->to;
			w->word = strRegister(w0->word);
			w->description = strRegister(w0->description);
			vectorSetElement(p_copy->words, w, i);
		}
	}
	p_copy->levels = NULL;
	for (l = p->levels; l; l = l->next) {
		String s = strRegister((String) l->item);

		p_copy->levels = listAppendElement(p_copy->levels, s);
	}
	p_copy->labels = NULL;
	for (l = p->labels; l; l = l->next) {
		String s = strRegister((String) l->item);

		p_copy->labels = listAppendElement(p_copy->labels, s);
	}

	if (p->verticesStructure == NULL)
		p_copy->verticesStructure = NULL;
	else
		p_copy->verticesStructure = vectorClone(p->verticesStructure);

	if (p->verticesLabels == NULL)
		p_copy->verticesLabels = NULL;
	else
		p_copy->verticesLabels = vectorClone(p->verticesLabels);

	p_copy->evalUnary = p->evalUnary;
	p_copy->evalBinary = p->evalBinary;
	p_copy->numberOfHits = p->numberOfHits;
	p_copy->evaluationTime = p->evaluationTime;

	if (p->statisticComment == NULL)
		p_copy->statisticComment = NULL;
	else
		p_copy->statisticComment = strRegister(p->statisticComment);

	p_copy->LVs = NULL;
	p_copy->lg = NULL;
	p_copy->violations = NULL;

	if (decorate && !parseIsAbstract(p)) {
		if (parseDecorate(p_copy, p->lg, NULL)) {
			parseEval(p_copy);
		}
	}

	p_copy->map = NULL;
	if (p->map) {
		p_copy->map = vectorClone(p->map);
	}
	p_copy->rmap = NULL;
	if (p->rmap) {
		p_copy->rmap = vectorClone(p->rmap);
	}

	p_copy->id = parseMakeId();

	return p_copy;
}

/* ----------------------------------------------------------------------
 Return new Word corresponding to an.

 Try very hard to set w->cat and w->description to fitting values. In
 particular, the description can later be used for selecting the correct
 lexeme nodes if the parse is ever decorated.
 */
Word anno2word(Annotation an) {

	Word w = newWord();

	w->from = an->from;
	w->to = an->to;
	w->word = an->word;
	w->description = strRegister(w->word);

	return w;

}

/* -------------------------------------------------------------------
 Convert an AnnoEntry to a Parse.

 All Annotations in AE are faithfully copied to produce a Parse that is
 as similar as possible. However, if the AnnoEntry lacks Annotations for
 some word/level pairs, the Parse is completed with default values (such
 as the first label found for that level).
 */
Parse parseFromAnno(AnnoEntry ae) {
	List l, m;
	Parse p = parseNew(ae->id);
	int nrlevels, nrlabels, nrwords, posword, poslevel, index;
#ifdef DEBUGPARSEFROMANNO
	cdgPrintf(CDG_WARNING, "translating annotation `%s'\n", ae->id);
#endif

	p->searchStrategy = strRegister("anno2parse");
	p->latticeId = strRegister(ae->lattice);
	p->comment = NULL;
	p->solutionNo = 1;
	p->noOfSolutions = 1;
	p->words = vectorNew(10);

	/* add levels and labels */
	for (m = ae->annos; m; m = m->next) {
		Annotation an = (Annotation) m->item;

		for (l = an->specs; l; l = l->next) {
			Specification s = (Specification) l->item;

			if (s->type == STDep) {
				p->levels = listAddUniqueElement(p->levels, s->kind);
				p->labels = listAddUniqueElement(p->labels, s->name);
			}
		}
	}

	/* If the parse contains the same set of levels as the current grammar,
	 ensure that they are also in the same order. */
	if (listSize(p->levels) == listSize(inputCurrentGrammar->levels)) {

		/* Is every level in P also known in the current grammar? */
		Boolean must_sort = TRUE;
		List l1;
		for (l1 = p->levels; l1 != NULL; l1 = l1->next) {
			String name = l1->item;
			if (!findLevel(name)) {
				must_sort = FALSE;
				break;
			}
		}

		if (must_sort) {
			listDelete(p->levels);
			p->levels = NULL;
			for (l1 = inputCurrentGrammar->levels; l1 != NULL; l1 = l1->next) {
				Level lev = l1->item;
				p->levels = listAppendElement(p->levels, lev->id);
			}
		}
	}

	/* add words */
	for (m = ae->annos; m; m = m->next) {
		Word w = anno2word(m->item);
		if (!w) {
			parseDelete(p);
			cdgPrintf(CDG_WARNING,
					"WARNING: can't translate annotation `%s'\n", ae->id);
			return NULL;
		}
		vectorAddElement(p->words, w);
	}

	nrlevels = listSize(p->levels);
	nrlabels = listSize(p->labels);
	nrwords = vectorSize(p->words);

	p->verticesStructure = vectorNew(nrwords * nrlevels);
	vectorSetAllElements(p->verticesStructure, (Pointer) -1);
	p->verticesLabels = vectorNew(nrwords * nrlevels);
	vectorSetAllElements(p->verticesStructure, NULL);

	/* add verticesStructure and verticesLabels */
	for (m = ae->annos, posword = 1; m; m = m->next, posword++) {
		Annotation an = (Annotation) m->item;

		for (l = an->specs; l; l = l->next) {
			Specification s = (Specification) l->item;

			if (s->type != STDep) {
				continue;
			}
			index = parseIndex(p, s->kind, (posword - 1));
			vectorSetElement(p->verticesLabels, s->name, index);
			if (s->position < -2 || s->position > listSize(ae->annos)) {
				cdgPrintf(
						CDG_ERROR,
						"ERROR: Annotation specifies illegal governor index %d\n",
						s->position);
				parseDelete(p);
				return NULL;
			}
			vectorSetElement(p->verticesStructure, (void *) (s->position - 1),
					index);

		}
	}

	/* Add trivial values for unspecified edges */
	poslevel = 0;
	for (l = p->levels; l != NULL; l = l->next) {

		/* try to find some label for this level */
		String levelname = l->item;
		Level level = NULL;
		String label;
		int i;
		Boolean warned = FALSE;
		for (m = inputCurrentGrammar->levels; m != NULL; m = m->next) {
			Level ld = m->item;
			if (!strcmp(levelname, ld->id)) {
				level = ld;
				break;
			}
		}
		label = strRegister(level ? level->labels->item : "");

		/* insert this label (the subordination is already -1) */
		for (i = 0; i < nrwords; i++) {
			index = i * nrlevels + poslevel;
			if (vectorSize(p->verticesLabels) <= index || !vectorElement(
					p->verticesLabels, index)) {
				vectorSetElement(p->verticesLabels, label, index);
				if (!warned) {
					cdgPrintf(CDG_WARNING,
							"WARNING: Annotation `%s' is incomplete.\n", ae->id);
				}
				warned = TRUE;
			}
		}
		poslevel++;
	}

	return p;
}

/* ----------------------------------------------------------------------
 Compute the mapping of Annotations from vector a onto Words from vector b
 for parseVerify().
 */
void parseVerifyComputeMap(Vector map, Vector a, Vector b) {
	int i, j;
	int la = vectorSize(a);
	int lb = vectorSize(b);

	Vector d = vectorNew((la + 1) * (lb + 1));

	/* Levenshtein distance matrix */
	Vector dir = vectorNew((la + 1) * (lb + 1));

	/* directions matrix, used to determine the mapping */

#define pos(x,y) ((x)*(lb+1)+(y))

	vectorSetElement(d, (void *) 0, pos(0, 0));
	for (i = 1; i <= la; i++)
		vectorSetElement(d, (void *) i, pos(i, 0));
	for (j = 1; j <= lb; j++)
		vectorSetElement(d, (void *) j, pos(0, j));
	for (i = 1; i <= la; i++)
		for (j = 1; j <= lb; j++) {
			String wa = ((Annotation) vectorElement(a, i - 1))->word;
			String wb = ((Word) vectorElement(b, j - 1))->word;
			int m1 = (int) vectorElement(d, pos(i - 1, j - 1)) + (strcasecmp(
					wa, wb) == 0 ? 0 : 1);
			int m2 = (int) vectorElement(d, pos(i, j - 1)) + 1;
			int m3 = (int) vectorElement(d, pos(i - 1, j)) + 1;

			if (m1 <= m2 && m1 <= m3) {
				vectorSetElement(d, (void *) m1, pos(i, j));
				vectorSetElement(dir, (void *) 1, pos(i, j));
			}
			if (m2 <= m1 && m2 <= m3) {
				vectorSetElement(d, (void *) m2, pos(i, j));
				vectorSetElement(dir, (void *) 2, pos(i, j));
			}
			if (m3 <= m1 && m3 <= m2) {
				vectorSetElement(d, (void *) m3, pos(i, j));
				vectorSetElement(dir, (void *) 3, pos(i, j));
			}
		}

#if 0
	for (i = 0; i <= la; i++) {
		for (j = 0; j <= lb; j++)
		cdgPrintf(CDG_INFO, "%d ", (int)vectorElement(d, pos(i, j)));
		cdgPrintf(CDG_INFO, "\n");
	}
#endif

	for (i = 0; i < la; i++)
		vectorSetElement(map, (void *) -1, i);
	i = la;
	j = lb;
	while (i >= 1 && j >= 1) {
		String wa = ((Annotation) vectorElement(a, i - 1))->word;
		String wb = ((Word) vectorElement(b, j - 1))->word;

		if (i == 1) {
			/* allow case mismatch (cf. inputSpuriousUppercase()) */
			if (strcasecmp(wa, wb) == 0) {
				vectorSetElement(map, (void *) (j - 1), i - 1);
			}
		} else {
			if (strcasecmp(wa, wb) == 0) {
				vectorSetElement(map, (void *) (j - 1), i - 1);
			}
		}

		switch ((int) vectorElement(dir, pos(i, j))) {
		case 1:
			i--;
			j--;
			break;
		case 2:
			j--;
			break;
		case 3:
			i--;
			break;
		}
	}
	for (i = la - 1; i >= 0 && (int) vectorElement(map, i) == -1; vectorSetElement(
			map, (void *) -2, i--))
		;

	/* remove duplicate mappings from the map:
	 an LV should not be matched against two annotations,
	 rather the missing LV should count as an error. */
	for (i = 1; i < vectorSize(map); i++) {
		if (vectorElement(map, i - 1) == vectorElement(map, i)
				&& vectorElement(map, i) > 0) {
			vectorSetElement(map, (Pointer) -1, i - 1);
		}
	}

	vectorDelete(d);
	vectorDelete(dir);
}

/* ----------------------------------------------------------------------
 Construct a new ParseVerification.
 */
ParseVerification parseVerificationNew(Parse p) {
	ParseVerification pv =
			(ParseVerification) memMalloc(sizeof(ParseVerificationStruct));

	pv->parse = NULL;
	pv->annoEntry = NULL;
	pv->tries = hashNew(10, 0.7, hashStringHashFunction,
			hashStringEqualFunction);
	pv->structHits = hashNew(10, 0.7, hashStringHashFunction,
			hashStringEqualFunction);
	pv->labelledHits = hashNew(10, 0.7, hashStringHashFunction,
			hashStringEqualFunction);
	pv->lexicalHits = hashNew(10, 0.7, hashStringHashFunction,
			hashStringEqualFunction);
	pv->strictHits = hashNew(10, 0.7, hashStringHashFunction,
			hashStringEqualFunction);
	pv->errors = NULL;

	return pv;
}

/* ----------------------------------------------------------------------
 * construct a new ParseError structure.
 */
ParseError parseErrorNew(ParseErrorType type) {
	ParseError pe = (ParseError) memMalloc(sizeof(ParseErrorStruct));

	pe->type = type;
	return pe;
}

/* ----------------------------------------------------------------------
 * destroy a ParseError
 */
void parseErrorDelete(ParseError pe) {
	memFree(pe);
}

/* ----------------------------------------------------------------------
 Destroy a ParseVerification.
 */
void parseVerificationDelete(ParseVerification pv) {
	hashDelete(pv->tries);
	hashDelete(pv->structHits);
	hashDelete(pv->labelledHits);
	hashDelete(pv->lexicalHits);
	hashDelete(pv->strictHits);
	listForEachDelete(pv->errors, parseErrorDelete);
	memFree(pv);
}

/* ----------------------------------------------------------------------
 Compare Parse to Annotation and print results.

 This function is exceedingly complex, although we make some pretty radical
 simplifying assumptions. Consider all of the differences that could exist
 between a Parse and an Annotation:

 - the structure may be different
 - the labels in analogous places may differ
 - the lexical reading of a word may differ
 - the phonetical form of a word may be different
 - indeed, the number of number of words may be different!
 This leads directly to the more general case:
 - the Annotation may contain less information than the Parse
 (labels, levels, words...) or more (levels or words).


 Here's what we make out of this mess:

 (1) In general, the Annotation is authoritative. Everything specified
 there is considered correct. If the Parse says something different,
 that is an error. If the Parse says nothing about something that the
 Annotation specifies, that is also an error. If information from the
 Parse is not specified in the Annotation, it is ignored.

 Thus, the measure computed here is always a measure of Recall, never
 Precision. This is usually a problem, since perfect recall can be
 achieved simply by specifying everything. However, a Parse can only
 specify one thing for each degree of freedom (with one exception,
 considered next).

 Then again, solution methods are perfectly capable of returning
 multiple Parses for a problem, so actually we would want to assess
 several Parses against one Annotation and retrieve a single measure of
 correctness. But this isn't possible with this function.

 (2) There is one exception to the rule that Parses must be unambiguous
 about everything: if all of its LVs are disjunctive in the modifier,
 the lexical reading at that point is ambiguous. This can only
 happen when two homonyms are identical in all features ever touched by
 the grammar, which should not happen (don't model features that you
 can't distinguish). It does happen massively when the syntax level,
 which should touch all features, is switched off.

 Strictly speaking, such Parses are placeholders for several similar
 solutions. But for purposes of parseVerify() we treat them as if they
 were unambiguous: if one of the lexical items matches the conditions
 specified by the annotation, the Parse is correct, otherwise it is an
 error.

 (3) If the sequence of words differs between Annotation and Parse, we
 match them up with the Levenshtein algorithm, but no word in the
 Annotation can correspond to more than one word in the Parse. The
 words that it considers as corresponding to each other are then
 compared. Words left over in the Parse are ignored. Words left over
 in the Annotation and not in the Parse count as errors. This means
 that if the Parse has two words in reverse order (recognizer
 error?), one of them will be considered missing.

 (4) For each word in the annotation, the corresponding LVs are checked
 whether they comply with it. The result is counted both globally and
 per-level. The following levels of compliance exist:

 (4) everything correct
 (3) everything correct except the lexical reading of the modifiee
 (2) label and structure correct
 (1) structure only correct

 Any LV that does not meet either of these definitions is an error
 under all measures. Recall that a missing LV is always considered an
 error.
 */
ParseVerification parseVerify(Parse p, AnnoEntry ae) {
	List l, path;
	int i, j, k;
	Vector wordHits;
	int gov1, gov2;
	int ignoredErrors = 0;

	/* compute word mapping from annotation to parse */
	Vector wordmap = vectorNew(20);
	List annolevels = NULL, missinglevels = NULL, missingwords = NULL;
	Vector annos = listToVector(ae->annos);
	int nrannowords = vectorSize(annos);
	Array matrix;
	char
			*measures[] = { "wrong", "structural", "labelled", "lexical",
					"strict" };
	ParseVerification pv = parseVerificationNew(p);
	ParseError pe;
	Hierarchy h = findHierarchy(inputFeatureHierarchy);

	/* initialize the parse verification */
	pv->parse = p;
	pv->annoEntry = ae;

	parseVerifyComputeMap(wordmap, annos, p->words);

	/* Pre-compute how well each word was recalled:
	 2 - lexically
	 1 - orthographically
	 0 - not at all   */
	wordHits = vectorNew(nrannowords);
	for (i = 0; i < nrannowords; i++) {
		Annotation a = vectorElement(annos, i);
		LexiconItem le = NULL;
		j = (int) vectorElement(wordmap, i);
		if (j > -1) {

			Word w = vectorElement(p->words, j);
#ifdef DEBUGPARSEVERIFY
			cdgPrintf(CDG_DEBUG, "DEBUG: anno: `%s'  word: `%s'\n", a->word, w->word);
#endif
			/* perfect match -- score 2 */
			vectorSetElement(wordHits, (Pointer) 2, i);

			/* find underlying lexical entry */
			if (!parseIsAbstract(p)) {
				Boolean found = FALSE;
				for (k = 0; k < vectorSize(p->lg->nodes); k++) {
					le = ((LexemNode) vectorElement(p->lg->nodes, k))->lexem;
					if (le->description == w->description) {
						found = TRUE;
						break;
					}
				}

				/* compare features */
				if (found) {
					for (l = a->specs; l != NULL; l = l->next) {
						Specification s = l->item;
						if (s->type == STTag) {
							Value annos = le->values[inputRegisterPath(path
									= scanPath(s->kind), inputCurrentGrammar)];
							String f = sprintValue(annos);

							listDelete(path);
							/* missing feature */
							if (annos == ErrorValue) {
								cdgPrintf(
										CDG_WARNING,
										"WARNING: word `%s': feature %s = %s not present in parse.\n",
										w->word, s->kind, s->name);
								/* lexical error -- score 1 */
								pe = parseErrorNew(PELexem);
								pe->data.lexError.modifier = i + 1; /* SMELL: or should it be j? */
								pe->data.lexError.feature = s->kind;
								pe->data.lexError.value = NULL;
								pe->data.lexError.annoValue = s->name;
								pv->errors = listPrependElement(pv->errors, pe);
								vectorSetElement(wordHits, (Pointer) 1, i);
								break;
							}

							/* wrong feature */
							if (inputCompareFeatures(f, s->name, h) > 1) {
								cdgPrintf(
										CDG_WARNING,
										"WARNING: word `%s': feature %s = %s should be %s.\n",
										w->word, s->kind, f, s->name);
								/* lexical error -- score 1 */
								pe = parseErrorNew(PELexem);
								pe->data.lexError.modifier = i + 1; /* SMELL: or should it be j? */
								pe->data.lexError.feature = s->kind;
								pe->data.lexError.value = f;
								pe->data.lexError.annoValue = s->name;
								pv->errors = listPrependElement(pv->errors, pe);
								vectorSetElement(wordHits, (Pointer) 1, i);
								break;
							}
						}
					}
				}
			}
		} else {
			/* word not present -- score 0 */
			vectorSetElement(wordHits, 0, i);
		}
	}

	/* find all levels in annotation */
	for (i = 0; i < nrannowords; i++) {
		Annotation a = vectorElement(annos, i);
		for (l = a->specs; l != NULL; l = l->next) {
			Specification s = l->item;
			if (s->type != STDep) {
				continue;
			}
			annolevels = listAddUniqueElement(annolevels, s->kind);
		}
	}

#ifdef DEBUGPARSEVERIFY
	cdgPrintf(CDG_DEBUG, "DEBUG: word map: ");
	for(i = 0; i < vectorSize(wordmap); i++) {
		int j = (int)vectorElement(wordmap, i);
		cdgPrintf(CDG_DEBUG, "%d ", j);
	}
	cdgPrintf(CDG_DEBUG, "\n");

	cdgPrintf(CDG_DEBUG, "DEBUG: word accuracy: ");
	for(i = 0; i < vectorSize(wordHits); i++) {
		int j = (int)vectorElement(wordHits, i);
		cdgPrintf(CDG_DEBUG, "%d ", j);
	}
	cdgPrintf(CDG_DEBUG, "\n");
#endif

	/* initialize verification matrix */
	matrix = arrayNew(nrannowords, listSize(annolevels), 0);
	for (i = 0; i < nrannowords; i++) {
		for (j = 0; j < listSize(annolevels); j++) {
			arraySetElement(matrix, (Pointer) -1, i, j);
		}
	}

	/* Now compare all edges in the annotation against the parse */
	for (i = 0; i < nrannowords; i++) {
		int posword = (int) vectorElement(wordmap, i);
		Annotation a = vectorElement(annos, i);
		for (l = a->specs; l != NULL; l = l->next) {
			Specification s = l->item;
			if (s->type == STDep) {
				int poslevel = listIndex(p->levels, s->kind) - 1;
				int posannolevel = listIndex(annolevels, s->kind) - 1;
				int parseindex = posword * listSize(p->levels) + poslevel;

				/* missing word -- score 0 */
				if (posword < 0) {
					if (!listContains(missingwords, a)) {
						cdgPrintf(CDG_WARNING,
								"WARNING: word `%s' not present in parse.\n",
								a->word);
						missingwords = listPrependElement(missingwords, a);
						pe = parseErrorNew(PEWord);
						pe->data.word = a->word;
						pv->errors = listPrependElement(pv->errors, pe);
					}
					arraySetElement(matrix, (Pointer) 0, i, posannolevel);
					continue;
				}

				/* missing level -- score 0 */
				if (poslevel == -1) {
					if (!listContains(missinglevels, s->kind)) {
						Level levelDecl = findLevel(s->kind);
						if (!levelDecl) {
							cdgPrintf(
									CDG_WARNING,
									"WARNING: level `%s' not present in parse.\n",
									s->kind);
							pe = parseErrorNew(PELevel);
							pe->data.level = s->kind;
							pv->errors = listPrependElement(pv->errors, pe);
						}
						missinglevels = listPrependElement(missinglevels,
								s->kind);
					}
					arraySetElement(matrix, (Pointer) 0, i, posannolevel);
					continue;
				}

				/* structural error -- score 0 */
				gov1 = (int) vectorElement(p->verticesStructure, parseindex);
				gov2 = s->position - 1;
				if (gov1 != gov2) {
					String s1 = "NIL", s2 = "NIL";
					if (gov1 >= 0 && gov1 < vectorSize(wordmap)) {
						int index = (int) vectorElement(wordmap, gov1);
						if (index >= 0) {
							s1 = ((Word) vectorElement(p->words, index))->word;
						}
					}
					if (gov2 >= 0) {
						s2 = ((Annotation) vectorElement(annos, gov2))->word;
					}

					cdgPrintf(
							CDG_WARNING,
							"WARNING: subordination %s: %s(%d)-->%s(%d) should be %s(%d)-->%s(%d)\n",
							s->kind, a->word, i + 1, s1, gov1 + 1, a->word, i
									+ 1, s2, gov2 + 1);

					pe = parseErrorNew(PEStruct);
					pe->data.structError.level = s->kind;
					pe->data.structError.modifier = i + 1;
					pe->data.structError.modifiee = gov1 + 1;
					pe->data.structError.annoModifiee = gov2 + 1;
					pe->data.structError.label = vectorElement(
							p->verticesLabels, parseindex);
					pe->data.structError.annoLabel = s->name;
					pv->errors = listPrependElement(pv->errors, pe);

					if (parseVerifyPartially && gov1 == -1) {
						ignoredErrors++;
						arraySetElement(matrix, (Pointer) 4, i, posannolevel);
					} else {
						arraySetElement(matrix, (Pointer) 0, i, posannolevel);
					}
					continue;
				}

				/* label error -- score 1 */
				if (vectorElement(p->verticesLabels, parseindex) != s->name) {
					cdgPrintf(CDG_WARNING,
							"WARNING: label %s: %s--`%s'--> should be `%s'\n",
							s->kind, a->word, vectorElement(p->verticesLabels,
									parseindex), s->name);
					arraySetElement(matrix, (Pointer) 1, i, posannolevel);
					pe = parseErrorNew(PELabel);
					pe->data.labelError.level = s->kind;
					pe->data.labelError.modifier = i + 1;
					pe->data.labelError.modifiee = gov1 + 1;
					pe->data.labelError.label = vectorElement(
							p->verticesLabels, parseindex);
					pe->data.labelError.annoLabel = s->name;
					pv->errors = listPrependElement(pv->errors, pe);
					continue;
				}

				/* lexical error in the modifier -- score 2 */
				if ((int) vectorElement(wordHits, i) < 2) {
					arraySetElement(matrix, (Pointer) 2, i, posannolevel);
					continue;
				}

				/* lexical error only in the modifiee -- score 3 */
				j = s->position - 1;
				if (j > 0 && (int) vectorElement(wordHits, j) < 2) {
					arraySetElement(matrix, (Pointer) 3, i, posannolevel);
					continue;
				}

				/* no error -- score 4 */
				arraySetElement(matrix, (Pointer) 4, i, posannolevel);
			}
		}
	}

#ifdef DEBUGPARSEVERIFY
	cdgPrintf(CDG_DEBUG, "DEBUG: matrix of compliance:\n");
	for(i = 0; i < nrannowords; i++) {
		for(j = 0; j < listSize(annolevels); j++) {
			cdgPrintf(CDG_DEBUG, "%3d ", arrayElement(matrix, i, j));
		}
		cdgPrintf(CDG_DEBUG, "\n");
	}
#endif

	/* Give matrix of information */
	cdgPrintf(CDG_INFO, "                   ");
	for (i = 1; i < 5; i++) {
		cdgPrintf(CDG_INFO, "%-14s", measures[i]);
	}
	cdgPrintf(CDG_INFO, "\n\n");

	i = 0;
	{
		int totaltries = 0;
		int totalhits[5] = { 0, 0, 0, 0, 0 };
		Hashtable hitsHash[5] = { 0, pv->structHits, pv->labelledHits,
				pv->lexicalHits, pv->strictHits };
		int n;
		for (l = annolevels; l != NULL; l = l->next) {
			String level = l->item;
			Level levelDcl = findLevel(level);
			float score;
			int tries = 0;
			int hits[5] = { 0, 0, 0, 0, 0 };

			if (!levelDcl || !levelDcl->useflag) {
				i++;
				continue;
			}

			cdgPrintf(CDG_INFO, "%10s:", level);
			for (j = 0; j < nrannowords; j++) {
				int k = (int) arrayElement(matrix, j, i);

				/* the annotation may omit specification of some edges */
				if (k == -1) {
					continue;
				}
				tries++;
				totaltries++;
				for (n = 1; n < 5; n++) {
					if (k >= n) {
						hits[n]++;
						totalhits[n]++;
					}
				}
			}
			hashSet(pv->tries, level, (Pointer) (tries + 1));
			for (n = 1; n < 5; n++) {
				score = 100.0 * hits[n] / tries; /* can't be 0, or we wouldn't have found it */
				cdgPrintf(CDG_INFO, "        %6.2f", score);
				hashSet(hitsHash[n], level, (Pointer) (hits[n] + 1));
			}
			i++;
			cdgPrintf(CDG_INFO, "\n");
		}

		hashSet(pv->tries, strRegister("overall"), (Pointer) (totaltries + 1));
		if (totaltries) {
			cdgPrintf(CDG_INFO, "%10s:", "overall");
			for (n = 1; n < 5; n++) {
				float score = 100.0 * totalhits[n] / totaltries;
				cdgPrintf(CDG_INFO, "        %6.2f", score);
				hashSet(hitsHash[n], strRegister("overall"),
						(Pointer) (totalhits[n] + 1));
			}
			cdgPrintf(CDG_INFO, "\n");
		}
		if (ignoredErrors) {
			cdgPrintf(CDG_WARNING,
					"\nWARNING: ignoring %d error(s) in %d tries\n",
					ignoredErrors, totaltries);
		}
	}

	arrayDelete(matrix);
	listDelete(annolevels);
	listDelete(missinglevels);
	listDelete(missingwords);
	vectorDelete(annos);
	vectorDelete(wordHits);
	vectorDelete(wordmap);

	return pv;
}

/* ----------------------------------------------------------------------
 Return the parse from cdgParses that has the id ID.
 */
Parse parseFind(String id) {
	return hashGet(cdgParses, id);
}

/* ----------------------------------------------------------------------
 Checks whether any regents are underspecified.
 */
Boolean parseIsUnderspecified(Parse p) {
	int i;

	for (i = 0; i < vectorSize(p->verticesStructure); i++) {
		if ((int) vectorElement(p->verticesStructure, i) == -2) {
			return TRUE;
		}
	}
	return FALSE;
}

/* ----------------------------------------------------------------------
 Return the creation date of P as a String.
 */
String parseDate(Parse p) {
	String result = asctime(localtime(&(p->dateOfCreation)));

	result[strlen(result) - 1] = ' '; /* instead of '\n' */
	return strRegister(result);
}

/* ----------------------------------------------------------------------
 Return LV of a decorated Parse.
 */
LevelValue parseGetLV(Parse p, int where) {
	int index;

	if (parseIsAbstract(p)) {
		return NULL;
	}

	index = (int) vectorElement(p->map, where);

	if (index >= 0)
		return vectorElement(p->LVs, index);
	else
		return NULL;
}

/* ----------------------------------------------------------------------
 Return lexeme nodes of a decorated Parse.
 */
List parseGetLexemNodes(Parse p) {
	List result = NULL;
	int i;
	if (!parseIsAbstract(p)) {
		Vector nodes = p->lg->nodes;
		for (i = 0; i < vectorSize(nodes); i++) {
			LexemNode ln = vectorElement(nodes, i);
			result = listPrependElement(result, ln);
		}
	}

	return result;
}

/* ----------------------------------------------------------------------
 Add P uniquely to the global list.
 */
void parseRegister(Parse p) {
	if (parseFind(p->id)) {
		return;
	}
	hashSet(cdgParses, p->id, p);
	parseMostRecentlyCreatedParse = p;
}

/* ----------------------------------------------------------------------
 Build a parse from a vector of LVs.
 */
Parse parseFromLVs(ConstraintNet net, Vector lvs, String strategy) {
	Parse p = parseNew(NULL);
	p->latticeId = net->lexemgraph->lattice->id;
	p->searchStrategy = strRegister(strategy);
	p->comment = NULL;
	p->solutionNo = 1;
	p->noOfSolutions = 1;
	p->optimal = FALSE;
	p->evalUnary = net->evalUnary;
	p->evalBinary = net->evalBinary;
	p->numberOfHits = 0;
	p->evaluationTime = 0;
	p->statisticComment = NULL;

	vectorSort(lvs, lvCompareByNaturalOrder);
	parseComputeStructure(p, lvs, net->lexemgraph->isDeletedNode);
	parseDecorate(p, net->lexemgraph, NULL);
	parseEval(p);

	return p;
}

/* ----------------------------------------------------------------------
 Return the minimal timepoint covered by P.
 */
int parseMinFrom(Parse p) {

	int i, result = 0;
	for (i = 0; i < vectorSize(p->words); i++) {
		result = min(result, ((Word)vectorElement(p->words, i))->from);
	}

	return result;

}

/* ----------------------------------------------------------------------
 Return the maximal timepoint covered by P.
 */
int parseMaxTo(Parse p) {

	int i, result = 0;
	for (i = 0; i < vectorSize(p->words); i++) {
		result = max(result, ((Word)vectorElement(p->words, i))->to);
	}

	return result;

}


/* ----------------------------------------------------------------------
 * Translate meta data of the lattice into individual SPECs.
 */
void transferMetaData(Arc arc, Annotation an) {
    if(!arc->info) {
        return;
    }
    Value val = arc->info;
    switch(val->type) {
    case VTConjunction:{
        List l;
        for (l = val->data.list; l != NULL; l = l->next) {
            Value v = (Value) l->item;
            if (v->type != VTAVNode) {
                cdgPrintf(CDG_WARNING, "WARNING: value ");
                printValue(CDG_WARNING, v, 0);
                cdgPrintf(CDG_WARNING, " in value ");
                printValue(CDG_WARNING, val, 0);
                cdgPrintf(CDG_WARNING,
                          " should be attribute/value pair, but isn't!\n\n");
                continue;
            }
            
            AVNode av = (AVNode)v->data.avnode;
            String kind = strCat("info:", av->attribute);
            String name = sprintValue(av->value);
            Specification s = newSpecification(STTag, kind, name, 0);
            inputSpecificationAdd(an, s);
        }
        break;
    }
    default:
        cdgPrintf(CDG_WARNING,
                  "WARNING: meta data of word '%s' is something other than a conjunction",
                  arc->word);
    }
}

/* ----------------------------------------------------------------------
 Turn P into an AnnoEntry.
 */
AnnoEntry parse2annotation(Parse p) {

	LexemNode ln = NULL;
	LevelValue lv;
	List l, m;
	int i, j;
	Word w;
	AnnoEntry result;
	Annotation a;
	Specification s;
	int nrLevels = listSize(p->levels);

	result = (AnnoEntry) memMalloc( sizeof(AnnoEntryStruct) );
	result->id = strRegister(p->latticeId);
	if (parseIsAbstract(p)) {
		result->lattice = p->latticeId;
	} else {
		result->lattice = p->lg->lattice->id;
	}
	result->annos = NULL;
	result->filename = NULL;
	result->lineNo = 0;

	/* iterate over words */
	for (i = 0; i < vectorSize(p->words); i++) {
		List desired_features;
		Boolean found;

		w = vectorElement(p->words, i);

		a = newAnnotation(w->from, w->to, strRegister(w->word));
		result->annos = listAppendElement(result->annos, a);

		if (!parseIsAbstract(p)) {
			ln = parseFindLn(p, w, NULL)->item;
			transferMetaData(ln->arc, a);

			/* include the distinctive features,
			 plus all that the user explicitly ordered */
			desired_features = listClone(ln->lexem->distinctive_features);
			for (l = inputAnnoCats; l != NULL; l = l->next) {
				found = FALSE;
				for (m = desired_features; m != NULL; m = m->next) {
					if (comparePath(l->item, m->item)) {
						found = TRUE;
						break;
					}
				}
				if (!found) {
					desired_features = listPrependElement(desired_features,
							l->item);
				}
			}

			for (l = desired_features; l != NULL; l = l->next) {
				Value v;
				List path;
				int k = inputLookupPath((List) l->item, inputCurrentGrammar);
				if (k < 0) {
					cdgPrintf(CDG_WARNING, "WARNING: can't lookup path `");
					printPath(CDG_WARNING, l->item);
					cdgPrintf(CDG_WARNING, "', ignoring\n");
					continue;
				}
				v = ln->lexem->values[k];
				path
						= listNthElement(inputCurrentGrammar->allAttributes, k
								+ 1);
				if (v == ErrorValue) {
					continue;
				}

				switch (v->type) {
				case VTString:
				case VTNumber:
				case VTList:
					s = newSpecification(STTag, sprintPath(path),
							sprintValue(v), 0);
					inputSpecificationAdd(a, s);
					break;
				default:
					break;
				}
			}
			listDelete(desired_features);

			// if the word is a virtual one, add that info to the annotation
			// TODO better add this info to the lexeme, so it will be processed here automatically
			if(ln->grapheme->isVirtual) {
				s = newSpecification(STTag, "virtual", "true", 0);
				inputSpecificationAdd(a, s);
			}
		}

		/* record structure */
		if (parseIsAbstract(p)) {
			for (j = 0; j < nrLevels; j++) {
				int index = i * nrLevels + j;
				s = newSpecification(STDep, listNthElement(p->levels, j + 1),
						vectorElement(p->verticesLabels, index),
						(int) vectorElement(p->verticesStructure, index) + 1);
				inputSpecificationAdd(a, s);
			}
		} else {
			for (j = 0; j < vectorSize(p->verticesStructure); j++) {
				lv = parseFindLV(p, j, NULL)->item;
				if (!listContains(lv->dependents, ln)) {
					lvDelete(lv);
					continue;
				}
				s = newSpecification(STDep, lv->level->id, lv->label,
						(int) vectorElement(p->verticesStructure, j) + 1);
				inputSpecificationAdd(a, s);
				lvDelete(lv);
			}
		}
	}

	return result;

}

/* ----------------------------------------------------------------------
 Calculate the subordination that would yield the best score for position
 WHERE if nothing else may be changed.

 If it isn't possible to improve the current score, the current
 subordination is returned even if other labels would yield the same
 score.
 */
int parseOptimizeStructure(Parse p, String levelname, int wordno) {
	Parse p2;
	int result = 0;
	Badness best;
	int where = parseIndex(p, levelname, wordno);
	Level level = findLevel(levelname);
	int i;
	int current = (int) vectorElement(p->verticesStructure, where);

	if (!level) {
		return current;
	}

	if (parseIsAbstract(p)) {
		return current;
	}

	p2 = parseClone(p, TRUE);
	best = worstBadness();
	for (i = p->lg->min; i <= p->lg->max; i++) {
		if (i != wordno) {
			int j = i;
			/* try NIL last */
			if (j == p->lg->max) {
				j = -1;
			}

			parseShiftEdge(p2, where, j, TRUE);
			if (bCompare(p2->badness, best)) {
				result = j;
				bCopy(best, p2->badness);
			}
		}
	}

	if (!bCompare(best, p->badness)) {
		result = current;
	}

	bDelete(best);
	parseDelete(p2);

	return result;

}

/* ----------------------------------------------------------------------
 Calculate the label that would yield the best score at position WHERE
 if nothing else may be changed.

 If it isn't possible to improve the current score, the current label is
 returned even if other labels would yield the same score.
 */
String parseOptimizeLabel(Parse p, String levelname, int wordno) {
	Parse p2;
	List l;
	String result = NULL;
	Badness best;
	int where = parseIndex(p, levelname, wordno);
	Level level = findLevel(levelname);

	String current = vectorElement(p->verticesLabels, where);

	if (!level) {
		return NULL;
	}

	if (parseIsAbstract(p)) {
		return NULL;
	}

	p2 = parseClone(p, TRUE);
	best = worstBadness();
	for (l = level->labels; l != NULL; l = l->next) {
		String label = l->item;
		parseSwapLabel(p2, where, label, TRUE);
		if (bCompare(p2->badness, best)) {
			result = label;
			bCopy(best, p2->badness);
		}
	}

	if (!bCompare(best, p->badness)) {
		result = current;
	}

	bDelete(best);
	parseDelete(p2);

	return result;

}

/* ----------------------------------------------------------------------
 Calculate the lexical variant that would yield the best score at position
 WHERE if nothing else may be changed.

 If it isn't possible to improve the current score, the current word is
 returned even if other words would yield the same score.
 */
String parseOptimizeWord(Parse p, int wordno) {
	if(wordno < 0) {
		return NULL;
	}

	Parse p2;
	LexemNode result = NULL;
	LexemNode current;
	Badness best;
	int i;
	Word w = vectorElement(p->words, wordno);

	if (parseIsAbstract(p)) {
		return NULL;
	}

	current = parseFindLn(p, w, NULL)->item;

	p2 = parseClone(p, TRUE);
	best = worstBadness();
	for (i = 0; i < vectorSize(p->lg->nodes); i++) {
		LexemNode ln = vectorElement(p->lg->nodes, i);
		if (ln->arc->from != wordno) {
			continue;
		}
		parseSwapWord(p2, wordno, ln->lexem->description, TRUE);
		if (bCompare(p2->badness, best)) {
			result = ln;
			bCopy(best, p2->badness);
		}
	}

	if (!bCompare(best, p->badness)) {
		result = current;
	}

	bDelete(best);
	parseDelete(p2);

	return result ? result->lexem->description : NULL;

}

/* ----------------------------------------------------------------------
 Return the lexicon item currently selected at timepoint I.
 */
LexiconItem parseLexiconItem(Parse p, int i) {
	Word w = vectorElement(p->words, i);
	LexemNode ln;

	if (parseIsAbstract(p)) {
		return NULL;
	}

	ln = parseFindLn(p, w, NULL)->item;
	return ln->lexem;
}

/* ----------------------------------------------------------------------
 Return the grapheme currently selected at timepoint I.
 */
GraphemNode parseGraphemeNode(Parse p, int i) {
	Word w = vectorElement(p->words, i);
	LexemNode ln;

	if (parseIsAbstract(p)) {
		return NULL;
	}

	ln = parseFindLn(p, w, NULL)->item;
	return ln->grapheme;
}

/* ----------------------------------------------------------------------
 Return a String describing the value of the word currently selected at
 timepoint I.
 */
String parseFormatValue(Parse p, int i) {
	Word w = vectorElement(p->words, i);
	LexemNode ln;
	LexiconItem li;

	if (parseIsAbstract(p)) {
		return NULL;
	}

	ln = parseFindLn(p, w, NULL)->item;
	li = ln->lexem;
	return formatValue(li->value, 0);

}

/* ----------------------------------------------------------------------
 Initialize parse module
 */
void parseInitialize(void) {
	setRegister("verifypartially"          , SET_BOOL, &parseVerifyPartially     , NULL, NULL,NULL, NULL);
	setRegister("parseTryOutLexemeVariants", SET_BOOL, &parseTryOutLexemeVariants, NULL, NULL,NULL, NULL);
}

/* ----------------------------------------------------------------------
 Finalize the parse module
 */
void parseFinalize(void) {
}

/* ----------------------------------------------------------------------
 Edge marking functions

 Edges are marked by appending "_disabled" to their lables.
 Marked edges are considered to be root edges. This is usefull
 when dealing with parses containing cycles.
 */

/* ----------------------------------------------------------------------
 Mark/unmark the edge from the word specified by wordno in parse p
 */
Boolean parseMarkEdge(Parse p, String levelname, int wordno, Boolean mark) {
	// Input Filter
	if (p == NULL)
		return FALSE;

	int max = vectorSize(p->words);
	if ((wordno < 0) || (wordno >= max))
		return FALSE;

	int wordIndex = parseIndex(p, levelname, wordno);
	String oldlabel = vectorElement(p->verticesLabels, wordIndex);

	Boolean marked = strIsSuffix(oldlabel, MARK_SUFFIX);

	String newlabel = NULL;
	if (mark != marked) {
		if (mark) {
			newlabel = strCat(oldlabel, MARK_SUFFIX);
		} else {
			newlabel = strSubStr(oldlabel, 0, strlen(oldlabel) - strlen(
					MARK_SUFFIX));
		}
		vectorSetElement(p->verticesLabels, newlabel, wordIndex);
	}

	return TRUE;
}

Boolean parseIsMarked(Parse p, String levelname, int wordno) {
	// Input Filter
	if (p == NULL)
		return FALSE;

	int max = vectorSize(p->words);
	if ((wordno < 0) || (wordno >= max))
		return FALSE;

	int wordIndex = parseIndex(p, levelname, wordno);
	String oldlabel = vectorElement(p->verticesLabels, wordIndex);

	return strIsSuffix(oldlabel, MARK_SUFFIX);
}

Boolean parseMarkOneEdgePerCycle(Parse p, String levelname) {

	if (p == NULL)
		return FALSE;
	if (levelname == NULL)
		return FALSE;

	int k, height;
	int max = vectorSize(p->words);
	int i, wordNo;

	for (i = 0; i < max; i++) {
		height = 0;
		wordNo = i;
		while (wordNo != -1 && height <= max) {
			k = parseIndex(p, levelname, wordNo);
			if (k == -1) {
				return FALSE;
			}
			if (parseIsMarked(p, levelname, wordNo)) {
				// marked edges are treated as if they would not exist
				wordNo = -1;
			} else {
				wordNo = (int) vectorElement(p->verticesStructure, k);
			}
			height++;
			// optimization
			if (wordNo == i) {
				height = max + 1;
				break;
			}
		}

		// cycle?
		if (height > max) {
			parseMarkEdge(p, levelname, wordNo, TRUE);
			return parseMarkOneEdgePerCycle(p, levelname);
		}
	}

	return TRUE;
}

void parseClearAllMarks(Parse p, String levelname) {

	if (p == NULL)
		return;

	int max = vectorSize(p->words);
	int i;
	for (i = 0; i < max; i++) {
		parseMarkEdge(p, levelname, i, FALSE);
	}
}

void parseMakeMarkedEdgesRootEdges(Parse p, String levelname, Boolean eval) {

	if (p == NULL)
		return;

	int max = vectorSize(p->words);
	int i;
	for (i = 0; i < max; i++) {
		if (parseIsMarked(p, levelname, i)) {
			parseShiftEdge(p, parseIndex(p, levelname, i), -1, eval);
		}
	}
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
