/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -- INCLUDES ------------------------------------------------------------- */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <blah.h>
#include <string.h>
#include "cdg.h"
#include "hook.h"
#include "lexemgraph.h"
#include "tagger.h"
#include "predict.h"

/* -- VARIABLES ------------------------------------------------------------ */

/* the string "POS".
 initialized in cdg.c */
String constStr_POS = NULL;
Boolean cdgPredictor = FALSE;

/* -- FUNCTIONS ------------------------------------------------------------ */

/* -------------------------------------------------------------------------
 Call an external program and record its predictions in the lexeme graph.
 */
int predictCall(Predictor p, String special, int slot, LexemGraph lg) {

	int i;
	int linecount = 0; //counts lines of outout
	char inputname[] = "/tmp/cdg-XXXXXX";
	char outputname[] = "/tmp/cdg-XXXXXX";
	int indescriptor;
	int outdescriptor;
	String call;
	FILE *input;
	FILE *output;

	/* fail on branching lattices */
	if (latticeBranches(lg->lattice)) {
		cdgPrintf(CDG_ERROR, "ERROR: can't process branching lattices\n");
		return -1;
	}

	/* open input and output and construct the invocation */
	indescriptor = mkstemp(inputname);
	if (NULL == (input = fdopen(indescriptor, "w"))) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't write input file `%s'!\n", inputname);
		return -1;
	}

	outdescriptor = mkstemp(outputname);
	if (NULL == (output = fdopen(outdescriptor, "r"))) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't read output file!\n");
		unlink(inputname);
		close(indescriptor);
		return -1;
	}

	/* write the lattice name as additional information */
	fprintf(input, "#sentence: %s\n", lg->lattice->id);

	/* iterate through words of the lattice
	 and write them to temporary file */
	for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
		GraphemNode gn = vectorElement(lg->graphemnodes, i);
		if (gn->isVirtual) {
			continue;
		}
		Arc a = (Arc) gn->arc;
		char *b;
		char *c;
		List l;
		String cat;
		LexiconItem favored = NULL;

		/* Words with embedded spaces confuse most programs.
		 Temporarily replace ' ' with an unprintable control character. */
		b = strCopy(a->word);
		while (NULL != (c = strchr(b, ' '))) {
			*c = 255;
		}

		/* print current word */
		fprintf(input, "%s", b);

		/* Try to find the preferred lexeme for the grapheme, even if `cat'
		 is not explicitly requested.

		 The reason for this is to find the most relevant value to print.
		 For instance, statistical analyzers probably want to normalize
		 words to their base forms. A predictor might ask for the value of
		 the feature `base'. But different homonyms can have different
		 values for that; for instance, the word `gefllt' may either be a
		 finite verb (VVFIN) with the base form `gefallen', or a participle
		 (VVPP) with the base form `fllen'. If POS tagging has made the
		 right decision, this will also disambiguate the base form. */
		if (-1 != taggerCategoryIndex && NULL
				!= (cat = gnHighestPrediction(gn))) {
			List l;
			for (l = gn->lexemes; l != NULL; l = l->next) {
				LexemNode ln = l->item;
				Value v = ln->lexem->values[taggerCategoryIndex];
				if (VTString == v->type && !strcmp(cat, v->data.string)) {
					favored = ln->lexem;
					break;
				}
			}
		}

		if (special == NULL) {
			/* Print additional information.

			 For instance, a PP attacher might want to know not just the surface
			 form but also the base form of each word. The List p->input
			 specifies the set of things that the predictor wants to know. The
			 token `cat' is special and refers to `the category that was scored
			 most highly by the POS tagger'. All other tokens are interpreted as
			 attribute names in the grammar, and the value of the first lexeme
			 of this grapheme which has that attribute will be printed.
			 */
			for (l = p->input; l != NULL; l = l->next) {
				String value = NULL;
				String s = l->item;
				List path = listPrependElement(NULL, s);
				int index = inputLookupPath(path, inputCurrentGrammar);
				listDelete(path);

				if (!strcmp("cat", s)) {

					/* print predicted POS tag */
					value = gnHighestPrediction(gn);
					/* This assumes that only the tagger was yet called! */

				} else if (-1 != index) {
					List m;

					if (favored) {
						Value v = favored->values[index];
						if (v->type == VTString) {
							value = v->data.string;
						}
					} else {
						for (m = gn->lexemes; m != NULL; m = m->next) {
							Value v;
							LexemNode ln = m->item;
							v = ln->lexem->values[index];

							if (v->type == VTString) {
								value = v->data.string;
								break;
							}
						}
					}
				}

				if (value) {
					fprintf(input, "\t%s", value);
				} else {
					fprintf(input, "\t-");
				}
			}
		} else {
			linecount++;

			// print Position in sentence, beginning with one
			fprintf(input, "\t%i", i + 1);
			predictPrintLexiconInfo(favored, input);

			List lexemes;
			for (lexemes = gn->lexemes; lexemes != NULL; lexemes
					= lexemes->next) {
				LexemNode ln = (LexemNode) lexemes->item;
				LexiconItem li = (LexiconItem) ln->lexem;

				//favored lexem is ignored because it was already written before
				if (li != favored) {
					/* print current word */
					fprintf(input, "%s", b);
					linecount++;

					// print Position in sentence, beginning with one
					fprintf(input, "\t%i", i + 1);

					predictPrintLexiconInfo(li, input);
				}
			}
		}
		memFree(b);

		fprintf(input, "\n");
	}
	fflush(input);

	/* call the program */
	cdgPrintf(CDG_INFO, "INFO: calling predictor `%s'\n", p->name);
	if (special == NULL) {
		call = strPrintf("%s < %s > %s", p->command, inputname, outputname);
	} else {
		call = strPrintf("%s %s %s", p->command, inputname, outputname);
	}
	if (system(call)) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't call predictor `%s'!\n", p->name);
		unlink(inputname);
		unlink(outputname);
		close(indescriptor);
		close(outdescriptor);
		return -1;
	}

	/* go through the response and record predictions */
	if (special == NULL || !strcmp(special, "-ppc3")) {
		//cdgPrintf(CDG_INFO, "CHB  predcit.c  predictSaveAsNormal  \n");
		//-ppc3: record predictions as if ppc is a standard predictor
		//this is important to recor slot -to homonym predictions
		predictSaveAsNormal(output, slot, lg);
	} else {
		//cdgPrintf(CDG_INFO, "CHB  predcit.c  predictSaveAsSpecial  \n");
		//record predictions in the special ppc way
		predictSaveAsSpecial(output, special, lg, linecount);
	}

	unlink(inputname);
	unlink(outputname);
	close(indescriptor);
	close(outdescriptor);

	return 0;
}

/* -------------------------------------------------------------------------
 The PPC-predictor requires special handling regarding how information is handed over
 */
int predictCallPPC(Predictor pred, String predictor, int slot, LexemGraph lg) {

	int i;
	int linecount = 0; //counts lines of outout
	int result = 0;
	char inputname[] = "/tmp/cdg-XXXXXX";
	//char inputname2[] = "/tmp/cdg-inPPC";
	char outputname[] = "/tmp/cdg-XXXXXX";
	//char outputname2[] = "/home/3baumgae/PPC_Output";


	int indescriptor;
	int outdescriptor;
	String call;
	FILE *input;
	FILE *output;

	/* fail on branching lattices */
	if (latticeBranches(lg->lattice)) {
		cdgPrintf(CDG_ERROR, "ERROR: can't process branching lattices\n");
		return -1;
	}

	/* open input and output and construct the invocation */
	indescriptor = mkstemp(inputname);
	//if(NULL == (input = fopen(inputname2, "w"))){
	if (NULL == (input = fdopen(indescriptor, "w"))) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't write input file `%s'!\n", inputname);
		return -1;
	}

	outdescriptor = mkstemp(outputname);
	//if(NULL == (output = fopen(outputname2, "r"))){
	if (NULL == (output = fdopen(outdescriptor, "r"))) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't read output file!\n");
		unlink(inputname);
		close(indescriptor);
		return -1;
	}

	/* write the lattice name as additional information into the temporary output file*/
	fprintf(input, "#sentence: %s\n\n", lg->lattice->id);

	/* iterate through words of the lattice
	 and write them to temporary file */
	for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
		GraphemNode gn = vectorElement(lg->graphemnodes, i);
		Arc a = (Arc) gn->arc;
		String cat;
		LexiconItem favored = NULL;

		char *b;
		char *c;

		/* Words with embedded spaces confuse most programs.
		 Temporarily replace ' ' with an unprintable control character. */
		b = strCopy(a->word);
		while (NULL != (c = strchr(b, ' '))) {
			*c = 255;
		}

		/* Try to find the preferred lexeme for the grapheme, even if `cat'
		 is not explicitly requested.

		 The reason for this is to find the most relevant value to print.
		 This LexiconItem is printed first for every word */
		if (-1 != taggerCategoryIndex && NULL
				!= (cat = gnHighestPrediction(gn))) {
			List l;
			for (l = gn->lexemes; l != NULL; l = l->next) {
				LexemNode ln = l->item;
				Value v = ln->lexem->values[taggerCategoryIndex];
				if (VTString == v->type && !strcmp(cat, v->data.string)) {
					favored = ln->lexem;
					break;
				}
			}
		}

		//cdgPrintf(CDG_INFO, "CHB preditct.c favored `%s'\n", favored->description);

		// print favored homonym first
		fprintf(input, "%s", b);

		linecount++;

		// print Position in sentence, beginning with one
		fprintf(input, "\t%i", i + 1);
		predictPrintLexiconInfo(favored, input);

		List lexemes;
		for (lexemes = gn->lexemes; lexemes != NULL; lexemes = lexemes->next) {

			LexemNode ln = (LexemNode) lexemes->item;
			LexiconItem li = (LexiconItem) ln->lexem;

			//favored lexem is ignored because it was already written before
			if (li != favored) {
				/* print current word */
				fprintf(input, "%s", b);
				linecount++;

				// print Position in sentence, beginning with one
				fprintf(input, "\t%i", i + 1);

				predictPrintLexiconInfo(li, input);
			}
		}

		memFree(b);

		fprintf(input, "\n");
	}

	fflush(input);

	/* call the program */
	cdgPrintf(CDG_INFO, "INFO: calling predictor `%s'\n", pred->name);

	call = strPrintf("%s %s %s", pred->command, inputname, outputname);
	//call = strPrintf("%s %s > %s", pred->command, inputname, outputname);


	if (system(call)) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't call predictor `%s'!\n", pred->name);
		unlink(inputname);
		unlink(outputname);
		close(indescriptor);
		close(outdescriptor);
		return -1;
	}

	/* go through the response and record predictions */

	if (!strcmp(predictor, "-ppc3")) {
		//cdgPrintf(CDG_INFO, "CHB  predcit.c  predictSaveAsNormal  \n");
		//record predictions as if ppc is a standard predictor
		//this is important to recor slot -to homonym predictions
		predictSaveAsNormal(output, slot, lg);
	} else {

		//cdgPrintf(CDG_INFO, "CHB  predcit.c  predictSaveAsSpecial  \n");
		//record predictions in the special ppc way
		predictSaveAsSpecial(output, predictor, lg, linecount);
	}

	unlink(inputname);
	unlink(outputname);
	close(indescriptor);
	close(outdescriptor);

	return result;
}

void predictPrintLexiconInfo(LexiconItem li, FILE *input) {

	List l;

	Value li_value = li->value;

	String description = li->description;

	fprintf(input, "\t%s", description);

	for (l = li_value->data.list; l; l = l->next) {

		Value list_value = l->item;

		if (list_value->type == VTAVNode) {

			AVNode node = list_value->data.avnode;

			String AVattribute = node->attribute;
			Value AVvalue = node->value;

			if (AVvalue->type == VTString) {

				String val = AVvalue->data.string;

				/* If the value is an empty string, this confuses the PPC when it reads the
				 temporary file. Therefore we change empty strings into ''.
				 */
				if (!strcmp("", val)) {
					val = "''";
				}

				fprintf(input, "\t%s\t%s", AVattribute, val);

			}
			if (AVvalue->type == VTNumber) {

				int val = AVvalue->data.number;
				fprintf(input, "\t%s\t%i", AVattribute, val);

			}
			if (AVvalue->type == VTList) {

				List listval;

				fprintf(input, "\t%s\t", AVattribute);

				fprintf(input, "<");

				for (listval = AVvalue->data.list; listval != NULL; listval
						= listval->next) {
					Value v = (Value) listval->item;

					if (v->type == VTString) {

						String val = v->data.string;
						fprintf(input, "%s,", val);

					}
					if (v->type == VTNumber) {

						int val = v->data.number;
						fprintf(input, "%i,", val);

					}

				}

				fprintf(input, ">");

			}

		}

	}

	fprintf(input, "\n");

	//     	String description = li->description;
	//
	//
	//     	fprintf(input, "\t%s", description);
	//
	//
	//
	//     	//print all attributes and their values
	//     	Value *values = li->values;
	//     	int i = 0;
	//     	int j = listSize(inputCurrentGrammar->allAttributes);
	//     	List attr = inputCurrentGrammar->allAttributes;
	//     	for(i=0; i<j;i++)
	//     	{
	//     		List p =  attr->item;
	//     		String attribute = p->item;
	//
	//     		if(values[i]->type == VTString){
	//
	//     			String val = values[i]->data.string;
	//
	// 			/* If the value is an empty string, this confuses the PPC when it reads the
	// 			   temporary file. Therefore we change empty strings into ''.
	// 			*/
	// 			if(!strcmp("", val))
	// 			{
	// 				val="''";
	//     			}
	//
	// 			fprintf(input, "\t%s\t%s", attribute, val);
	//
	//     		}
	//     		if(values[i]->type == VTNumber){
	//
	//     			int val = values[i]->data.number;
	//     			fprintf(input, "\t%s\t%i", attribute, val);
	//
	//     		}
	//     		if(values[i]->type == VTList){
	//
	//     			List listval;
	//
	//     			fprintf(input, "\t%s\t", attribute);
	//
	//     			fprintf(input, "<");
	//
	//     			for(listval = values[i]->data.list; listval !=NULL; listval= listval->next)
	//     			{
	//   					Value v = (Value) listval->item ;
	//
	//   					;
	//
	//   					if(v->type == VTString){
	//
	//     					String val = v->data.string;
	//     					fprintf(input, "%s,", val);
	//
	//     			}
	//     				if(v->type == VTNumber){
	//
	//     					int val = v->data.number;
	//     					fprintf(input, "%i,", val);
	//
	//     				}
	//
	//
	//
	//     			}
	//
	//     			fprintf(input, ">");
	//
	//     		}
	//     		attr = attr->next;
	//
	//
	//
	//     	}
	//
	//
	//
	//
	//     	fprintf(input, "\n");

}

int predictSaveAsNormal(FILE *output, int slot, LexemGraph lg) {

	int gi = 0;
	char *line = NULL;
	size_t len = 0;
	int result = 0;

	/* go through the response and record predictions */
	while (getline(&line, &len, output) != -1) {

		GraphemNode gn;
		String w;
		char *b;
		char *p;

		//ignore empty lines
		if (!strcmp("\n", line)) {
			continue;
		}

		/* Misbehaving programs might print more lines of output than input,
		 guard against this */
		if (gi >= vectorSize(lg->graphemnodes)) {
			//if there exist too many lines of output, ignore them
			break;

			/*cdgPrintf(CDG_ERROR,
			 "ERROR: Too many lines of output! Processing stopped for this sentence.\n");
			 return -1;*/
		}
		gn = (GraphemNode) vectorElement(lg->graphemnodes, gi);

		w = strtok(line, " \t\n");
		if (!w) {
			cdgPrintf(CDG_ERROR, "ERROR: can't find word\n");
			return -1;
		}

		/* Undo space escaping that we might have done */
		b = strCopy(w);
		while (NULL != (p = strchr(b, 255))) {
			*p = ' ';
		}

		//cdgPrintf(CDG_INFO,"CHB predcit.c word %s.\n", b);


		if (!cdgPredictor) {
			if (strcmp(b, gn->arc->word)) {
				cdgPrintf(CDG_ERROR,
						"ERROR: word mismatch! `%s' should be `%s'\n", b,
						gn->arc->word);
				result = -1;
				break;
			}
		}

		/* loop through all tokens in a response line */
		String key = NULL, value = NULL;
		do {

			key = strtok(NULL, " \t\n");
			if (key) {
				value = strtok(NULL, " \t\n");

				/* read one prediction */
				if (key && value) {
					Number f;
#ifdef DEBUGPREDICTCALL
					cdgPrintf(CDG_DEBUG, "DEBUG: %s  %s  %s\n", b, key, value);
#endif
					if (strspn(value, "+-e.0123456789") == strlen(value) && 1
							== sscanf(value, "%lf", &f)) {
						gnWriteNumber(gn, slot, strRegister(key), f);

					} else {
						gnWriteString(gn, slot, strRegister(key), strRegister(
								value));

					}
				}
			}
		} while (key && value);
		memFree(b);
		gi++;
	}

	if (line) {
		memFree(line);
	}

	return result;

}

int predictSaveAsSpecial(FILE *output, String predictor, LexemGraph lg,
		int linecount) {

	int gi;
	char *line = NULL;
	size_t len = 0;

	gi = 0;
	while (getline(&line, &len, output) != -1) {

		GraphemNode gn;
		//GraphemNode gnold = NULL;
		String w;
		char *b;
		char *p;

		//ignore empty lines
		if (!strcmp("\n", line)) {
			continue;
		} else {

			//The output file of the PPC also contains debug information
			//These lines should be ignored
			//Therefore we stop after we read all relevant information

			if (linecount <= 0) {
				break;
			}
			linecount--;

		}

		w = strtok(line, " \t\n");
		if (!w) {
			cdgPrintf(CDG_ERROR, "ERROR: can't find word\n");
			return -1;
		}

		/* Undo space escaping that we might have done */
		b = strCopy(w);
		while (NULL != (p = strchr(b, 255))) {
			*p = ' ';
		}

		//very ugly workaround for not reading useless lines that PPC puts into output file
		// has to be fixed soon
		/*if(!strncmp("====", b, 4))
		 {
		 break;
		 }*/

		//cdgPrintf(CDG_INFO,"CHB predcit.c word %s gi %i .\n", b, gi);

		//Second token is always the position of the word in the sentence
		String position = strtok(NULL, " \t\n");

		//Check if the second token is really a number
		//exit if it is not
		if (strspn(position, "0123456789") != strlen(position)) {
			cdgPrintf(CDG_ERROR,
					"ERROR: Word-position in PPC-output is not a number\n");
			return -1;
		}

		Number pos;
		sscanf(position, "%lf", &pos);

		gi = pos - 1;

		//check if the position provided by PPC is really the position of a word in the sentence
		if (gi >= vectorSize(lg->graphemnodes)) {
			cdgPrintf(CDG_ERROR,
					"ERROR: Word-position in PPC-output too large for sentence\n");
			return -1;
		}

		gn = (GraphemNode) vectorElement(lg->graphemnodes, gi);

		/*if(gi>0)
		 {
		 gnold = (GraphemNode) vectorElement(lg->graphemnodes, gi-1);

		 }*/

		//Third token is always the unique description for a lexeme
		String description = strtok(NULL, " \t\n");

		/* loop through all tokens in a response line */
		String key = NULL, value = NULL;
		do {

			key = strtok(NULL, " \t\n");
			if (key) {
				value = strtok(NULL, " \t\n");

				/* read one prediction */
				if (key && value) {
					Number f;
#ifdef DEBUGPREDICTCALL
					cdgPrintf(CDG_DEBUG, "DEBUG: %s  %s  %s\n", b, key, value);
#endif
					if (strspn(value, "+-e.0123456789") == strlen(value) && 1
							== sscanf(value, "%lf", &f)) {

						/*if(gnold)
						 {
						 Value v = gnReadPredictionSpecial(gnold,predictor,"der_ART",key);
						 double d = v->data.number;
						 cdgPrintf(CDG_INFO, "CHB predict.c number gnold %s %lf\n",key, d);
						 }*/

						//cdgPrintf(CDG_INFO, "CHB  predcit.c  predictor  description key number %s %s %s %f \n", predictor,  description, key, f);
						gnWriteNumberSpecial(gn, strRegister(predictor),
								strRegister(description), strRegister(key), f);
						//Value v = gnReadPredictionSpecial(gn,predictor,description,key);
						//	 double d = v->data.number;
						//cdgPrintf(CDG_INFO, "CHB predict.c predictor description key number %s %s %s %lf\n",predictor,description,key, d);


					} else {
						gnWriteStringSpecial(gn, strRegister(predictor),
								strRegister(description), strRegister(key),
								strRegister(value));
						//Value v = gnReadPredictionSpecial(gn,predictor,description,key);

						//String s = v->data.string;
						//	cdgPrintf(CDG_INFO, "CHB  predict.c  string %s %s %s\n",b, key, s);


						/*if(gnold)
						 {
						 Value v = gnReadPredictionSpecial(gnold,predictor,"der_ART",key);
						 String s = v->data.string;
						 cdgPrintf(CDG_INFO, "CHB  predict.c  string gnold %s %s\n",key, s);
						 }*/

					}
				}
			}
		} while (key && value);
		memFree(b);
	}

	if (line) {
		memFree(line);
	}

	return 0;

}

/* ----------------------------------------------------------------------
 Find a Predictor by its name
 the returned value is either 0 (for the POS tagger)
 or one more than an index into the cdgPredictors vector
 or -1 if nothing can be found
 */
int predictFindPredictor(String name) {
	int i;
	if (name == constStr_POS)
		return 0;
	for (i = 0; i < vectorSize(cdgPredictors); i++) {
		if (((Predictor) vectorElement(cdgPredictors, i))->name == name)
			return i + 1;
	}
	return -1;
}

/* ----------------------------------------------------------------------
 Create a new Predictor.
 */
Predictor predictNew(String name, String command, List features) {
	Predictor result = memMalloc(sizeof(PredictorStruct));
	result->name = strRegister(name);
	result->input = features;
	result->command = command;
	return result;
}

/* ----------------------------------------------------------------------
 Deallocate a Predictor.
 */
void predictFree(Predictor p) {
	cdgFreeString(p->name);
	listDelete(p->input);
	cdgFreeString(p->command);
	memFree(p);
}

/* ----------------------------------------------------------------------
 Deallocate all Predictors.
 */
void predictFinalize() {
	int i;
	for (i = 0; i < vectorSize(cdgPredictors); i++) {
		Predictor p = vectorElement(cdgPredictors, i);
		predictFree(p);
	}
	vectorDelete(cdgPredictors);
	cdgPredictors = NULL;
}

/* ----------------------------------------------------------------------
 Call all predictors.
 */
void predictAll(LexemGraph lg) {
	int i;

	for (i = 0; i < vectorSize(cdgPredictors); i++) {

		Predictor pred = (Predictor) vectorElement(cdgPredictors, i);
		String special;
		special = hashGet(specialPredictors, pred->name);

		if (special == NULL || !strcmp(special, "-ppc4")) {
			// -ppc4: record predictions for slot-to-slot relations
			predictCall(pred, NULL, i + 1, lg);
		} else if (!strcmp(special, "-cdg")) {
			cdgPredictor = TRUE;
			predictCall(pred, NULL, i + 1, lg);
			cdgPredictor = FALSE;
		} else {
			// -ppc1: record predictions for homonym-to-homonym relations    	
			// -ppc2: record predictions for homonym-to-slot relations    	
			// -ppc3: record predictions for slot-to-homonym relations    	    	
			predictCall(pred, special, i + 1, lg);
		}
	}
}

/* ------------------------------------------------------------------- */
/* ------------------------------------------------------------------- */

