/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * this program is distributed in the hope that it will be useful, but
 * without any warranty, to the extent permitted by law; without even the
 * implied warranty of merchantability or fitness for a particular purpose.
 *
 * $id: skel.c,v 1.8 2003/05/14 15:07:59 micha exp $
 */

/* -------------------------------------------------------------------------
 * @addtogroup Scache Scache - Cache structures for binary LV scores
 * @author Michael Schulz
 * @date somewhen in 1998
 *
 * A score cache serves much the same purpose as the score matrices in
 * the edges of a constraint net, except that it can be used even if the
 * constraint net has no edges in it. This simply means that the binary
 * score of a pair of LVs will be computed when it is used for the first
 * time rather than upon initializing the net.
 *
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include "scache.h"
#include "hook.h"
#include "set.h"

/* ---------------------------------------------------------------------- */
/** determines the index of a given pair */
#define indexOfPair(x,y) ((x<y) ?  (((y*(y-1))>>1)+x) : (((x*(x-1))>>1)+y))
/** debug flag */
#define DEBUG_SCORECACHE 0

/* ----------------------------------------------------------------------
 * implements the CDG variable \b cache.
 * If this flag is set, binary constraints are never evaluated twice on the
 * same pair of LVs. The function evalBinary() will then simply call
 * scGetScore() instead of evalConstraint().
 *
 * This variable is exported because it's cheaper for the calling function
 * to check a boolean variable than calling a cache function which does
 * nothing again and again.
 */
Boolean scUseCache=FALSE;

/* ----------------------------------------------------------------------
 * return an initial score-cache
 *
 * This function returns a pointer to a new cache. The fields
 * \b count and \b hits are initialized to 0. All elements of the
 * underlying Vector are set to -1.
 */
ScoreCache scNew(int noValues)
{
  ScoreCache cache;
  int capacity = (((noValues*(noValues-1))>>1)+noValues-1); //indexOfPair(noValues, noValues - 1);

  cache = (ScoreCache) memMalloc(sizeof (ScoreCacheStruct));
  cache->count = 0;
  cache->size = 0;
  cache->capacity = capacity;
  cache->hits = 0;

  cache->data = (Number *) memMalloc(sizeof (Number) * capacity);
#if 0
  for (i = 0; i < capacity; i++) {
    cache->data[i] = -1.0;
  }
#endif

  return cache;
}

/* ----------------------------------------------------------------------
 * free memory from a cache
 *
 * This function deallocates a cache.
 */
void scDelete(ScoreCache cache)
{
  if(!cache) { return; }

   memFree(cache->data);
   memFree(cache);
}

/* ----------------------------------------------------------------------
 * retrieve score from cache if available, else return -1.0
 *
 * This function returns the mutual score for the LVs \b a and
 * \b b from the score. If it is not yet known, -1 is returned.
 */
Number scGetScore(ScoreCache cache, LevelValue a, LevelValue b)
{
  Number score;
  int index;

  if (!cache || a->no < 0 || b->no < 0
      || a->indexWRTNet == -1 || a->indexWRTNet == -1) { // frobbing hack
    return -1.0;
  }

  index = indexOfPair(a->no, b->no);
  if (index >= cache->size) {
    return -1.0;
  }

  score = cache->data[index];
  if (score != -1.0) {
    cache->hits++;
  }

  return score;
}

/* ----------------------------------------------------------------------
 * store score in cache
 *
 * This function registers a binary score in the cache. It behaves as follows:
 *
 * - If \b cache is \b NULL the function returns.
 * - If either of the LVs still has its \b no set to -1,
 *   \b registerValue() is applied to it.
 * - A unique index is computed from the fields \b no of both LVs.
 * - The underlying Vector is resize if necessary.
 * - \b score is entered in the corresponding cell.
 */
void scSetScore(ScoreCache cache, LevelValue a, LevelValue b, Number score)
{
  int index;
  int i;

  if (!cache
      || a->indexWRTNet == -1 || b->indexWRTNet == -1) { // frobbing hack
    return;
  }

  if (a->no < 0) {
    a->no = cache->count++;
  }
  if (b->no < 0) {
    b->no = cache->count++;
  }

  index = indexOfPair(a->no, b->no);

  /*
   * resizing should only happen if we don't know the maximum number
   * of levelvalues in a constraintnet in advance, ie. in incremental
   * computations
   */
  if (index >= cache->capacity) {
    int old;

    old = cache->capacity;
    while (index >= cache->capacity) {
      cache->capacity *= 2;
    }

#if DEBUG_SCORECACHE
    cdgPrintf(CDG_DEBUG,
	      "DEBUG: resizing score-cache from %d to %d "
	      "because of levelvalues %d, %d (index %d)\n",
	      old, cache->capacity, a->no, b->no, index);
#endif

    cache->data = (Number *) memRealloc(cache->data,
					sizeof (Number) * cache->capacity);

#if 0
    for (i = old; i < cache->capacity; i++) {
      cache->data[i] = -1.0;
    }
#endif
  }

  /*
   * if we want a bucket behind the used size then initialize the unused part
   */
  if (cache->size < index) {
    for (i = cache->size; i < index; i++) {
      cache->data[i] = -1.0;
    }
    cache->size = index;
  }

#if DEBUG_SCORECACHE
  else {
    Number s;

    s = cache->data[index];
    if (s != -1.0 && s != score) {
      cdgPrintf(CDG_ERROR,
		"ERROR: overwriting binary score %f with %f\n", s, score);
      cdgPrintf(CDG_ERROR, "       a=%d b=%d index=%d\n", a->no, b->no, index);
    }
  }
#endif

  cache->data[index] = score;
}

/* ----------------------------------------------------------------------
 * initialize the module \b Scache. It registers
 * the module's CDG variables.
 */
void scInitialize()
{
  setRegister("cache", SET_BOOL, &scUseCache, NULL, NULL, NULL, NULL);
}
/** @} */
