/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Jochen Hagenstroem (see also AUTHORS and THANKS for more)
 * Purpose: interface CDG to an external POS tagger
 * Birth: 2001/02/07
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <signal.h>
#include <blah.h>
#include "hook.h"
#include "set.h"
#include "lexemgraph.h"
#include "tagger.h"
#include "virtualnode.h"

/* -- VARIABLES ------------------------------------------------------------ */

/* exported globals */
int taggerCategoryIndex = -1;
String taggerCategoryPath = NULL;

/* only internally visible */
static String taggerCommand = NULL;
static Boolean taggerIgnoreImpossible = TRUE;
static Boolean taggerNormalizeToOne = TRUE;
static Boolean is_up = TRUE;


/* -- TYPE DEFINITIONS ----------------------------------------------------- */
typedef struct {
	Number score;
	String tag;
} TagScoreStruct;
typedef TagScoreStruct *TagScore;

/* -- FUNCTIONS ------------------------------------------------------------ */

/* -------------------------------------------------------------------------
 Set taggerCategoryIndex.

 Looks up the feature index of the feature specified as
 `taggerCategoryPath' and stores it into taggerCategoryIndex.
 */
Boolean taggerLookupCategoryIndex() {
	List path;

	taggerCategoryIndex = -1;
	if (!taggerCategoryPath || !*taggerCategoryPath) {
		cdgPrintf(CDG_WARNING, "WARNING: no category path\n");
		return FALSE;
	}
	path = scanPath(taggerCategoryPath);
	taggerCategoryIndex = inputLookupPath(path, inputCurrentGrammar);
	inputFreePath(path);

	return taggerCategoryIndex >= 0;
}

/**
 * normalize all tag score in the list via the given factor
 */
void normalizeTagScore(List tss, float factor) {
	List l;
	for (l = tss; l != NULL; l = l->next) {
		TagScore ts = l->item;
		ts->score /= factor;
	}
	return;
}

/*
 * delete the list and all tagger scores it contains
 */
void taggerDeleteScoreList(List tss) {
	List l= tss;
	for(;l;l=l->next) {
		TagScore ts = l->item;
		memFree(ts);
	}
	listDelete(tss);
}

/**
 * calculates feedback to be given to the tagger,
 * used in incremental parsing to inform the tagger about tags selected by the parser
 *
 * the feedback list consists of all the selected tags, but
 *
 * - if a tag the tag history differs from the selected tags, feedback list ends after that tag
 * - for a delay of n, the last n elements are ignored
 */
List taggerCalculateFeedback(List selectedTags, List tagHistory, int delay) {
	List l =selectedTags;
	List old = tagHistory;
	List feedback= listNew();
	Boolean changed= FALSE;

	// for all tokens
	while(l && !changed) {
		String tag =l->item;
		if(old) {
			String oldTag= old->item;
			if(strcmp(tag,oldTag)!=0) {
				// the tag has changed,
				// so just use the unchanged prefix as feedback,
				// as the rest should be retagged
				cdgPrintf(CDG_DEBUG,"DEBUG: selected tag changed from %s to %s, retagging ... \n", oldTag, tag);
				changed = TRUE;
			}
			old = old->next;
		}

		if(listSize(l)>delay) {
			feedback= listAppendElement(feedback,l->item);
		}

		l=l->next;
	}

	return feedback;
}


/* -------------------------------------------------------------------------
 Decorate a lexeme graph with part-of-speech scores.
 
 Calls the specified external tagger and assigns predictions to the
 underlying grapheme node.
 */
Boolean taggerTag(LexemGraph lg, List feedback) {

	int gi;
	List l;
	Boolean result = TRUE;
	List fb = feedback;
	char inputname[] = "/tmp/cdg-XXXXXX";
	char outputname[] = "/tmp/cdg-XXXXXX";
	int indescriptor;
	int outdescriptor;
	String call;
	FILE *input;
	FILE *output;
	char buffer[1024];

	if (!is_up) {
		return FALSE;
	}

	/* check if wordgraph is linear, only then tagging can be done */
	if (latticeBranches(lg->lattice)) {
		cdgPrintf(CDG_ERROR, "ERROR: can't tag branching lattices\n");
		return FALSE;
	}

	/* check for valid category index */
	if (taggerCategoryIndex < 0 && !taggerLookupCategoryIndex()) {
		cdgPrintf(CDG_ERROR, "ERROR: can't lookup category index\n");
		return FALSE;
	}

	/* open input and output and construct the tagger call */
	indescriptor = mkstemp(inputname);
	if (NULL == (input = fdopen(indescriptor, "w"))) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't write tagger input file `%s'!\n",
				inputname);
		return FALSE;
	}

	outdescriptor = mkstemp(outputname);
	if (NULL == (output = fdopen(outdescriptor, "r"))) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't read tagger output file!\n");
		unlink(inputname);
		close(indescriptor);
		return FALSE;
	}

	if(lg->lattice->lookahead) {
		l = listAppendList(listClone(lg->lattice->arcs), lg->lattice->lookahead);
	} else {
		l = lg->lattice->arcs;
	}

	/* iterate through words of the lattice
	 and write them to temporary file */
	for ( ; l; l = l->next) {
		Arc a = (Arc) l->item;
		char b[1024];
		char *p;

		/* Words with embedded spaces confuse most taggers.
		 Temporarily replace ' ' with an unprintable control character. */
		strncpy(b, a->word, 1024);
		while (NULL != (p = strchr(b, ' '))) {
			*p = 255;
		}

		if(fb) {
			// pass the current word plus feedback to the tagger
			fprintf(input, "%s %s\n", b, (String)fb->item);
			fb=fb->next;
		} else {
			// pass only the current word to the tagger
			fprintf(input, "%s\n", b);
		}
	}
	fflush(input);

	/* call the tagger */
	call = strPrintf("%s < %s > %s", taggerCommand, inputname, outputname);
	if (system(call)) {
		cdgPrintf(CDG_ERROR, "ERROR: Can't call tagger!\n");
		unlink(inputname);
		unlink(outputname);
		close(indescriptor);
		close(outdescriptor);
		return FALSE;
	}

	/* go through tagger response and set tag score accordingly */
	gi = 0;
	while (fgets(buffer, 1024, output)) {

		GraphemNode gn;
		String w;
		String t;
		List lexCats, tagCats = NULL;
		Number total = 0.0;
		Number best = 0.0;
		char b[1024];
		char *p;

		/* No newline? That means the tagger wrote a line that is longer than
		 our buffer. Ignore in distress. */
		if (strcspn(buffer, "\n") == strlen(buffer)) {
			cdgPrintf(CDG_WARNING,
					"WARNING: Tagger output too wide! Tagging ignored for this line.\n");
			continue;
		}

		/* Misbehaving taggers might print more lines of output than input,
		 guard against this */
		if (gi >= (lg->ve?lg->ve->vStartID:vectorSize(lg->graphemnodes)) ) {
			if(lg->lattice->lookahead) {
				// a longer output is to be expected, as we gave more words, just stop the loop
				break;
			} else {
				cdgPrintf(CDG_ERROR,
						"ERROR: Too many lines of tagger output! Tagging stopped for this sentence.\n");
				return FALSE;
			}
		}

		gn = (GraphemNode) vectorElement(lg->graphemnodes, gi);
		if (!gn) {
			cdgPrintf(CDG_ERROR, "ERROR: invalid graphem node\n");
			result = FALSE;
			break;
		}
		w = strtok(buffer, " \t\n");
		if (!w) {
			cdgPrintf(CDG_ERROR, "ERROR: can't find word\n");
			result = FALSE;
			break;
		}

		/* Undo space escaping that we might have done */
		strncpy(b, w, 1024);
		while (NULL != (p = strchr(b, 255))) {
			*p = ' ';
		}

		if (strcmp(b, gn->arc->word)) {
			cdgPrintf(CDG_ERROR, "ERROR: word mismatch %s %s\n", b,
					gn->arc->word);
			result = FALSE;
			break;
		}

		// don't overwrite tagger scores for tokens that are part of the feedback we gave to the tagger
		if(!feedback || gi >= listSize(feedback)) {
			// determine and write tag scores

			lexCats = lgQueryCat(lg, gn);

			/* loop through all tokens in a response line */
			for (t = strRegister(strtok(NULL, " \t\n")); t != strRegister(""); t
					= strRegister(strtok(NULL, " \t\n"))) {
				String sc = strtok(NULL, " \t\n");
				Number score = 1.0;
				TagScore ts;

				/* read one tagscore checking for well-formedness of tagger output */
				if (sc && 1 != sscanf(sc, "%lf", &score)) {
					cdgPrintf(CDG_WARNING, "WARNING: can't read score from %s\n",
							sc);
					continue;
				}

				cdgPrintf(CDG_DEBUG, "DEBUG: %s  %s  %4.3e\n", b, t, score);
				if (!listContains(lexCats, t)) {
					if (taggerIgnoreImpossible) {
						cdgPrintf(CDG_DEBUG,
								"WARNING: category `%s' not in lexicon\n", t);
						continue;
					}
				}
				best = max(best, score);
				total += score;
				ts = memMalloc(sizeof(TagScoreStruct));
				ts->score = score;
				ts->tag = t;
				tagCats = listPrependElement(tagCats, ts);
			}

			/* Normalize away the probabilities spent on impossible categories */
			if (taggerIgnoreImpossible) {
				normalizeTagScore(tagCats,total);
				best /= total;
			}

			/* Convert probabilities to scores */
			if (taggerNormalizeToOne) {
				normalizeTagScore(tagCats,best);
			}

			/* if any of the predictions match any of the possibilities in our
			 lexicon, write those scores and set the others to 0 */
			if (tagCats) {

				for (l = lexCats; l != NULL; l = l->next) {
					String cat = l->item;
					gnWriteNumber(gn, 0, cat, 0.0);
				}
				for (l = tagCats; l != NULL; l = l->next) {
					TagScore ts = l->item;
					gnWriteNumber(gn, 0, ts->tag, ts->score);
					memFree(ts);
				}
				listDelete(tagCats);
			} else {
				/* Otherwise just leave the predictions of 1 everywhere. */
				cdgPrintf(CDG_WARNING, "WARNING: none of the suggested tags "
					"for `%s' is in your lexicon.\n", b);
			}
			listDelete(lexCats);
		} else {
			cdgPrintf(CDG_DEBUG,"DEBUG: didn't read tagger score as token was part of feedback\n");
			String fbtag = listNthElement(feedback,gi+1);
			cdgPrintf(CDG_DEBUG,"fed back tag: %s, old score: %.3f\n",fbtag ,gnReadPrediction(gn,0,fbtag)->data.number);
		}//end if(!feedback ...

		gi++;
	}
	unlink(inputname);
	unlink(outputname);
	close(indescriptor);
	close(outdescriptor);

	return result;
}

/* -------------------------------------------------------------------------
 * taggerInitialize
 */
void taggerInitialize() {
	taggerCategoryPath = strRegister("cat");
	taggerLookupCategoryIndex();

	is_up = FALSE;

	taggerCommand = strRegister("");
	setRegister("taggerCommand", SET_STRING, &taggerCommand, NULL, NULL, NULL,
			NULL);
	/* callback function on set to get pathindex  */
	setRegister("taggerCategoryPath", SET_STRING, &taggerCategoryPath, NULL,
			NULL, (VoidFunction *) taggerLookupCategoryIndex, NULL);

	setRegister("taggerNormalizeToOne", SET_BOOL, &taggerNormalizeToOne, NULL,
			NULL, NULL, NULL);
	setRegister("taggerIgnoreImpossible", SET_BOOL, &taggerIgnoreImpossible,
			NULL, NULL, NULL, NULL);

}

/* -------------------------------------------------------------------------
 Return resources allocatged by taggerInitialize().
 */
void taggerFinalize() {
}

/* ----------------------------------------------------------------------
 Return tagger status.

 If the tagger is in use, i.e. `newnet' commands will cause tagging to be
 done, TRUE is returned, otherwise FALSE.
 */
Boolean taggerUp(void) {
	return is_up;
}

/* ----------------------------------------------------------------------
 Toggle tagger status.
 */
int taggerStartStop(void) {
	is_up = !is_up;
	cdgPrintf(CDG_INFO, "INFO: tagger st%sed\n", is_up ? "art" : "opp");
	return 0;
}

/* ------------------------------------------------------------------- */
/* ------------------------------------------------------------------- */

