/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    
*/

/* ----------------------------------------------------------------------
 * @defgroup BitString BitStrings
 * 
 * Implementation of a string of bits.
 * 
 * A bitstring is a memory efficient implementation of a sequence of
 * Boolean information with arbitrary but fixed length. 
 *
 * @{
 */


/* -- INCLUDES ------------------------------------------------------- */
#include <stdio.h>
#include <string.h>
#include "blah.h"

/* --MACROS ---------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * defines our version of a BYTE definition
 */
#define BITS_PER_BYTE 8

/* ----------------------------------------------------------------------
 * defines the size of an unsigned long
 */
#define BYTES_PER_LONG sizeof(unsigned long)

/* ----------------------------------------------------------------------
 * defines the number of bits in an unsigned long integer 
 */
#define BITS_PER_LONG (BITS_PER_BYTE*BYTES_PER_LONG)

/* ----------------------------------------------------------------------
 * return code for size error messages 
 */
#define E_SIZEMISMATCH -1

/* -- TYPE DEFINITIONS ----------------------------------------------- */

/* ----------------------------------------------------------------------
 * internal structure of a string of bits.
 */

struct BitStringStruct {
#ifdef BITSTRINGS_DEBUG
  int magic;                    /**< magic number to identify bitstrings */
  int magic;		        /**< magic number to identify bitstrings */
#endif
  int size;                     /**< size of bitstring in bits */
  int length;                   /**< length of array; 
length=capacity/BITS_PER_LONG */
  unsigned long mask;           /**< mask for the last unsigned long */
  unsigned long *bits;          /**< array of longs */
};

/* -- VARIABLES ------------------------------------------------------ */
#ifdef BITSTRINGS_DEBUG
static int magic_constant=-321934;
#endif

/* -- FUNCTIONS ------------------------------------------------------ */

#ifdef BITSTRINGS_DEBUG
 static void print_long_as_binary(unsigned long l)
{
  int i;

  for (i=0; i<BITS_PER_LONG; i++)
    {
      printf("%d", l & (1<<i) ? 1 : 0);
    }
}

static void check_magic_function(BitString bs, char *file, int no)
{
  if (bs->magic!=magic_constant)
    {
      fprintf(stderr, "%s:%d: not a bitstring\n", file, no);
      abort();
    }
}
/* ----------------------------------------------------------------------
 * a magic checking routine used when compiled with -DBITSTRINGS_DEBUG
 */
#define check_magic(bs) check_magic_function(bs, __FILE__, __LINE__)

#else /* #ifdef BITSTRINGS_DEBUG */

/* ----------------------------------------------------------------------
 * this is defined to a null operation when debugging is switched off.
 *
 * Switch it on by compiling with -DBITSTRINGS_DEBUG
 */
#define check_magic(bs)

#endif /* #ifdef BITSTRINGS_DEBUG */


/* ----------------------------------------------------------------------
 * creates a new bitstring with an initial size of \a size bits.
 *
 * @param size specifies the size of the new bit vector to be created.
 * @returns a new bit vector of the specified size.
 */
BitString bitNew(int size)
{
  BitString bs;
  int s;
  int i;

  if(size <= 0) {
    size = 1;
  }
  
  s=size/BITS_PER_LONG;
  if (0 != size%BITS_PER_LONG) { s++; }

  bs=(BitString)memMalloc(sizeof(BitStringStruct));
#ifdef BITSTRINGS_DEBUG
  bs->magic=magic_constant;
#endif
  bs->size=size;
  bs->length=s;

  /*
   * bs->mask has all significant bits in the last unsigned long
   * set and all unsignificant bits cleared.
   */
  bs->mask = ~(unsigned long)0;
  for (i=bs->size%BITS_PER_LONG; i<BITS_PER_LONG; i++)
    {
      /* clear ith bit */
      bs->mask &= ~(1 << i);
    }
  bs->bits=(unsigned long *)memMalloc(s*BYTES_PER_LONG);
  /* memset(bs->bits, 0x55, s*BYTES_PER_LONG); */
  memset(bs->bits, 0, s*BYTES_PER_LONG);
  
  return(bs);
}

/* ----------------------------------------------------------------------
 * creates a new bitstring which is a complete clone of bs.
 *
 * @param bs the bit string that has to be cloned
 * @returns a copy of the bitstring  after cloning.
 */
BitString bitClone(BitString bs)
{
  BitString new;

  check_magic(bs);
  
  new=bitNew(bs->size);
  memcpy(new->bits, bs->bits, bs->length*BYTES_PER_LONG);

  return(new);
}

/* ----------------------------------------------------------------------
 * copies information from src to dst.
 *
 * dst and src must be bitstrings of the same size
 *
 * @param dst destination BitString
 * @param src source Bitstring 
 * @returns 1 on success, 0 on failure (e. g. , due to size mismatch)
 *
 */
BitString bitCopy(BitString dst, BitString src)
{
  check_magic(dst);
  check_magic(src);

  if (dst->size!=src->size) { return NULL; }
  memcpy(dst->bits, src->bits, src->length*BYTES_PER_LONG);
  
  return dst;
}


/* ----------------------------------------------------------------------
 * deletes bitstring \a bs and frees memory.
 *
 * @param bs the bitstring that has to be deleted.
 */
void bitDelete(BitString bs)
{
  check_magic(bs);

  memFree(bs->bits);
  memFree(bs);
}

/* ----------------------------------------------------------------------
 * returns the size of bitstring \a bs.
 *
 * @param bs the bitstring whose size has to be determined.
 * @returns the size of the bitstring bs
 */
int bitSize(BitString bs)
{
  check_magic(bs);

  return bs->size;
}

/* ----------------------------------------------------------------------
 * prints a bitstring to stdout.
 *
 * @param bs the BitString to be printed
 */
void bitPrint(BitString bs)
{
  int i;

  check_magic(bs);

  for(i=0; i<bs->size; i++)
    {
      printf("%d", bitGet(bs, i) ? 1 : 0);
    }
}

/* ----------------------------------------------------------------------
 * resize bs to size \a size; sets length, mask and size.
 *
 * @param bs the bitstring that has to be resized.
 * @param size the size to which the bitstring has to be resized.
 * @returns the new resized bitstring.
 */
static void resize(BitString bs, int size)
{
  int i;

  if (size>bs->length*BITS_PER_LONG)
    {
      int l;

#ifdef BITSTRINGS_DEBUG
      fprintf(stderr, "resize: re-alloc necessary\n");
#endif
      l=size/BITS_PER_LONG;
      if (0 != size%BITS_PER_LONG) { l++; }
      bs->bits=(unsigned long *)memRealloc(bs->bits, l*BYTES_PER_LONG);
      /* clear those bits that are new */
      memset(&bs->bits[bs->length], 0, (l-bs->length)*BYTES_PER_LONG);
      bs->length=l;
    }

  bs->size=size;

  /*
   * bs->mask has all significant bits in the last unsigned long
   * set and all unsignificant bits cleared.
   */
  bs->mask = ~(unsigned long)0;
  for (i=bs->size%BITS_PER_LONG; i<BITS_PER_LONG; i++)
    {
      /* clear ith bit */
      bs->mask &= ~(1 << i);
    }
}

/* ----------------------------------------------------------------------
 * sets a bit.
 *
 * if \a no is larger than current size the bitstring is enlarged.
 * intermediate bits are cleared.
 * Bits are numbered from zero upwards.
 *
 * @param bs the bitstring in which a bit has to be set.
 * @param no the number of the bit that has to be set
 */
void bitSet(BitString bs, int no)
{
  check_magic(bs);

  if (bs->size<=no) { resize(bs, no+1); }
  bs->bits[no/BITS_PER_LONG] |= (1<<(no%BITS_PER_LONG));
}


/* ----------------------------------------------------------------------
 * clears a bit.
 *
 * if no is larger than current size the bitstring is enlarged.
 * intermediate bits are cleared.
 * Bits are numbered from zero upward
 *
 * @param bs the bitstring from which a bit has to be cleared.
 * @param no the bit number that has to be cleared.
 */
void bitClear(BitString bs, int no)
{
  check_magic(bs);

  if (bs->size<=no) { resize(bs, no+1); }
  bs->bits[no/BITS_PER_LONG] &= ~(1<<(no%BITS_PER_LONG));
}


/* ----------------------------------------------------------------------
 * sets all the bits to one.
 *
 * @param bs the bitstring whose bits are to be set.
 * @returns the new bitstring after setting all its bits to 1.
 */
void bitSetAll(BitString bs)
{
  check_magic(bs);

  memset(bs->bits, ~0, bs->length*BYTES_PER_LONG);
  bs->bits[bs->length-1] &= bs->mask;
}

/* ----------------------------------------------------------------------
 * sets all bits to zero.
 *
 * @param bs the bitstring whose bits are to be cleared.
 * @returns the new bitstring after setting all its bits to zero.
 */
void bitClearAll(BitString bs)
{
  check_magic(bs);

  memset(bs->bits, 0, bs->length*BYTES_PER_LONG);
}

/* ----------------------------------------------------------------------
 * computes a logical \a AND between two sets of bits
 *
 * sets all the bits in bitstring a that are set to 0 in bitstring b to 0.
 * bitstring b is never modified.
 * the bitstrings must be of the same size.
 *
 * @param a the first set of bits
 * @param b the second set of bits
 * @returns a on success and NULL on failure.
 */
BitString bitAnd(BitString a, BitString b)
{
  int i;

  check_magic(a);
  check_magic(b);

  if (a->size!=b->size) { return NULL; }

  for (i=0; i<a->length; i++)
    {
      a->bits[i] &= b->bits[i];
    }

  return a;
}


/* ----------------------------------------------------------------------
 * computes a logical \a OR between two sets of bits
 *
 * sets all the bits in bitstring a that are set to 1 in bitstring b to 1.
 * bitstring b is never modified.
 * the bitstrings must be of the same size.
 *
 * @param a the first set of bits
 * @param b the second set of bits
 * @returns a on success and NULL on failure.
 */
BitString bitOr(BitString a, BitString b)
{
  int i;

  check_magic(a);
  check_magic(b);

  if (a->size!=b->size) { return NULL; }

  for (i=0; i<a->length; i++)
    {
      a->bits[i] |= b->bits[i];
    }

  return a;
}

/* ----------------------------------------------------------------------
 * checks whether two sets have bits switched on in common
 *
 * @param a the first set
 * @param b the second set of bits
 * @returns TRUE on the first common bit (i.e. two bits at the same position
 * which are switched on); otherwise false is returned, that there are no
 * common bits switched on.
 * @requires a->size == b->size
 */
Boolean bitCheck(BitString a, BitString b)
{
  int i;
  
  check_magic(a);
  check_magic(b);

  if (a->size != b->size) {
    return FALSE;
  }

  for (i = 0; i < a->length; i++) {
    if (a->bits[i] & b->bits[i]) {
      return TRUE;
    }
  }

  return FALSE;
}

/* ----------------------------------------------------------------------
 * test for a bit
 *
 * This function checks whether a bitstring has bit at a certain
 * position switched on.
 *
 * @param bs the bitstring
 * @param no the position where to look at
 * @returns the state of the bit at the given position
 */
Boolean bitGet(BitString bs, int no)
{
  check_magic(bs);
#ifdef BITSTRINGS_DEBUG
  if (bs->size <= no) {
    fprintf(stderr, "bitGet: no >= bs->size\n");
    abort();
  }
#endif

  return (bs-> bits[no / BITS_PER_LONG] & (1 << (no % BITS_PER_LONG)) ? TRUE : 
FALSE);
}

/* ----------------------------------------------------------------------
 * checks whether all bits are cleared
 *
 * @param bs the bitstring we are talking about
 * @returns FALSE on the first bit being switched on, or TRUE if all are
 * switched off
 */
Boolean bitIsAllCleared(BitString bs)
{
  int i;

  check_magic(bs);

  for (i = 0; i < bs->length; i++) {
    if (bs->bits[i]) {
      return (FALSE);
    }
  }

  return (TRUE);
}

/* ----------------------------------------------------------------------
 * checks whether all bits are set
 *
 * @see bitIsAllCleared
 * @param bs the current bitstring
 * @returns FALSE on the first bit being swithced off, or TRUE of all bits
 * are switched on
 */
Boolean bitIsAllSet(BitString bs)
{
  int i;

  check_magic(bs);

  for (i = 0; i < bs->length - 1; i++) {
    if (~bs->bits[i]) {
      return (FALSE);
    }
  }

  return bs->bits[bs->length - 1] == bs->mask ? TRUE : FALSE;
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
