#!/usr/bin/env perl
# cdg/utils/cda2ps.pl
use warnings;
use strict;

BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0); 
  require 'cdgsetup.pl';
}
use CDG;

use PostScript::Simple;

## ---------------------------------------------------------------------
## Variables
## ---------------------------------------------------------------------
our $mainlevel = 'SYN';

our @words;
our @edges; # modifiee
our @labels;
our @children;
our @x;
our @level;

# variables for distance
my $y_step = 100;
my $treeHeight = 0;

# vars for ...
my $annos;
my $annoEntry;
my $l;
my $id;
my $cursor;
my $filename = shift;
## ---------------------------------------------------------------------


## ---------------------------------------------------------------------
## Calculate the y coordinate for a given timepoint
## ---------------------------------------------------------------------
sub y_coord {
  my $n = shift;
  40 + $y_step * $level[$n];
}

## ---------------------------------------------------------------------
## Get word height
## ---------------------------------------------------------------------
sub wordHeight {
  my $x = shift;
  my $result = 0;
  while($x) {
    $x = $edges[$x];
    $result++;
    if($result > $#edges) {
      warn "cyclical tree!\n";
      return $#edges;
    }
  }
  return $result;
}


## ---------------------------------------------------------------------
## Sibling check
## ---------------------------------------------------------------------
sub siblingCheck {

  my $down = shift;
  my $up = $edges[$down];
  my @l = $down<$up ? ($down..$up-1) : ($up-1..$down);
  for(@l) {
    return 0 if ! defined $edges[$_];
    return 0 if $edges[$_] != $up;
  }
  return 1;
}

## ---------------------------------------------------------------------
## get the values from the annotation file:
## edges, words, labels ...
## ---------------------------------------------------------------------
sub setValues {
  $annoEntry = hashGet($annos, ListStruct_item_get $l);
  $id = AnnoEntryStruct_id_get $annoEntry;
  print "$id\n";

  @words = @edges = @labels = @children = @x = @level = ();
  $cursor = 0;

  # Annotation structures
  for(my $m = AnnoEntryStruct_annos_get $annoEntry;
    $m;
    $m = ListStruct_next_get $m) {

    my $annoStruct = ListStruct_item_get $m;
    my $to = AnnotationStruct_to_get $annoStruct;
    my $from = AnnotationStruct_from_get $annoStruct;
    my $word = AnnotationStruct_word_get $annoStruct;
    my $specs = AnnotationStruct_specs_get $annoStruct;

    # specifications
    for (my $ss = $specs;
         $ss;
         $ss = ListStruct_next_get($ss)) {
      my $s = ListStruct_item_get($ss);

      if (SpecificationStruct_type_get($s) eq $CDG::STDep) {
        my $level = SpecificationStruct_kind_get $s;
        if($mainlevel eq $level) {

          $words[$to] = $word;
          $edges[$to] = SpecificationStruct_position_get $s;
          $labels[$to] = SpecificationStruct_name_get $s;
          $cursor += 10*length($words[$to]);
          $x[$to] = $cursor;
          $cursor += 10*length($words[$to]);
          push @{$children[$edges[$to]]}, $to;

        }
      }

    }
  }
}
    

## ---------------------------------------------------------------------
## Calculate where to put the nodes and labels
## ---------------------------------------------------------------------
sub layoutTree {
    # Calculate the tree height
    for(1..$#edges) {
      my $y = wordHeight($_);
      $treeHeight = $y if $treeHeight < $y;
      $level[$_] = $y;
    }

    # Calculate the height of the nodes
    for(1..$#edges) {
      if(!$children[$_] && !$edges[$_]) { # isolated node
	$level[$_] = 0
      }

      elsif (!$edges[$_]) { # root node
	$level[$_] = $treeHeight
      }

      elsif (!$children[$_]) { # leaf node
	if (siblingCheck($_)) {
	  $level[$_] = 0;
	} else {
	  $level[$_] = $treeHeight - $level[$_]
	}
      }

      else { # inner node
	$level[$_] = $treeHeight - $level[$_] ;
      }
    }
    $level[0] = 0;
}


## ---------------------------------------------------------------------
## Calculate pixel values for the tree
## and draw it as ps.
## ---------------------------------------------------------------------
sub createPS {
    my $scale = 0.8;
    my $p = new PostScript::Simple
    ( units => "bp",
      xsize => $scale * ($cursor+20),
      ysize => $scale * 100 * (1+$treeHeight),
      colour => 1,
      reencode => 'ISOLatin1Encoding',
      eps => 1);
    $p->setfont('Helvetica-Bold-iso', 16);

    for (1..$#edges) {
      my $x = $x[$_] - 5*length($words[$_]);
      my $x1 = $x[$_];
      my $y1 = y_coord($_);
      my $x2 = $edges[$_] ? $x[$edges[$_]] : $x[$_];
      my $y2 = y_coord($edges[$_]);
      my $edge = $_;

      # calculate root edge
      if ($y2 < $y1) {
	$y2 = $y1 + 40;
      }

      # paint word
      $p->text($scale * $x,$scale * 10,$words[$_]);

      # paint node
      $p->circle({filled=>1}, $scale * $x[$_],$scale * $y1,4);

      # paint edge
      $p->line($scale*$x1, $scale*$y1,
	       $scale*$x2, $scale*$y2);

      # paint label
      $p->text($scale* (10 + $x1+0.5*($x2-$x1)),
	       $scale* ($y1+0.5*($y2-$y1)),
	       $labels[$_]);

      # draw node and line for root node
      if (! $edges[$edge] ) {
        my $y3 = $y1 + $y_step/2;
	$p->circle({filled=>1}, $scale * $x[$_],$scale * $y3,4);
	$p->line($scale*$x1,$scale*$y1,
		 $scale*$x1,$scale*$y3);
      }

      # draw vertical line
      # TODO: the line begins in the mid of the node instead of outside
      $p->setcolour(217,217,217); # grey
      $p->line($scale*$x1,$scale*$y1,
               $scale*$x1,$scale*40);
      $p->setcolour("black");     # switch back to black
    }
    
    # write postscript file  
    $p->output("$id.ps") or die "Can't write `$id.ps': $!\n";
}


## ---------------------------------------------------------------------
## draw one cda
## ---------------------------------------------------------------------
sub drawGraphic {

    # Get the values from the annotation
    setValues();

    # Layout tree
    layoutTree();

    # Create postscript graphic
    createPS();
}


## ---------------------------------------------------------------------
## MAIN
## ---------------------------------------------------------------------
sub main {
  die "Usage: $0 foo.cda\n" unless -f $filename;
  
  # Load cda files
  commandEval "load $filename";

  # get the annotations
  $annos = InputStruct_annotations_get($CDG::inputCurrentGrammar);

  # create the graphics for each annotation
  for($l = hashListOfKeys $annos; $l; $l = ListStruct_next_get $l) {

      # create the graphic for one annotation
      drawGraphic();

  }
}


## ---------------------------------------------------------------------
## ---------------------------------------------------------------------
main();
