#!/usr/bin/perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

# parse a text and generate a graph from the cdgp result

use warnings;
use GD;
use Data::Dumper;
use Getopt::Long;

our $debug = 1;
###############################################################################
sub writeDebug {
  print (STDERR "Debug - $_[0]\n") if $debug;
}

###############################################################################
sub getTempFileName {
  my $name = shift;
  $name = "" if ! $name;

  my $temp_dir = -d '/tmp' ? '/tmp' : $ENV{TMPDIR} || $ENV{TEMP};
  my $base_name = sprintf("%s/$name-%d-%d-0000", $temp_dir, $$, time());
  my $count = 0;
  while (-e $base_name && $count < 100) {
    $count++;
    $base_name =~ s/-(\d+)$/"-" . (1 + $1)/e;
  }
  if ($count == 100) {
    return undef;
  } else {
    return $base_name;
  }
}

###############################################################################
sub parse {
  
  my $help = 0;
  GetOptions(
      "syntax|h|help|?" => \$help,
      "o:s" => \$outputFile,
      "solver:s" => \$solver ); 
  $help = 1 if ! @ARGV;

  my $helptext = <<EOT;
text2png [options] 'Parse and graph this sentence, please.'

Options:
  -o	  : specify output png file. default is 'parseResult.png'
  -solver : the solver used for parsing. either 'gls' 'frobbing'.
	    default is 'gls'
Note: 
  The environment variable \$CDGPATH needs to be present!

EOT

  die "$helptext" if $help;
  die "Environment variable 'CDGPATH' not present!" if ! $ENV{CDGPATH};
 
  $outputFile = 'parseResult.png' if ! $outputFile;
  my $cdgPath = $ENV{CDGPATH};
  $solver = "gls" if (! $solver);
  $solver = "frobbing execute=q" if ($solver eq "frobbing");
  my $tokIn = &getTempFileName("tokInput"); # input for the tokenizer
  my $tokOut = &getTempFileName("tokOutput"); #output of the tokenizer
  my $cdgpIn = &getTempFileName("cdgpInput"); # input for cdgp
  my $lexiconFile = "$cdgPath/grammar/deutsch/deutsch.cdg";
  my $cdgpBinary = "$cdgPath/cdgp/cdgp";
  
  my $inputText = join(' ', @ARGV);
  
  open (TOKIN, ">$tokIn");
  print TOKIN $inputText;
  close TOKIN;
  
  # tokenize text
  my $tokenizerPath = "$cdgPath/utils/tokenize.pl";
  my $tokReturn = `$tokenizerPath $tokIn > $tokOut`;
  &writeDebug("calling the tokenizer failed. return code: $tokReturn")if $tokReturn;

# if ($debug) {
#   open (TOKOUT, "<$tokOut");
#   while (<TOKOUT>) {
#     &writeDebug("tokenizerOutput=$_");
#   }
#   close TOKOUT;
# }

  # generate cdgp input
  open (CDGPIN, ">$cdgpIn");
  print CDGPIN <<END;
load $lexiconFile
load $tokOut
set info on
newnet s1
$solver
printparse
quit
END
  close CDGPIN;

  my $cdgpReturn = `$cdgpBinary -q -i $cdgpIn`;

  # get resulting syntax edges out of cdgp output
  my @rawEdges = grep (/[0-9]{3}\s+SYN:\s.*$/, split ("\n", $cdgpReturn));
  
  #get resulting violations out of cdgp output
  $cdgpReturn =~ /Violations: *\n(.*)?$/s;
#   &writeDebug("violations: $1");
  my @rawViolations = ();
  @rawViolations = split("\n", $1) if $1; # there may be no violations at all

  # construct a hash of all edges
  my $levelRegex = qr/([0-9]{3})/;
  my $labelRegex = qr/--([A-Z0-9]*)?-->/;
  my $posRegex = qr/(\(\d+-\d+\))?/;
  my $wordRegex = qr/(.+?)/;

  my $edgeRegex = $levelRegex . '\s*SYN:\s*' . $wordRegex . $posRegex . $labelRegex . $wordRegex . $posRegex . '$';
  our %edges;
  foreach my $rawEdge (@rawEdges) {
    $rawEdge =~ /$edgeRegex/;
    $edges{$1} = { fromWord => $2,
		   fromPos  => $3,
		   label    => $4,
		   toWord   => $5,
		   toPos    => $6};
#    if ($6) {
#      &writeDebug("matched: $2, $3, $4, $5, $6.");
#    } else {
#      &writeDebug("matched: $2, $3, $4, $5, ' ',.");
#    }
  }

  # adjust some fields in the edge has 
  foreach my $ce (keys %edges) {
    # separate word an category
    if (! ($edges{$ce}{fromWord} eq '_')) {
      if ($edges{$ce}{fromWord} =~ /^(.+?)_(.+)$/) {
	$edges{$ce}{fromWord} = $1;
	$edges{$ce}{fromCat} = $2;
      }
    }
    if (! ($edges{$ce}{toWord} eq '_')) {
      if ($edges{$ce}{toWord} =~ /^(.+?)_(.+)$/) {
	$edges{$ce}{toWord} = $1;
	$edges{$ce}{toCat} = $2;
      }
    }

    # only use the second number of the position thing
    $edges{$ce}{fromPos} =~ /\([0-9]+\-([0-9]+)\)/;
    $edges{$ce}{fromPos} = $1;
    if ($edges{$ce}{toPos}) {
      $edges{$ce}{toPos} =~ /\([0-9]+\-([0-9]+)\)/;
      $edges{$ce}{toPos} = $1;
    }
  }
  
  # construct a hash of violations
  # for now we only put violations in it which are concerning on single
  # edge
  my $violationsRegex = qr/([0-9]{3}) : ([0-9]{3})?(?: : )?([0-9.e-]+) : (.*)/;
  our %violations;
  foreach my $rawViolation (@rawViolations) {
    $rawViolation =~ /$violationsRegex/;
    if ($1 && !$2) {
      $violations{$1} = { value => $3,
			  name  => $4};
    }
#    if ($2) {
#      &writeDebug("matched violations: $1, $2, $3, $4.");
#    } else {
#      &writeDebug("matched violations: $1, ' ', $3, $4.");
#    }
  }

  my $result = "Edges:\n" . join ("\n", @rawEdges) . "\n\nViolations:\n" .
    join("\n", @rawViolations);
  
  # remove temporary files
  unlink($tokIn);
  unlink($tokOut);
  unlink($cdgpIn);

  &graph;
 
  return $result;
#  return $cdgpReturn;
}
###############################################################################
# return the level for a text node depending on the number of its children
sub getLevel {
  my $currentNode = $_[0]; # hash key for node to process
  my $textNodes = $_[1]; # reference to a hash of all text nodes
  if (!@{$textNodes->{$currentNode}{children}}) {
    return 0;
  } else {
    &writeDebug("node $currentNode has children: " . @{$textNodes->{$currentNode}{children}});
    my $level = 0;
    foreach my $child (@{$textNodes->{$currentNode}{children}}) {
      $newLevel = &getLevel($child, $textNodes);
      $level = $newLevel + 1 if ($newLevel >= $level);
      &writeDebug("determined level for node $currentNode: $level");
    }
    return $level;
  }
}
###############################################################################
# sort function for text nodes
sub sortNodes {
#  $a =~ /\(([0-9]+)\-[0-9]+\)/;
#  my $aPos = $1;
#  $b =~ /\(([0-9]+)\-[0-9]+\)/;
#  my $bPos = $1;
#
#  return -1 if ($aPos <  $bPos);
#  return  0 if ($aPos == $bPos);
#  return  1 if ($aPos >  $bPos);

  return $a <=> $b;
}
###############################################################################
# returns the key of the highest node of a given %tNodes hash
sub getHighest {
  my (%tNodes) = @_;
  my ($highest) = keys %tNodes; # begin somewhere
  foreach my $node (keys %tNodes) {
    $highest = $node if ($tNodes{$node}{level} > $tNodes{$highest}{level});
  }
  return $highest;
}
###############################################################################
sub graph {
  # generate a hash for all words
  # $word{wordpos} -> { word
  #		        level
  #                     @children }

  my %tNodes;
  foreach my $edge (keys %edges) {
    my $pos = $edges{$edge}{fromPos};
    $tNodes{$pos} = { word => $edges{$edge}->{fromWord},
	             level => 0,
		     children => [ ]};
  }
  
  # do some modifications to the %edge and %tNodes hash
  foreach my $ce (keys %edges) {
    # collect children
    my $pos = $edges{$ce}{toPos};
    push @{$tNodes{$pos}->{children}}, $edges{$ce}{fromPos} if $pos;
    # store parent
    if ($pos) {
      $tNodes{$edges{$ce}{fromPos}}{parent} = $pos;
    } else {
      $tNodes{$edges{$ce}{fromPos}}{parent} = 'NIL';
    }
    # make NIL the same position as its child
    if ($edges{$ce}{toWord} eq 'NIL') {
      $edges{$ce}{toPos} = $edges{$ce}{fromPos};
    }
  }

  # get level for each text node
  foreach my $tNode (keys %tNodes) {
    $tNodes{$tNode}{level} = &getLevel($tNode, \%tNodes);
  }

  # raise levels: we have to raise the level of a node if:
  # - he has a sister with a higher level which is on his right (left) and
  #   their parent is on the right (left) side of her.
  # - it is connected to NIL and there is an edge crossing it.
  

  foreach my $cn (keys %tNodes) {
    if ($tNodes{$cn}{parent} ne 'NIL') { # skip nodes connected to nil first
      # get sister with the highest level which is relevant for raising
      my @sisters = @{$tNodes{$tNodes{$cn}{parent}}->{children}};
      # &writeDebug("Found sisters for Node $cn:" . join(', ', @sisters));
      my $hSis = $cn; # we begin with our selves
      foreach my $sis (@sisters) {
	if ($tNodes{$sis}{level} > $tNodes{$hSis}{level}) {
	  if (($cn < $tNodes{$cn}{parent} && $sis < $tNodes{$cn}{parent}) || 
	    ($cn > $tNodes{$cn}{parent} && $sis > $tNodes{$cn}{parent})) { 
	  $hSis = $sis;
	  }
	}
      }
      if ($hSis != $cn) {
	$tNodes{$cn}{level} = $tNodes{$hSis}{level};
	&writeDebug("raised level of node $cn (" . $tNodes{$cn}{word} . ")"
	  . "auf level " . $tNodes{$hSis}{level});
      }
    } else { # this is a node going to nil
      # first get the level of the highest node in the net
      my $newLevel = $tNodes{getHighest(%tNodes)}{level};
      # then raise the node, if it is necessary
      my $crossing = 0;
      foreach my $ce (keys %edges) {
	my $from = $edges{$ce}{fromPos};
	my $to = $edges{$ce}{toPos};
	if ( ($from < $cn && $cn < $to) ||
	     ($from > $cn && $cn > $to) ) {
	  $crossing = 1;
	  last;
	}
      }
      $tNodes{$cn}{level} = $newLevel if $crossing;
    }
  }

            ######### generate image #########

  my $tFont = gdGiantFont; # used font for text nodes
  my $lFont = gdSmallFont; # used font for labels
  my $tSpacing = 10; # space between text nodes
  my $imMargW = 20; # width of image margin
  my $levelH = 30; # height of each level
  my $gNSize = 4; # size of a graphical node
  
#  my $imW = our $imH = 300; # width and heigth of the image
  
  my $imW = $imMargW;
  my $imH = $imMargW;

  # count positions for all text nodes and compute
  # image height and width
  my $highestLevel = 0;
  foreach my $cn (sort sortNodes keys %tNodes) {
    $tNodes{$cn}{lIndex} = $imW;
    $imW += length ($tNodes{$cn}{word}) * $tFont->width;
    $tNodes{$cn}{rIndex} = $imW;
    $imW += $tSpacing;
    $highestLevel = $tNodes{$cn}{level} if ($tNodes{$cn}{level} > $highestLevel);
    $tNodes{$cn}{middle} = $tNodes{$cn}{lIndex} + int(($tNodes{$cn}{rIndex} - 
      $tNodes{$cn}{lIndex}) / 2);
  }
  $imW += $imMargW;
  $imH += ($highestLevel + 1) * $levelH + $imMargW; # we need to add 1 level to have
						    # enough room for root nodes
  
  
  # create a new image
  my $im = new GD::Image($imW, $imH);

  # allocate some colors
  my $white = $im->colorAllocate(255,255,255);
  my $black = $im->colorAllocate(0,0,0);       
  my $red = $im->colorAllocate(255,0,0);      
  my $blue = $im->colorAllocate(0,0,255);
  my $grey = $im->colorAllocate(200,200,200);

  # set colors for specific elements
  my $tColor = $black; # color for text nodes
  my $eColor = $black; # color for edges
  my $labelColor = $blue; # color for edge labels
  my $vColor = $red; # color for violations

  # draw text line
  foreach my $cn (keys %tNodes) {
    $im->string($tFont, $tNodes{$cn}{lIndex}, $imH - $imMargW, $tNodes{$cn}{word}, $tColor);
  }
  
  # draw graphical nodes for text
  foreach my $cn (keys %tNodes) {
    my $x1 = my $x2 = $tNodes{$cn}{middle};
    my $y1 = $imH - $tFont->height - $imMargW;
    my $y2 = $y1 - $levelH * $tNodes{$cn}{level};
    $im->arc($x1, $y1, $gNSize, $gNSize, 0, 360, $black);
    $im->fill($x1, $y1, $black);
    $im->arc($x2, $y2, $gNSize, $gNSize, 0, 360, $black);
    $im->fill($x2, $y2, $black);
  }

  # draw vertical lines above text nodes if they have children
  foreach my $cn (keys %tNodes) {
    if ($tNodes{$cn}{level} > 0) {
      my $x1 = $tNodes{$cn}{lIndex} + int(($tNodes{$cn}{rIndex} - $tNodes{$cn}{lIndex}) / 2);
      my $y1 = $imH - $tFont->height - $imMargW;
      my $x2 = $x1;
      my $y2 = $y1 - $levelH * $tNodes{$cn}{level};
      $im->line($x1, $y1, $x2, $y2, $grey);
    }
  }

  # draw edges, inner nodes and labels
  foreach my $ce (keys %edges) {
    # we change the color of an edge if it is a violation
    my $c = $violations{$ce} ? $vColor : $eColor;
    my $fromPos = $edges{$ce}{fromPos};
    my $toPos = $edges{$ce}{toPos};
    my $x1 = $tNodes{$fromPos}{middle};
    my $y1 = $imH - $imMargW - $tFont->height - $levelH * $tNodes{$fromPos}{level};
    my $x2 = $tNodes{$toPos}{middle};
    my $y2;
    # raise level of toPos if this edge is going to NIL
    if ($toPos eq $fromPos) {
      $y2 = $imH - $imMargW - $tFont->height - $levelH * ($tNodes{$toPos}{level} + 0.5);
    } else {
      $y2 = $imH - $imMargW - $tFont->height - $levelH * $tNodes{$toPos}{level};
    }	
    # draw node
    $im->arc($x1, $y1, $gNSize, $gNSize, 0, 360, $black);
    $im->fill($x1, $y1, $black);
    $im->arc($x2, $y2, $gNSize, $gNSize, 0, 360, $black);
    $im->fill($x2, $y2, $black);
    # draw line
    $im->line($x1, $y1, $x2, $y2, $c);
    # draw label
    my $xMiddle = $x1 + ($x2 - $x1)/ 2;
    my $yMiddle = $y1 + ($y2 - $y1)/ 2;
    my $label = $edges{$ce}{label};
    my $xLabel; # the upper left corner of the label
    my $yLabel; # ...
    # the position of the label depends on the direction of the edge
    if ($fromPos < $toPos) { # from lower left to upper right
      $xLabel = $xMiddle - ($lFont->width * length($label));
      $yLabel = $yMiddle - $lFont->height;
    } elsif ($fromPos == $toPos) { # vertical edge
      $xLabel = $xMiddle - ($lFont->width * length($label));
      $yLabel = $yMiddle - ($lFont->height / 2);
    } elsif ($fromPos > $toPos) { # from upper left to lower right
      $xLabel = $xMiddle;
      $yLabel = $yMiddle - $lFont->height;
    }  
    $im->string($lFont, $xLabel, $yLabel, $label, $labelColor);
  }

#  foreach my $key (sort keys %tNodes) {
#    &writeDebug(Dumper($tNodes{$key}));
#  }

  # make sure we are writing to a binary stream
  binmode STDOUT;

  # Convert the image to PNG and generate file
  open(IMAGE , ">$outputFile");
  print IMAGE $im->png;
  close IMAGE;
}
###############################################################################
print STDOUT &parse() . "\n";
