# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllLevels - manage CDG grammar levels 
##
##
## \author Michael Daum
# ----------------------------------------------------------------------------
class AllLevels {
  inherit DataBrowser

  
  # public methods
  public method init_data {}
  public method getMainlevelId {}
  public method getLabels {levelId}; ## \type TclString
  public method getAllLevelIds {};
  public method getAllSectionIds {};

  ## replaces getLevel
  public method getCData {id};## \type Level
  
  public method refreshrow {row level}; ## \type TclNumber, Level

  constructor {args} {}; ## \type TclList

  # private methods
  private method _init_data {}
  private method displaybutton_action {}
  private method showbutton_action {}
  private method usebutton_action {}
  private method editbutton_action {}


};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllLevels::constructor {args} {
    set _idColumnIndex 0
    itk_component add displaybutton {
	button $itk_component(childsite).display\
	    -anchor w \
	    -text "Display" \
	    -command [code $this displaybutton_action]
    } { }

    itk_component add showbutton {
	button $itk_component(childsite).show\
	    -anchor w \
	    -text "Show" \
	    -command [code $this showbutton_action]
    } { }

  itk_component add usebutton {
    button $itk_component(childsite).use\
	-anchor w \
	-text "Use" \
	-command [code $this usebutton_action]
  } { }

  itk_component add editbutton {
    button $itk_component(childsite).edit\
	-anchor w \
	-text "Edit" \
	-command [code $this editbutton_action]
  } { }

  # packing
  pack \
      $itk_component(displaybutton) \
      $itk_component(showbutton) \
      $itk_component(usebutton) \
      $itk_component(editbutton) \
      -side left -padx 3 

  # get the initial data
  _init_data
  
  # register the helpmessages
  .cdgmain help sethelpstr \
      $itk_component(table) \
      "Click selects row, Ctrl-Click selects more rows, NW cell selects all" \
      $itk_component(displaybutton) "Print out this level in the shell" \
      $itk_component(showbutton) "(Don't) display this level in constraint nets" \
      $itk_component(usebutton) "(Don't) use this level in future calculations" \
      $itk_component(editbutton) "Edit the declaration of this level" 

  eval itk_initialize $args
}



## ----------------------------------------------------------------------------
## call _init_data if necessary.
## ----------------------------------------------------------------------------
body AllLevels::init_data {} {
  global inputCurrentGrammar

  if {[InputStruct_xcdglevelsToDate_get $inputCurrentGrammar]} {
      return
  }

  .cdgmain busy compute "Init Levels" [code $this _init_data]

}

## ----------------------------------------------------------------------------
## get data from the cdg tool
## ----------------------------------------------------------------------------
body AllLevels::_init_data {} {
  global inputCurrentGrammar

  set table $itk_component(table)
  $table configure -state normal
  $table clear all
  $table delete rows 1 [$table cget -rows]
  $table tag config default
  $table tag config colored -bg gray80 
  $table width 0 8 1 13 2 4 3 4 4 4 5 4 6 4 7 4 8 4
  
  $table setCell row 0,0 { "" "File" "Labels" \
      "Unary" "Binary" "shown" "used" "avg. score" }

  _setCount [listSize [InputStruct_levels_get $inputCurrentGrammar]]
  set r 1
  for {set l [InputStruct_levels_get $inputCurrentGrammar]} \
    {$l != "NULL"} {set l [ListStruct_next_get $l]} {
    set level [ListStruct_item_get $l]
    refreshrow $r $level
    incr r
  }
  
  setSelection
  InputStruct_xcdglevelsToDate_set $inputCurrentGrammar 1

  $table configure -state disabled
}


## ----------------------------------------------------------------------------
## fill a row with the values from a specific level
## ----------------------------------------------------------------------------
body AllLevels::refreshrow {row level} {
  global inputCurrentGrammar

  set table $itk_component(table)

  # level name
  set level_id [LevelStruct_id_get $level]

  # base name of defining file  
  regexp "\[^/\]*$" [LevelStruct_filename_get $level] basename
  
  # nr of labels
  set labels [listSize [LevelStruct_labels_get $level]]

  # nr of unary/binary constraints
  set unaries 0
  set binaries 0
  set scores 0
  set h [InputStruct_constraints_get $inputCurrentGrammar]
  set m [hashListOfKeys $h]
  for {set l $m} { $l != "NULL" } { set l [ListStruct_next_get $l] } {
    set constraint [hashGet $h [ListStruct_item_get $l]]
    set i 0; set found 0
    for { set m [ConstraintStruct_vars_get $constraint] } \
        { $m != "NULL" } \
        { set m [ListStruct_next_get $m] } \
        {
          incr i
          set varinfo [ListStruct_item_get $m]
          set id [VarInfoStruct_levelname_get $varinfo]
          if {$id == $level_id} {set found 1}
        }
    
    if {$found == 1} {
      set score [ConstraintStruct_penalty_get $constraint]
      set scores [expr $scores + $score]
      if {$i == 1} {incr unaries}
      if {$i == 2} {incr binaries}
    }
  }
  listDelete $m

  # display flag
  set shown [LevelStruct_showflag_get $level]
  if { $shown == 0 } { set shown no }
  if { $shown == 1 } { set shown yes }

  # usage flag
  set used [LevelStruct_useflag_get $level]
  if { $used == 0 } { set used no }
  if { $used == 1 } { set used yes }

  # average score
  set avgscore 0

  # count may be uninitialized when this is called
  set nrconstraints [expr $unaries + $binaries]
  if {$nrconstraints>0} {
    set avgscore [format "%4.3f" [expr $scores / $nrconstraints]]
  }

  $itk_component(table) setCell row $row,0 [list \
      $level_id $basename $labels $unaries $binaries $shown $used $avgscore ]
}




## ----------------------------------------------------------------------------
## toggle usage of selected levels.
## ----------------------------------------------------------------------------
body AllLevels::usebutton_action {} {

  if {$_selection == ""} return

  foreach id $_selection { commandEval "uselevel $id" }
  foreach id $_selection { refreshid $id }
}

## ----------------------------------------------------------------------------
## display the level in the shell
## ----------------------------------------------------------------------------
body AllLevels::displaybutton_action {} {
  foreach id $_selection { commandEval "level $id" }
}

## ----------------------------------------------------------------------------
## toggle display of selected levels
## ----------------------------------------------------------------------------
body AllLevels::showbutton_action {} {
  foreach id $_selection { commandEval "showlevel $id" }
  foreach id $_selection { refreshid $id }
}

## ----------------------------------------------------------------------------
## call editor for the declaration
## ----------------------------------------------------------------------------
body AllLevels::editbutton_action {} {
  # avoid flickering
  if {$_selection == ""} return
  
  # build list of filenames
  set target ""
  foreach id $_selection {

    # I suppose most editors grok "+<linenum> file" nowadays...
    set level [findLevel $id]
    set linenum [LevelStruct_lineNo_get $level]
    set filename [LevelStruct_filename_get $level]
    append target "+$linenum $filename "
  }

  # build command line
  regsub -all "%f" "/bin/sh -c \"[.cdgmain cget -editor]\"" $target command

  # execute
  if {[catch {eval exec $command >/dev/null &} errMsg]} {
    ::cmd::Puts "ERROR: $errMsg"
  }

}


## ----------------------------------------------------------------------------
## get all labels of a certain level.
## ----------------------------------------------------------------------------
body AllLevels::getLabels {levelId} {
  global inputCurrentGrammar

  set labels ""

  set levelDecl "NULL"
  for {set l [InputStruct_levels_get $inputCurrentGrammar]} \
      {$l != "NULL"} {set l [ListStruct_next_get $l]} {
    set levelDecl [ListStruct_item_get $l]
    set thisLevelId [LevelStruct_id_get $levelDecl]
    if {$levelId == $thisLevelId} {
      break
    }
  }

  if {$levelDecl != "NULL"} {
    for {set l [LevelStruct_labels_get $levelDecl]} {$l != "NULL"} \
        {set l [ListStruct_next_get $l]} {
      set item [ListStruct_item_get $l]
      lappend labels [pointer2string $item ]
    }
  }

  return $labels
}

## ----------------------------------------------------------------------------
## get the name of the main level.
## ----------------------------------------------------------------------------
body AllLevels::getMainlevelId {} {
  global inputCurrentGrammar

  set levelDecl [inputGetMainlevel $inputCurrentGrammar]
  if {$levelDecl == "NULL"} {
    return ""
  } else {
    return [LevelStruct_id_get $levelDecl]
  }
}

## ----------------------------------------------------------------------------
## get pointer to a level declaration.
## ----------------------------------------------------------------------------
body AllLevels::getCData {id} {
  return [findLevel $id]
}



## ----------------------------------------------------------------------------
## Return list of all levels' names
## ----------------------------------------------------------------------------
body AllLevels::getAllLevelIds {} {
  global inputCurrentGrammar
  for {set l [InputStruct_levels_get $inputCurrentGrammar]} \
      {$l != "NULL"} \
      {set l [ListStruct_next_get $l]} {
        set i [LevelStruct_id_get [ListStruct_item_get $l]]
        lappend result $i
      }
  return $result
}

## ----------------------------------------------------------------------------
## Return list of all sections' names
## ----------------------------------------------------------------------------
body AllLevels::getAllSectionIds {} {
  global inputSections

  for {set l $inputSections} {$l != "NULL"} {set l [ListStruct_next_get $l]} {
    set i [SectionStruct_id_get [ListStruct_item_get $l]]
    lappend result $i
  }
  return $result
}
