# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## InfoSystem - a central Place for info/error message handling & storage
##
## The InfoSystem serves multiple registered Listeners. Each Message posted
## is delivered to all of them using insert(). Additionally, the InfoSystem
## has an internal message storage, so no message gets lost. The storage can
## be read using insertCompleteStorage().
## The InfoSystem itself implements the InfoSystemListener Interface, so multiple
## InfoSystem instances can be chained in a tree-like manner.
##
## \author Bjrn Engelmann (see also AUTHORS and THANKS for more)
##-----------------------------------------------------------------------------
class InfoSystem {
	inherit InfoSystemListener

	constructor {args} {}; ## \type TclList

	## internal message storage
	private variable storage [list]

	## list of all registered listeners
	private variable listeners [list]

	public method insertMessage {message}; ## \type TclString
	public method addListener {listener}; ## \type InfoSystemListener
	public method removeListener {listener}; ## \type InfoSystemListener
	public method clearListeners {}

	public method messageCount {}
	public method insertCompleteStorage {listener}; ## \type InfoSystemListener

}


## -----------------------------------------------------------------------------
## post a Message to the InfoSystem
## the message is
## 1) put into the internal storage
## 2) posted to all listeners that registered with this InfoSystem
## \param message the Message to be posted
## -----------------------------------------------------------------------------
body InfoSystem::insertMessage {message} {
	lappend storage $message
	
	foreach listener $listeners {
		$listener insertMessage $message
	}
	
}

## -----------------------------------------------------------------------------
## register a Listener with the InfoSystem
## as long as registered the Listener will receive any message posted to the InfoSystem
## via his insertMessage - method
## \param listener  the Listener to be registered
## -----------------------------------------------------------------------------
body InfoSystem::addListener {listener} {
	lappend listeners $listener
}

## -----------------------------------------------------------------------------
## unregister a Listener
## the Listener will not recieve any Messages from this InfoSystem no matter
## how much times it was registered before
## \param listener   the Listener to be removed
## -----------------------------------------------------------------------------
body InfoSystem::removeListener {listener} {

	for {set i [expr [llength $listeners] - 1]} {$i > -1} {set i [expr $i-1]} {
		if {[lindex $listeners $i] == $listener} {
			set listeners [lreplace $listeners $i $i]
		}
	}
}

## -----------------------------------------------------------------------------
## unregister all Listeners
## after calling this method no Listener will receive any message from the InfoSystem
## any more.
## -----------------------------------------------------------------------------
body InfoSystem::clearListeners {} {
	set listeners [list]
}

## -----------------------------------------------------------------------------
## how many messages are there in the storage ?
## \returns  the Number of Messages in the internal storage
## -----------------------------------------------------------------------------
body InfoSystem::messageCount {} {
	return [llength $storage]
}

## -----------------------------------------------------------------------------
## send the complete internal storage to a specific Listener
## listener must not neccessarily be registered to this InfoSystem
## it is not registered afterward either
## \param listener  the Listener the messages are to be inserted into
## -----------------------------------------------------------------------------
body InfoSystem::insertCompleteStorage {listener} {

	foreach message $storage {
		$listener insertMessage $message
	}

}

