# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## Tree Editor - the grafical tree editor.
## 
## A VisParses object instanciate one toplevel window that embedds a set
## of ParseTree objects to be displayed. There are two common situations that
## determine the set of parses: (1) displaying annotations, (2) displaying
## parser solutions. In the first case one VisParses object can fetch all
## annotations loaded into the system. In the second case one VisParses object
## displays all solutions a specific parsing flavour has found for one wordgraph.
## So in the latter case another parser cannot reuse the same VisParses object
## to visualize its solution record also, but needs a new VisParses object of
## its own, giving you two separate toplevel windows. I think this is the best
## compromise here between lots of toplevel windows and embedded objects in
## separate tabs of one window.
##
## As you see below we don't provide a separate destructor to destruct
## embedded ParseTree objects but add ParseTree objects directly as itk-components
## to the VisParses widget. So they get destructed automatically.
##
## \author Michael Daum, Kilian A. Foth, Dietmar Fnning 
## (see also AUTHORS and THANKS for more)
##
## $Id visparses.tcl $
## ----------------------------------------------------------------------------
class TreeEditor {
   inherit ::itk::Widget

#   itk_option define -iconname iconName IconName "xcdg"
   itk_option define -interactive interactive Interactive 0

   public method addParse {parse}; ## \type ParseTree
   public method removeParse {parse}; ## \type ParseTree
   public method clear {};
   public method view {no}; ## \type TclNumber
   public method center {}
   public method readWord {}
   public method getBadness {}
   public method getGeometry {}
   public method setGeometry {geometry}; ## \type TclString
   public method update {}
   public method getParent {}   
   public method print {}
   public method loadAction {}
   public method saveAction {}
   public method saveAsAction {}
   public method mirrorAction {}
   public method verifyAction {}
   public method showcycleAction {}
   public method breakcycleAction {}
   public method nextAction {}
   public method previousAction {}
   public method refresh {}
   public method undo {} ; # mapped to edit-menu
   public method redo {} ; # mapped to edit-menu
   public method computeViolas {}
   public method getCurrentParse {}
   public method spaceHandler {}
   public method BackspaceHandler {}
   public method zoomIn {}
   public method zoomOut {}
   public method setOrientation {orient}; ## \type Orientation
   public method getOrientation {}
   public method toggleOrientation {}
   public method updateVisual {}
   public method _switchLevel {x}; ## \type TclNumber
   public method paint {}
   public method reeval {}
   public method redraw {}
   public method chkRedrawGfx {}
   public method getSize {};
   public method toggleEval {};
   public method toggleRedraw {};
   public method togglePaint {};
   public method retry {};
   public method adjust_tabs {};

   constructor {parent args} {}; ## \type TreeEditorParent, TclList


   private variable _noPages 0
   private variable _userInput ""
   # DEBUG
   public variable currentOrient "horizontal"

  ## array mapping Parses to page-numbers
  private variable _pageOfParse  ;## \type TclArray

  # array mapping Parse to ParseTree
  private variable _parse2tree
   
   # the toplevel widget in which
   # the editor is embedded
   private variable _parent
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body TreeEditor::constructor {parent args} {

  set _parent $parent
  
  global queue env
  set queue ""

  # toolbar
  itk_component add toolbar {
    frame $itk_interior.toolbar \
      -borderwidth 0 \
      -relief flat
  } {}
  set toolbar $itk_component(toolbar)

  # prev button
  set allImages [image names]
  if {[lsearch -exact $allImages prevImage] < 0} {
    image create photo prevImage -file $env(XCDG_LIB)/back.gif
  }
  itk_component add prevButton {
    button $itk_component(toolbar).prevButton \
      -image prevImage \
      -command [code $this previousAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(prevButton) <Enter> {%W configure -relief raised}
  bind $itk_component(prevButton) <Leave> {%W configure -relief flat}

  # next button
  if {[lsearch -exact $allImages nextImage] < 0} {
    image create photo nextImage -file $env(XCDG_LIB)/forward.gif
  }
  itk_component add nextButton {
    button $itk_component(toolbar).nextButton \
      -image nextImage \
      -command [code $this nextAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(nextButton) <Enter> {%W configure -relief raised}
  bind $itk_component(nextButton) <Leave> {%W configure -relief flat}

  # save button
  if {[lsearch -exact $allImages saveImage] < 0} {
    image create photo saveImage -file $env(XCDG_LIB)/floppy.gif
  }
  itk_component add saveButton {
    button $itk_component(toolbar).saveButton \
      -image saveImage \
      -command [code $this saveAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(saveButton) <Enter> {%W configure -relief raised}
  bind $itk_component(saveButton) <Leave> {%W configure -relief flat}

  # mirror button
  if {[lsearch -exact $allImages mirrorImage] < 0} {
    image create photo mirrorImage -file $env(XCDG_LIB)/mirror.gif
  }
  itk_component add mirrorButton {
    button $itk_component(toolbar).mirrorButton \
      -image mirrorImage \
      -command [code $this mirrorAction] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(mirrorButton) <Enter> {%W configure -relief raised}
  bind $itk_component(mirrorButton) <Leave> {%W configure -relief flat}

  # zoom in button
  if {[lsearch -exact $allImages zoomInImage] < 0} {
    image create photo zoomInImage -file $env(XCDG_LIB)/zoom_in.gif
  }
  itk_component add zoomInButton {
    button $itk_component(toolbar).zoomInButton \
      -image zoomInImage \
      -command [code $this zoomIn] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(zoomInButton) <Enter> {%W configure -relief raised}
  bind $itk_component(zoomInButton) <Leave> {%W configure -relief flat}

  # zoom out button
  if {[lsearch -exact $allImages zoomOutImage] < 0} {
    image create photo zoomOutImage -file $env(XCDG_LIB)/zoom_out.gif
  }
  itk_component add zoomOutButton {
    button $itk_component(toolbar).zoomOutButton \
      -image zoomOutImage \
      -command [code $this zoomOut] \
      -borderwidth 0 \
      -relief flat \
      -activebackground gray90 \
  } {}
  bind $itk_component(zoomOutButton) <Enter> {%W configure -relief raised}
  bind $itk_component(zoomOutButton) <Leave> {%W configure -relief flat}

  # redraw delay button
  if {[lsearch -exact $allImages delayRedrawActiveImage] < 0} {
    image create photo delayRedrawActiveImage -file $env(XCDG_LIB)/redraw-active.gif
  }
  if {[lsearch -exact $allImages delayRedrawInactiveImage] < 0} {
    image create photo delayRedrawInactiveImage -file $env(XCDG_LIB)/redraw-inactive.gif
  }
  itk_component add redrawDelayButton {
    button $itk_component(toolbar).redrawDelayButton \
      -image delayRedrawActiveImage \
      -relief flat \
      -command [code $this toggleRedraw] \
      -borderwidth 0 \
      -activebackground gray90 \
  } {}
  bind $itk_component(redrawDelayButton) <Enter> {%W configure -relief raised}
  bind $itk_component(redrawDelayButton) <Leave> {%W configure -relief flat}

  # paint button
  if {[lsearch -exact $allImages paintImage] < 0} {
    image create photo paintImage -file $env(XCDG_LIB)/paint.gif
  }
  itk_component add paintButton {
    button $itk_component(toolbar).paintButton \
      -image paintImage \
      -relief flat \
      -command [code $this togglePaint] \
      -borderwidth 0 \
      -activebackground gray90 \
  } {}
  bind $itk_component(paintButton) <Enter> {%W configure -relief raised}
  bind $itk_component(paintButton) <Leave> {%W configure -relief flat}

  # retry button
  if {[lsearch -exact $allImages retryImage] < 0} {
    image create photo retryImage -file $env(XCDG_LIB)/retry.gif
  }
  itk_component add retryButton {
    button $itk_component(toolbar).retryButton \
      -image retryImage \
      -relief flat \
      -command [code $this retry] \
      -borderwidth 0 \
      -activebackground gray90 \
  } {}
  bind $itk_component(retryButton) <Enter> {%W configure -relief raised}
  bind $itk_component(retryButton) <Leave> {%W configure -relief flat}
  
  # id label 
  itk_component add idLabel {
    label $itk_component(toolbar).idLabel \
      -text "Parse " \
  } {}

  # score label
  itk_component add scoreLabel {
    label $itk_component(toolbar).scoreLabel \
      -text "Score: 00/00/0.0" \
  } {}
    
  # the tabnotebook containing the ParseTree objects
  itk_component add tbn {
    iwidgets::tabnotebook $itk_interior.tbn -auto true -equaltabs true
  } {
    keep -width -height
  }

  # text box for parsing of user input 
  itk_component add textbox {
     iwidgets::entryfield $itk_interior.entry \
         -textvariable [scope _userInput] \
         -labeltext "Type here:"         \
         -labelpos w                     \
       }
  set box [$itk_component(textbox) component entry]

  # Take control of the textbox
  bind $box <space> [code "$this spaceHandler; break"]
  bind $box <BackSpace> [code "$this BackspaceHandler; break"]

  set top_hull [$_parent component hull]
  bind $top_hull <Shift-Tab> [code $this _switchLevel -1]
  bind $top_hull <Tab> [code $this _switchLevel 1]
  bind $top_hull <q> [list destroy $top_hull]
  bind $top_hull <c> [code $this showcycleAction]
  bind $top_hull <m> [code $this mirrorAction]
  bind $top_hull <v> [code $this verifyAction]
  bind $top_hull <n> [code $this nextAction]
  bind $top_hull <p> [code $this previousAction]
  bind $top_hull <s> [code $this saveAction]
  bind $top_hull <KP_Enter> [code $this toggleOrientation]
  bind $top_hull <KP_Add> [code $this zoomIn]
  bind $top_hull <KP_Subtract> [code $this zoomOut]
  bind $top_hull <Control-z> [code $this undo]
  bind $top_hull <Control-r> [code $this redo]

   # packing
#  pack $itk_component(menuframe) -fill x -side top
  pack $itk_component(toolbar) -fill x -side top -padx 5
  pack $itk_component(prevButton) \
       $itk_component(nextButton) \
       $itk_component(saveButton) \
       $itk_component(mirrorButton) \
       $itk_component(zoomInButton) \
       $itk_component(zoomOutButton) \
       $itk_component(redrawDelayButton) \
       $itk_component(paintButton) \
       $itk_component(retryButton) \
       -side left -padx 3 
  pack $itk_component(scoreLabel) -side right -padx 3
  pack $itk_component(idLabel) -side right -padx 3

#   pack $itk_component(filebutton) \
#         $itk_component(editbutton) \
#        	$itk_component(viewbutton) \
#        	$itk_component(parsebutton) \
#        	-side left -padx 5 -pady 5
  pack $itk_component(tbn) -fill both -side top -expand 1



  eval itk_initialize $args

  if {$itk_option(-interactive)} {
    pack $itk_component(textbox) -fill both -side top
  }
   
}

## ----------------------------------------------------------------------------
## return the geometry of my parent
## ----------------------------------------------------------------------------
body TreeEditor::getGeometry {} {
  return [$_parent getGeometry]
}

## ----------------------------------------------------------------------------
## set the geometry of my parent
## ----------------------------------------------------------------------------
body TreeEditor::setGeometry {geometry} {
  $_parent setGeometry $geometry
}

## ----------------------------------------------------------------------------
## add a parse on a separate page.
## returns the newly constructed ParseTree
## ----------------------------------------------------------------------------
body TreeEditor::addParse {parse} {

   # check whether this Parse is already visible
   set index [lsearch -exact [array names _parse2tree] $parse]
   if {-1 != $index} {
     return $_parse2tree($parse)
   }

   incr _noPages
   set page [$itk_component(tbn) add \
    -label "$_noPages" \
    -command [code $this updateVisual]]

   itk_component add parseTree$_noPages {
      ParseTree $page.parseTree$_noPages $parse $this
   } {
      keep -background -cursor
   }
   set result $itk_component(parseTree$_noPages)
   pack $result -fill both -side top -expand 1

   $result addUndoListener $this ;# register for updating menu buttons
   set _pageOfParse($parse) $_noPages
   set _parse2tree($parse) $result
   
   return $result
}

## ----------------------------------------------------------------------------
## make a ParseTree invisible.
## This does not destroy the underlying Parse.
## ----------------------------------------------------------------------------
body TreeEditor::removeParse {parse} {

   if {![info exists _pageOfParse($parse)]} {
      error "parse $parse is not visualized here"
   }
   set page $_pageOfParse($parse)
   unset _pageOfParse($parse)
   unset _parse2tree($parse)

   set parsetree $itk_component(parseTree$page)
   destroy $parsetree

   # since the normal label search operations do not work properly, search yourself
   set index end
   for {set i 0} {$i < [llength [$itk_component(tbn) childsite]]} {incr i} {
   	set label [$itk_component(tbn) pageconfigure $i -label]
	if { $page == [lindex $label end] } {
		set index $i
		break
	}
   }

   $itk_component(tbn) delete $index
}

## ----------------------------------------------------------------------------
## TODO docu
## ----------------------------------------------------------------------------
body TreeEditor::clear {} {
  foreach parse [array names _pageOfParse] {
    removeParse $parse
  }
  
  set _noPages 0
}


## ----------------------------------------------------------------------------
## delegate the view command to the \c tbn component
## ----------------------------------------------------------------------------
body TreeEditor::view {no} {
   $itk_component(tbn) view $no
}

## ----------------------------------------------------------------------------
## returns the parsetree object that is currently selected
## ----------------------------------------------------------------------------
body TreeEditor::getCurrentParse {} {

  set index [$itk_component(tbn) index select]
  if {$index < 0} {
    return ""
  }

  incr index
  return $itk_component(parseTree$index)
}

## ----------------------------------------------------------------------------
## print the visible level.
## ----------------------------------------------------------------------------
body TreeEditor::print {} {

  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  
  set fileName [tk_getSaveFile \
	    -parent $itk_component(hull) \
            -title "Print to file" \
            -defaultextension .ps \
            -initialfile [$parse parse getId]]

  regsub ".ps$" $fileName "" fileName
  set fileName "$fileName.ps"
  
  set mainlevel [$parse mainlevelname]
  set synCanvas [[$parse component $mainlevel] component canvas]

  scan "[$synCanvas bbox all]" "%f %f %f %f" x1 y1 x2 y2

  set width [expr $x2 - $x1]
  set height [expr $y2 - $y1]
  $synCanvas postscript -file $fileName -width $width -height $height \
    -pagewidth 28c -pageheight 20c -rotate 1 -pageanchor nw -pagex 1c -pagey 1c

  ::cmd::Puts "INFO: written postscript to file '$fileName'\n";
}

## ----------------------------------------------------------------------------
## load a parse directly into the editor.
## \todo this is not implelented yet here, but in AllParses
## ----------------------------------------------------------------------------
body TreeEditor::loadAction {} {
   .cdgmain not_yet [$_parent component hull]
}

## ----------------------------------------------------------------------------
## writes the tree to file as an annotation.
## \todo The name of the file to write to is determined by parsing the
## \c -title property. Actually this should be done better by paying a round
## of new properties.
## ----------------------------------------------------------------------------
body TreeEditor::saveAction {} {

  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse writeToFile [$parse parse getId]
}

## ----------------------------------------------------------------------------
## writes the tree to file under a new name.
## ----------------------------------------------------------------------------
body TreeEditor::saveAsAction {} {

  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  set id [tk_getSaveFile \
	    -parent $itk_component(hull) \
            -title "Save as new file" \
            -defaultextension .cda \
            -initialfile [$parse parse getId]]

  regsub ".cda$" $id "" id

  if {$id != ""} {  
    $parse writeToFile $id 
  }
}


## ----------------------------------------------------------------------------
## call parseMirror() eventually.
## ----------------------------------------------------------------------------
body TreeEditor::mirrorAction {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse mirror
}

## ----------------------------------------------------------------------------
## visual version of the `verify' command
## ----------------------------------------------------------------------------
body TreeEditor::verifyAction {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse verify
}

## ----------------------------------------------------------------------------
## Switch to next or previous level in the parsetree.
## ----------------------------------------------------------------------------
body TreeEditor::_switchLevel {x} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }

  if {1 == $x} {
    $parse forwardLevel
  } elseif {-1 == $x} {
    $parse backwardLevel
  }
}

## ----------------------------------------------------------------------------
## Display cycles on a level in red ink.
## ----------------------------------------------------------------------------
body TreeEditor::showcycleAction {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse showcycles
}

## ----------------------------------------------------------------------------
## Break cycles in a parse.
## ----------------------------------------------------------------------------
body TreeEditor::breakcycleAction {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse breakcycles
}

## ----------------------------------------------------------------------------
## Ensure the selected tab is visible.
## ----------------------------------------------------------------------------
body TreeEditor::adjust_tabs {} {

  set tbn $itk_component(tbn)
  set tabs [$tbn component tabset]
  set start [lindex [$tabs configure -start] 4]
  set width [lindex [$tabs configure -width] 4]
  set left [lindex [$tabs tabconfigure select -left] 2]

  # This code is only executed when there actually is more than one
  # tab, so we can fearlessly assume that there *is* a second tab.
  set first [lindex [$tabs tabconfigure 0 -left] 2]
  set second [lindex [$tabs tabconfigure 1 -left] 2]
  set tabwidth [expr $second - $first]
  
  if {$left >= [expr $width - $tabwidth] || $left <= $tabwidth} {
    $tbn configure -start [expr -$tabwidth * [$tabs index select]]
  }

}


## ----------------------------------------------------------------------------
## switch to the next tab in the tree editor.
## Switching to a next tab means that we select the next tree in a set of 
## trees loaded into the editor. In case of solution parses we simply
## switch to the next parse in the editor. In case of annotation parses
## we switch to the next annotation or generate new annotation parses for the
## next lattice in the set of annotated lattices.
## ----------------------------------------------------------------------------
body TreeEditor::nextAction {} {

  set parse [getCurrentParse]  
  if {$parse == ""} {
    return
  }
  set latticeId [$parse parse getLatticeId]

  if {[$_parent getAutosave]} {
    saveAction
  }

  set tbn $itk_component(tbn)
  set nextLatticeId ""

  if {[$tbn index select] == [$tbn index end] && $_parent != "" && \
      $latticeId != ""} {
    set nextLatticeId [.cdgmain wordgraphs getNext $latticeId] 
    if {$nextLatticeId == ""} {
      return
    }
    set lat [findLattice $nextLatticeId]
    if {$lat == "NULL"} {
      ::cmd::Puts "ERROR: no lattice `$nextLatticeId' found."
      return
    }
    for {set l [findAnnotations $lat 1]} {$l != "NULL" } {set l [ListStruct_next_get $l]} {
      set id [AnnoEntryStruct_id_get [ListStruct_item_get $l]]
      set parse [.cdgmain parses getCData $id]
      addParse $parse
    }
  }
    

  catch {$tbn next}
  adjust_tabs

}



## ----------------------------------------------------------------------------
## display the previous annotation.
## ----------------------------------------------------------------------------
body TreeEditor::previousAction {} {
    catch {$itk_component(tbn) prev}
  adjust_tabs
}


## ----------------------------------------------------------------------------
## compute the violations of the selected ParseTree.
## ----------------------------------------------------------------------------
body TreeEditor::computeViolas {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse computeViolas
}


## ----------------------------------------------------------------------------
## undo the last action of the selected page.
## ----------------------------------------------------------------------------
body TreeEditor::undo {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse undo
}


## ----------------------------------------------------------------------------
## undo the last action of the selected page.
## ----------------------------------------------------------------------------
body TreeEditor::redo {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse redo
}

## ----------------------------------------------------------------------------
## refresh the display of the selected VisParses.
## ----------------------------------------------------------------------------
body TreeEditor::refresh {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse drawAll
  getBadness
}


## ----------------------------------------------------------------------------
## center on the middle of the application.
## ----------------------------------------------------------------------------
body TreeEditor::center {} {
   scan [wm geometry .] "%dx%d+%d+%d" appwd appht appx appy
   set wd [expr $appwd  - 80]
   set ht [expr $appht - 80 ]
   set x [expr ($appwd - $wd) / 2 + $appx]
   set y [expr ($appht - $ht)  / 2 + $appy]
   setGeometry "$wd\x$ht+$x+$y"
}

## ----------------------------------------------------------------------------
## gets another word of typed input.
## Hangs until the word is supplied.
## ----------------------------------------------------------------------------
body TreeEditor::readWord {} {
  global queue

  # busy waiting for next input
  if {$queue == ""} {
    vwait queue
  }

  set word [lindex $queue 0]
  set queue [lreplace $queue 0 0]

  return $word
}

## ----------------------------------------------------------------------------
## invoked by space in the textbox.
## ----------------------------------------------------------------------------
body TreeEditor::spaceHandler {} {

  global queue
  set box $itk_component(textbox)

  # Spaces are very special, since they trigger actions in the
  # background. But that only happens if the space
  # comes right after a word, not after another space.
  if {[regexp {[ 	]$} $_userInput]} return

  # And space at the beginning of a line doesn't make sense.
  if {$_userInput == ""} return

  # get last space-delimited word from the text
  if {![regexp {([^ ]+)[ 	]*$} $_userInput foo word]} {
    #puts "WARNING: `$_userInput' doesn't match the regexp."
  }

  # deliver word to the queue
  #puts "DEBUG: queueing `$word'"
  lappend queue $word
  
  # finally insert the space
  $box insert end " "

}

## ----------------------------------------------------------------------------
## invoked by Backspace in the textbox.
## ----------------------------------------------------------------------------
body TreeEditor::BackspaceHandler {} {

  global queue
  set box $itk_component(textbox)

  # backspace on an empty line is inoperable.
  if {$_userInput == ""} return

  # If the last space after a word is erased, the previous word is
  # withdrawn from the parse (but it remains in the textbox for
  # further editing).
  if {[regexp {[^ ] $} $_userInput]} {

    # if the queue has a word in it (not a take-back request),
    # we simply remove the last element (so strictly speaking, it
    # isn't really a queue...)

    if {$queue != "" && ![regexp {\$\#\$} $queue]} {
      #puts "DEBUG: retracting `[lindex $queue end]'"
      set nextToLast [expr [llength $queue] - 2]
      set queue [lrange $queue 0 $nextToLast]
    } else {

      # a special symbol is queued to denote `take back a word'
      set word "$\#$"
      #puts "DEBUG: queueing `$word'"
      lappend queue $word

      # and we also switch to the previous tab in the notebook
      set tbn $itk_component(tbn) 
      $tbn configure -auto true
      set previoustab [expr [$tbn view] -1]
      if {$previoustab >= 0} {
        $tbn view $previoustab
      }
    }

  }

  # process the backspace
  $box delete [expr [string length $_userInput] - 1]

}

## ----------------------------------------------------------------------------
## Display all canvases of the selected parse smaller.
## ----------------------------------------------------------------------------
body TreeEditor::zoomIn {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse zoom 1
}

## ----------------------------------------------------------------------------
## display all canvases of the selected parse bigger.
## ----------------------------------------------------------------------------
body TreeEditor::zoomOut {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse zoom -1
}

## ----------------------------------------------------------------------------
## set the orientation of the split of the current parsetree.
## ----------------------------------------------------------------------------
body TreeEditor::setOrientation {orient} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  
  if {$orient != ""} { 
    set currentOrient $orient 
  }
  $parse setOrientation $currentOrient
}

## ----------------------------------------------------------------------------
## get the orientation of the split of the current parsetree.
## ----------------------------------------------------------------------------
body TreeEditor::getOrientation {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  return [$parse getOrientation]
}

## ----------------------------------------------------------------------------
## Toggle the orientation of the split of the current parsetree.
## ----------------------------------------------------------------------------
body TreeEditor::toggleOrientation {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  set current [getOrientation] 
  if {"horizontal" == $current} {
    $parse setOrientation "vertical"
  } else {
    $parse setOrientation "horizontal"
  }
}


## ----------------------------------------------------------------------------
## get the score of current parsetree
## ----------------------------------------------------------------------------
body TreeEditor::getBadness {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  if {[scan [$parse parse getBadness] "%d %d %f" hard soft score] > 0} {
    set badnessString [format "Score: %d/%d/%4.3e" $hard $soft $score]
  } else {
    set badnessString "Score: n.a."
  }
  $itk_component(scoreLabel) configure -text $badnessString

  $itk_component(idLabel) configure -text "Parse: [$parse parse getId]"
}

## ----------------------------------------------------------------------------
## update the visual parts which depends on the selected parsetree
## ----------------------------------------------------------------------------
body TreeEditor::updateVisual {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  set method [$parse parse getSearchStrategy]
  set lattice [$parse parse getLatticeId]
  set id [$parse parse getId]
  
  # If the id is machine-generated, it just says stuff like `parse0'.
  # That is uninteresting, and is already expressed by the tab
  # numbers. But the distinction between `s1' and `auto-s1' is
  # informative, and should be shown.
  if {[regexp {^parse[0-9]+$} $id]} {
    set id $lattice
  }

  if {$_parent != ""} {
  	$_parent setTitle "$method:$id"
  }
  if {$lattice != ""} {
      ## Correcting the focus in the wordgraph table
   .cdgmain wordgraphs setSelection $lattice
  }

  getOrientation
  getBadness
}

## ----------------------------------------------------------------------------
## TODO: docu
## ----------------------------------------------------------------------------
body TreeEditor::update {} {

  if {$_parent == ""} {
    return
  }

  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }

  $_parent setUndo [$parse canUndo]
  $_parent setRedo [$parse canRedo]
}

## ----------------------------------------------------------------------------
## returns my parent component this treeeditor is embeded in
## ----------------------------------------------------------------------------
body TreeEditor::getParent {} {
  return $_parent
}   

## ----------------------------------------------------------------------------
## Notify parsetrees to paint their trees
## ----------------------------------------------------------------------------
body TreeEditor::paint {} {

  set i 1
  while {$i <= $_noPages} {
    set parsetree $itk_component(parseTree$i)
    if {$parsetree != ""} {
      $parsetree paintTabs
    }
    incr i
  }
}

## ----------------------------------------------------------------------------
## Notify current ParseTree to redraw
## ----------------------------------------------------------------------------
body TreeEditor::redraw {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse drawAll
}

## ----------------------------------------------------------------------------
## Check which graphic file you have to use for the auto redraw Button
## ----------------------------------------------------------------------------
body TreeEditor::chkRedrawGfx {} {
  if {[$_parent getRedraw] == 1} {
  	$itk_component(redrawDelayButton) configure -image delayRedrawActiveImage
  } else {
  	$itk_component(redrawDelayButton) configure -image delayRedrawInactiveImage
  }
}

## ----------------------------------------------------------------------------
## Notify current ParseTree to eval
## ----------------------------------------------------------------------------
body TreeEditor::reeval {} {
  set parse [getCurrentParse]
  if {$parse == ""} {
    return
  }
  $parse evalParse
}


## ----------------------------------------------------------------------------
## returns how many pages are present
## ----------------------------------------------------------------------------
body TreeEditor::getSize {} {
  return [$itk_component(tbn) index end]
}   

## ----------------------------------------------------------------------------
## Toggle automatic redrawing.
## ----------------------------------------------------------------------------
body TreeEditor::toggleRedraw {} {
  set x [$_parent getRedraw]
  $_parent setRedraw [expr 1 - $x]
}

## ----------------------------------------------------------------------------
## Toggle automatic evaluation.
## ----------------------------------------------------------------------------
body TreeEditor::toggleEval {} {
  set x [$_parent getEval]
  $_parent setEval [expr 1 - $x]
}

## ----------------------------------------------------------------------------
## Toggle tree painting.
## ----------------------------------------------------------------------------
body TreeEditor::togglePaint {} {
  set x [$_parent getPaintMode]
  $_parent setPaintMode [expr 1 - $x]
}

## ----------------------------------------------------------------------------
## Restart the parser while keeping edits that the user has made.
## ----------------------------------------------------------------------------
body TreeEditor::retry {} {
  set cur [getCurrentParse]
  if {$cur != ""} {
    $cur retry
  }
}


