# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## CdgProcess - interface for a cdgp process.
## This class wraps arround a cdgp command spawned in the background and tries
## to provide all you need to handle a cdgp job. The class communicates with
## the cdgp job connecting to its STDIN and reads back all its output on STDOUT.
## \ingroup YadaScheduler
##
## \author Michael Daum 
##
## ----------------------------------------------------------------------------
class CdgProcess {
  inherit Process

  # variables ----------------------------------------------------------------

  ## the file where to store the xml output of the logfile
  public variable xmlFile	""
  
  ## if you like to call cdgp with an init file (-i commandline option), put
  ## that filename in here.
  public variable initFile	""

  ## binary location.
  ## This variable stores the absolute path to the cdgp binary that we wrap arround.
  public variable cdgp		""

  ## name of the job which is told
  public variable job		""

  constructor {args} {}; ## \type TclList
  destructor {}


  # methods ------------------------------------------------------------------
  public method start {{command ""}}; ## \type TclString
  public method stop {}

};

## ----------------------------------------------------------------------------
## constructor.
## This constructor builds a cdgp wrapper object. If you don't specify a
## cdgp binary yourself (with -cdgp "/here/it/is/cdgp"), we assume the binary
## to be in \c "$env(YADA_TMP)/cdgp-[pid]".
## ----------------------------------------------------------------------------
body CdgProcess::constructor {args} {
  global env

  eval configure $args

  # did we already provide a command or not
  if {$cdgp == ""} {
    set cdgp $env(YADA_CDGP)
  }
}

## ----------------------------------------------------------------------------
## destructor.
## ----------------------------------------------------------------------------
body CdgProcess::destructor {} {
}

## ----------------------------------------------------------------------------
## start the cdgp process.
## This forks the cdgp in the background, which is actually done in Process::start().
## \param command this lets you overwrite the command given in CdgProcess::cdgp.
## ----------------------------------------------------------------------------
body CdgProcess::start {{command ""}} {

  # prepair the command line
  if {$command == ""} {
    if {![file exists $cdgp]} {
      return -1;
    }
    set command "$cdgp"
  }
  
  if {$xmlFile != ""} {
    # strip off gz 
    regsub {^(.*)\.gz$} $xmlFile {\1} clearXmlFile
    append command " -x $clearXmlFile"
  } 

  if {$initFile == ""} {
    append command " -i none"
  } else {
    append command " -i $initFile"
  }

  # start the process
  set pid [chain $command]
  
  return $pid
}

## ----------------------------------------------------------------------------
## stop the cdgp process.
## This method stops a started cdgp process and gzips the xmllogfile. Afterwards
## Process::stop() is called to finalize everything. There  the job scheduled
## to be \a consumed.
## ----------------------------------------------------------------------------
body CdgProcess::stop {} {
  # compress xmlFile
  if {$xmlFile != ""} {
    regsub {^(.*)\.gz$} $xmlFile {\1} clearXmlFile
    if {[file exists $clearXmlFile]} {
      if {[catch {exec gzip -f $clearXmlFile} errMsg]} {
	print "ERROR: $errMsg\n$::errorInfo\n"
      } else {

	# wait until the file is zipped
	set timeout 10
        while {$timeout > 0 && ! [file exists $xmlFile]} {
	  incr timeout -1
	  after 1000
	}

        # inform about zipped file
	if {[file exists $xmlFile]} {
	  _print "\nfinished $job\n"
	}
        # else ...
      }
    }
  } 
  
  # Process::stop and consume
  chain
}
