#!/bin/sh
#\
exec tclsh "$0" "$@" 

package require Itcl
namespace import -force itcl::*

# ----------------------------------------------------------------------------
# cdg-server
# ----------------------------------------------------------------------------
class CdgServer {

  # variables ----------------------------------------------------------------
  public variable cdgpBinary $env(YADA_CDGP)
  public variable yadadir $env(YADA_LIB)
  public variable timeout 500

  private variable processCount 0 
  private variable allProcesses
  private variable mostCurrentCdgProcess ""
  private variable NEWregex {^NEW ([^\s]*)(?: ([^\s]*) ([^\s]*) (.*))?$}
  private variable TELLregex {TELL(?:\s+([^\s]*))?\s+(.*)$}
  private variable KILLregex {^KILL\s+(?:(\d\d?)\s+)?(.*)$}
  private variable newRegex 

  # methods ------------------------------------------------------------------
  constructor {args} {}; ## \type TclList
  
  public method stopEvent {cdgProcess}
  public method getProcess {process}
  public method handle_NEW {{name ""} {xmlFile ""} {logFile ""} {cdgp ""}}
  public method handle_TELL {{process ""} input_for_cdgp}
  public method handle_KILL {process signo}
  public method handle_STATUS {}
  public method handle_QUIT {}
  public method eventLoop {}
}

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body CdgServer::constructor {args} {
  source $yadadir/Process.tcl
  source $yadadir/CdgProcess.tcl
  eval configure $args
}

## ---------------------------------------------------------------------------
## event if a cdg process stops 
## ---------------------------------------------------------------------------
body CdgServer::stopEvent {cdgProcess} {
  set pid [eval $cdgProcess getPid]
  foreach {internal_pid external_pid} [array get allProcesses] {
    if {$external_pid == $pid} {
      # delete the stopped cdg process from the table
      array unset allProcesses $internal_pid
      if {$internal_pid == $mostCurrentCdgProcess} {
        # not nice, there should be a list or info that most current is deleted: 
        set mostCurrentCdgProcess "" ;
      }
    }
  }
}

## ---------------------------------------------------------------------------
## check whether the given process name is a process or
## get most current process if specified process-name is an empty string
## \returns a process or an empty string if getProcess failed
## ---------------------------------------------------------------------------
body CdgServer::getProcess {process} {

  # get most current process if no process is given
  if {$process == ""} {
    if {$mostCurrentCdgProcess == ""} {
      puts "no cdg-processes available or most current process not available any longer"
    } else {
      set process $mostCurrentCdgProcess
    }
  }

  # check wether process is available
  if {$process != ""} {
    if {[array names allProcesses -exact $process] == ""} { 
      puts "specified process '$process' not available"
      set process ""
    }
  }

  return $process
}

## ---------------------------------------------------------------------------
## handle server command NEW: start a cdgp process
## ---------------------------------------------------------------------------
body CdgServer::handle_NEW {{name ""} {xmlFile ""} {logFile ""} {cdgp ""}} {
  if {$cdgp != ""} {
    set cdgp $cdgpBinary
  }
  
  # check the name of the cdg-process
  if {$name == ""} {
    set name $processCount
  }

  foreach key [array names allProcesses] {
    if {$key == $name} {
      puts "Process with name `$name' already exists."
      return
    }
  }

  # start the cdg-process
  set cdgProcess [CdgProcess $name \
    -cdgp $cdgp \
    -initFile "none" \
    -xmlFile $xmlFile \
    -logFile $logFile \
    -showProgress 1 \
    -sync 1 \
    -stopCallback [code $this stopEvent] \
    -job $name \
  ]
  
  set pid [$cdgProcess start $cdgpBinary]
  if {$pid <= 0} {
    puts "ERROR: cannot start cdgp process '$cdgpBinary'"
    continue
  }

  # report the process
  incr processCount
  set mostCurrentCdgProcess $name
  set allProcesses($name) $pid
  
  puts "New cdgp-process started."
  puts "server-internal Pid: $name"
  puts "external Pid:        $pid"

  if {$name != ""} {
    puts "Started job $name."
  }

  return $name
}

## ---------------------------------------------------------------------------
## handle server command TELL:
## tell a cdg-process a command
## if specified process or most current process is available
## (leave out process is only possible if command consists of merely one word)
## ---------------------------------------------------------------------------
body CdgServer::handle_TELL {{process ""} input} {
  set process [getProcess $process]
  if {$process == ""} {
    return
  }

puts "telling process '$process': $input"
  if {[catch {$process tell $input} errMsg]} {
    puts "ERROR: $errMsg"
  }
}

## ---------------------------------------------------------------------------
## handle server command KILL: kill a cdg-process with signal $signo
## ---------------------------------------------------------------------------
body CdgServer::handle_KILL {process signo} {
  set process [getProcess $process]
  if {$process == ""} {
    return
  }

  set pid $allProcesses($process)  
  if {[catch {exec kill \-$signo $pid} errMsg]} {
    puts "ERROR: $errMsg"
  } else {
    puts "Cdg-process '$process' is killed with signal $signo."
  }
}

## ---------------------------------------------------------------------------
## handle STATUS: print all processes with name and id
## ----------------------------------------------------------------------------
body CdgServer::handle_STATUS {} {
  puts "ID on Server\tID"
  foreach {intern extern} [array get allProcesses] {
    puts "$intern\t\t$extern"
  }
}

## ---------------------------------------------------------------------------
## handle server command QUIT: quit the cdg-server
## ---------------------------------------------------------------------------
body CdgServer::handle_QUIT {} {
  foreach process [array names allProcesses] {
  if {[catch {"$process" tell "quit"} errMsg]} {
      puts "ERROR: $errMsg"
    }
  }
  array unset allProcesses
}


## ---------------------------------------------------------------------------
## MAIN
## ---------------------------------------------------------------------------
body CdgServer::eventLoop {} {

    fconfigure stdin -blocking 0

  # read input line permanentely
  while {1} {
    set task [gets stdin]
    if {$timeout < 0} { handle_QUIT; break }

    if {$task == ""} {
       after 1000; incr timeout -1
       update
       continue
    }

    # execute commands of the server
    switch -regexp -- $task \
      ^H$ - \
      ^HELP$ {
	puts "available commands:
	H or HELP:		display this help
	NEW:			start a cdgp-process
	TELL \[process\] cmd:	tell the command `cmd' to an existing cdg process
			        process default: most current process (if `cmd' is only one word)
	KILL \[process\] signo:	kill a cdgp-process with signal number `signo'
	                        process default: most current process
	STATUS:			gives a list over all processes 
	CPU:			CPU load 
	QUIT:			quit the server"
      } \
      ^NEW$ {
        handle_NEW
      } \
      $NEWregex {
          regexp $NEWregex $task dummy job xml log cdgp
	  handle_NEW $job $xml $log $cdgp
      } \
      $TELLregex {
        regexp $TELLregex $task dummy process input
	handle_TELL $process $input
      } \
      $KILLregex {
	regexp $KILLregex $task dummy signo process
	handle_KILL $signo $process
      } \
      ^STATUS$ {
        handle_STATUS
      } \
      ^CPU$ {
	set cpu "/bin/cat /proc/loadavg"
	puts [eval exec $cpu]
      } \
      ^Q$ - \
      ^QUIT$ {
	handle_QUIT; break
      } \
      default {
	  puts "no available command: '$task'"
      }

  }
}

# create an object of itself and initiate the event loop
CdgServer cdgserver
cdgserver eventLoop
