# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaExperiment - bundled information about a experiment.
## A YadaExperiment is the central object of the YADA application. It brings
## together all the parts that are needed to start YadaJob objects starting the parser
## process in the background and consuming the produced data.
##
## So a YadaExperiment 
## - has a unique name,
## - needs a YadaGrammar, 
## - needs a command to fork the parser
## - specifies a YadaMachine where all jobs are started on,
## - a script to be played to the parser itself and
## - a unique data directory where the results of the experiments are stored
##   (two files per wordgraph: the standard output and the xml output data)
##
## All YadaExperiment objects are accessed via the YadaExperiments configuration
## document. YadaGrammars and YadaMachines that are available are configured separately.
##
## \ingroup YadaConfiguration
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaExperiment {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------

  ## the name of the experiment
  #public variable name ""

  ## the type of the experiment. 
  ## The type of the experiment specifies the parsing flavour used.
  ## Possible sensefull values by now are 
  ## - annotation (which is actually now parsing flavour but a way to compute
  ##   parses via anno2parse())
  ## - arcconsistency 
  ## - frobbing 
  ## - genetic
  ## - gls 
  ## - incremental completion
  ## - isearch 
  ## - netsearch 
  ## - pruning 
  ## - wfst
  public variable type ""

  ## path to cdgp binary file
  public variable command ""

  ## path where to store all computed data files
  public variable dataDir ""

  ## YadaGrammar to appy this experiment.
  public variable grammarName ""

  ## YadaMachine where to schedule resulting YadaJobs on.
  public variable machineName "localhost"

  ## CDGP script to feed into the STDIN of a CdgProcess.
  public variable script ""

  ## flag indicating whether this experiment is currently waiting for 
  ## a YadaJob to finish.
  public variable isActive 1

  ## flag indicating whether all data has been computed for this
  ## experiment, so no more YadaJobs are needed.
  public variable isComplete 0

  ## flag indicating whether all statistics in all data has been
  ## collected or not.
  public variable isUpToDate 0

  ## tcl command used to print a message. 
  ## \see print()
  public variable printCommand ""

  ## total scores collected in the experiments data files. 
  public variable totalScore 0

  ## labelled hits collected in the experiments data files. 
  public variable labelledHits 0

  ## strict hits collected in the experiments data files. 
  public variable strictHits 0

  ## lexical hits collected in the experiments data files. 
  public variable lexicalHits 0

  ## struct hits collected in the experiments data files. 
  public variable structHits 0

  ## total number of tries collected in the experiments data files. 
  public variable totalTries 0

  ## total time collected in the experiments data files. 
  public variable totalTime 0

  ## total solution time collected in the experiments data files. 
  public variable totalSolTime 0

  ## total soft time collected in the experiments data files. 
  public variable totalSoftTime 0

  # methods ------------------------------------------------------------------
  public method clone {args}; ## \type TclList
  public method check {}
  public method print {message}; ## \type TclString
  public method resetStatistics {}
  public method getXmlFileName {wgName}; ## \type TclString
  public method getLogFileName {wgName}; ## \type TclString
  public method getDocFileName {}
  public method getLogFile {wgName}; ## \type TclString
  public method getXmlFile {wgName}; ## \type TclString
  public method getFileName {} ; ## overriding virtual method
  public method toDom {}
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {experimentNode}; ## \type domNode
  public method deleteData {}

  constructor {args} {}; ## \type TclList

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaExperiment::constructor {args} {
  global env
  eval configure $args

  # check obliatory options
  if {"$name" == ""} {
    error "ERROR: unnamed experiment"
  }

  # create default data store in
  # YADA_DATA/machineName/grammarName/type/name
  if {"$dataDir" == ""} {
    #set dataDir [file join $::env(YADA_DATA) $machineName $grammarName $type $name]   
    set dataDir "<none>"
  }

  # security filter for dataDir
  if {$dataDir != "<none>"} {
    set securityFilter {[\\\*\?\~\^\$\@\%\`\"\'\&\;\|\<\>\x00-\x1F]} ;# from twiki
    regsub -all $securityFilter $dataDir "" dataDir
  }

  # check if the named experiment already exists
  foreach experiment [find object -class YadaExperiment] {
    if {$experiment == $this} {
      continue
    }
    if {[$experiment cget -name] == $name} {
      error "ERROR: experiment '$name' already exists"
    }
    if {[$experiment cget -dataDir] == $dataDir && $dataDir != "<none>"} {
      error "ERROR: experiment '$name' uses same store as '[$experiment cget -name]': $dataDir"
    }
  }

  # create data store
  if {$dataDir != "<none>" && ![file exists "$dataDir"] && $dataDir != ""} {
    file mkdir "$dataDir"
    set isComplete 0
  }
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaExperiment::clone {args} {
  set clone [eval newExperiment \
    -name {$name} \
    -type {$type} \
    -command {$command} \
    -grammarName {$grammarName} \
    -machineName {$machineName} \
    -script {$script} \
    -isActive {$isActive} \
    $args ]

  return $clone
}

## ----------------------------------------------------------------------------
## check
## ----------------------------------------------------------------------------
body YadaExperiment::check {} {
  if {$name == "<none>"} {
    return 1
  }
  
  if {$dataDir == "" || $grammarName == ""} {
    return 0
  }

  # check if the experiment is complete
  set grammar [.main getGrammar $grammarName]
  set selection [$grammar getSelection]
  set isComplete 1
  foreach wgName $selection {
    set file [getXmlFileName $wgName]
    if {![file exists $file]} {
      set isComplete 0
      break
    }
  }

  return $isComplete
}

## ----------------------------------------------------------------------------
## print
## ----------------------------------------------------------------------------
body YadaExperiment::print {message} {
  if {$printCommand != ""} {
    $printCommand $message
  }
}

## ----------------------------------------------------------------------------
## getXmlFileName
## ----------------------------------------------------------------------------
body YadaExperiment::getXmlFileName {wgName} {
  return [file join $dataDir ${wgName}.xml.gz]
}

## ----------------------------------------------------------------------------
## getLogFileName
## ----------------------------------------------------------------------------
body YadaExperiment::getLogFileName {wgName} {
  return [file join $dataDir ${wgName}.log.gz]
}

## ----------------------------------------------------------------------------
## getDocFileName
## ----------------------------------------------------------------------------
body YadaExperiment::getDocFileName {} {
  return [file join $dataDir yada.xml]
}

## ----------------------------------------------------------------------------
## getLogFile
## ----------------------------------------------------------------------------
body YadaExperiment::getLogFile {wgName} {
  set fileName [getLogFileName $wgName]
  if {![file exists $fileName]} {
    return ""
  }
  set fd [open "|zcat $fileName" r]
  set data [read $fd]
  close $fd

  return "$data"
}

## ----------------------------------------------------------------------------
## getXmlFile
## ----------------------------------------------------------------------------
body YadaExperiment::getXmlFile {wgName} {
  set fileName [getXmlFileName $wgName]
  if {![file exists $fileName]} {
    return ""
  }
  set fd [open "|zcat $fileName" r]
  set data [read $fd]
  close $fd

  return "$data"
}


## ----------------------------------------------------------------------------
## resetStatistics
## ----------------------------------------------------------------------------
body YadaExperiment::resetStatistics {} {
  set totalScore 0
  set labelledHits 0
  set structHits 0
  set lexicalHits 0
  set strictHits 0
  set totalTries 0
  set totalTime 0
  set totalSolTime 0
  set totalSoftTime 0
  set noSolutions 0
  set isUpToDate 0
}

## ----------------------------------------------------------------------------
## callback for script.
## ----------------------------------------------------------------------------
configbody YadaExperiment::script {
  regsub -all {^\s+} $script "" script
  regsub -all {\n\s+} $script "\n" script
}


## ----------------------------------------------------------------------------
## get the filename where the persistent state is saved in
## ----------------------------------------------------------------------------
body YadaExperiment::getFileName {} {
  global env
  set fileName [file join $env(YADA_ETC) experiments $name.xml]
  return [getSecureFileName $fileName]
}

## ----------------------------------------------------------------------------
## get a dom representation of this experiment
## @returns a new dom document
## ----------------------------------------------------------------------------
body YadaExperiment::toDom {} {
#  global env
  set document [dom createDocument "yada"]
  set root [$document documentElement]

  set experimentNode [$document createElement "experiment"]

  $experimentNode setAttribute "name" $name
  $experimentNode setAttribute "type" $type
  $experimentNode setAttribute "command" $command
  $experimentNode setAttribute "dataDir" $dataDir
  $experimentNode setAttribute "isActive" $isActive

  set grammarNode [$document createElement "grammarRef"]
  $grammarNode setAttribute "name" $grammarName
#  $grammarNode setAttribute "xmlns:xlink" "http://www.w3.org/1999/xlink"
#  $grammarNode setAttribute "xlink:type" "simple"
#  $grammarNode setAttribute "xlink:href" [file join $env(YADA_ETC) ${grammarName}_grammar.xml].gz 

  set machineNode [$document createElement "machineRef"]
  $machineNode setAttribute "name" $machineName
#  $machineNode setAttribute "xmlns:xlink" "http://www.w3.org/1999/xlink"
#  $machineNode setAttribute "xlink:type" "simple"
#  $machineNode setAttribute "xlink:href" [file join $env(YADA_ETC) ${machineName}_machine.xml].gz

  set scriptNode [$document createElement "script"]
  set scriptNodeText [$document createTextNode "$script"]

  $root appendChild $experimentNode
  $experimentNode appendChild $grammarNode 
  $experimentNode appendChild $machineNode 
  $experimentNode appendChild $scriptNode
  $scriptNode appendChild $scriptNodeText

  return $document
}

## ----------------------------------------------------------------------------
## initialize this experiment from a dom document
## @param document containing a /yada/experiment
## ----------------------------------------------------------------------------
body YadaExperiment::fromDom {document} {
  set rootNode [$document documentElement]
  set experimentNode [$rootNode selectNodes "/yada/experiment"]
  fromDomNode $experimentNode
}

## ----------------------------------------------------------------------------
## initialize this experiment from a dom node. 
## @param experimentNode containing the experiment attributes
## ----------------------------------------------------------------------------
body YadaExperiment::fromDomNode {experimentNode} {

  set grammarNode [$experimentNode selectNodes "./grammarRef"] 
  set scriptNode [$experimentNode selectNodes "./script"]
  set machineNode [$experimentNode selectNodes "./machineRef"]

  configure \
    -name [$experimentNode getAttribute "name"] \
    -type [$experimentNode getAttribute "type"] \
    -command [$experimentNode getAttribute "command"] \
    -dataDir [$experimentNode getAttribute "dataDir"] \
    -grammarName [$grammarNode getAttribute "name"] \
    -machineName [$machineNode getAttribute "name"] \
    -script [$scriptNode text] \
    -isActive [$experimentNode getAttribute "isActive"]
}

## ----------------------------------------------------------------------------
## delete all data computed using this experiment
## ----------------------------------------------------------------------------
body YadaExperiment::deleteData {} {
  global env

  if {[string first $env(YADA_DATA) $dataDir] != 0 && $dataDir != "<none>"} {
    # never reach
    error "invalid data directory '$dataDir', should have prefix '$env(YADA_DATA)'"
  }
  file delete -force -- $dataDir
}
