import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/** Parse a CDG annotation file and return its thematic relations. **/
public class SentencesParser {

    final static Pattern LINE_PATTERN = Pattern
	    .compile("(\\d\\d\\d)\\s+(\\S+ \\S+ \\S+ \\S+ (\\S+) \\S+ (\\S+) \\S+ (\\S+) (\\S+)\\.)");

    private final Map<String, String> _base;

    public SentencesParser(String lexikon) {
	_base = mapToBases(lexikon);
    }

    private Map<String, String> mapToBases(String lexikon) {
	final Pattern basePattern = Pattern
		.compile("^([^':\\s]+)\\s*:=\\s*\\[.*base:([^,]+).*\\];$");

	final Map<String, String> base = new HashMap<String, String>();

	try {
	    BufferedReader in = new BufferedReader(
		    new InputStreamReader(new FileInputStream(lexikon), Charset
			    .forName("ISO-8859-1")));
	    String line;
	    while ((line = in.readLine()) != null) {
		Matcher matcher = basePattern.matcher(line);
		if (matcher.matches()) {
		    base.put(matcher.group(1), matcher.group(2));
		}
	    }
	    in.close();
	} catch (IOException e) {
	    System.err
		    .println("Couldn't open grammar for base form extraction.");
	    e.printStackTrace();
	}

	return base;
    }

    private String getBase(String word) {
	// Special cases for compound words and unusual spellings
	if (word.equals("Reiseleiterin")) {
	    word = "Leiterin";
	} else if (word.equals("Filmdiva")) {
	    word = "Diva";
	} else if (word.equals("Blumenstrau")) {
	    word = "Strau";
	} else if (word.equals("Schachspielerin")) {
	    word = "Spielerin";
	} else if (word.equals("Quizmasterin")) {
	    word = "Masterin";
	} else if (word.equals("Choreografen")) {
	    word = "Choreograf";
	}
	if (_base.containsKey(word)) {
	    return _base.get(word);
	} else {
	    Logger.getLogger("SentencesParser").warning(
		    "No base form found for word: " + word);
	    return word;
	}
    }

    /**
     * Read a line from <code>in</code> and return thematic relations for a
     * genitive and dative reading of the sentence.
     * 
     * @param in
     *            A buffered reader, ready for reading a line from the
     *            van-Kampen-corpus.
     * @return The thematic relations, or <code>null</null> at end of stream.
     */
    public Pair<ThematicRelations, ThematicRelations> parseLine(
	    BufferedReader in) {
	String line = null;

	try {
	    line = in.readLine();
	} catch (IOException e) {
	    System.err.println("Error while reading the cda file.");
	    e.printStackTrace();
	}

	if (line == null) {
	    return null;
	}

	Logger.getLogger("SentencesParser").finer("parsing line: " + line);

	Matcher matcher = LINE_PATTERN.matcher(line);
	if (!matcher.matches()) {
	    System.err.println("Line doesn't match pattern: " + line);
	    return null;
	}
	final String id = matcher.group(1);
	final String sentence = matcher.group(2);
	final String person1 = matcher.group(3); // concept: Human.m.f
	final String person2 = matcher.group(4);
	final String object = matcher.group(5); // concept: Physical.Object
	final String verb = matcher.group(6);

	// Create the thematic relations for genitive reading.
	final ThematicRelations genitiveReading = new ThematicRelations(id,
		sentence);

	genitiveReading.addWord(6, getBase(person1));
	genitiveReading.addBaseConcept(6, BaseConcept.PERSON);
	genitiveReading.addNumber(6, Number.sg);
	genitiveReading.addRole(6, Role.AGENT, 11);

	genitiveReading.addWord(8, getBase(person2));
	genitiveReading.addBaseConcept(8, BaseConcept.PERSON);
	genitiveReading.addNumber(8, Number.sg);
	genitiveReading.addRole(8, Role.OWNER, 6);

	genitiveReading.addWord(10, getBase(object));
	genitiveReading.addBaseConcept(10, BaseConcept.OBJECT);
	genitiveReading.addNumber(10, Number.sg);
	genitiveReading.addRole(10, Role.THEME, 11);

	genitiveReading.addWord(11, getBase(verb));
	genitiveReading.addSemVal(11, SemVal.ag_th);

	// Create the thematic relations for dative reading.
	final ThematicRelations dativeReading = new ThematicRelations(id,
		sentence);

	dativeReading.addWord(6, getBase(person1));
	dativeReading.addBaseConcept(6, BaseConcept.PERSON);
	dativeReading.addNumber(6, Number.sg);
	dativeReading.addRole(6, Role.AGENT, 11);

	dativeReading.addWord(8, getBase(person2));
	dativeReading.addBaseConcept(8, BaseConcept.PERSON);
	dativeReading.addNumber(8, Number.sg);
	dativeReading.addRole(8, Role.RECIPIENT, 11);

	dativeReading.addWord(10, getBase(object));
	dativeReading.addBaseConcept(10, BaseConcept.OBJECT);
	dativeReading.addNumber(10, Number.sg);
	dativeReading.addRole(10, Role.THEME, 11);

	dativeReading.addWord(11, getBase(verb));
	dativeReading.addSemVal(11, SemVal.ag_re_th);

	return new Pair<ThematicRelations, ThematicRelations>(genitiveReading,
		dativeReading);
    }

    /**
     * Read a line from <code>in</code> and return thematic relations for a
     * genitive reading of the sentence.
     * 
     * @param in
     *            A buffered reader, ready for reading a line from the
     *            van-Kampen-corpus.
     * @return The thematic relations.
     */
    public ThematicRelations parseWordgraphForGenitiveReading(BufferedReader in) {
	return parseLine(in).x;
    }

    /**
     * Read a line from <code>in</code> and return thematic relations for a
     * dative reading of the sentence.
     * 
     * @param in
     *            A buffered reader, ready for reading a line from the
     *            van-Kampen-corpus.
     * @return The thematic relations.
     */
    public ThematicRelations parseWordgraphForDativeReading(BufferedReader in) {
	return parseLine(in).y;
    }
}
