import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.logging.Logger;

import org.semanticweb.owl.inference.OWLClassReasoner;
import org.semanticweb.owl.inference.OWLReasonerException;
import org.semanticweb.owl.model.AxiomType;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLDeclarationAxiom;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLInverseObjectPropertiesAxiom;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectPropertyExpression;
import org.semanticweb.owl.model.OWLObjectValueRestriction;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyChangeException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.OWLOntologyStorageException;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.owl.util.ToldClassHierarchyReasoner;
import org.semanticweb.owl.vocab.OWLRDFVocabulary;

/**
 * Get the relevant information for building an A-Box from the T-Box.
 */
public class TBoxParser {

    private OWLOntology _tBox;

    private OWLOntologyManager _manager;

    private OWLClass _entityConcept;

    private OWLClass _lexicalisationConcept;

    private Map<String, List<OWLClass>> _conceptForLexicalisation;

    private Map<Role, OWLObjectPropertyExpression> _property;

    private Map<BaseConcept, OWLClass> _conceptForBaseConcept;

    private Map<Number, OWLClass> _conceptForNumber;

    private Map<SemVal, OWLClass> _conceptForSemVal;

    private OWLClassReasoner _reasoner;

    private OWLObjectPropertyExpression _hasLexicalisationProperty;

    public TBoxParser(OWLOntologyManager manager, OWLOntology tBox) {
	_manager = manager;
	_tBox = tBox;
	_entityConcept = getEntityConcept();
	_lexicalisationConcept = getLexicalisationConcept();
	_hasLexicalisationProperty = getHasLexicalisationProperty();
	_conceptForLexicalisation = mapLexicalisationsToConcepts();
	_property = mapRolesToProperties();
	_conceptForBaseConcept = mapBaseConcepts();
	_conceptForNumber = mapNumbersToConcepts();
	_reasoner = new ToldClassHierarchyReasoner(_manager);
	Set<OWLOntology> singletonSet = new HashSet<OWLOntology>();
	singletonSet.add(tBox);
	try {
	    _reasoner.loadOntologies(singletonSet);
	    _conceptForSemVal = mapSemValsToConcepts();
	} catch (OWLReasonerException e) {
	    System.err.println("Error using ToldClassHierarchyReasoner.");
	    e.printStackTrace();
	}
    }

    private OWLObjectPropertyExpression getHasLexicalisationProperty() {
	for (OWLInverseObjectPropertiesAxiom axiom : _tBox
		.getAxioms(AxiomType.INVERSE_OBJECT_PROPERTIES)) {
	    OWLObjectPropertyExpression property = axiom.getFirstProperty();
	    if (property.toString().equals("has_Lexicalisation")) {
		return property;
	    }
	}
	throw new NoSuchElementException(
		"ObjectProperty has_Lexicalisation missing from T-Box.");
    }

    private OWLClass getLexicalisationConcept() {
	for (OWLSubClassAxiom axiom : _tBox.getAxioms(AxiomType.SUBCLASS)) {
	    if (axiom.getSubClass().toString().equals("Lexicalisation")) {
		return axiom.getSubClass().asOWLClass();
	    }
	}
	throw new NoSuchElementException(
		"Concept Lexicalisation missing from T-Box.");
    }

    private Map<BaseConcept, OWLClass> mapBaseConcepts() {
	Map<String, OWLClass> concepts = new HashMap<String, OWLClass>();
	for (OWLSubClassAxiom axiom : _tBox.getAxioms(AxiomType.SUBCLASS)) {
	    if (axiom.getSuperClass().toString().equals("Concrete")) {
		OWLDescription description = axiom.getSubClass();
		concepts.put(description.toString(), description.asOWLClass());
	    }
	}

	Map<BaseConcept, OWLClass> baseConcepts = new HashMap<BaseConcept, OWLClass>();

	for (BaseConcept baseConcept : BaseConcept.values()) {
	    baseConcepts
		    .put(baseConcept, concepts.get(baseConcept.conceptName));
	    Logger.getLogger("TBoxParser").fine(
		    "Mapped BaseConcept: " + baseConcept + " to concept: "
			    + concepts.get(baseConcept.conceptName));
	}

	return baseConcepts;
    }

    /**
     * @return Entity.Concept
     */
    private OWLClass getEntityConcept() {
	OWLClass thing = _manager.getOWLDataFactory().getOWLClass(
		OWLRDFVocabulary.OWL_THING.getURI());
	for (OWLDescription subClass : thing.getSubClasses(_tBox)) {
	    if (subClass.toString().equals("Entity.Concept")) {
		return subClass.asOWLClass();
	    }
	}
	throw new NoSuchElementException("Entity.Concept missing from T-Box.");
    }

    /**
     * @return A mapping from lexicalisations to concepts that have them.
     */
    private Map<String, List<OWLClass>> mapLexicalisationsToConcepts() {
	Map<String, List<OWLClass>> concepts = new HashMap<String, List<OWLClass>>();

	for (OWLSubClassAxiom axiom : _tBox.getAxioms(AxiomType.SUBCLASS)) {
	    OWLDescription subClass = axiom.getSubClass();
	    if (subClass.isAnonymous()) {
		continue;
	    }
	    OWLClass concept = subClass.asOWLClass();

	    OWLDescription superClass = axiom.getSuperClass();
	    OWLObjectValueRestriction restriction = null;

	    if (superClass instanceof OWLObjectValueRestriction) {
		restriction = (OWLObjectValueRestriction) superClass;
	    } else if (superClass instanceof OWLObjectIntersectionOf) {
		OWLObjectIntersectionOf intersection = (OWLObjectIntersectionOf) superClass;
		for (OWLDescription desc : intersection.getOperands()) {
		    if (desc instanceof OWLObjectValueRestriction) {
			restriction = (OWLObjectValueRestriction) desc;
		    }
		}
	    }

	    if (restriction != null
		    && restriction.getProperty().toString().equals(
			    "has_Lexicalisation")) {
		String lexicalisation = restriction.getValue().toString();
		if (!concepts.containsKey(lexicalisation)) {
		    concepts.put(lexicalisation, new LinkedList<OWLClass>());
		}
		concepts.get(lexicalisation).add(concept);
	    }

	}

	return concepts;
    }

    /**
     * @return a mapping from roles to object properties that represent them
     */
    private Map<Role, OWLObjectPropertyExpression> mapRolesToProperties() {
	Map<Role, OWLObjectPropertyExpression> propertyForRole = new HashMap<Role, OWLObjectPropertyExpression>();

	Map<String, OWLObjectPropertyExpression> propertyForName = new HashMap<String, OWLObjectPropertyExpression>();
	for (OWLInverseObjectPropertiesAxiom axiom : _tBox
		.getAxioms(AxiomType.INVERSE_OBJECT_PROPERTIES)) {
	    for (OWLObjectPropertyExpression property : axiom.getProperties()) {
		propertyForName.put(property.toString(), property);
	    }
	}

	for (Role role : Role.values()) {
	    String name = "is_" + role + "_for";
	    propertyForRole.put(role, propertyForName.get(name));
	}

	return propertyForRole;
    }

    /**
     * @return a mapping from grammatical numbers to matching concepts
     */
    private Map<Number, OWLClass> mapNumbersToConcepts() {

	Map<String, OWLClass> concepts = new HashMap<String, OWLClass>();
	for (OWLDeclarationAxiom axiom : _tBox.getAxioms(AxiomType.DECLARATION)) {
	    OWLEntity entity = axiom.getEntity();
	    if (entity.isOWLClass()) {
		concepts.put(entity.toString(), entity.asOWLClass());
	    }
	}

	Map<Number, OWLClass> numbersToConcepts = new HashMap<Number, OWLClass>();

	for (Number number : Number.values()) {
	    numbersToConcepts.put(number, concepts.get(number.conceptName));
	    Logger.getLogger("TBoxParser").fine(
		    "Mapped number: " + number + " to concept: "
			    + concepts.get(number.conceptName));
	}

	return numbersToConcepts;
    }

    /**
     * @return a mapping from semantic values to matching concepts
     * @throws OWLReasonerException
     */
    private Map<SemVal, OWLClass> mapSemValsToConcepts()
	    throws OWLReasonerException {
	OWLClass thing = _manager.getOWLDataFactory().getOWLClass(
		OWLRDFVocabulary.OWL_THING.getURI());
	OWLClass situationConcept = null;
	for (OWLDescription subClass : thing.getSubClasses(_tBox)) {
	    if (subClass.toString().equals("Situation.Concept")) {
		situationConcept = subClass.asOWLClass();
	    }
	}
	if (situationConcept == null) {
	    throw new NoSuchElementException(
		    "Situation.Concept missing from T-Box.");
	}

	Map<String, OWLClass> concepts = new HashMap<String, OWLClass>();
	for (OWLDescription subClass : situationConcept.getSubClasses(_tBox)) {
	    for (Set<OWLClass> semValConcepts : _reasoner
		    .getSubClasses(subClass)) {
		for (OWLClass semValConcept : semValConcepts) {
		    Logger.getLogger("TBoxParser").fine(
			    "Adding this semantic value concept: "
				    + semValConcept);
		    concepts.put(semValConcept.toString(), semValConcept);
		}
	    }
	}

	Map<SemVal, OWLClass> semValsToConcepts = new HashMap<SemVal, OWLClass>();

	for (SemVal semVal : SemVal.values()) {
	    if (!concepts.containsKey(semVal.conceptName)) {
		throw new NoSuchElementException(
			"No concept for semantic value: " + semVal);
	    }
	    semValsToConcepts.put(semVal, concepts.get(semVal.conceptName));
	    Logger.getLogger("TBoxParser").fine(
		    "Mapped sem_val: " + semVal + " to concept: "
			    + concepts.get(semVal.conceptName));
	}

	return semValsToConcepts;
    }

    /**
     * @return the ontology URI of the T-Box
     */
    public URI getURI() {
	return _tBox.getURI();
    }

    /**
     * Log a warning that the <code>word</code> from sentence
     * <code>sentenceID</code> will be represented by a generalized
     * <code>concept</code>.
     */
    private void logGeneralization(OWLClass concept, String word,
	    String sentenceID) {
	Logger.getLogger("TBoxParser").warning(
		"Using generalized concept " + concept + " for word " + word
			+ " from sentence: " + sentenceID);
    }

    /**
     * @return the concept whose superclass is <code>numberConcept</code>.
     */
    private OWLClass findConceptWithMatchingNumber(String word,
	    List<OWLClass> concepts, OWLClass numberConcept, String sentence) {
	for (OWLClass concept : concepts) {
	    try {
		Set<Set<OWLClass>> superClasses = _reasoner
			.getSuperClasses(concept);
		for (Set<OWLClass> superClassesSet : superClasses) {
		    if (superClassesSet.contains(numberConcept)) {
			return concept;
		    }
		}
	    } catch (OWLReasonerException e) {
		System.err.println("Error querying reasoner about T-Box.");
		e.printStackTrace();
	    }
	}
	return askUserAboutConcept(word, concepts, sentence);
    }

    /**
     * @return the concept whose superclass is <code>semValConcept</code>.
     */
    private OWLClass findConceptWithMatchingSemVal(String word,
	    List<OWLClass> concepts, OWLClass semValConcept, String sentenceID) {
	for (OWLClass concept : concepts) {
	    try {
		Set<Set<OWLClass>> superClasses = _reasoner
			.getSuperClasses(concept);
		for (Set<OWLClass> superClassesSet : superClasses) {
		    if (superClassesSet.contains(semValConcept)) {
			return concept;
		    }
		}
	    } catch (OWLReasonerException e) {
		System.err.println("Error querying reasoner about T-Box.");
		e.printStackTrace();
	    }
	}
	return createNewConcept(word, semValConcept, sentenceID);
    }

    /**
     * @return the concept the user chose for representing the word.
     */
    private OWLClass askUserAboutConcept(String word, List<OWLClass> concepts,
	    String sentence) {
	System.out.println("Sentence: " + sentence);
	System.out.println("Choose a concept for individual: " + word);
	int n = concepts.size();
	for (int i = 0; i < n; ++i) {
	    System.out.format("%2d: %s\n", i, concepts.get(i));
	}
	int index = -1;
	BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
	while (index < 0 || index >= n) {
	    System.out.format("Enter a number [0-%d]: ", n - 1);
	    try {
		index = Integer.parseInt(in.readLine());
	    } catch (IOException e) {
		// index is unchanged, new iteration.
	    } finally {
		System.out.println();
	    }
	}
	Logger.getLogger("TBoxParser").info(
		"Using concept: " + concepts.get(index));
	return concepts.get(index);
    }

    /**
     * Create a mapping from the individuals of the thematic relations to
     * concepts that have their words as lexicalisation.
     * 
     * @param relations
     *            The thematic relations.
     * @return The mapping.
     */
    public Map<Integer, OWLClass> mapIndividualsToConcepts(
	    ThematicRelations relations) {
	Map<Integer, OWLClass> mapping = new HashMap<Integer, OWLClass>();

	for (int id : relations.getIndividuals()) {
	    String word = relations.getWord(id);
	    OWLClass concept;

	    if (_conceptForLexicalisation.containsKey(word)) {
		List<OWLClass> concepts = _conceptForLexicalisation.get(word);
		if (relations.hasSemVal(id)) {
		    SemVal semVal = relations.getSemVal(id);
		    concept = findConceptWithMatchingSemVal(word, concepts,
			    _conceptForSemVal.get(semVal), relations.getID());
		} else if (concepts.size() == 1) {
		    concept = concepts.get(0);
		} else if (relations.hasNumber(id)) {
		    Number number = relations.getNumber(id);
		    concept = findConceptWithMatchingNumber(word, concepts,
			    _conceptForNumber.get(number), relations
				    .getSentence());
		} else {
		    concept = askUserAboutConcept(word, concepts, relations
			    .getSentence());
		}
	    } else {
		if (relations.hasSemVal(id)) {
		    concept = createNewConcept(word, _conceptForSemVal
			    .get(relations.getSemVal(id)), relations.getID());
		} else if (relations.hasBaseConcept(id)) {
		    concept = createNewConcept(word, _conceptForBaseConcept
			    .get(relations.getBaseConcept(id)), relations
			    .getID());
		} else if (relations.hasNumber(id)) {
		    concept = _conceptForNumber.get(relations.getNumber(id));
		    logGeneralization(concept, word, relations.getID());
		} else {
		    concept = _entityConcept;
		    logGeneralization(concept, word, relations.getID());
		}
	    }
	    mapping.put(id, concept);
	}
	return mapping;
    }

    private OWLClass createNewConcept(String word, OWLClass baseConcept,
	    String sentenceID) {
	OWLDataFactory factory = _manager.getOWLDataFactory();

	// Create the lexicalisation individual
	URI lexicalisationURI = URI.create(String.format("%s#%s", _tBox
		.getURI(), word));
	OWLIndividual lexicalisation = factory
		.getOWLIndividual(lexicalisationURI);

	// Create the concept
	String uriString = _tBox.getURI() + "#";
	if (baseConcept == _conceptForSemVal.get(SemVal.ag)) {
	    uriString += "Null." + word;
	} else if (baseConcept == _conceptForSemVal.get(SemVal.ag_re)) {
	    uriString += "Jmd." + word;
	} else if (baseConcept == _conceptForSemVal.get(SemVal.ag_th)) {
	    uriString += "Etw." + word;
	} else if (baseConcept == _conceptForSemVal.get(SemVal.ag_re_th)) {
	    uriString += "Jmd.Etw." + word;
	} else {
	    uriString += word + "_Concept";
	}
	URI conceptURI = URI.create(uriString);
	OWLClass concept = factory.getOWLClass(conceptURI);

	// Create a ObjectValueRestriction for assigning the lexicalisation
	OWLObjectValueRestriction hasLexicalisation = factory
		.getOWLObjectValueRestriction(_hasLexicalisationProperty,
			lexicalisation);

	// Collect all subclasses of the base concept that should be disjoint
	Set<OWLDescription> subClasses = baseConcept.getSubClasses(_tBox);
	subClasses.add(concept);

	try {
	    // Put the lexicalisation into the T-Box
	    _manager.addAxiom(_tBox, factory.getOWLClassAssertionAxiom(
		    lexicalisation, _lexicalisationConcept));
	    _manager.addAxiom(_tBox, factory
		    .getOWLDeclarationAxiom(lexicalisation));

	    // Put the concept into the T-Box
	    _manager.addAxiom(_tBox, factory.getOWLSubClassAxiom(concept,
		    baseConcept));
	    _manager.addAxiom(_tBox, factory.getOWLDeclarationAxiom(concept));

	    // Assign the lexicalisation to the concept
	    _manager.addAxiom(_tBox, factory.getOWLSubClassAxiom(concept,
		    hasLexicalisation));

	    // Make the subclasses of the base concept disjoint
	    _manager.addAxiom(_tBox, factory
		    .getOWLDisjointClassesAxiom(subClasses));
	} catch (OWLOntologyChangeException e) {
	    System.err.println("Could not extend the T-Box.");
	    e.printStackTrace();
	    logGeneralization(baseConcept, word, sentenceID);
	    return baseConcept;
	}
	try {
	    // Save the changes to the T-Box
	    _manager.saveOntology(_tBox);

	} catch (OWLOntologyStorageException e) {
	    System.err.println("Could not save the extended T-Box.");
	    e.printStackTrace();
	}

	Logger.getLogger("TBoxParser").info(
		"Created new concept " + concept + " for word " + word
			+ " from sentence: " + sentenceID);

	return concept;
    }

    /**
     * @param role
     *            a kind of thematic relation.
     * @return the property representing this kind of thematic relations.
     */
    public OWLObjectPropertyExpression getPropertyForRole(Role role) {
	return _property.get(role);
    }

}
