import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import java.util.regex.Pattern;

/** Representation of the thematic relations in a sentence. **/
public class ThematicRelations {

    private String _id;

    private Map<Integer, String> _word;

    private Set<Integer> _individuals;

    private Map<Integer, Pair<Role, Integer>> _role;

    private String _sentence;

    private Map<Integer, Number> _number;

    private Map<Integer, SemVal> _semVal;

    private Map<Integer, BaseConcept> _baseConcept;

    public ThematicRelations(String id, String sentence) {
	_id = id;
	_word = new HashMap<Integer, String>();
	_individuals = new HashSet<Integer>();
	_role = new HashMap<Integer, Pair<Role, Integer>>();
	_sentence = sentence;
	_number = new HashMap<Integer, Number>();
	_semVal = new HashMap<Integer, SemVal>();
	_baseConcept = new HashMap<Integer, BaseConcept>();
    }

    /**
     * @return the sentence ID that this thematic relations were built from.
     */
    public String getID() {
	return _id;
    }

    /**
     * @param id
     *            A word ID.
     * @return The word for the given ID.
     **/
    public String getWord(int id) {
	return _word.get(id);
    }

    /**
     * @param id
     *            A word ID.
     * @param word
     *            The word for the given ID.
     */
    public void addWord(int id, String word) {
	if (Pattern.compile("[]").matcher(word).find()) {
	    word = replaceUmlauts(word);
	}
	_word.put(id, word);
    }

    private String replaceUmlauts(String word) {
	word = Pattern.compile("").matcher(word).replaceAll("Ae");
	word = Pattern.compile("").matcher(word).replaceAll("Oe");
	word = Pattern.compile("").matcher(word).replaceAll("Ue");
	word = Pattern.compile("").matcher(word).replaceAll("ae");
	word = Pattern.compile("").matcher(word).replaceAll("oe");
	word = Pattern.compile("").matcher(word).replaceAll("ue");
	return Pattern.compile("").matcher(word).replaceAll("ss");
    }

    /**
     * Add a thematic relation.
     * 
     * @param dependant
     *            The ID of the dependant word.
     * @param role
     *            The role of the relation.
     * @param regent
     *            The ID of the regent word.
     */
    public void addRole(int dependant, Role role, int regent) {
	_role.put(dependant, new Pair<Role, Integer>(role, regent));

	_individuals.add(dependant);
	_individuals.add(regent);
	Logger.getLogger("ThematicRelations").fine(
		String.format("Added role %s from dependant %d to regent %d.",
			role, dependant, regent));
    }

    /**
     * @return The set of individuals.
     * 
     *         It contains the word IDs of all words taking part in a thematic
     *         relation.
     */
    public Set<Integer> getIndividuals() {
	return _individuals;
    }

    /**
     * @param id
     *            A word ID.
     * @return If the word ID belongs to the dependant word of a thematic
     *         relation.
     */
    public boolean hasRole(int id) {
	return _role.containsKey(id);
    }

    /**
     * @param id
     *            The word ID of a dependant word.
     * @return The role of its thematic relation.
     */
    public Role getRole(int id) {
	return _role.get(id).x;
    }

    /**
     * @param id
     *            The word ID of a dependant word.
     * @return The word ID of the regent word of its thematic relation.
     */
    public int getRegent(int id) {
	return _role.get(id).y;
    }

    /**
     * @return A string representing the sentence the .cda file resulted from.
     *         It contains the words from the ID lines, seperated by spaces.
     */
    public String getSentence() {
	return _sentence;
    }

    /**
     * Add a base concept.
     * 
     * @param id
     *            The ID of the word.
     * @param semVal
     *            The base concept to use for it.
     */
    public void addBaseConcept(int id, BaseConcept baseConcept) {
	_baseConcept.put(id, baseConcept);
    }

    /**
     * @param id
     *            A word ID.
     * @return If the word ID has been assigned a base concept.
     */
    public boolean hasBaseConcept(int id) {
	return _baseConcept.containsKey(id);
    }

    /**
     * @param id
     *            The word ID of a word which has been assigned a base concept.
     * @return the base concept to use for it.
     */
    public BaseConcept getBaseConcept(int id) {
	return _baseConcept.get(id);
    }

    /**
     * Add a grammatical number.
     * 
     * @param id
     *            The ID of the word.
     * @param semVal
     *            Its grammatical number.
     */
    public void addNumber(int id, Number number) {
	_number.put(id, number);
    }

    /**
     * @param id
     *            A word ID.
     * @return If the word ID has been assigned a grammatical number.
     */
    public boolean hasNumber(int id) {
	return _number.containsKey(id);
    }

    /**
     * @param id
     *            The word ID of a word which has been assigned a grammatical
     *            number.
     * @return Its grammatical number.
     */
    public Number getNumber(int id) {
	return _number.get(id);
    }

    /**
     * Add a semantic value.
     * 
     * @param id
     *            The ID of the word.
     * @param semVal
     *            Its semantic value.
     */
    public void addSemVal(int id, SemVal semVal) {
	_semVal.put(id, semVal);
    }

    /**
     * @param id
     *            A word ID.
     * @return If the word ID has been assigned a semantic value.
     */
    public boolean hasSemVal(int id) {
	return _semVal.containsKey(id);
    }

    /**
     * @param id
     *            The word ID of a word which has been assigned a semantic
     *            value.
     * @return Its semantic value.
     */
    public SemVal getSemVal(int id) {
	return _semVal.get(id);
    }

}
